/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: Adm.java,v 1.43 2003/09/06 00:08:47 ehardesty Exp $
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas.adm;


import org.objectweb.jonas_ear.deployment.api.EarDeploymentDescException;
import java.rmi.RemoteException;
import java.net.URLClassLoader;
import java.net.MalformedURLException;
import java.util.Enumeration;
import java.util.Properties;
import java.util.Vector;
import java.io.FileOutputStream;
import java.io.File;
import java.io.IOException;
import javax.naming.InitialContext;
import javax.naming.NamingEnumeration;
import javax.naming.NamingException;
import javax.naming.NameClassPair;
import javax.rmi.PortableRemoteObject;
import org.objectweb.jonas.common.JProp;
import org.objectweb.jonas.common.LogManagement;
import org.objectweb.jonas.container.EJBServiceImpl;
import org.objectweb.jonas.ear.EarServiceException;
import org.objectweb.jonas.ear.EarServiceImplMBean;
import org.objectweb.jonas.jtm.TransactionServiceImpl;
import org.objectweb.jonas.mail.MailServiceImplMBean;
import org.objectweb.jonas.naming.NamingManager;
import org.objectweb.jonas.resource.ResourceServiceException;
import org.objectweb.jonas.resource.ResourceServiceImplMBean;
import org.objectweb.jonas.service.ServiceException;
import org.objectweb.jonas.service.ServiceManager;
import org.objectweb.jonas.web.AbsJWebContainerServiceImplMBean;
import org.objectweb.jonas.web.JWebContainerServiceException;
import org.objectweb.jonas_ejb.container.Container;
import org.objectweb.jonas_ejb.genic.GenIC;
import org.objectweb.jonas_ear.deployment.xml.Ejb;
import org.objectweb.common.Cmd;
import org.objectweb.jonas_ear.deployment.api.EarDeploymentDesc;
import java.net.URL;

/**
 * This class implements a remote interface used for administering
 *      the server.
 * @author Philippe Coq, Philippe Durieux
 * Contributor(s): Marc Dutoo, 
 * JOnAS 2.4 Murad Meghani (Murad.Meghani@compuware.com) killServer and stopServer
 * JOnAS 2.5 2002.06 Florent Benoit & Ludovic Bert : Methods for wars and ear files
 * Dean Jennings : add deployFile
 */
public class Adm extends PortableRemoteObject implements AdmInterface {

    public static final String ADMNAME_SUFFIX = "_Adm";

    private ServiceManager sm = null;
    private EJBServiceImpl ejbserv  = null;
    private TransactionServiceImpl tm = null;
    private JProp jonasProperties = null;
    private InitialContext ictx = null;
    private boolean isEJBContainer;

    // Server State
    public static final int NOT_READY = 0;
    public static final int READY = 1;
    public static final int STOPPED = 2;
    private int serverState = NOT_READY;

    /**
     * Reference to the web container service
     */
    private AbsJWebContainerServiceImplMBean webContainerService = null;

    /**
     * Reference to the ear service
     */
    private EarServiceImplMBean earService = null;

    /**
     * Reference to the resource service
     */
    private ResourceServiceImplMBean resourceService = null;

    /**
     * Reference to the mail service
     */
    private MailServiceImplMBean mailService = null;

    /**
     * Adm constructor
     */
    public Adm(JProp jp) throws RemoteException, NamingException, ServiceException, Exception {

        jonasProperties = jp;

        // Get Service references
        sm = ServiceManager.getInstance();

        // Get an Initial Context and
        // registers the JOnAS Server Name to JNDI
        String name = null;
        name = jonasProperties.getValue("jonas.name", "jonas") + ADMNAME_SUFFIX;
        try {
            ictx = NamingManager.getInstance().getInitialContext();
            ictx.rebind(name, this);
        } catch (NamingException e) {
            throw e;
        }
    }

    private TransactionServiceImpl getTM() {
        if (tm == null) {
            try {
                tm = (TransactionServiceImpl) sm.getTransactionService();
            } catch (ServiceException e) {
            }
        }
        return tm;
    }

    /** 
     * Get the web container service
     * We can use methods provided by the interface for MBean
     */
    private AbsJWebContainerServiceImplMBean getWebContainerService() {
        if (webContainerService == null) {
            try {
                webContainerService = (AbsJWebContainerServiceImplMBean) sm.getWebContainerService();
            } catch (ServiceException e) {
                //not started
            }
        }
        return webContainerService;
    }

    /**
     * Get the mail service
     * We can use methods provided by the interface for MBean
     */
    private MailServiceImplMBean getMailService() {
        if (mailService == null) {
            try {
                mailService = (MailServiceImplMBean) sm.getMailService();
            } catch (Exception e) {
                //not started
            }
        }
        return mailService;
    }

    /**
     * Get the ear service
     * We can use methods provided by the interface for MBean
     */
    private EarServiceImplMBean getEarService() {
        if (earService == null) {
            try {
                earService = (EarServiceImplMBean) sm.getEarService();
            } catch (Exception e) {
                //not started
            }
        }
        return earService;
    }

    /**
     * Get the resource service
     * We can use methods provided by the interface for MBean
     */
    private ResourceServiceImplMBean getResourceService() {
        if (resourceService == null) {
            try {
                resourceService = (ResourceServiceImplMBean) sm.getResourceService();
            } catch (Exception e) {
                //not started
            }
        }
        return resourceService;
    }

    // ------------------------------------------------------------------
    //  AdmInterface implementation
    // ------------------------------------------------------------------

    /**
     * get Topics. 
     * Assumes that all Loggers are TopicalLoggers.
     */
    public String [] getTopics() throws RemoteException {
        return LogManagement.getInstance().getTopics();
    }

    /**
     * get Topic Level
     */
    public String getTopicLevel(String topic) throws RemoteException {
        return LogManagement.getInstance().getTopicLevel(topic);
    }

    /**
     * set Topic Level
     */
    public void setTopicLevel(String topic, String l) throws RemoteException {
        LogManagement.getInstance().setTopicLevel(topic, l);
    }
	
    /**
     * Create a container and load beans in it
     * @param fileName name of the ejb-jar or xml file
     */
    public void addBeans(String fileName) throws RemoteException {
        if (!isEJBContainer) {
            return;
        }
        ejbserv.createContainer(fileName);
    }


    /**
     * Deploy a given ear file with the help of the ear service.
     * @param fileName the name of the ear file.
     * @throws RemoteException if rmi call failed.
     * @throws EarServiceException if the deployment failed.
     */
    public void addEar(String fileName) throws RemoteException, EarServiceException {
        if (getEarService() != null) {
            getEarService().deployEarMBean(fileName);
        } else {
            throw new EarServiceException("Can't add the Ear file " + fileName + ". The ear service is not started");
        }
    }

    /**
     * Test if the specified filename is already deployed or not
     * @param fileName the name of the ear file.
     * @return true if the ear is deployed, else false.
     * @throws RemoteException if rmi call failed.
    */
    public boolean isEarLoaded(String fileName) throws RemoteException, EarServiceException {
        boolean isLoaded = false;
        if (getEarService() != null) {
            isLoaded = getEarService().isEarLoaded(fileName);
        } else {
            throw new EarServiceException("Can't test if the ear file " + fileName + " is deployed or not. The ear service is not started");
        }
        return isLoaded;
    }

    /**
     * Deploy a given rar file with the help of the resource service.
     * @param fileName the name of the rar file.
     * @throws RemoteException if rmi call failed.
     * @throws ResourceServiceException if the deployment failed.
     */
    public void addRar(String fileName) throws RemoteException, ResourceServiceException {
        if (getResourceService() != null) {
            getResourceService().deployRarMBean(fileName);
        } else {
            throw new ResourceServiceException("Can't add the Rar file " + fileName + ". The resource service is not started");
        }
    }

    /**
     * Test if the specified filename is already deployed or not
     * @param fileName the name of the rar file.
     * @return true if the rar is deployed, else false.
     * @throws RemoteException if rmi call failed.
     * @throws ResourceServiceException if unable to get resource service
    */
    public boolean isRarLoaded(String fileName) throws RemoteException, ResourceServiceException {
        boolean isLoaded = false;
        if (getResourceService() != null) {
            isLoaded = getResourceService().isRarLoaded(fileName);
        } else {
            throw new ResourceServiceException("Can't test if the rar file " + fileName + " is deployed or not. The rar service is not started");
        }
        return isLoaded;
    }

    /**
     * Deploy a given war file with the help of the web container service.
     * @param fileName the name of the war file.
     * @throws RemoteException if rmi call failed.
     * @throws JWebContainerServiceException if the deployment failed.
     */
    public void addWar(String fileName) throws RemoteException, JWebContainerServiceException {
        if (getWebContainerService() != null) {
            getWebContainerService().registerWarMBean(fileName);
        } else {
            throw new JWebContainerServiceException("Can't add the war file " + fileName + ". The web container service is not started");
        }
    }

    /**
     * Test if the specified filename is already deployed or not
     * @param fileName the name of the war file.
     * @return true if the war is deployed, else false.
     * @throws RemoteException if rmi call failed.
     */
    public boolean isWarLoaded(String fileName) throws RemoteException, JWebContainerServiceException {
        boolean isLoaded = false;
        if (getWebContainerService() != null) {
            isLoaded = getWebContainerService().isWarLoaded(fileName);
        } else {
            throw new JWebContainerServiceException("Can't test if the war file " + fileName + " is deployed or not. The war service is not started");
        }
        return isLoaded;
    }

    /**
     * UnDeploy a given ear file with the help of the ear service.
     * @param fileName the name of the ear file.
     * @throws RemoteException if rmi call failed.
     * @throws EarServiceException if the undeployment failed.
     */
    public void removeEar(String fileName) throws RemoteException, EarServiceException {
        if (getEarService() != null) {
            getEarService().unDeployEarMBean(fileName);
            // needed here to run also the distributed garbage collector
            runGC();
        } else {
            throw new EarServiceException("Can't remove of the Ear file " + fileName + ". The ear service is not started");
        }
    }

    /**
     * UnDeploy a given rar file with the help of the resource service.
     * @param fileName the name of the rar file.
     * @throws RemoteException if rmi call failed.
     * @throws ResourceServiceException if the undeployment failed.
     */
    public void removeRar(String fileName) throws RemoteException, ResourceServiceException {
        if (getResourceService() != null) {
            getResourceService().unDeployRarMBean(fileName);
            // needed here to run also the distributed garbage collector
            runGC();
        } else {
            throw new ResourceServiceException("Can't remove of the Rar file " + fileName + ". The resource service is not started");
        }
    }

    /**
     * UnDeploy a given war file with the help of the web container service.
     * @param fileName the name of the war file.
     * @throws RemoteException if rmi call failed.
     * @throws JWebContainerServiceException if the undeployment failed.
     */
    public void removeWar(String fileName) throws RemoteException, JWebContainerServiceException {
        if (getWebContainerService() != null) {
            getWebContainerService().unRegisterWarMBean(fileName);
            // needed here to run also the distributed garbage collector
            runGC();
        } else {
            throw new JWebContainerServiceException("Can't remove of the war file " + fileName + ". The web container service is not started");
        }
    }


    /**
     * Remove the container identified by fileName and remove all beans in it
     * @param fileName name of the ejb-jar or xml file 
     */
    public void removeBeans(String fileName) throws RemoteException {
        if (!isEJBContainer) {
            return;
        }
        ejbserv.removeContainer(fileName);
        // needed here to run also the distributed garbage collector
        runGC();
    }

    /**
     * returns true if beans are already loaded in server.
     * @param fileName name of the ejb-jar or xml file
     */
    public boolean isLoaded(String fileName) throws RemoteException {
        return (ejbserv.getContainer(fileName) != null);
    }

    /**
     * List beans of all JOnAS containers
     */
    public String [] listBeans() throws RemoteException {

        if (!isEJBContainer) {
            return new String[0];
        }

        Container[] lcont = ejbserv.listContainers();
        Vector lbeans = new Vector();
        for (int j = 0; j < lcont.length; j++) {
            String[] lbn = lcont[j].listBeanNames();
            String contname = lcont[j].getName();
            for (int k = 0; k < lbn.length; k++) {
                String s = new String(contname + ": " + lbn[k]);
                lbeans.addElement(s);
            } 
        }
        String[] ret = new String[lbeans.size()];
        lbeans.copyInto(ret);

        return ret;
    }

    /**
     * List JNDI context
     */
    public Vector listContext() throws RemoteException {
        String name = null;
        Vector ret  = new Vector();
        try {
            NamingEnumeration ne;
            ne = NamingManager.getInstance().getInitialContext().list("");
            for (Enumeration e = ne; e.hasMoreElements(); ) {
                name = ((NameClassPair) e.nextElement()).getName();
                ret.addElement(name);
            }
        } catch (Exception ex) {
            throw new RemoteException("Cannot list JNDI context", ex);
        }
        return ret;
    }

    /**
     * List Environment (configuration properties provided by the configuration file).
     */
    public Properties listEnv() {
        return jonasProperties.getConfigFileEnv();
    }

    /**
     * Stop the Server without stopping the JVM
     */
    public void stopServer() throws RemoteException {
        try {
            serverState = STOPPED;
            // keep registry alive a little while (for jonas admin) in case of error at launch
            Thread.sleep(5000);
            sm.stopServices();
        } catch (ServiceException e) {
            throw new RemoteException("Cannot stop services: ", e);
        } catch (InterruptedException e) {
        }
    }
 

    /**
     * Stop  the Server and  stop the JVM
     */
    public void killServer() throws RemoteException {
        stopServer();
        System.exit(0);
    }

    /**
     * To test if the server is ready
     * @return int 0=not ready, 1=ready, 2=stopped
     */
    public int getServerState() throws RemoteException {
        return serverState;
    }

    /**
     * To test if the server is an EJB container
     */
    public boolean isEJBContainer() throws RemoteException {
        return isEJBContainer;
    }

    /**
     * set the default value for transaction timeout
     */
    public void setTransactionTimeout(int timeout) throws RemoteException {
        getTM().setCurrentTimeOut(new Integer(timeout));
    }

    /**
     * run the garbage collector
     */
    public void runGC() throws RemoteException {
        Runtime.getRuntime().gc();
    }
    
    /**
     * sync all entity instances outside transactions
     * @param passivate passivate instances after synchronization.
     */
    public void syncAllEntities(boolean passivate) throws RemoteException {
        if (!isEJBContainer) {
            return;
        }
        ejbserv.syncAllEntities(passivate);
    }


    // ------------------------------------------------------------------
    //  other public methods
    // ------------------------------------------------------------------

    /**
     * server is ready
     */
    public void serverReady(boolean isEJB) {
        serverState = READY;
        isEJBContainer = isEJB;
        if (isEJBContainer) {
            ejbserv = (EJBServiceImpl) sm.getEjbService();
        }
    }

    /**
     * Deploy file (GenIC), needed for Ishmael to work. The file is in the directory
     * specified by the property "jonas.service.deployment.directory"
     * @param type type of the file (EJB, WAR, EAR, RAR)
     * @param bfile
     * @param filename basename of the file to be deployed
     * @throws RemoteException
     * @throws EarServiceException
     * @throws ResourceServiceException
     * @throws JWebContainerServiceException
     * @author Dean Jennings. <da_jennings@bigpond.com>
     */
    public void deployFile(int type, byte[] bfile, String filename) throws RemoteException , EarServiceException, JWebContainerServiceException {
        try {
            String directory = jonasProperties.getValue("jonas.service.deployment.directory");
            File file = new File(directory + filename);
            
            FileOutputStream out = new FileOutputStream(file);
            out.write(bfile);
            out.close();
            if (type == TYPE_EJB) {
                String args[] = new String[1];
                args[0] = directory + filename;
                GenIC.main(args);
                addBeans(directory + filename);
            } else if (type == TYPE_WAR) {
                addWar(file.getAbsolutePath());
            } else if (type == TYPE_EAR) {
                unpackAndCompileEar(file);
                addEar(file.getAbsolutePath());
            } else if (type == TYPE_RAR) {
                addRar(file.getAbsolutePath());
            }
        } catch (Exception e) {
            e.printStackTrace();
            throw new RemoteException(e.toString(), e);
        }
    }

    private void unpackAndCompileEar(File file) throws EarServiceException, IOException, EarDeploymentDescException {
        URL earUrl[] = new URL[1];

        try {
            earUrl[0] = file.toURL();
        } catch (MalformedURLException e) {
            String err = "Invalid ear file name '" + file;
            throw new EarServiceException(err, e);
        }
 
        //Create classLoader
        //parent classloader is the current classloader
        ClassLoader currentLoader = Thread.currentThread().getContextClassLoader();
        URLClassLoader loaderCls = new URLClassLoader(earUrl, currentLoader);

        EarDeploymentDesc desc = EarDeploymentDesc.getInstance(earUrl[0].getFile(), loaderCls);

        Ejb[] ejbTags = desc.getEjbTags();
        /*
         * Example of a jar command:
         *   jar -uf ../output/ejbjars/sb.jar -C /tmp/genic1547.tmp a/b/C1.class
         *                                    -C /tmp/genic1547.tmp a/b/C2.class
         *                                    .....
         */
        String javaHomeBin = System.getProperty("java.home", "");
        if (!("".equals(javaHomeBin))) {
            javaHomeBin = javaHomeBin + File.separator + ".."
                + File.separator + "bin" + File.separator;
        }
        
        Cmd cmd = null;
        cmd = new Cmd(javaHomeBin + "jar");
        cmd.addArgument("-xf");
        cmd.addArgument(file.getAbsolutePath());
        
        for (int i = 0; i < ejbTags.length; i++) {
            cmd.addArgument(ejbTags[i].getValue());
        }

        boolean exitCmd = cmd.run();
        // Analyse the result of the jar command
        if (!exitCmd) {
            throw new EarServiceException("Failed when extracting the the ejb jar "
                                          + "in the given jar file '" + file + "'.");
        }

        for (int i = 0; i < ejbTags.length; i++) {
            String args[] = new String[1];
            args[0] = ejbTags[i].getValue();
            GenIC.main(args);
        }

        cmd = new Cmd(javaHomeBin + "jar");
        cmd.addArgument("-uf");
        cmd.addArgument(file.getAbsolutePath());
        
        for (int i = 0; i < ejbTags.length; i++) {
            cmd.addArgument(ejbTags[i].getValue());
        }
        exitCmd = cmd.run();
        // Analyse the result of the jar command
        if (!exitCmd) {
            throw new EarServiceException("Failed when extracting the the ejb jar "
                                          + "in the given jar file '" + file + "'.");
        }

        for (int i = 0; i < ejbTags.length; i++) {
            File del = new File(ejbTags[i].getValue());
            del.delete();
        }

    }

}
