/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial Developer : Sauthier Guillaume
 * --------------------------------------------------------------------------
 * $Id: WsGenTask.java,v 1.1 2003/10/20 13:52:24 sauthieg Exp $
 * --------------------------------------------------------------------------
*/
package org.objectweb.jonas.ant;

import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.Project;
import org.apache.tools.ant.DirectoryScanner;
import org.apache.tools.ant.taskdefs.MatchingTask;
import org.apache.tools.ant.types.FileSet;
import org.apache.tools.ant.types.Path;
import org.apache.tools.ant.taskdefs.Java;
import org.apache.tools.ant.AntClassLoader;

import java.io.File;

public class WsGenTask extends MatchingTask {

    /** Bootstrap class name. */
    private static final String BOOTSTRAP_CLASS = "org.objectweb.jonas.server.Bootstrap";

    /** WsGen class name (JOnAS 3.3). */
    private static final String WSGEN_CLASS = "org.objectweb.jonas_ws.wsgen.WsGen";

    /** destination directory */
    private File destination   = null;

    /** source directory */
    private File source        = null;

    /** validation of XML files ? */
    private boolean validation = true;

    /** name of javac command */
    private String javac       = null;

    /** list of javac options */
    private String javac_opts  = null;

    /** will WsGen keep already generated files ? */
    private boolean keep_gen   = false;

    /** Will WsGen create configuration files ? */
    private boolean no_config  = false;

    /** verbose mode */
    private boolean verbose    = false;

    /** debug mode */
    private boolean debug      = false;

    /** $JONAS_ROOT */
    private File jonas_root    = null;

    /** $JONAS_BASE */
    private File jonas_base    = null;

    public void setDestdir(File d) {
        destination = d;
    }

    public void setSrcdir(File s) {
        source = s;
    }

    public void setValidation(boolean v) {
        validation = v;
    }

    public void setJavac(String j) {
        javac = j;
    }

    public void setJavacopts(String opts) {
        javac_opts = opts;
    }

    public void setKeepgen(boolean k) {
        keep_gen = k;
    }

    public void setNoconfig(boolean c) {
        no_config = c;
    }

    public void setVerbose(boolean v) {
        verbose = v;
    }

    public void setDebug(boolean b) {
        debug = b;
    }

    public void setJonasroot(File jr) {
        jonas_root = jr;
    }

    public void setJonasbase(File jb) {
        jonas_base = jb;
    }

    /** 
     * Execute the WsGen Ant Task.
     */
    public void execute() throws BuildException {

        // init the wsgen task
        initWsGenTask();

        // execute
        DirectoryScanner ds = getDirectoryScanner(source);
        ds.scan();
        String[] files = ds.getIncludedFiles();

        for (int i = 0; i < files.length; i++) {
            
            Java wsgen = createWsGen(source + File.separator + files[i]);
            
            // calling WsGen task
            log("Calling  WsGen task for '" + source + File.separator + files[i] + "'.", Project.MSG_VERBOSE);
            
            if (wsgen.executeJava() != 0) {
                throw new BuildException("WsGen reported an error.");
            }
        }

    }

    private Java createWsGen(String filename) throws BuildException {

        Java wsgenTask = null;  // WsGen task 
        boolean error = false;  // true if an error occurs during the GenIC call

        wsgenTask = (Java) getProject().createTask("java");
        wsgenTask.setTaskName("wsgen");
        wsgenTask.setFork(true);

        // jonas root 
        wsgenTask.createJvmarg().setValue("-Dinstall.root=" + jonas_root);

        // jonas base
        wsgenTask.createJvmarg().setValue("-Djonas.base=" + jonas_base);

        // java policy file
        String jonasConfigDir = jonas_root + File.separator + "conf";            
        File javaPolicyFile = new File(jonasConfigDir, "java.policy");
        if (javaPolicyFile.exists()) {
            wsgenTask.createJvmarg().setValue("-Djava.security.policy="
                                              + javaPolicyFile.toString());
        }

        // The bootstrap class must launch the GenIC class
        wsgenTask.createArg().setValue(WSGEN_CLASS);

        wsgenTask.createArg().setValue("-d");
        wsgenTask.createArg().setFile(destination);

        // add ow_jonas_bootstrap.jar
        String bootJar = jonas_root + File.separator + "lib" + File.separator + "common"
            + File.separator + "ow_jonas_bootstrap.jar";
        Path classpath = new Path(getProject(), bootJar);
      
        log("Using classpath: " + classpath.toString(), Project.MSG_VERBOSE);
        wsgenTask.setClasspath(classpath);

        if (!checkBootstrapClassName(classpath)) {
            log("Cannot find bootstrap class in classpath.", Project.MSG_ERR);
            throw new BuildException("Bootstrap class not found, please check the classpath.");
        } else {
            wsgenTask.setClassname(BOOTSTRAP_CLASS);
        }

        // keepgenerated
        if (keep_gen) {
            wsgenTask.createArg().setValue("-keepgenerated");
        }

        // noconfig
        if (no_config) {
            wsgenTask.createArg().setValue("-noconfig");
        }

        // novalidation
        if (!validation) {
            wsgenTask.createArg().setValue("-novalidation");
        }

        // javac
        if (javac != null) {
            wsgenTask.createArg().setValue("-javac");
            wsgenTask.createArg().setLine(javac);
        }

        // javacopts
        if (javac_opts != null && !javac_opts.equals("")) {
            wsgenTask.createArg().setValue("-javacopts");
            wsgenTask.createArg().setLine(javac_opts);
        }

        // verbose
        if (verbose) {
            wsgenTask.createArg().setValue("-verbose");
        }

        // debug
        if (debug) {
            wsgenTask.createArg().setValue("-debug");
        }

        // input file to process by GenIC
        wsgenTask.createArg().setValue(filename);

        return wsgenTask;
        
    }

    private void initWsGenTask() throws BuildException {
        // try to define jonas_root if not set
        if (jonas_root == null ) {
            // get ant property
            String jr = getProject().getProperty("jonas.root");
            jonas_root = new File(jr);
            
            if (jr == null) {
                throw new BuildException("attribute jonasroot is necessary.");
            }
        }

        // use jonas_root as jonas_base if not set
        if (jonas_base == null) {
            jonas_base = jonas_root;
        }

        // default destination : project directory
        if (destination == null) {
            destination = getProject().getBaseDir();
        }

        // javac javac_opts
        // by default WsGen already handle them

    }
    /**
     * Check the bootstrap class name to use in the given classpath.
     *
     * @param classpath classpath where the boostrap class must be searched.
     * @return true if the bootstrap is available in the classpath
     */
    private boolean checkBootstrapClassName(Path classpath) {
        log("Looking for bootstrap class in classpath: " + classpath.toString(), Project.MSG_VERBOSE);
        AntClassLoader cl = new AntClassLoader(classpath.getProject(), classpath);
        try {
            cl.loadClass(BOOTSTRAP_CLASS);
            log("Found Bootstrap class '" + BOOTSTRAP_CLASS
                + "' in classpath.", Project.MSG_VERBOSE);
        } catch (ClassNotFoundException cnf1) {
            log("Bootstrap class '" + BOOTSTRAP_CLASS
                + "' not found in classpath.", Project.MSG_VERBOSE);
            return false;
        }
        return true;
    }

}
