/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or 1any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer: Florent BENOIT
 * --------------------------------------------------------------------------
 * $Id: ClientContainer.java,v 1.5 2003/09/22 14:19:16 benoitf Exp $
 * --------------------------------------------------------------------------
 */
package org.objectweb.jonas.client;

import org.objectweb.carol.util.configuration.CarolConfiguration;

import org.objectweb.jonas.common.Log;
import org.objectweb.jonas.ear.lib.EarClassPathManager;
import org.objectweb.jonas.ear.lib.EarClassPathManagerException;
import org.objectweb.jonas.ear.lib.EarFileManager;

import org.objectweb.jonas.ear.lib.JarList;
import org.objectweb.jonas.ear.lib.JarListException;

import org.objectweb.jonas.naming.NamingManager;
import org.objectweb.jonas.security.auth.callback.NoInputCallbackHandler;

import org.objectweb.jonas_client.deployment.api.ClientContainerDeploymentDesc;
import org.objectweb.jonas_client.deployment.api.ClientContainerDeploymentDescException;
import org.objectweb.jonas_client.deployment.api.EjbRefDesc;
import org.objectweb.jonas_client.deployment.api.EnvEntryDesc;
import org.objectweb.jonas_client.deployment.api.ResourceEnvRefDesc;
import org.objectweb.jonas_client.deployment.api.ResourceRefDesc;
import org.objectweb.jonas_client.deployment.lib.ClientDeploymentDescManager;

import org.objectweb.jonas_ear.deployment.api.EarDeploymentDesc;
import org.objectweb.jonas_ear.deployment.api.EarDeploymentDescException;
import org.objectweb.jonas_ear.deployment.xml.*;

import org.objectweb.jonas_ejb.deployment.lib.EjbDeploymentDescManager;

import org.objectweb.jonas_lib.naming.ContainerNaming;

import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

import java.io.File;
import java.io.IOException;

import java.lang.reflect.Method;

import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;

import java.util.ArrayList;
import java.util.StringTokenizer;
import java.util.jar.Attributes;

import java.util.jar.JarFile;
import java.util.jar.Manifest;

import javax.naming.Context;

import javax.naming.LinkRef;
import javax.naming.NamingException;
import javax.naming.Reference;
import javax.naming.StringRefAddr;

import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.login.LoginContext;


/**
 * Defines the class use for the client container This class analyze the ear or
 * the jar client and launch the client
 *
 * @author Florent Benoit
 */
public class ClientContainer {

    /**
     * Name of the carol file
     */
    private static final String CAROL_FILE = "carol.properties";

    /**
     * Reference to the EjbDeploymentDescManager
     */
    private static EjbDeploymentDescManager ejbDDManager = null;

    /**
     * Reference to the ClientDeploymentDescManager
     */
    private static ClientDeploymentDescManager clientDDManager = null;

    /**
     * Main class to use to launch the application client
     */
    private String mainClass = null;

    /**
     * Temporary directory
     */
    private String tmpDir = null;

    /**
     * Jar client to use (if many)
     */
    private String jarClient = null;

    /**
     * Classpath for the application client
     */
    private String classpath = null;

    /**
     * trace.properties file to use instead of the default file.
     */
    private String clientTraceFile = null;

    /**
     * Reference on the NamingManager.
     */
    protected ContainerNaming naming;

    /**
     * carol.properties file to use instead of the default file.
     */
    private String carolFile = null;

    /**
     * Arguments used by the client
     */
    private String[] args = null;

    /**
     * Class loader of the EAR
     */
    private URLClassLoader earClassLoader = null;

    /**
     * Extra Arguments
     */
    private ArrayList appArgs = null;

    /**
     * URLs resolved in the case of the extension mechanism in the Ear case
     */
    private URL[] extensionsURLs = null;

    /**
     * Logger to use
     */
    private Logger logger = null;

    /**
     * Constructor for a Client container
     *
     * @param args the arguments of the instance of the client container
     */
    private ClientContainer(String[] args) {
        this.args = args;

        appArgs = new ArrayList();

        // Init the ejb deployment desc manager.
        ejbDDManager = EjbDeploymentDescManager.getInstance();

        clientDDManager = ClientDeploymentDescManager.getInstance();
    }

    /**
     * Call the Log class to instanciate the client container logger
     */
    private void initLogger() {
        // init the logger
        this.logger = Log.getLogger(Log.JONAS_CLIENT_PREFIX);
    }

    /**
     * Main method of the Client container
     *
     * @param args the arguments of the client container
     */
    public static void main(String[] args) {
        // Retrieve command line parameters
        ClientContainer cc = new ClientContainer(args);

        try {
            cc.start();
        } catch (Exception e) {
            System.err.println("There was the following error : " + e.getMessage());
            e.printStackTrace();
        }
    }

    /**
     * Start the client container
     * @throws Exception if it fails
     */
    private void start() throws Exception {
        analyzeArgs();

        // Use the specified traceclient.properties
        if (clientTraceFile != null) {
            File tClient = new File(clientTraceFile);

            if (!tClient.exists()) {
                throw new ClientContainerException("The file '" + clientTraceFile + "' was not found.");
            }

            if (!tClient.isFile()) {
                throw new ClientContainerException("The file '" + clientTraceFile + "' is not a valid file. Maybe a directory ?");
            }

            System.setProperty("jonas.client.trace.file", clientTraceFile);
        }

        initLogger();

        // Get the filename
        String fileName = null;

        try {
            fileName = (String) appArgs.get(0);
        } catch (IndexOutOfBoundsException ioobe) {
            throw new ClientContainerException("You haven't specify a jar or an ear file as argument. See the Usage.");
        }

        // Test if this is an ear or a jar file.
        if (!(fileName.toLowerCase().endsWith(".jar") || fileName.toLowerCase().endsWith(".ear"))) {
            throw new ClientContainerException("The filename '" + fileName + "' is not a valid filename. It must be an ear file or a jar file");
        }

        // Build file and test if it exists
        File argFile = new File(fileName);

        if (!argFile.exists()) {
            throw new ClientContainerException("The specified file '" + fileName + "' doesn't exists.");
        }

        File clientJarFile = null;

        // Unpack and analyze EAR file if it is an ear
        if (fileName.toLowerCase().endsWith(".ear")) {
            clientJarFile = extractAndAnalyzeEar(argFile);
        } else {
            // Client jar is the given file
            clientJarFile = argFile;
        }

        // Carol initialisation (property)
        System.setProperty("javax.rmi.CORBA.PortableRemoteObjectClass", "org.objectweb.carol.rmi.multi.MultiPRODelegate");
        System.setProperty("java.naming.factory.initial", "org.objectweb.carol.jndi.spi.MultiOrbInitialContextFactory");
        System.setProperty("org.omg.PortableInterceptor.ORBInitializerClass.org.objectweb.jonas_tm.ots.OTSORBInitializer", "");
        System.setProperty("org.omg.PortableInterceptor.ORBInitializerClass.org.objectweb.jonas.security.interceptors.iiop.SecurityInitializer", "");

        // Build a classloader for the initialisation of carol
        // Priority 
        //    1/ file as argument
        //    2/ file in the jar client
        //    3/ file in client.jar
        URL urlCarolFile = null;

        if (carolFile != null) {
            File fCarol = new File(carolFile);

            if (!fCarol.exists()) {
                throw new ClientContainerException("The file '" + carolFile + "' was not found.");
            }

            if (!fCarol.isFile()) {
                throw new ClientContainerException("The file '" + carolFile + "' is not a valid file. Maybe a directory ?");
            }

            if (!fCarol.getName().equals(CAROL_FILE)) {
                throw new ClientContainerException("The file '" + carolFile + "' must be named '" + CAROL_FILE + "'.");
            }

            try {
                urlCarolFile = fCarol.getParentFile().toURL();
            } catch (MalformedURLException mue) {
                throw new ClientContainerException("Error when building an URL for the file '" + fCarol + "'.");
            }
        }

        ArrayList sysCP = getSystemClassPath();
        URL[] urlCarol = null;
        int urlInd = 0;

        if (urlCarolFile != null) {
            // classpath + url of client + url of argument file
            urlCarol = new URL[sysCP.size() + 2];
            urlCarol[urlInd] = urlCarolFile;
            urlInd++;
        } else {
            // classpath + url of client
            urlCarol = new URL[sysCP.size() + 1];
        }

        try {
            urlCarol[urlInd] = clientJarFile.toURL();
            urlInd++;
            int classPathSize = sysCP.size();
            for (int j = 0; j < classPathSize; j++) {
                urlCarol[urlInd] = new File(((String) sysCP.get(j))).toURL();
                urlInd++;
            }
        } catch (MalformedURLException mue) {
            throw new ClientContainerException("Error when building an URL for the classpath.");
        }

        ClassLoader oldCL = Thread.currentThread().getContextClassLoader();
        URLClassLoader newCL = new URLClassLoader(urlCarol, null);
        Thread.currentThread().setContextClassLoader(newCL);

        // Init carol with the properties file
        CarolConfiguration.init(Log.getLoggerFactory());
        Thread.currentThread().setContextClassLoader(oldCL);

        // Carol configuration is done
        // Extract Main-Class to use in the jar from the manifest
        Manifest manifest = new JarFile(clientJarFile).getManifest();

        if (manifest == null) {
            throw new ClientContainerException("No manifest was found inside the file" + clientJarFile);
        }

        // Extract attributes
        Attributes attributes = manifest.getMainAttributes();

        if (attributes == null) {
            throw new ClientContainerException("No attributes were found in the manifest of the file '" + clientJarFile + "'.");
        }

        mainClass = attributes.getValue(Attributes.Name.MAIN_CLASS);

        if (mainClass == null) {
            throw new ClientContainerException("No main class was found inside the Manifest  of the file '"
                                               + clientJarFile + "'. This attribute is required to launch the application client.");
        }

        if (logger.isLoggable(BasicLevel.DEBUG)) {
            logger.log(BasicLevel.DEBUG, "Using Main-Class :" + mainClass);
        }

        // Convert file to URL
        URL clientJarURL = null;

        try {
            clientJarURL = clientJarFile.toURL();
        } catch (MalformedURLException mue) {
            throw new ClientContainerException("Error when building an URL with the file '" + clientJarFile + "'.");
        }

        // Build the urls for the classloader
        URL[] urlsClient = null;

        // URLs for the classloader
        if (extensionsURLs != null) {
            // There were URLs with the extension mechanism in the EAR
            urlsClient = new URL[extensionsURLs.length + 1];

            for (int i = 0; i < extensionsURLs.length; i++) {
                urlsClient[i] = extensionsURLs[i];

                if (logger.isLoggable(BasicLevel.DEBUG)) {
                    logger.log(BasicLevel.DEBUG, "Adding " + extensionsURLs[i] + " to the urls of the client");
                }
            }

            urlsClient[extensionsURLs.length] = clientJarURL;
        } else {
            if (logger.isLoggable(BasicLevel.DEBUG)) {
                logger.log(BasicLevel.DEBUG, "Only one url for urls of client");
            }

            // No extension or jar case.
            urlsClient = new URL[1];
            urlsClient[0] = clientJarURL;
        }

        // Build classloader
        URLClassLoader clientClassloader = new URLClassLoader(urlsClient, Thread.currentThread().getContextClassLoader());
        Thread.currentThread().setContextClassLoader(clientClassloader);

        // Get the deployment descriptor from file
        ClientContainerDeploymentDesc clientDD = null;

        try {
            clientDD = clientDDManager.getDeploymentDesc(clientJarURL, clientClassloader, earClassLoader);
        } catch (ClientContainerDeploymentDescException e) {
            String err = "Cannot read the deployment descriptors '" + clientJarURL + "'";
            error(err + ": " + e);
            throw new ClientContainerException(err + e.getMessage());
        }

        // Populate the java:comp/env (ENC) environment.
        try {
            setClientEnvironment(clientDD);
        } catch (Exception e) {
            //populating environment failed.
            String err = "Error when populating ";
            error(err + e.getMessage());
            throw new ClientContainerException(err, e);
        }

        // JAAS
        String jaasFile = clientDD.getJaasFile();
        String jaasEntry = clientDD.getJaasEntry();
        String username = clientDD.getUsername();
        String password = clientDD.getPassword();

        if (logger.isLoggable(BasicLevel.DEBUG)) {
            logger.log(BasicLevel.DEBUG, "Using jaas file = " + jaasFile);
        }

        String jaasConfigFile = null;

        if (jaasFile != null) {
            // Use JAAS
            jaasConfigFile = "jar:" + clientJarURL.toExternalForm() + "!/" + jaasFile;
            System.setProperty("java.security.auth.login.config", jaasConfigFile);
        }

        CallbackHandler ch = null;

        if ((username != null) && (password != null)) {
            ch = new NoInputCallbackHandler(username, password);
            info("Using the login/password specified in the jonas-client.xml file with a specific CallbackHandler");
        } else {
            // Is there a callbackHandler specified ?
            String ddCallbackHandler = clientDD.getCallbackHandler();

            if (ddCallbackHandler != null) {
                if (logger.isLoggable(BasicLevel.DEBUG)) {
                    logger.log(BasicLevel.DEBUG, "Using '" + ddCallbackHandler + "' class as CallbackHandler.");
                }

                Class clazz = null;

                //try to invoke this class
                try {
                    clazz = clientClassloader.loadClass(ddCallbackHandler);
                } catch (Exception e) {
                    throw new ClientContainerException("There was an error while trying to instantiate the class '"
                                                       + ddCallbackHandler + "' which is specified in the application.xml as CallbackHandler class", e);
                }

                try {
                    ch = (CallbackHandler) clazz.newInstance();
                } catch (Exception e) {
                    throw new ClientContainerException("Error while triyng to cast the class '"
                                                       + ddCallbackHandler + "' to CallbackHandler interface, maybe the specified class doesn't implement this interface.", e);
                }
            }
        }

        // Use JAAS
        if (ch != null) {
            if (jaasFile == null) {
                throw new ClientContainerException("You have defined that you want use a CallbackHandler but you haven't specify the jaas file to use for the JAAS configuration.");
            }

            if (jaasEntry == null) {
                throw new ClientContainerException("You have defined that you want use a CallbackHandler but you haven't specify the jaas entry to use from the jaas config file.");
            }

            info("Using JAAS loginContext '" + jaasEntry + "' from the file '" + jaasConfigFile + "'.");

            try {
                LoginContext lc = new LoginContext(jaasEntry, ch);
                lc.login();
            } catch (Exception e) {
                String err = "Can not use the JAAS authentication";
                error(err);
                throw new ClientContainerException(err, e);
            }
        }

        // Invoke client
        // Launch the "class_to_run" by using our classloader.
        Class clazz = clientClassloader.loadClass(mainClass);
        Class[] argList = new Class[] {args.getClass()};
        Method meth = clazz.getMethod("main", argList);

        // Remove name of the file from arguments
        String[] newArgs = new String[appArgs.size() - 1];
        String txtArgs = "";

        for (int i = 0; i < newArgs.length; i++) {
            newArgs[i] = (String) appArgs.get(i + 1);
            txtArgs += (newArgs[i] + " ");
        }

        if (logger.isLoggable(BasicLevel.DEBUG)) {
            logger.log(BasicLevel.DEBUG, "Starting the application client with the arguments '" + txtArgs + "'.");
        }

        info("Starting client...");

        Object appli = meth.invoke(null, new Object[] {newArgs});

        if (logger.isLoggable(BasicLevel.DEBUG)) {
            logger.log(BasicLevel.DEBUG, "End of main method");
        }

    }

    /**
     * Set the environment of this client
     * @param clientDD deployment descriptor used for build the environment
     * @throws NamingException if the creation of the environment fails
     */
    private void setClientEnvironment(ClientContainerDeploymentDesc clientDD) throws NamingException {
        if (logger.isLoggable(BasicLevel.DEBUG)) {
            logger.log(BasicLevel.DEBUG, "");
        }

        //Init the naming manager
        try {
            naming = NamingManager.getInstance();
        } catch (NamingException e) {
            throw new ClientContainerException("Error when getting the reference to the Naming manager",
                e);
        }

        Context javaCtx = naming.createEnvironmentContext("ClientContainer");
        naming.setComponentContext(javaCtx);

        Context envCtx = javaCtx.createSubcontext("comp/env");

        // Bean Environment
        EnvEntryDesc[] envt = clientDD.getEnvEntryTags();

        for (int i = 0; i < envt.length; i++) {
            // get information in descriptor
            String name = envt[i].getName();
            Object obj = envt[i].getValue();

            // register object in JNDI
            if (logger.isLoggable(BasicLevel.DEBUG)) {
                logger.log(BasicLevel.DEBUG,
                    "Binding object " + name + " -> " + obj);
            }

            envCtx.rebind(name, obj);
        }

        // Resource References
        ResourceRefDesc[] resref = clientDD.getResourceRefTags();

        for (int i = 0; i < resref.length; i++) {
            // get information in descriptor
            String name = resref[i].getName();
            String resname = resref[i].getJndiName();
            String type = resref[i].getType().getName();

            // build the LinkRef that will be registered:
            // FactoryClassName = null, size = 1, refAddr = resname.
            // register object in JNDI
            if (logger.isLoggable(BasicLevel.DEBUG)) {
                logger.log(BasicLevel.DEBUG,
                    "Linking resource " + name + " -> " + resname);
            }

            if (type.equalsIgnoreCase("java.net.URL")) {
                // Specify the factory to use with the right URL
                Reference ref = new Reference("java.net.URL",
                        "org.objectweb.jonas_lib.naming.factory.URLFactory",
                        null);
                StringRefAddr refAddr = new StringRefAddr("url", resname);
                ref.add(refAddr);
                envCtx.rebind(name, ref);
            } else {
                LinkRef lref = new LinkRef(resname);
                envCtx.rebind(name, lref);
            }
        }

        // Resource Environment References
        ResourceEnvRefDesc[] resEnvref = clientDD.getResourceEnvRefTags();

        for (int i = 0; i < resEnvref.length; i++) {
            // get information in descriptor
            String name = resEnvref[i].getName();
            String resname = resEnvref[i].getJndiName();
            LinkRef lref = new LinkRef(resname);

            if (logger.isLoggable(BasicLevel.DEBUG)) {
                logger.log(BasicLevel.DEBUG,
                    "Linking resource environment " + name + " -> " + resname);
            }

            envCtx.rebind(name, lref);
        }

        // EJB References
        EjbRefDesc[] ejbref = clientDD.getEjbRefTags();

        for (int i = 0; i < ejbref.length; i++) {
            // get information in descriptor
            String name = ejbref[i].getName();
            String ejbname = null;
            String ejblink = ejbref[i].getEjbLink();
            ejbname = ejbref[i].getJndiName();

            LinkRef lref = new LinkRef(ejbname);

            if (logger.isLoggable(BasicLevel.DEBUG)) {
                logger.log(BasicLevel.DEBUG, "Linking ejb " + name + " -> " + ejbname);
            }

            envCtx.rebind(name, lref);
        }
    }

    /**
     * Analyze arguments and extract parameters for the client container
     * @throws Exception if there is an error when analyzing arguments
     */
    private void analyzeArgs() throws Exception {
        for (int argn = 0; argn < args.length; argn++) {
            String arg = args[argn];

            try {
                if (arg.equals("-tmpDir")) {
                    tmpDir = args[++argn];

                    continue;
                }

                if (arg.equals("-jarClient")) {
                    jarClient = args[++argn];

                    continue;
                }

                if (arg.equals("-traceFile")) {
                    clientTraceFile = args[++argn];

                    continue;
                }

                if (arg.equals("-carolFile")) {
                    carolFile = args[++argn];

                    continue;
                }

                if (arg.equals("-cp")) {
                    classpath = args[++argn];

                    continue;
                }

                if (arg.equals("--help") || arg.equals("-help") || arg.equals("-h") || arg.equals("-?")) {
                    usage();
                    System.exit(1);
                }

                // Add argument to the application arguments
                appArgs.add(arg);
            } catch (ArrayIndexOutOfBoundsException aioobe) {
                // The next argument is not in the array
                throw new ClientContainerException("A required parameter was missing after the argument" + arg);
            }
        }
    }

    /**
     * Print the usage of this client
     */
    private void usage() {
        System.out.println("Usage of this client :");
        System.out.println("-------------------------------------------------------------------");
        System.out.println(" -jarClient   : Specify the client jar to use of the ear if many.");
        System.out.println(" -traceFile   : Specify the configuration file to use for the traces\n"
                           + "                of this client instead of the default file\n"
                           + "                (traceclient.properties) present in client.jar.");
        System.out.println(" -carolFile   : Specify the carol.properties file to use instead of \n" + "                the default carol.properties file of the client.jar");
        System.out.println(" -tmpDir      : Specify the temp directory where unpack the ear.");
        System.out.println(" -cp          : Specify the classpath to use for the jar client.");
        System.out.println("-------------------------------------------------------------------");
        System.out.println("  --help  : Display this help.");
        System.out.println("  -help   : Display this help.");
        System.out.println("  -h      : Display this help.");
        System.out.println("  -?      : Display this help.");
        System.out.println("-------------------------------------------------------------------");
    }

    /**
     * Extract the client of an ear and analyze ear too
     * @param earFile ear to be analyzed
     * @return the file of the client which was extracted
     * @throws Exception if the analyze and/or extract fails
     */
    private File extractAndAnalyzeEar(File earFile) throws Exception {
        if (logger.isLoggable(BasicLevel.DEBUG)) {
            logger.log(BasicLevel.DEBUG, "");
        }

        URL earUrl = null;

        try {
            earUrl = earFile.toURL();
        } catch (MalformedURLException mue) {
            throw new ClientContainerException("Can not build an url with the filename '" + earFile + "'.");
        }

        // Create classLoader
        URL[] arrURL = new URL[1];
        arrURL[0] = earUrl;

        // parent classloader is the current classloader
        ClassLoader currentLoader = Thread.currentThread().getContextClassLoader();
        URLClassLoader loaderCls = new URLClassLoader(arrURL, currentLoader);

        EarDeploymentDesc earDD = null;

        if (logger.isLoggable(BasicLevel.DEBUG)) {
            logger.log(BasicLevel.DEBUG, "Getting the deployment descriptor of the file" + earFile.getPath());
        }

        try {
            earDD = EarDeploymentDesc.getInstance(earFile.getPath(), loaderCls);
        } catch (EarDeploymentDescException e) {
            String err = "Error in the Deployment descriptor '" + earFile + "': " + e;

            //error( err);
            throw new ClientContainerException(err, e);
        }

        // Get the tags from the Deployment descriptor
        Ejb[] ejbTags = earDD.getEjbTags();
        Java[] clientTags = earDD.getClientTags();
        AltDd[] altDDEjbs = earDD.getAltDDEjbs();
        AltDd[] altDDClients = earDD.getAltDDClients();

        // Check if all modules are inside the EAR file
        // no relatives mode like ../../file1.jar
        File fEar = null;
        File tmpFile = null;

        try {
            fEar = earFile.getCanonicalFile();
        } catch (IOException ioe) {
            String err = "Error : Can not get canonical file for the file '" + earFile + "'.";
            throw new ClientContainerException(err);
        }

        try {
            for (int i = 0; i < ejbTags.length; i++) {
                tmpFile = new File(fEar, ejbTags[i].getValue());
                tmpFile = tmpFile.getCanonicalFile();
                if (!tmpFile.getPath().startsWith(fEar.getPath())) {
                    String err = "Error : The ejb-jar file " + ejbTags[i].getValue() + " is not inside the ear file " + fEar;
                    throw new ClientContainerException(err);
                }
            }

            for (int i = 0; i < clientTags.length; i++) {
                tmpFile = new File(fEar, clientTags[i].getValue());
                tmpFile = tmpFile.getCanonicalFile();

                if (!tmpFile.getPath().startsWith(fEar.getPath())) {
                    String err = "Error : The war file " + clientTags[i].getValue() + " is not inside the ear file " + fEar;
                    throw new ClientContainerException(err);
                }
            }
        } catch (IOException ioe) {
            String err = "Error while trying to get the canonical file of " + tmpFile;
            throw new ClientContainerException(err, ioe);
        }

        //Changing array into JarList
        JarList ejbsList = new JarList(ejbTags);
        JarList clientsList = new JarList(clientTags);

        // Temporary directory
        String tempDir = null;

        if (tmpDir != null) {
            // use specific directory
            tempDir = tmpDir;
            info("Use your specified temp directory '" + tempDir + "'.");
        } else {
            // use default
            tempDir = System.getProperty("java.io.tmpdir");
        }

        if (logger.isLoggable(BasicLevel.DEBUG)) {
            logger.log(BasicLevel.DEBUG, "Using temp directory '" + tempDir + "'.");
        }

        // Can we write to ?
        File tmpFileDir = new File(tempDir);

        if (!tmpFileDir.exists() || !tmpFileDir.isDirectory()) {
            throw new ClientContainerException("The temp directory '" + tempDir + "' doesn't exist or is not a directory.");
        }

        if (!tmpFileDir.canWrite()) {
            throw new ClientContainerException("Can not write to the temporary directory '" + tempDir + "'.");
        }

        // Unpack the ear file and get the unpacked dir
        URL dirUnpackURL = null;

        try {
            dirUnpackURL = EarFileManager.unpackEar(earUrl, tmpFileDir.toURL(), false);
        } catch (Exception e) {
            String err = "Error while unpacking the file '" + earUrl + "'";
            throw new ClientContainerException(err, e);
        }

        // Ear is unpacked, now analyze manifest Class-path:
        EarClassPathManager earCPManager = null;

        try {
            earCPManager = new EarClassPathManager(clientsList, dirUnpackURL);
        } catch (EarClassPathManagerException e) {
            String err = "Error while creating the Ear class path manager of the ear : '" + earUrl + "'";
            error(err + " : " + e.getMessage());
            throw new ClientContainerException(err, e);
        }

        URL[] classpathURLs = null;

        //Get the urls of the ear class path manager
        try {
            classpathURLs = earCPManager.getResolvedClassPath();
        } catch (EarClassPathManagerException e) {
            String err = "Error while trying to resolve the classpath of the ejbjars and wars of the ear : '" + earUrl + "'";
            error(err + " : " + e.getMessage());
            throw new ClientContainerException(err, e);
        }

        if (logger.isLoggable(BasicLevel.DEBUG)) {
            logger.log(BasicLevel.DEBUG, "EAR : ejbs = " + ejbsList);
            logger.log(BasicLevel.DEBUG, "EAR : clientUrls = " + clientsList);
        }

        currentLoader = Thread.currentThread().getContextClassLoader();
        earClassLoader = new URLClassLoader(new URL[0], currentLoader);

        //Extract the urls of the jarList
        URL[] jarUrls = null;
        URL[] clientUrls = null;

        try {
            jarUrls = ejbsList.getURLs(dirUnpackURL.toExternalForm());
            clientUrls = clientsList.getURLs(dirUnpackURL.toExternalForm());
        } catch (JarListException e) {
            String err = "Error while geting the Urls from jarlist of the ear : '" + earUrl + "'";

            //error( err + " : " + e.getMessage());
            throw new ClientContainerException(err, e);
        }

        //Fill Alt-DD for Ejbs and Clients
        String altdd = null;
        File fAltDD = null;

        //Transorm the array altDDWebs into an array with the absolute URL to the file
        URL[] clientsAltDDs = new URL[altDDClients.length];

        for (int i = 0; i < altDDClients.length; i++) {
            if (altDDClients[i] != null) {
                altdd = altDDClients[i].getValue();

                if (altdd != null) {
                    try {
                        fAltDD = new File(new URL(dirUnpackURL.toExternalForm() + File.separator + altdd).getFile());
                        clientsAltDDs[i] = fAltDD.getCanonicalFile().toURL();
                    } catch (MalformedURLException e) {
                        String err = "Can't build URL for alt-dd '" + altdd;
                        error(err + "': " + e.getMessage());
                        throw new ClientContainerException(err, e);
                    } catch (IOException ioe) {
                        String err = "Can't get canonicalFile() for the file '" + fAltDD;
                        error(err + "': " + ioe.getMessage());
                        throw new ClientContainerException(err, ioe);
                    }
                }
            }
        }

        URL[] ejbsAltDDs = new URL[altDDEjbs.length];

        for (int i = 0; i < altDDEjbs.length; i++) {
            if (altDDEjbs[i] != null) {
                altdd = altDDEjbs[i].getValue();

                if (altdd != null) {
                    try {
                        fAltDD = new File(new URL(dirUnpackURL.toExternalForm() + File.separator + altdd).getFile());
                        ejbsAltDDs[i] = fAltDD.getCanonicalFile().toURL();
                    } catch (MalformedURLException e) {
                        String err = "Can't build URL for alt-dd '" + altdd;
                        error(err + "': " + e.getMessage());
                        throw new ClientContainerException(err, e);
                    } catch (IOException ioe) {
                        String err = "Can't get canonicalFile() for the file '" + fAltDD;
                        error(err + "': " + ioe.getMessage());
                        throw new ClientContainerException(err, ioe);
                    }
                }
            }
        }

        ejbDDManager.setAvailableEjbJarsAndAltDDs(earClassLoader, jarUrls, ejbsAltDDs);
        clientDDManager.setAltDD(earClassLoader, clientUrls, clientsAltDDs);

        // Construct the ejb classloader for all the ejb-jars of the same
        // ear application. Because there is one ejb classloader for all
        // the ejb-jars of the same ear application.
        extensionsURLs = new URL[jarUrls.length + classpathURLs.length];
        System.arraycopy(jarUrls, 0, extensionsURLs, 0, jarUrls.length);
        System.arraycopy(classpathURLs, 0, extensionsURLs, jarUrls.length, classpathURLs.length);

        if (logger.isLoggable(BasicLevel.DEBUG)) {
            logger.log(BasicLevel.DEBUG, "Extensions urls :");

            for (int ii = 0; ii < extensionsURLs.length; ii++) {
                logger.log(BasicLevel.DEBUG, "url[" + ii + "] = " + extensionsURLs[ii]);
            }
        }

        // Choose the jar client
        if (clientUrls.length == 0) {
            throw new ClientContainerException("No java client was found in the application.xml file of the Ear '" + earUrl + "'.");
        }

        File fClient = null;

        // User has specify an application client to execute
        // Search it
        if (jarClient != null) {
            int f = 0;
            File ff = null;
            boolean found = false;

            while (f < clientUrls.length && !found) {
                ff = new File(clientUrls[f].getFile());

                if (ff.getPath().endsWith(jarClient)) {
                    found = true;
                    fClient = ff;

                    if (logger.isLoggable(BasicLevel.DEBUG)) {
                        logger.log(BasicLevel.DEBUG, "Found a matching client with the name " + ff);
                    }
                }

                f++;
            }

            if (!found) {
                throw new ClientContainerException("No client with the name '" + jarClient + "' was found in this Ear file");
            }
        } else {
            // Take first
            fClient = new File(clientUrls[0].getFile());

            // Warning if many
            if (clientUrls.length > 1) {
                warn("There are " + clientUrls.length + " clients in this ear, choosing the first one : " + fClient.getName());
            }
        }

        info("Use the application client '" + fClient + "' of the Ear '" + earUrl + "'.");

        return fClient;
    }

    /**
     * @return the system class path
     */
    private ArrayList getSystemClassPath() {
        String cpValue = System.getProperty("java.class.path");

        if (logger.isLoggable(BasicLevel.DEBUG)) {
            logger.log(BasicLevel.DEBUG, "java.class.path = " + cpValue);
        }

        String sep = File.pathSeparator;
        ArrayList h = new ArrayList();
        StringTokenizer st = new StringTokenizer(cpValue, sep);

        while (st.hasMoreTokens()) {
            h.add(st.nextToken());
        }

        return h;
    }

    /**
     * Display an info message with the logger
     * @param s the message
     */
    private void info(String s) {
        logger.log(BasicLevel.INFO, s);
    }

    /**
     * Display an error message with the logger
     * @param s the message
     */
    private void error(String s) {
        logger.log(BasicLevel.ERROR, s);
    }

    /**
     * Display a warn message with the logger
     * @param s the message
     */
    private void warn(String s) {
        logger.log(BasicLevel.WARN, s);
    }
}
