/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Florent BENOIT & Ludovic BERT
 * --------------------------------------------------------------------------
 * $Id: EarServiceImpl.java,v 1.22 2003/10/13 09:48:07 benoitf Exp $
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas.ear;

import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Date;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.StringTokenizer;
import java.util.Vector;

import javax.management.InstanceAlreadyExistsException;
import javax.management.MBeanRegistrationException;
import javax.management.MBeanServer;
import javax.management.NotCompliantMBeanException;
import javax.naming.Context;
import javax.naming.NamingException;

import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

import org.objectweb.jonas.common.JModule;
import org.objectweb.jonas.common.JProp;
import org.objectweb.jonas.common.Log;
import org.objectweb.jonas.container.EJBService;
import org.objectweb.jonas.ear.lib.EarClassPathManager;
import org.objectweb.jonas.ear.lib.EarClassPathManagerException;
import org.objectweb.jonas.ear.lib.EarCleaner;
import org.objectweb.jonas.ear.lib.EarCleanerException;
import org.objectweb.jonas.ear.lib.EarDeployerLog;
import org.objectweb.jonas.ear.lib.EarDeployerLogException;
import org.objectweb.jonas.ear.lib.EarFileManager;
import org.objectweb.jonas.ear.lib.EarFileManagerException;
import org.objectweb.jonas.ear.lib.JarList;
import org.objectweb.jonas.ear.lib.JarListException;
import org.objectweb.jonas.jmx.JmxService;
import org.objectweb.jonas.jmx.JonasObjectName;
import org.objectweb.jonas.naming.CompNamingContext;
import org.objectweb.jonas.resource.ResourceService;
import org.objectweb.jonas.resource.ResourceServiceException;
import org.objectweb.jonas.service.AbsServiceImpl;
import org.objectweb.jonas.service.ServiceException;
import org.objectweb.jonas.service.ServiceManager;
import org.objectweb.jonas.web.JWebContainerService;
import org.objectweb.jonas.web.JWebContainerServiceException;
import org.objectweb.jonas.ws.WSServiceException;
import org.objectweb.jonas.ws.WebServicesService;

import org.objectweb.jonas_ejb.deployment.lib.EjbDeploymentDescManager;

import org.objectweb.jonas_web.deployment.lib.WebDeploymentDescManager;

import org.objectweb.jonas_ear.deployment.api.EarDeploymentDesc;
import org.objectweb.jonas_ear.deployment.api.EarDeploymentDescException;
import org.objectweb.jonas_ear.deployment.xml.AltDd;
import org.objectweb.jonas_ear.deployment.xml.Connector;
import org.objectweb.jonas_ear.deployment.xml.ContextRoot;
import org.objectweb.jonas_ear.deployment.xml.Ejb;
import org.objectweb.jonas_ear.deployment.xml.SecurityRole;
import org.objectweb.jonas_ear.deployment.xml.Web;


/**
 * JOnAS EAR Service Implementation class.
 * This class provides an implementation of the ear service.
 *
 * @author Florent Benoit
 * @author Ludovic Bert
 *
 * Contributor(s):
 * Adriana Danes: highlight configuration properties
 * Eric Hardesty: added ability to include rar files in an ear
 */
public class EarServiceImpl
    extends AbsServiceImpl
    implements EarService, EarServiceImplMBean {

    /**
     * The name of the JONAS_BASE directory.
     */
    protected static final String JONAS_BASE = JProp.getJonasBase();

    /**
     * The name of the apps directory.
     */
    protected static final String APPS_DIR = JONAS_BASE + File.separator 
        + "apps";

    /**
     * Property for ear to deploy at runtime
     */
    protected static final String DESCRIPTORS = "jonas.service.ear.descriptors";

    /**
     * Directories property for autoload
     */
    protected static final String AUTOLOADDIR = "jonas.service.ear.autoloaddir";

    /**
     * Property for parsing with validation or not
     */
    protected static final String PARSINGWITHVALIDATION = "jonas.service.ear.parsingwithvalidation";

    /**
     * Property for the class of the EAR service
     */
    protected static final String CLASS = "jonas.service.ear.class";

    /**
     * Logger for this service.
     */
    private static Logger logger = null;

    /**
     * Reference to the EjbDeploymentDescManager
     */
    private static EjbDeploymentDescManager ejbDDManager = null;

    /**
     * Reference to the WebbDeploymentDescManager
     */
    private static WebDeploymentDescManager webDDManager = null;

    /**
     * the name of the current jonas server
     */
    private static String nameOfServer = null;

    /**
     * Reference to the Ear cleaner
     */
    private static EarCleaner earCleaner = null;

    /**
     * Reference to a MBean server.
     */
    private MBeanServer mbeanServer = null;

    /**
     * List of the ear names to load when starting the EarService
     */
    private Vector earNames = new Vector();

    /**
     * Reference to the Ejb service.
     */
    private EJBService ejbService = null;

    /**
     * Reference to the WebContainer service.
     */
    private JWebContainerService webContainerService = null;

    /**
     * Reference to the WebServices service.
     */
    private WebServicesService wsService = null;

    /**
     * Reference to the Resource service.
     */
    private ResourceService resourceService = null;

    /**
     * List of loaded ear
     */
    private Hashtable ears = null;

    /**
     * List of autoloaded directories.
     */
    private ArrayList autoloadDirectories = new ArrayList();

    /**
     * Reference on the EarDeployerLog which is the class that manage the
     * accesses to the log file (to remove the ear).
     */
    private EarDeployerLog earDeployerLog = null;



    /**
     * Init the EAR service.
     * @param ctx the configuration of the Ear service.
     * @throws ServiceException if the initialization failed.
     */
    protected void doInit(Context ctx) throws ServiceException {

        //we init the logger
        logger = Log.getLogger(Log.JONAS_EAR_PREFIX);

        // Get the JMX Server via JMX Service
        try {
            mbeanServer =
                ((JmxService) ServiceManager.getJmxService()).getJmxServer();
        } catch (ServiceException e) {
            // the JMX service may not be started
            mbeanServer = null;
        }

        // Get the EJB service
        try {
            ejbService = (EJBService) ServiceManager.getEjbService();
        } catch (ServiceException e) {
            //not started
            ejbService = null;
        }

        // Get the Web Container service
        try {
            webContainerService =
                (JWebContainerService) ServiceManager.getWebContainerService();
        } catch (ServiceException e) {
            //not started
            webContainerService = null;
        }

        // Get the WebServices service
        try {
            wsService =
                (WebServicesService) ServiceManager.getWebServicesService();
        } catch (ServiceException e) {
            //not started
            wsService = null;
        }

        // Get the Resource service
        try {
            resourceService = (ResourceService) ServiceManager.getResourceService();
        } catch (ServiceException e) {
            //not started
            resourceService = null;
        }

        // Set the XML parsing mode to no validation
        String parsingMode = "false";
        try {
            parsingMode = (String) ctx.lookup(PARSINGWITHVALIDATION);
        } catch (NamingException e) {
            ;
            /* No problem if there is no value for 'parsingwithvalidation'
               (false by default) */
        }
        EarDeploymentDesc.setParsingWithValidation("true".equalsIgnoreCase(parsingMode));
        if ("false".equalsIgnoreCase(parsingMode)) {
            logger.log(BasicLevel.DEBUG, "Ear XML parsing without validation");
        } else {
            logger.log(BasicLevel.DEBUG, "Ear XML parsing with validation");
        }

        // Init the ear names to be loaded when starting
        String descsValue = null;
        try {
            descsValue = (String) ctx.lookup(DESCRIPTORS);
        } catch (NamingException e) {
            ;
            // No problem if there is no value for 'descriptors'
            //(no ear to load)
        }
        if (descsValue != null) {
            StringTokenizer st = new StringTokenizer(descsValue, ",");
            while (st.hasMoreTokens()) {
                String fileName = st.nextToken().trim();
                earNames.add(fileName);
            }
        }

        // Add the ears of the jonas.service.ear.autoloaddir property
        String dirValue = null;
        ArrayList autoDirs = new ArrayList();
        try {
            dirValue = (String) ctx.lookup(AUTOLOADDIR);
        } catch (NamingException e) {
            ;
            // No problem if there is no value for 'autoloaddir'
            //(no ear to load)
        }
        if (dirValue != null) {
            StringTokenizer st = new StringTokenizer(dirValue, ",");
            while (st.hasMoreTokens()) {
                String dirName = st.nextToken().trim();
                addEars(dirName);
                autoDirs.add(dirName);
            }
        }
        // Build autoload directories
        File oFile;
        Iterator it = autoDirs.iterator();
        while (it.hasNext()) {
            String dirName = (String) it.next();
            try {
                oFile = new File(APPS_DIR, dirName);
                if (!oFile.exists()) {
                    oFile = new File(dirName);
                }
                if (oFile.exists()) {
                    autoloadDirectories.add(oFile.getCanonicalPath());
                }
            } catch (Exception e) {
                String err = "Error when trying to verify Application EAR autoload directory : " + dirName;
                logger.log(BasicLevel.ERROR, err + " " + e.getMessage());
            }
        }

        //create hashtable of the ears ( Url of an ear ---> Ear structure)
        ears = new Hashtable();

        // Init the ejb deployment desc manager.
        ejbDDManager = EjbDeploymentDescManager.getInstance();

        // Init the web deployment desc manager.
        webDDManager = WebDeploymentDescManager.getInstance();

        // Init the rar deployment desc manager.
        //rarDDManager = RarDeploymentDescManager.getInstance();


        JProp jp = null;
        try {
            jp = JProp.getInstance();
        } catch (Exception e) {
            String err = "Error when trying to get jonas name ";
            logger.log(BasicLevel.ERROR, err + " " + e.getMessage());
            throw new ServiceException(err, e);
        }
        EarServiceImpl.nameOfServer = jp.getValue("jonas.name", "jonas");

    }
    /**
     * Stop the EAR service.
     * @throws ServiceException if the stop failed.
     */
    protected void doStop() throws ServiceException {

        //undeploy the loaded ears files
        URL earFileName = null;
        //For each ear file, we delegate the operation to the
        //unDeployEar method
        for (Enumeration earEntries = ears.keys(); earEntries.hasMoreElements();) {
            earFileName = (URL) earEntries.nextElement();
            //Try to undeploy the ear
            try {
                Context ctx = new CompNamingContext(earFileName.getFile());
                ctx.rebind("filename", earFileName);
                unDeployEar(ctx);
            } catch (Exception e) {
                //We don't stop the process of undeploying the ears.
                //We only display an error in the logger.
                String err = "Error when undeploying the ear :" + earFileName;
                logger.log(BasicLevel.ERROR, err + e.getMessage());
            }
        }

        if (mbeanServer != null) {
            try {
                // unregister EarService MBean : EarServiceImplMBean
                mbeanServer.unregisterMBean(JonasObjectName.earService());
            } catch (Exception e) {
                logger.log(BasicLevel.ERROR, "Cannot stop the EarService: " 
                           + e.getMessage());
            }
        }

        //The service is stopped
        logger.log(BasicLevel.DEBUG, "EarService stopped");

    }

    /**
     * Deploy an EAR file with sending JAR file to the EJB container and WAR
     * file to the WEB container and RAR file to the resource service.
     * @param ctx the context which contains the configuration in order to
     * deploy an EAR.
     * @throws EarServiceException if the deployment of the EAR failed.
     */
    public void deployEar(Context ctx) throws EarServiceException {

        //We have one of the ejb or web services launched
        if ((webContainerService == null) && (ejbService == null)) {
            throw new EarServiceException("The ear service requires that at least the service ejb or web is launched for deploying an ear file.");
        }

        //We get the fileName
        // Get the file name and check if file exists
        String fileName;
        try {
            fileName = (String) ctx.lookup("filename");
        } catch (NamingException e) {
            throw new EarServiceException("Error during performing lookup a fileName", e);
        }

        File f = null;
        try {
            //We search the file
            f = new File(fileName).getCanonicalFile();

            if (!f.exists()) {
                boolean found = false;
                String earFileName = null;
                if (fileName.toLowerCase().endsWith(".ear")) {
                    // In case of the name is a ear file name, check also in
                    // the JONAS_BASE/apps directory
                    earFileName = APPS_DIR + File.separator + fileName;
                    f = new File(earFileName).getCanonicalFile();
                    found = f.exists();
                }
                if (found) {
                    fileName = earFileName;
                } else {
                    String err = "deployEar: The file " + fileName + " was not found neither in the current directory nor in the " + APPS_DIR + " directory";
                    logger.log(BasicLevel.ERROR, err);
                    throw new EarServiceException(err);
                }
            }
        } catch (IOException e) {
            String err = "Error when trying to get the canonical file from " + fileName;
            logger.log(BasicLevel.ERROR, err + " " + e.getMessage());
            throw new EarServiceException(err, e);
        }

        //Make the url for the specified fileName and the earRootUrl
        URL[] earUrl = new URL[1];
        URL earRootUrl = null;

        try {
            earUrl[0] = f.toURL();
            earRootUrl = (new File(APPS_DIR + File.separator + nameOfServer)).toURL();
        } catch (MalformedURLException e) {
            String err = "Invalid ear file name '" + fileName;
            logger.log(BasicLevel.ERROR, err + "': " + e.getMessage());
            throw new EarServiceException(err, e);
        }


        //Check if the ear is already deployed or not
        if (ears.get(earUrl[0]) != null) {
            String err = "The ear file : " + f.getName() + " is already deployed ('" 
                + earUrl[0].getFile() + "'). You must undeploy the application "
                + "before a new deployment.";
            logger.log(BasicLevel.ERROR, err);
            throw new EarServiceException(err);
        }


        //Create classLoader
        //parent classloader is the current classloader
        ClassLoader currentLoader = Thread.currentThread().getContextClassLoader();
        URLClassLoader loaderCls = new URLClassLoader(earUrl, currentLoader);

        EarDeploymentDesc earDD = null;
        logger.log(BasicLevel.DEBUG, "Getting the deployment descriptor of the file" + f.getName());
        try {
            earDD = EarDeploymentDesc.getInstance(earUrl[0].getFile(), loaderCls);
        } catch (EarDeploymentDescException e) {
            String err = "Error in the Deployment descriptor '" + fileName + "': " + e;
            logger.log(BasicLevel.ERROR, err);
            throw new EarServiceException(err);
        }

        //We get the tags from the Deployment descriptor
        Web[] webTags = earDD.getWebTags();
        Ejb[] ejbTags = earDD.getEjbTags();
        Connector[] connectorTags = earDD.getConnectorTags();
        AltDd[] altDDEjbs = earDD.getAltDDEjbs();
        AltDd[] altDDWebs = earDD.getAltDDWebs();
        AltDd[] altDDConnectors = earDD.getAltDDConnectors();
        SecurityRole[] securityRoles = earDD.getSecurityRoles();

        //Check if all modules are inside the EAR file
        //no relatives mode like ../../file1.jar

        File fEar = new File(earUrl[0].getFile());
        File tmpFile = null;

        try {
            File fCanonicEar = fEar.getCanonicalFile();
            for (int i = 0; i < ejbTags.length; i++) {
                tmpFile = new File(fEar, ejbTags[i].getValue());
                tmpFile = tmpFile.getCanonicalFile();
                if (!tmpFile.getPath().startsWith(fCanonicEar.getPath())) {
                    String err =  "Error : The ejb-jar file " + ejbTags[i].getValue()  + " is not inside the ear file " + fEar;
                    logger.log(BasicLevel.ERROR, err);
                    throw new EarServiceException(err);
                }
            }

            for (int i = 0; i < webTags.length; i++) {
                tmpFile = new File(fEar, webTags[i].getWebUri().getValue());
                tmpFile = tmpFile.getCanonicalFile();
                if (!tmpFile.getPath().startsWith(fCanonicEar.getPath())) {
                    String err =  "Error : The war file " + webTags[i].getWebUri().getValue() + " is not inside the ear file " + fEar;
                    logger.log(BasicLevel.ERROR, err);
                    throw new EarServiceException(err);
                }
            }
            for (int i = 0; i < connectorTags.length; i++) {
                tmpFile = new File(fEar, connectorTags[i].getValue());
                tmpFile = tmpFile.getCanonicalFile();
                if (!tmpFile.getPath().startsWith(fCanonicEar.getPath())) {
                    String err =  "Error : The rar file " + connectorTags[i].getValue()  + " is not inside the ear file " + fEar;
                    logger.log(BasicLevel.ERROR, err);
                    throw new EarServiceException(err);
                }
            }

        } catch (IOException ioe) {
            String err =  "Error while trying to get the canonical file of " + tmpFile;
            logger.log(BasicLevel.ERROR, err + " : " + ioe.getMessage());
            throw new EarServiceException(err, ioe);
        }


        //Changing array into JarList
        JarList ejbsList =  new JarList(ejbTags);
        JarList warsList =  new JarList(webTags);
        JarList connectorsList =  new JarList(connectorTags);

        //Unpack the ear file and get the unpacked dir
        URL dirUnpackURL = null;
        try {
            dirUnpackURL = EarFileManager.unpackEar(earUrl[0], earRootUrl);
        } catch (EarFileManagerException e) {
            String err =  "Error while unpacking the file '" + earUrl[0] + "'";
            logger.log(BasicLevel.ERROR, err + " : " + e.getMessage());
            throw new EarServiceException(err, e);
        }


        //the file is unpacked. Log it only if the ear is a file and not a directory
        if (new File(earUrl[0].getFile()).isFile()) {
            try {
                earDeployerLog.addEntry(earUrl[0].getFile(), dirUnpackURL);
            } catch (EarDeployerLogException e) {
                String err =  "Error while adding the " + earUrl[0] + " entry in the log file";
                logger.log(BasicLevel.ERROR, err + " : " + e.getMessage());
                throw new EarServiceException(err, e);
            }
        }

        //We have successfully unpacked the ear file, now we analyze manifest Class-path:
        EarClassPathManager earCPManager = null;
        try {
            earCPManager = new EarClassPathManager(ejbsList, warsList, dirUnpackURL);
        } catch (EarClassPathManagerException e) {
            String err =  "Error while creating the Ear class path manager of the ear : '" + earUrl[0] + "'";
            logger.log(BasicLevel.ERROR, err + " : " + e.getMessage());
            throw new EarServiceException(err, e);
        }

        URL[] classpathURLs = null;
        //Get the urls of the ear class path manager
        try {
            classpathURLs = earCPManager.getResolvedClassPath();
        } catch (EarClassPathManagerException e) {
            String err =  "Error while trying to resolve the classpath of the ejbjars and wars of the ear : '" + earUrl[0] + "'";
            logger.log(BasicLevel.ERROR, err + " : " + e.getMessage());
            throw new EarServiceException(err, e);
        }


        /* We have the urls to load at our level
           We can now create our classLoader
        */
        logger.log(BasicLevel.DEBUG, "Creating the EAR classLoader");
        currentLoader = Thread.currentThread().getContextClassLoader();
        URLClassLoader earClassLoader = new URLClassLoader(new URL[0], currentLoader);

        //Extract the urls of the jarList
        URL[] jarUrls = null;
        URL[] warUrls = null;
        URL[] connectorUrls = null;
        try {
            jarUrls = ejbsList.getURLs(dirUnpackURL.toExternalForm());
            warUrls = warsList.getURLs(dirUnpackURL.toExternalForm());
            connectorUrls = connectorsList.getURLs(dirUnpackURL.toExternalForm());
        } catch (JarListException e) {
            String err =  "Error while geting the Urls from jarlist of the ear : '" + earUrl[0] + "'";
            logger.log(BasicLevel.ERROR, err + " : " + e.getMessage());
            throw new EarServiceException(err, e);
        }

        //Extract context roots for wars
        //They are in webTags. Dump into warsContextRoots
        String[] warsContextRoots = new String[webTags.length];
        ContextRoot ctxRoot = null;
        for (int i = 0; i < webTags.length; i++) {
            ctxRoot = webTags[i].getContextRoot();
            if (ctxRoot != null) {
                warsContextRoots[i] = ctxRoot.getValue();
            }
        }

        //Fill Alt-DD for Ejbs and Wars and Rars
        String altdd = null;
        File fAltDD = null;

        //Transorm the array altDDWebs into an array with the absolute URL to the file
        URL[] warsAltDDs = new URL[altDDWebs.length];
        for (int i = 0; i < altDDWebs.length; i++) {
            if (altDDWebs[i] != null) {
                altdd = altDDWebs[i].getValue();
                if (altdd != null) {
                    try {
                        fAltDD = new File(new URL(dirUnpackURL.toExternalForm() + File.separator + altdd).getFile());
                        warsAltDDs[i] = fAltDD.getCanonicalFile().toURL();
                    } catch (MalformedURLException e) {
                        String err = "Can't build URL for alt-dd '" + altdd;
                        logger.log(BasicLevel.ERROR, err + "': " + e.getMessage());
                        throw new EarServiceException(err, e);
                    } catch (IOException ioe) {
                        String err = "Can't get canonicalFile() for the file '" + fAltDD;
                        logger.log(BasicLevel.ERROR, err + "': " + ioe.getMessage());
                        throw new EarServiceException(err, ioe);
                    }
                }
            }
        }

        URL[] ejbsAltDDs = new URL[altDDEjbs.length];
        for (int i = 0; i < altDDEjbs.length; i++) {
            if (altDDEjbs[i] != null) {
                altdd = altDDEjbs[i].getValue();
                if (altdd != null) {
                    try {
                        fAltDD = new File(new URL(dirUnpackURL.toExternalForm() + File.separator + altdd).getFile());
                        ejbsAltDDs[i] = fAltDD.getCanonicalFile().toURL();
                    } catch (MalformedURLException e) {
                        String err = "Can't build URL for alt-dd '" + altdd;
                        logger.log(BasicLevel.ERROR, err + "': " + e.getMessage());
                        throw new EarServiceException(err, e);
                    } catch (IOException ioe) {
                        String err = "Can't get canonicalFile() for the file '" + fAltDD;
                        logger.log(BasicLevel.ERROR, err + "': " + ioe.getMessage());
                        throw new EarServiceException(err, ioe);
                    }
                }
            }
        }

        URL[] connectorsAltDDs = new URL[altDDConnectors.length];
        for (int i = 0; i < altDDConnectors.length; i++) {
            if (altDDConnectors[i] != null) {
                altdd = altDDConnectors[i].getValue();
                if (altdd != null) {
                    try {
                        fAltDD = new File(new URL(dirUnpackURL.toExternalForm() + File.separator + altdd).getFile());
                        connectorsAltDDs[i] = fAltDD.getCanonicalFile().toURL();
                    } catch (MalformedURLException e) {
                        String err = "Can't build URL for alt-dd '" + altdd;
                        logger.log(BasicLevel.ERROR, err + "': " + e.getMessage());
                        throw new EarServiceException(err, e);
                    } catch (IOException ioe) {
                        String err = "Can't get canonicalFile() for the file '" + fAltDD;
                        logger.log(BasicLevel.ERROR, err + "': " + ioe.getMessage());
                        throw new EarServiceException(err, ioe);
                    }
                }
            }
        }

        //Extract roles name from the securityRole array
        String[] roleNames = new String[securityRoles.length];
        String affRoleNames = "";
        for (int i = 0; i < securityRoles.length; i++) {
            roleNames[i] = securityRoles[i].getRoleName().getValue();
            affRoleNames += roleNames[i] + ";";
        }

        logger.log(BasicLevel.DEBUG, "role names = " + affRoleNames);

        /*
          Now this is the deployment step with sending
          + jars and wars to web services service
          + rars to the rar service
          + jars to the ejb service
          + wars to the web container service
        */

        /**
         * Ejb ClassLoader is needed for WebServices deployment
         * so We create it in advance ...
         */
        // Set the list of the ejb-jar that can do an ejb-link in this
        // ear application. This array can be empty.
        // The alternate urls are given

        ejbDDManager.setAvailableEjbJarsAndAltDDs(earClassLoader, jarUrls, ejbsAltDDs);
        webDDManager.setAltDD(earClassLoader, warUrls, warsAltDDs);

        // Construct the ejb classloader for all the ejb-jars of the same
        // ear application. Because there is one ejb classloader for all
        // the ejb-jars of the same ear application.
        URL[] urls = new URL[jarUrls.length + classpathURLs.length];
        System.arraycopy(jarUrls, 0, urls, 0, jarUrls.length);
        System.arraycopy(classpathURLs, 0, urls, jarUrls.length, classpathURLs.length);
        URLClassLoader ejbClassLoader = new URLClassLoader(urls, earClassLoader);

        // Only if the WebServicesService is started
        if (wsService != null) {
            try {
                Hashtable ctxRoots = new Hashtable();
                for (int w = 0; w < webTags.length; w++) {
                    ctxRoots.put(warUrls[w],
                                 webTags[w].getContextRoot().getValue());
                }
                CompNamingContext contctx = null;
                try {
                    contctx = new CompNamingContext(dirUnpackURL.getFile());
                    contctx.rebind("unpackDir", dirUnpackURL.toExternalForm());
                    contctx.rebind("jarUrls", jarUrls);
                    contctx.rebind("warUrls", warUrls);
                    contctx.rebind("earURL", earUrl[0]);
                    contctx.rebind("ejbClassLoader", ejbClassLoader);
                    contctx.rebind("earClassLoader", earClassLoader);
                    contctx.rebind("warCtxRootMapping", ctxRoots);
                } catch (NamingException e) {
                    String err = "Can not bind params for the WebServices service, Can't deploy Web Services Endpoint ";
                    throw new WSServiceException(err, e);
                }
                wsService.deployWebServices(contctx);
            } catch (ServiceException se) {
                String err = "Error during the deployment of the WebServices of the Ear file " + fileName;
                logger.log(BasicLevel.ERROR, err + " : " + se.getMessage());
                throw new EarServiceException(err, se);
            }
        }

        // Set the list of the rars that can be referenced in this
        // ear application. This array can be empty.
        // The alternate urls are given

        //rarDDManager.setAvailableRarsAndAltDDs(earClassLoader, connectorUrls, connectorsAltDDs);

        //Only if the connectorUrls is not empty
        //And if the service is started
        if (resourceService != null && (connectorUrls.length > 0)) {
            try {
                CompNamingContext contctx = null;
                try {
                    contctx = new CompNamingContext(dirUnpackURL.getFile());
                    contctx.rebind("earUrl", earUrl[0]);
                    contctx.rebind("urls", connectorUrls);
                    contctx.rebind("earClassLoader", earClassLoader);
                    contctx.rebind("altDDs", connectorsAltDDs);
                } catch (NamingException e) {
                    String err = "Can not bind params for the resource service, Can't deploy rars ";
                    throw new ResourceServiceException(err, e);
                }
                resourceService.deployRars(contctx);
            } catch (ServiceException e) {
                String err = "Error during the deployment of the rars files of the Ear file " + fileName;
                logger.log(BasicLevel.ERROR, err + "': " + e.getMessage());
                //remove DDesc
                //resourceService.removeCache(earClassLoader);
                throw new EarServiceException(err, e);
            }
        }

        // ejbClassLoader created above

        //Only if the jarUrls is not empty
        //And if the service is started
        if (ejbService != null && (jarUrls.length > 0)) {
            try {
                CompNamingContext contctx = null;
                try {
                    contctx = new CompNamingContext(dirUnpackURL.getFile());
                    contctx.rebind("earRootUrl", dirUnpackURL);
                    contctx.rebind("earUrl", earUrl[0]);
                    contctx.rebind("jarURLs", jarUrls);
                    contctx.rebind("earClassLoader", earClassLoader);
                    contctx.rebind("ejbClassLoader", ejbClassLoader);
                    contctx.rebind("roleNames", roleNames);
                } catch (NamingException e) {
                    String err = "Can not bind params for the ejb service, Can't deploy jars ";
                    throw new EarServiceException(err, e);
                }
                ejbService.deployJars(contctx);
            } catch (ServiceException e) {
                String err = "Error during the deployment of the jars files of the Ear file " + fileName;
                logger.log(BasicLevel.ERROR, err + "': " + e.getMessage());
                //remove DDesc
                ejbService.removeCache(earClassLoader);
                throw new EarServiceException(err, e);
            }
        }
        /* We have deploy the ejb-jar files if the service is active
           And we have the ejbClassLoader*/

        /* We're going to deploy wars
           Parent ClassLoader for web container is ejbclassloader if ejbService is active,
           and a classloader with the visibility of jar file if ejbservice is not active.
        */

        ClassLoader parentWarClassLoader = ejbClassLoader;

        //Only if the warUrls is not empty
        //And if the service is started
        if (webContainerService != null && (warUrls.length > 0)) {
            try {
                CompNamingContext contctx = null;
                try {
                    contctx = new CompNamingContext(dirUnpackURL.getFile());
                    contctx.rebind("earURL", earUrl[0]);
                    contctx.rebind("urls", warUrls);
                    contctx.rebind("parentClassLoader", parentWarClassLoader);
                    contctx.rebind("earClassLoader", earClassLoader);
                    contctx.rebind("altDDs", warsAltDDs);
                    contctx.rebind("contextRoots", warsContextRoots);
                } catch (NamingException e) {
                    String err = "Can not bind params for the web container service, Can't deploy wars ";
                    throw new EarServiceException(err, e);
                }
                webContainerService.deployWars(contctx);

            } catch (JWebContainerServiceException e) {
                String err = "Error during the deployment of the wars file of the Ear file " + fileName;
                logger.log(BasicLevel.ERROR, err + "': " + e.getMessage());
                //remove DDesc
                webContainerService.removeCache(earClassLoader);
                err = "Undeploy the jars loaded from this ear";
                logger.log(BasicLevel.DEBUG, err);
                try {
                    if (ejbService != null && (jarUrls.length > 0)) {
                        //remove DDesc
                        ejbService.removeCache(earClassLoader);
                        //remove the ejbs
                        ejbService.unDeployJars(jarUrls);
                    }
                } catch (ServiceException se) {
                    err = "Error during the undeployment of the jars file of the Ear file " + fileName;
                    logger.log(BasicLevel.ERROR, err + "': " + se.getMessage());
                    throw new EarServiceException(err, e);
                }

                throw new EarServiceException(err, e);
            }
        }
        logger.log(BasicLevel.DEBUG, "We store the rars/wars/jars associated to the url :" + earUrl[0]);

        //Remove the Deployment descriptors
        if (resourceService != null) {
            //resourceService.removeCache(earClassLoader);
        }

        if (ejbService != null) {
            ejbService.removeCache(earClassLoader);
        }

        if (webContainerService != null) {
            webContainerService.removeCache(earClassLoader);
        }

        //Store the ejb-jar, war and the rar urls associated with the ear URL in case of success deploy
        Ear ear = new Ear(earUrl[0], earDD.getXmlContent(), jarUrls, warUrls, connectorUrls);
        try {
            // Register Ear MBean : EarMBean
            if (mbeanServer != null) {
                mbeanServer.registerMBean(ear, JonasObjectName.ear(earUrl[0].getFile()));
            }
        } catch (Exception eare) {
            logger.log(BasicLevel.ERROR, "Can not register the MBean for the ear" + earUrl[0] + " :" + eare.getMessage());
        }
        ears.put(earUrl[0], ear);

        logger.log(BasicLevel.INFO, "Ear " + earUrl[0] + " available.");

        //set the timestamp dir
        File dirUnpackFile = new File(dirUnpackURL.getFile());
        dirUnpackFile.setLastModified((new Date()).getTime());

    }

    /**
     * Deploy an EAR by delegating the operation to the deployEar() method.
     * This is used for JMX management.
     * @param fileName the fileName of the ear which must be be deployed.
     * @throws RemoteException if rmi call failed.
     * @throws EarServiceException if the deployment of the EAR failed.
     */
    public void deployEarMBean(String fileName) throws RemoteException, EarServiceException {
        Context ctx = null;
        try {
            ctx = new CompNamingContext(fileName);
            ctx.rebind("filename", fileName);
        } catch (NamingException e) {
            String err = "Error when deploying the ear file ";
            logger.log(BasicLevel.ERROR, err + e.getMessage());
            throw new EarServiceException(err, e);
        }
        deployEar(ctx);
    }

    /**
     * Start the EAR service.
     * @throws ServiceException if the startup failed.
     */
    protected void doStart() throws ServiceException {
        //url
        URL earApps = null;
        try {
            earApps = new File(APPS_DIR + File.separator + nameOfServer).toURL();
        } catch (MalformedURLException mue) {
            throw new ServiceException("Error when trying to get the URL of the jonasroot/apps directory" , mue);
        }


        File fLog = new File(earApps.getFile() + File.separator + nameOfServer + ".log");
        if (!fLog.exists()) {
            try {
                //create
                fLog.getParentFile().mkdirs();
                fLog.createNewFile();
            } catch (IOException e) {
                throw new ServiceException("cannot create the log file" + fLog);
            }
        }

        //get the logger
        try {
            earDeployerLog = new EarDeployerLog(fLog);
        } catch (EarDeployerLogException e) {
            throw new ServiceException("Can not get an EarDeployerLog" , e);
        }

        //run the cleaner
        earCleaner = EarCleaner.getInstance(earApps, earDeployerLog);

        //remove all unused ear when starting
        try {
            earCleaner.removeUnusedDirectories();
        } catch (EarCleanerException e) {
            throw new ServiceException("Can not remove unused directories" , e);
        }
        earCleaner.start();

        // Deploy all ear which are in descriptors section
        Iterator it = earNames.iterator();
        while (it.hasNext()) {
            String fileName = (String) it.next();
            Context contctx = null;
            try {
                contctx = new CompNamingContext(fileName);
                contctx.rebind("filename", fileName);
            } catch (NamingException e) {
                throw new ServiceException("Cannot start the EarService" , e);
            }
            deployEar(contctx);
        }

        try {
            // Register EarService MBean : EarServiceImplMBean
            if (mbeanServer != null) {
                mbeanServer.registerMBean(this, JonasObjectName.earService());
            }
        } catch (InstanceAlreadyExistsException iae) {
            throw new ServiceException("Cannot start the EarService Already Exists" , iae);
        } catch (MBeanRegistrationException mbre) {
            throw new ServiceException("Cannot start the EarService (MBean registration error)" , mbre);
        } catch (NotCompliantMBeanException ncmbe) {
            throw new ServiceException("Cannot start the EarService (MBean Not compliant error)" , ncmbe);
        }
    }

    /**
     * Undeploy an EAR by sending the request to the EJB container and to the
     * WEB container and the Resource service.
     * @param ctx the context which contains the configuration in order to
     * undeploy an EAR.
     * @throws EarServiceException if the undeployment of the EAR failed.
     */
    public void unDeployEar(Context ctx) throws EarServiceException {

        //Get the file param which we want to undeploy
        URL fileName = null;
        try {
            fileName = (URL) ctx.lookup("filename");
        } catch (NamingException e) {
            throw new EarServiceException("Trying to remove the ear file but there is no filename specified", e);
        }

        //This ear is in loaded EARs files ?
        Ear ear = (Ear) ears.get(fileName);

        //Not loaded
        if (ear == null) {
            throw new EarServiceException("Trying to remove the ear file " + fileName.getFile()
                                          + " but this file was not found in the loaded Ear files");
        }

        //undeploy wars if the web container service is started
        URL[] warsToUndeploy = ear.getWars();
        if (webContainerService != null && (warsToUndeploy.length > 0)) {
            webContainerService.unDeployWars(warsToUndeploy);
        }

        //undeploy jars if the ejb service is started
        URL[] jarsToUndeploy = ear.getEjbJars();
        if (ejbService != null && (jarsToUndeploy.length > 0)) {
            ejbService.unDeployJars(jarsToUndeploy);
        }

        //undeploy rars if the resource service is started
        URL[] rarsToUndeploy = ear.getRars();
        if (resourceService != null && (rarsToUndeploy.length > 0)) {
            resourceService.unDeployRars(rarsToUndeploy);
        }

        File f = new File(fileName.getFile());
        //Undeploy successful, we remove the ear


        //Only if there is no Exception !
        ears.remove(fileName);
        //unregister mbean
        if (mbeanServer != null) {
            try {
                // unregister Ear MBean : EarMBean
                mbeanServer.unregisterMBean(JonasObjectName.ear(fileName.getFile()));
            } catch (Exception e) {
                logger.log(BasicLevel.ERROR, "Cannot remove the MBean for the ear " + fileName.getFile() + " : " + e.getMessage());
            }
        }

        logger.log(BasicLevel.INFO, "Ear " + f.getName() + " no longer available.");

    }

    /**
     * Undeploy an EAR by delegating the operation to the unDeployEar() method.
     * This is used for JMX management.
     * @param fileName the fileName of the ear which must be be undeployed.
     * @throws RemoteException if rmi call failed.
     * @throws EarServiceException if the undeployment of the EAR failed.
     */
    public void unDeployEarMBean(String fileName) throws RemoteException, EarServiceException {

        /* We have only the name of the file, not its associated path, so we look in
           the current directory and in the ear applications directory
        */

        //We check if the url exists in the hashtable
        //compare: First in the current dir, after in the apps directory

        URL url = null;
        URL earUrl = null;
        boolean found = false;
        try {
            Enumeration e = ears.keys();
            url = new File(fileName).getCanonicalFile().toURL();
            while (e.hasMoreElements() && !found) {
                earUrl = (URL) e.nextElement();
                if (earUrl.equals(url)) {
                    //We have found
                    found = true;
                }
            }

            if (fileName.toLowerCase().endsWith(".ear") && !found) {
                // In case of the name is an ear check also in
                // the JONAS_BASE/apps directory
                String earFileName = APPS_DIR + File.separator + fileName;
                e = ears.keys();
                url = new File(earFileName).getCanonicalFile().toURL();
                while (e.hasMoreElements() && !found) {
                    earUrl = (URL) e.nextElement();
                    if (earUrl.equals(url)) {
                        //We have found
                        found = true;
                    }
                }
            }
            if (!found) {
                String err = "Cannot undeploy the ear '" + fileName + "', it is not deployed.";
                logger.log(BasicLevel.ERROR, err);
                throw new EarServiceException(err);
            }
        } catch (MalformedURLException mue) {
            String err = "Error when trying to get the url from" + fileName;
            logger.log(BasicLevel.ERROR, err + mue.getMessage());
            throw new EarServiceException(err, mue);
        } catch (IOException ioe) {
            String err = "Error when trying to get the canonical file fron " + fileName;
            logger.log(BasicLevel.ERROR, err + ioe.getMessage());
            throw new EarServiceException(err, ioe);
        }

        //We've got the file, now we bind the params
        Context ctx = null;
        try {
            ctx = new CompNamingContext(fileName);
            //The fileName is an url
            ctx.rebind("filename", url);
        } catch (NamingException e) {
            String err = "Error when binding parameters";
            logger.log(BasicLevel.ERROR, err + e.getMessage());
            throw new EarServiceException(err, e);
        }

        //Call the function
        unDeployEar(ctx);

    }

    /**
     * @return current number of ears deployed in the JOnAS server
     */
    public Integer getCurrentNumberOfEars() {
        return new Integer(ears.size());
    }

    /**
     * Return the list of installed Applications.
     * The EAR files or the directories with expanded Applications are searched
     * in JONAS_BASE/apps and all Applications directories 'autoload'.
     * @return The list of EAR files or the directories with expanded Applications found
     * @throws Exception if the list can't be retrieved
     */
    public List getInstalledEars() throws Exception {
        // get EAR files found in JONAS_BASE/apps
        ArrayList al = JModule.getInstalledContainersInDir(APPS_DIR, JModule.EAR_EXTENSION
                                                           , JModule.EAR_CHILD_DIR, JModule.EAR_CONFIRM_FILE);
        // get EAR files found in all autoload directories
        Iterator it = autoloadDirectories.iterator();
        while (it.hasNext()) {
            String name = (String) it.next();
            al.addAll(JModule.getInstalledContainersInDir(name, JModule.EAR_EXTENSION,
                                                          JModule.EAR_CHILD_DIR, JModule.EAR_CONFIRM_FILE));
        }
        return al;
    }

    /**
     * This method is added temporarily. It will disapear when Ears will have their associated MBeans
     * (when Ears will become manageable)
     * @return the names of the ears currently deployed in the JOnAS server
     */
    public Set getEarNames() {
        HashSet names = new HashSet();
        URL earUrl = null;
        for (Enumeration e = ears.keys(); e.hasMoreElements();) {
            earUrl = (URL) e.nextElement();
            names.add(earUrl.getFile());
        }
        return names;
    }

    /**
     * Add recursively the ears of the specified directory.
     * If the dir has a relative path, this path is relative from where
     * the Application Server is launched. If the dir is not found it will
     * be searched in $JONAS_BASE/apps/ directory.
     * @param dirPath the path to the directory containing the ears to
     * load.
     */
    private void addEars(String dirPath) {
        boolean found = false;

        // Look the directory relative to where the JOnAS server is launch or
        // if the path is absolute.
        File dir = new File(dirPath);
        found = dir.isDirectory();

        if (!found) {
            // Look the directory relative to the $JONAS_BASE/apps directory
            dir = new File(APPS_DIR + File.separator + dirPath);
            found = dir.isDirectory();
        }

        if (found) {
            addEarsFrom(dir);
        } else {
            String err = "Warning: Cannot load dir: '" + dirPath + "' ";
            err += "is not a directory or directory doesn't exist";
            logger.log(BasicLevel.WARN, err);
        }
    }

    /**
     * Add the ears of the specified directory.
     * @param dir the directory from which the ears are loaded.
     * @throws EarServiceException if the argument is not a directory
     */
    private void addEarsFrom(File dir) throws EarServiceException {
        try {
            if (dir.isDirectory()) {
                File[] files = dir.listFiles();
                for (int i = 0; i < files.length; i++) {
                    if (files[i].isFile()) {
                        if (files[i].getPath().toLowerCase().endsWith(".ear")) {
                            earNames.add(files[i].getCanonicalPath());
                        }
                    } else {
                        addEarsFrom(files[i]);
                    }
                }
            } else {
                String err = "Cannot load dir: '" + dir.getPath();
                err += "' is not a directory";
                logger.log(BasicLevel.ERROR, err);
                throw new EarServiceException(err);
            }
        } catch (IOException e) {
            String err = "Invalid file name '" + dir.getPath();
            logger.log(BasicLevel.ERROR, err);
            throw new EarServiceException(err, e);
        }
    }

    /**
     * Test if the specified filename is already deployed or not
     * @param fileName the name of the ear file.
     * @return true if the ear is deployed, else false.
     */
    public boolean isEarLoaded(String fileName) {

        URL url = null;
        boolean isLoaded = false;
        try {
            // Absolute filename
            try {
                url = new File(fileName).getCanonicalFile().toURL();
                //Check if the ear is already deployed or not
                if (ears.get(url) != null) {
                    isLoaded = true;
                } else {
                    // Not found force to test in relative Apps directory
                    url = null;
                }
            } catch (Exception e) {
                url = null;
            }
            // Relative filename
            if (url == null) {
                url = new File(APPS_DIR + File.separator + fileName).getCanonicalFile().toURL();
                //Check if the ear is already deployed or not
                if (ears.get(url) != null) {
                    isLoaded = true;
                }
            }
        } catch (Exception e) {
            String err = "Can not found if the ear is deployed or not";
            logger.log(BasicLevel.ERROR, err);
            return false;
        }

        return isLoaded;
    }

    /**
     * Return the list of all loaded Applications.
     *
     * @return The list of deployed Applications
     */
    public List getDeployedEars() {
        ArrayList al = new ArrayList();
        Ear oEar;
        URL oURL;
        for (Enumeration enumEars = ears.elements(); enumEars.hasMoreElements();) {
            oEar = (Ear) enumEars.nextElement();
            oURL = oEar.getEarUrl();
            al.add(oURL.getFile());
        }
        return al;
    }

    /**
     * Return the list of installed Applications ready to deploy.
     *
     * @return The list of deployable Applications
     * @throws Exception
     */
    public List getDeployableEars() throws Exception {
        List al = getInstalledEars();
        al.removeAll(getDeployedEars());
        return al;
    }

    /**
     * Return the list of "autoload" directories for web applications.
     * @return The list of all "autoload" directories
     */
    public List getAutoloadDirectories() {
        ArrayList al = new ArrayList();
        Iterator it = autoloadDirectories.iterator();
        while (it.hasNext()) {
            String fileName = (String) it.next();
            try {
                al.add(new File(fileName).toURL().getPath());
            } catch (Exception e) {
                ;
                // none
            }
        }
        return al;
    }

    /**
     * Return the Apps directory.
     * @return The Apps directory
     */
    public String getAppsDirectory() {
        String sRet = null;
        try {
            sRet = (new File(APPS_DIR)).toURL().getPath();
        } catch (Exception e) {
            ;
            // none
        }
        return sRet;
    }


}
