/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * ObjectWeb Connector: an implementation of JCA Sun specification along
 *                      with some extensions of this specification.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Based on LArrayPool in ObjectWeb common
 *
 * --------------------------------------------------------------------------
 * $Id: HArrayPool.java,v 1.6 2003/09/18 16:38:37 ehardesty Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.resource.pool.lib;

import java.util.HashSet;
import java.util.Hashtable;
import java.util.LinkedList;
import java.util.Iterator;

import javax.resource.spi.ManagedConnection;
import org.objectweb.jonas.resource.pool.api.Pool;
import org.objectweb.jonas.resource.pool.api.PoolItemStats;
import org.objectweb.jonas.resource.pool.api.PoolMatchFactory;

/**
 * The class <b>HArrayPool</b> implements a Pool as a HashSet of ManagedConnections,
 * managing free/active resources.
 * 
 * Updated to use an LRU list of free resources
 *
 * Author: Eric HARDESTY
 */
public class HArrayPool implements Pool {

    private int minSize = 0;
    private int maxSize = -1;
    private int initSize = -1;
    private long timeout = 0;
    private int curSize = 0;
    private int freeSize = 0;
    private PoolMatchFactory matchFactory = null;
    private HashSet freeList = null;
    private LinkedList linkList = null;
    private HashSet activeList = null;
    private Hashtable infoList = null;
    
    private long maxage = 0;

    public HArrayPool() {
        freeList = new HashSet();
        linkList = new LinkedList();
        activeList = new HashSet();
        infoList = new Hashtable();
        (new HArrayPoolMonitor(this)).start();
    }
        
    // IMPLEMENTATION OF METHODS FROM THE Pool INTERFACE
        
    /**
     * @see org.objectweb.jonas.resource.pool.api.Pool#check4MaxAge
     */
    public void check4MaxAge() {

        if (maxage == 0) {
            return;
        }
        if (!freeList.isEmpty()) {
            Iterator it = freeList.iterator();
            long curTime = System.currentTimeMillis();
            while (it.hasNext()) {
                ManagedConnection res = (ManagedConnection) it.next();
                PoolItemStats pis = (PoolItemStats) infoList.get(res);
                if (pis != null && curTime > pis.getMaxAgeTimeout()) {
                    try {
                        matchFactory.releaseResource(res);
                    } catch (Exception ex) { }
                    freeList.remove(res);
                    try {
                        res.destroy();
                    } catch (Exception ex) { }
                    linkList.remove(res);
                    curSize--;
                    freeSize--;
                }
            }
        }
    }
    
    /**
     * @see org.objectweb.jonas.resource.pool.api.Pool#getTimeout
     */
    public long getTimeout() {
        return timeout;
    }
    
    /**
     * @see org.objectweb.jonas.resource.pool.api.Pool#getMinSize
     */
    public int getMinSize() {
        return minSize;
    }
    
    /**
     * @see org.objectweb.jonas.resource.pool.api.Pool#getMatchFactory
     */
    public PoolMatchFactory getMatchFactory() {
        return matchFactory;
    }
    
    /**
     * @see org.objectweb.jonas.resource.pool.api.Pool#getMaxSize
     */
    public int getMaxSize() {
        return maxSize;
    }
    
    /** 
     * @see org.objectweb.jonas.resource.pool.api.Pool#getResource
     */
    public synchronized Object getResource(Object hints)
        throws Exception {
        
        if (matchFactory == null) {
            throw new Exception("The matchFactory mandatory!!");
        }
        ManagedConnection res = null;
        if (!freeList.isEmpty()) {
            try {
                res = (ManagedConnection) matchFactory.matchResource(freeList, hints);
            } 
            catch (Exception ex) { }
        }
        if (res == null) {
            if (maxSize < 0 || curSize < maxSize) {
                res = (ManagedConnection) matchFactory.createResource(hints);
                activeList.add(res);
                curSize++;
            } else if (freeSize > 0) {
                res = (ManagedConnection) linkList.getFirst();
                matchFactory.releaseResource(res);
                res.destroy();
                freeList.remove(res);
                linkList.remove(res);
                infoList.remove(res);
                curSize--;
                freeSize--;
                // Create a new one and return it
                res = (ManagedConnection) matchFactory.createResource(hints);
                activeList.add(res);
                curSize++;
            } else {
                throw new Exception("No more resources are available in the pool. (C:"+curSize+
                                     " M:"+maxSize+" F:"+freeSize+")");
            }
            infoList.put(res, new PoolItemStats());
        }
        else {
            freeList.remove(res);
            linkList.remove(res);
            activeList.add(res);
            freeSize--;
        }
//        printLists();
        setItemStats(res);
        return res;
    }
        
    public int getSize() {
        return curSize;
    }
        
    /**
     * @see org.objectweb.jonas.resource.pool.api.Pool#releaseResource
     */
    public synchronized void releaseResource(Object resource, boolean destroy)
        throws Exception {
        
        ManagedConnection res = (ManagedConnection) resource;
        if (matchFactory == null) {
            throw new Exception("The matchFactory mandatory!!");
        }
        if (activeList == null) {
            throw new Exception("No active resources to releases!!");
        }
        if (!activeList.contains(res)) {
            throw new Exception("Attempt to release inactive resource!!");
        }
        activeList.remove(res);
        if ((maxSize < 0 || curSize <= maxSize) && !destroy) {
            freeList.add(res);
            linkList.addLast(res);
            freeSize++;
            PoolItemStats pis = (PoolItemStats) infoList.get(res);
            if (pis != null) {
                pis.setTotalConnectionTime(System.currentTimeMillis() - pis.getStartTime());
            }
        }
        else {
            matchFactory.releaseResource(res);
            res.destroy();
            infoList.remove(res);
            curSize--;
        }

        while (curSize < minSize) {
            res = (ManagedConnection) matchFactory.createResource(null);
            freeList.add(res);
            linkList.addLast(res);
            infoList.put(res, new PoolItemStats());
            curSize++;
            freeSize++;
        }
//        printLists();
    }
        
    /**
     * @see org.objectweb.jonas.resource.pool.api.Pool#setInitSize
     */
    private void setInitSize() throws Exception {
        if (initSize <= 0) return;
        int initsize;
        if (maxSize < 0) { 
            initsize = initSize;
        } else {
            initsize = initSize < maxSize ? initSize : maxSize;
        }
        initsize = initsize < minSize ? 0 : initsize - minSize;
        
        ManagedConnection res;
        for (int i=0;i<initsize;i++) {
            res = (ManagedConnection) matchFactory.createResource(null);
            freeList.add(res);
            linkList.addLast(res);
            infoList.put(res, new PoolItemStats());
            curSize++;
            freeSize++;
        }
    }
    

    /**
     * @see org.objectweb.jonas.resource.pool.api.Pool#setMaxAge
     */
    public void setMaxAge(long _maxAge) {
        maxage = _maxAge;
    }
    
    /**
     * @see org.objectweb.jonas.resource.pool.api.Pool#setInitSize
     */
    public synchronized void setInitSize(int initsize) throws Exception {
        if (matchFactory == null)
            throw new Exception("The matchFactory mandatory!!");
        if (initSize < 0 && initsize >= 0) {
            initSize = initsize;
            setInitSize();
        }
    }

    /**
     * @see org.objectweb.jonas.resource.pool.api.Pool#setMatchFactory
     */
    public synchronized void setMatchFactory(PoolMatchFactory pmf) {
        matchFactory = pmf;
    }
    
    /**
     * @see org.objectweb.jonas.resource.pool.api.Pool#setMaxSize
     */
    public synchronized void setMaxSize(int maxsize) throws Exception {
        if (matchFactory == null) {
            throw new Exception("The matchFactory mandatory!!");
        }
        if ((maxsize < 0) || (maxsize < minSize) || (maxsize == maxSize)) {
            return;
        }
        if (maxsize < maxSize) {
            int cursize = curSize;
            int freesize = freeSize;
            if (maxsize < cursize) {
                // Removes as many free cells as possible
                int nbRemove = cursize - maxsize;
                if (freeList != null) {
                    while (!freeList.isEmpty() && nbRemove > 0) {
                            ManagedConnection res = (ManagedConnection) linkList.getFirst();
                            matchFactory.releaseResource(res);
                            res.destroy();
                            freeList.remove(res);
                            linkList.remove(res);
                            infoList.remove(res);
                            nbRemove--;
                            cursize--;
                            freesize--;
                    }
                }
            }
            curSize = cursize;
            freeSize = freesize;
        }
        maxSize = maxsize;
    }
        
    /**
     * @see org.objectweb.jonas.resource.pool.api.Pool#setMinSize
     */
    public synchronized void setMinSize(int minsize) throws Exception {
        if (matchFactory == null) {
            throw new Exception("The matchFactory mandatory!!");
        }
        if ((minsize < 0) || (maxSize < 0) || (minsize > maxSize) || (minsize == minSize)) {
            return;
        }
        if (minsize < minSize) {
            minSize = minsize;
            return;
        }
        ManagedConnection res;
        while (minsize > curSize) {
            res = (ManagedConnection) matchFactory.createResource(null);
            freeList.add(res);
            linkList.addLast(res);
            infoList.put(res, new PoolItemStats());
            curSize++;
            freeSize++;
        }
        minSize = minsize;
    }
        
    /**
     * @see org.objectweb.jonas.resource.pool.api.Pool#setTimeout
     */
    public synchronized void setTimeout(long crto) {
    }
    
    private void setItemStats(Object res) {
        PoolItemStats pis = (PoolItemStats) infoList.get(res);
        if (pis != null) { 
            pis.incrementUses();
            long sTime = System.currentTimeMillis();
            pis.setStartTime(sTime);
            if (maxage > 0) {
                pis.setMaxAgeTimeout(sTime + maxage);
            }
        }
    }
    
//      void printLists() {
//              if (maxSize != 200)
//                      return;
//              System.err.println("minSize=" + minSize + ", maxSize=" + maxSize
//                                                 + ", curSize=" + curSize + ", freeSize=" +freeSize);
//              System.err.print("activeList:");
//              if (activeList == null)
//                      System.err.print(" null");
//              else {
//                      Iterator it = activeList.iterator();
//                      while (it.hasNext())
//                              System.err.print(" " + (ManagedConnection) it.next());
//              }
//              System.err.println(".");
//              System.err.print("freeList:");
//              if (freeList == null)
//                      System.err.print(" null");
//              else {
//                      Iterator it = freeList.iterator();
//                      while (it.hasNext())
//                              System.err.print(" " + (ManagedConnection) it.next());
//              }
//              System.err.println(".");
//      }

}
