/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: JonasSecurityServiceImpl.java,v 1.26 2003/08/19 12:24:48 benoitf Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.security;

//import java
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.StringReader;
import java.net.URL;
import java.net.MalformedURLException;
import java.util.Enumeration;
import java.util.Hashtable;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.management.MBeanServer;

import java.util.*;
import java.security.Guard;
import java.security.NoSuchAlgorithmException;

//import apache
import org.apache.commons.digester.Digester;

// import sax
import org.xml.sax.InputSource;

import org.objectweb.jonas.common.JProp;
import org.objectweb.jonas.common.Log;
import org.objectweb.jonas_ejb.deployment.api.BeanDesc;
import org.objectweb.jonas_ejb.deployment.api.MethodDesc;
import org.objectweb.jonas_ejb.lib.BeanNaming;
import org.objectweb.jonas.jmx.JmxService;
import org.objectweb.jonas.jmx.JonasObjectName;

import org.objectweb.jonas.service.AbsServiceImpl;
import org.objectweb.jonas.service.ServiceException;
import org.objectweb.jonas.service.ServiceManager;

import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;

import org.objectweb.jonas.security.realm.factory.JResource;
import org.objectweb.jonas.security.realm.factory.JResourceDS;
import org.objectweb.jonas.security.realm.factory.JResourceLDAP;
import org.objectweb.jonas.security.realm.factory.JResourceMemory;
import org.objectweb.jonas.security.realm.lib.HashHelper;

 
/**
 * Security Service implementation
 *
 * @author Jeff Mesnil (initial developper)
 * @author Philippe Coq
 * @author John Ellis   JOnAS 2.2 Add the possibility to plug
 *                                other implementations of the security service.
 * @author Joe Gittings JOnAS 2.2
 *                      Joe Gittings coding for method signature into MethodGuard.add
 */

public class JonasSecurityServiceImpl extends AbsServiceImpl implements SecurityService, JonasSecurityServiceImplMBean {

    /**
     * Logger which is used
     */
    private static Logger logger = null;
    
    /**
     * Associate a MethodGuard to a container
     */
    private Hashtable methodGuards = new Hashtable();

    /**
     * Associate a RoleGuard to a container
     */
    private Hashtable roleGuards = new Hashtable();

    /**
     * Security service configuration properties
     */
    public static final String CLASS = "jonas.service.security.class";

    /**
     * JONAS_ROOT
     */
    private static final String JONAS_ROOT = JProp.getInstallRoot();

    /**
     * JONAS_ROOT/xml
     */
    private String xmlDir = null;

    /**
     * Reference to a MBean server.
     */
    private MBeanServer mbeanServer = null;

    /**
     * Relative path of the realm configuration file
     */
    protected static final String CONFIG_FILE = "conf" + File.separator + "jonas-realm.xml";

    /**
     * DTD of the realm configuration file
     */
    protected static final String CONFIG_FILE_DTD = "jonas-realm_1_0.dtd";

    /**
     * List of Resource
     */
    private Hashtable jResources = new Hashtable();

   /**
    * Initial Context for Naming
    */
    private Context ictx = null;

    /**
     * Name of the realm resource to use
     */
    private String ejbRealm = null;

    /**
     * Key of the ejb realm in the jonas.properties file
     */
    private static final String EJB_REALM = "jonas.service.security.ejbrealm";

    /**
     * Xml Header
     */
    private static final String HEADER_XML = 
        "<?xml version='1.0' encoding='utf-8'?>\n"
        + "<!DOCTYPE jonas-realm PUBLIC\n"
        + "          \"-//ObjectWeb//DTD JOnAS realm 1.0//EN\"\n"
        + "          \"http://www.objectweb.org/jonas/dtds/jonas-realm_1_0.dtd\">\n";


    // -------------------------------------------------------------------
    // JOnAS Service Implementation
    // -------------------------------------------------------------------

    /**
     * Init the Service.
     * Configuration information is passed thru a Context object.
     * @param ctx the configuration of the Security service.
     * @throws ServiceException if the initialization failed.
     */
    public void doInit(Context ctx) throws ServiceException {
        if (logger == null) {
            logger = Log.getLogger(Log.JONAS_SECURITY_PREFIX);
        }

        //Get the initial Context
        try {
            ictx = new InitialContext();
        } catch (NamingException e) {
            logger.log(BasicLevel.ERROR, "Cannot create initial context during the mail service initializing");
            throw new ServiceException("Cannot create initial context during the mail service initializing", e);
        }

        // Get the JMX Server via JMX Service
        try {
            mbeanServer = ((JmxService) ServiceManager.getJmxService()).getJmxServer();
        } catch (ServiceException e) {
            // the JMX service may not be started
            mbeanServer = null;
        }
        
        // Get the realm name to use
        try {
            ejbRealm = (String) ctx.lookup(EJB_REALM);
        } catch (NamingException e) {
            logger.log(BasicLevel.ERROR,  "No realm is set in your jonas.properties file for the key :" + EJB_REALM + ". You won't be able to authenticate if you don't use a client running inside the web container.");
        }

        logger.log(BasicLevel.DEBUG,  "JonasSecurityService initialized");
    }

    /**
     * Add the Resource (memory, ldap, datasource,...)
     * @param jResource  an instance of JResource or subclasses (JResourceMemory, JResourceDS,...)
     * @throws Exception if the resource name already exists
     */
    public void addJResource(JResource jResource) throws Exception {
        if (jResources.get(jResource.getName()) != null) {
            throw new Exception("The resource name " + jResource.getName() + " already exists !");
        }

        jResources.put(jResource.getName(), jResource);
        // bind the resource into the jndi
        try {
            ictx.rebind(jResource.getName(), jResource);
            logger.log(BasicLevel.DEBUG, "jResource " + jResource.getName() + " bound into the registry.");
        } catch (NamingException e) {
            // nothing
        }

        try {
            // register security factory mbean 
            if (jResource instanceof JResourceMemory) {
                mbeanServer.registerMBean(jResource, JonasObjectName.securityMemoryFactory(jResource.getName()));
            } else if (jResource instanceof JResourceDS) {
                mbeanServer.registerMBean(jResource, JonasObjectName.securityDatasourceFactory(jResource.getName()));
            } else if (jResource instanceof JResourceLDAP) {
                mbeanServer.registerMBean(jResource, JonasObjectName.securityLdapFactory(jResource.getName()));
            } 
        } catch (ServiceException se) {
            // Jmx Service not available, do nothing
        } catch (Exception e) {
            logger.log(BasicLevel.ERROR, "Can not register the MBean for the resource " + jResource.getName() + " : " + e.getMessage());
            throw new ServiceException("Can not register the MBean for the resource " + jResource.getName() + " : " + e.getMessage());
        }

    }



    /**
     * Remove the Resource (memory, ldap, datasource,...)
     * @param resourceName name of the resource
     * @throws Exception if the resource name does not exist
     */
    public void removeJResource(String resourceName) throws Exception {
        JResource jResource = getJResource(resourceName);
        if (jResource == null) {
            throw new Exception("The resource name " + resourceName + " doesn't exist !");
        }
        jResources.remove(resourceName);

        // remove the resource into the jndi
        try {
            ictx.unbind(resourceName);
            logger.log(BasicLevel.DEBUG, "jResource " + resourceName + " remove from the registry.");
        } catch (NamingException e) {
            // nothing
        }

        try {
            // Remove mbeans of the resources
            jResource.removeMBeans();

            // register security factory mbean 
            if (jResource instanceof JResourceMemory) {
                mbeanServer.unregisterMBean(JonasObjectName.securityMemoryFactory(resourceName));
            } else if (jResource instanceof JResourceDS) {
                mbeanServer.unregisterMBean(JonasObjectName.securityDatasourceFactory(resourceName));
            } else if (jResource instanceof JResourceLDAP) {
                mbeanServer.unregisterMBean(JonasObjectName.securityLdapFactory(resourceName));
            } 
        } catch (ServiceException se) {
            // Jmx Service not available, do nothing
        } catch (Exception e) {
            logger.log(BasicLevel.ERROR, "Can not unregister the MBean for the resource " + resourceName + " : " + e.getMessage());
            throw new ServiceException("Can not unregister the MBean for the resource " + resourceName + " : " + e.getMessage());
        }

    }


    /**
     * Start the Service
     * Initialization of the service is already done.
     * @throws ServiceException if the stop failed.
     */
    public void doStart() throws ServiceException {
        try {
            // register security service mbean 
            mbeanServer.registerMBean(this, JonasObjectName.securityService());
        } catch (ServiceException se) {
            // Jmx Service not available, do nothing
        } catch (Exception e) {
            logger.log(BasicLevel.ERROR, "SecurityService: Cannot start the Security service:\n" + e);
            throw new ServiceException("SecurityService: Cannot start the Security service", e);
        }

        xmlDir = JONAS_ROOT + File.separator + "xml" +  File.separator;
        createRealm();

        //Which resource use for the mapping for java client
        setRoleMechanismRealm(ejbRealm);
    }


    /**
     * Stop the Service
     */
    public void doStop() {
        
        // Unregister MBean
        try {
            // Unregister Security Service MBean : SecurityServiceImplMBean 
            ((JmxService) ServiceManager.getJmxService()).getJmxServer().unregisterMBean(JonasObjectName.securityService());
        } catch (ServiceException se) {
            // Jmx Service not available, do nothing
        } catch (Exception e) {
            logger.log(BasicLevel.ERROR, "Cannot stop the security service:\n" + e);
            throw new ServiceException("Cannot stop the security service", e);
        }
    }
 
    /**
     * Return a resource by giving its name
     * @param name the wanted Resource
     * @return a JResouce
     */
    public JResource getJResource(String name) {
        return ((JResource) jResources.get(name));
    }


    /**
     * Parse the xml file and create all the JResource
     * @throws ServiceException if a JResource can't be created
     */
    private void createRealm() throws ServiceException {

        // Create the digester for the parsing of the jonas-realm.xml file.
        Digester realmDigester = createRealmDigester();
     
        // Execute the digester for the parsing of the jonas-realm.xml file.
        File configFile = null;
        try {
            configFile = getConfigFile();
        } catch (FileNotFoundException e) {
            logger.log(BasicLevel.ERROR, "Cannot find config file " + configFile);
            throw new ServiceException(e.getMessage(), e);
        }

        try {
            InputSource is = new InputSource("file://" + configFile.getAbsolutePath());
            FileInputStream fis = new FileInputStream(configFile);
            is.setByteStream(fis);
            realmDigester.push(this);
            realmDigester.parse(is);
            fis.close();
        } catch (org.xml.sax.SAXException saxe) {
            logger.log(BasicLevel.ERROR, "Error when parsing the XML of the file " + configFile);
            throw new ServiceException(saxe.getMessage(), saxe);
        } catch (Exception e) {
            logger.log(BasicLevel.ERROR, "Error when reading config file " + configFile);
            throw new ServiceException(e.getMessage(), e);
        }
    }

    
    /**
     * Return a File object representing the jonas-realm.xml configuration file.
     * @return a File object representing the jonas-realm.xml configuration file.
     * @throws FileNotFoundException if the configuration file is not found.
     */
    protected File getConfigFile() throws FileNotFoundException {
        String fileName = System.getProperty("jonas.base");
        fileName = fileName + File.separator + CONFIG_FILE;
        File file = new File(fileName);
        if (!file.exists()) {
            String err = "Can't find configuration file : " + fileName;
            throw new FileNotFoundException(err);
        }
        return (file);
    }


    // -------------------------------------------------------------------
    // SecurityService Implementation
    // -------------------------------------------------------------------

    /**
     * Returns the MethodGuard for the given Container
     * @param containerName The container name
     * @return Guard the MethodGuard for the given Container
     */
    public Guard getMethodGuard(String containerName) {
        MethodGuard mg = (MethodGuard) methodGuards.get(containerName);
        if (mg == null) {
            mg = new MethodGuard(containerName);
            methodGuards.put(containerName, mg);
        }
        return mg;
    }
    
    /**
     * Returns the RoleGuard for the given Container
     * @param containerName The container name
     * @return Guard the RoleGuard for the given Container
     */
    public Guard getRoleGuard(String containerName) {
        RoleGuard rg = (RoleGuard) roleGuards.get(containerName);
        if (rg == null) {
            rg = new RoleGuard(containerName);
            roleGuards.put(containerName, rg);
        }
        return rg;
    }

    /**
     * Set the security attributes for the bean being deployed.
     * @param containerName The container the bean is deployed in
     * @param bd The bean descriptor
     */
    public void deployBean(String containerName, BeanDesc bd) {
        setMethodRoleMapping(containerName, bd);
        setRoleRefMapping(containerName, bd);
    }
    
    // -------------------------------------------------------------------
    // Other methods
    // -------------------------------------------------------------------

    /**
     * Set the role mapping for the method for a container
     * @param containerName name of the container
     * @param bd the bean descriptor
     */
    private void setMethodRoleMapping(String containerName, BeanDesc bd) {
        MethodGuard methodGuard = (MethodGuard) methodGuards.get(containerName);
        // deployment of the method guards
        for (Iterator i = bd.getMethodDescIterator(); i.hasNext();) {
            MethodDesc md = (MethodDesc) i.next();
            methodGuard.add(BeanNaming.getSignature(bd.getEjbName(), md.getMethod()), md.getRoleName());
        }
        methodGuards.put(containerName, methodGuard);
    }
 
    /**
     * Set the role mapping for the role-ref for a container
     * @param containerName name of the container
     * @param bd the bean descriptor
     */
    private  void setRoleRefMapping(String containerName, BeanDesc bd) {
        // deployment of the security-role-ref  mapping
        //RoleMechanism rm = new RoleMechanism();
        RoleGuard roleGuard = (RoleGuard) roleGuards.get(containerName);
        String[] roles = bd.getSecurityRoleNames();
        for (int i = 0; i < roles.length; i++) {
            roleGuard.add(roles[i], bd.getRoleLink(roles[i]));
        }
        roleGuards.put(containerName, roleGuard);
    }   

    /**
     * Set the realm to use for the association between a principal and the roles
     * @param realm the name of the realm to use
     * @throws ServiceException if the realm can't be associated to the role mechanism
     */
    private void setRoleMechanismRealm(String realm) throws ServiceException {
        RoleMechanism rm = RoleMechanism.getInstance();
        try {
            rm.setRealmResource(realm);
        } catch (Exception e) {
            throw new ServiceException(e.getMessage());
        }
    }


    /**
     * Create and configure the Digester that will be used for the xml
     * parsing of the JOnAS realm file.
     * @return Digester the digester containing the rules for the xml parsing
     * of the JOnAS realm file (jonas-realm.xml)
     */
    protected Digester createRealmDigester() {

        // Initialize the digester
        Digester digester = new Digester();
        File realmDtd = new File(xmlDir + CONFIG_FILE_DTD);
        URL realmDtdURL = null;
        boolean validating = true;
        try {
            realmDtdURL = realmDtd.toURL();
            digester.register("-//ObjectWeb//DTD JOnAS realm 1.0//EN", realmDtdURL.toExternalForm());
        } catch (MalformedURLException e) {
            // Can't locate URL for DTD, no validation
            logger.log(BasicLevel.INFO, "Can not locate URL for DTD validation, no validation will be done." + e.getMessage());
            validating = false;
        }

        digester.setValidating(validating);
        digester.setErrorHandler(new JErrorHandler());
        addMemoryRules(digester, "jonas-realm/jonas-memoryrealm/");
        addDSRules(digester, "jonas-realm/jonas-dsrealm/");
        addLDAPRules(digester, "jonas-realm/jonas-ldaprealm/");
        return (digester);
    }


    /**
     * Add the rules for the memory in the jonas-realm.xml file
     * @param digester the digester to use
     * @param prefix prefix of the rules
     */
    protected void addMemoryRules(Digester digester, String prefix) {

        // Configure the actions we will be using
        digester.addObjectCreate(prefix + "memoryrealm",
                                 "org.objectweb.jonas.security.realm.factory.JResourceMemory");

        digester.addSetProperties(prefix + "memoryrealm");
        digester.addSetNext(prefix + "memoryrealm",
                            "addJResource",
                            "org.objectweb.jonas.security.realm.factory.JResource");

        digester.addObjectCreate(prefix + "memoryrealm/roles/role",
                                 "org.objectweb.jonas.security.realm.principals.Role");
        digester.addSetProperties(prefix + "memoryrealm/roles/role");
        digester.addSetNext("memoryrealm/roles/role",
                            "addRole",
                            "org.objectweb.jonas.security.realm.principals.Role");

        digester.addObjectCreate(prefix + "memoryrealm/groups/group",
                                 "org.objectweb.jonas.security.realm.principals.Group");
        digester.addSetProperties(prefix + "memoryrealm/groups/group");
        digester.addSetNext(prefix + "memoryrealm/groups/group",
                            "addGroup",
                            "org.objectweb.jonas.security.realm.principals.Group");

        digester.addObjectCreate(prefix + "memoryrealm/users/user",
                                 "org.objectweb.jonas.security.realm.principals.User");
        digester.addSetProperties(prefix + "memoryrealm/users/user");
        digester.addSetNext(prefix + "memoryrealm/users/user",
                            "addUser",
                            "org.objectweb.jonas.security.realm.principals.User");
    }


    /**
     * Add the rules for the datasource in the jonas-realm.xml file
     * @param digester the digester to use
     * @param prefix prefix of the rules
     */
    protected void addDSRules(Digester digester, String prefix) {

        // Configure the actions we will be using
        digester.addObjectCreate(prefix + "dsrealm",
                                 "org.objectweb.jonas.security.realm.factory.JResourceDS");

        digester.addSetProperties(prefix + "dsrealm");
        digester.addSetNext(prefix + "dsrealm",
                            "addJResource",
                            "org.objectweb.jonas.security.realm.factory.JResource");
    }

    /**
     * Add the rules for the ldap in the jonas-realm.xml file
     * @param digester the digester to use
     * @param prefix prefix of the rules
     */
    protected void addLDAPRules(Digester digester, String prefix) {

        // Configure the actions we will be using
        digester.addObjectCreate(prefix + "ldaprealm",
                                 "org.objectweb.jonas.security.realm.factory.JResourceLDAP");

        digester.addSetProperties(prefix + "ldaprealm");
        digester.addSetNext(prefix + "ldaprealm",
                            "addJResource",
                            "org.objectweb.jonas.security.realm.factory.JResource");
    }


    /**
     * String representation of the LDAP realm
     * @return the xml representation of the LDAP realm
     */
    public String toXML() {

        // Required values

        StringBuffer xml = new StringBuffer(HEADER_XML);
        xml.append("<!--\n");
        xml.append(" - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -\n");
        xml.append(" - Define a jonas-realm.xml file for JOnAS realms\n");
        xml.append(" - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -\n");
        xml.append(" -->\n");

        xml.append("<jonas-realm>\n");

        // Memory realms
        xml.append("  <!--\n");
        xml.append("       -=  MEMORY REALM =-\n");
        xml.append("       Define the users, groups and roles\n");
        xml.append("  -->\n");
        xml.append("  <jonas-memoryrealm>\n");
        for (Enumeration e = jResources.elements(); e.hasMoreElements(); ) {
                Object o = e.nextElement();
                if (o instanceof JResourceMemory) {
                    xml.append(o.toString());
                    xml.append("\n");
                }
        }
        xml.append("  </jonas-memoryrealm>\n");


        // Datasource realms
        xml.append("  <!--\n");
        xml.append("       -=  DATASOURCE REALM =-\n");
        xml.append("       Define the configuration to use datas from a datasource\n");
        xml.append("  -->\n");
        xml.append("  <jonas-dsrealm>\n");
        for (Enumeration e = jResources.elements(); e.hasMoreElements(); ) {
                Object o = e.nextElement();
                if (o instanceof JResourceDS) {
                    xml.append(o.toString());
                    xml.append("\n");
                }
        }
        xml.append("  </jonas-dsrealm>\n");

        // Ldap realms
        xml.append("  <!--\n");
        xml.append("       -=  LDAP REALM =-\n");
        xml.append("       Define the configuration to use datas from an ldap server\n");
        xml.append("  -->\n");
        xml.append("  <jonas-ldaprealm>\n");
        for (Enumeration e = jResources.elements(); e.hasMoreElements(); ) {
                Object o = e.nextElement();
                if (o instanceof JResourceLDAP) {
                    xml.append(o.toString());
                    xml.append("\n");
                }
        }
        xml.append("  </jonas-ldaprealm>\n");



        xml.append("</jonas-realm>\n");


        return xml.toString();
    }


    /**
     * Encrypt a string with an algorithm
     * @param string the string to encode
     * @param algo algorithm to apply on the given string
     * @return the encoded string
     * @throws NoSuchAlgorithmException One reason could be a bad algorithm
     */
    public String encryptPassword(String string, String algo) throws NoSuchAlgorithmException {
        String encrypt = HashHelper.hashPassword(string, algo);
        // Prefix with algorithm
        return "{" + algo.toUpperCase() + "}" + encrypt;
    }


    /**
     * Check if the given algorithm is a valid algorithm
     * @param algo algorithm to apply on the given string
     * @return true if it is a valid algorithm
     */
    public boolean isValidAlgorithm(String algo) {
        boolean b = true;
        try {
            encryptPassword("test", algo);
        } catch (NoSuchAlgorithmException nsae) {
            b = false;
        }
        return b;
    }

    /**
     * Add JResources with a given xml configuration
     * @param xml xml representation of the resources to add
     * @throws Exception if the resources can't be added
     */
    public void addResources(String xml) throws Exception {

        // Create the digester for the parsing of the jonas-realm.xml file.
        Digester realmDigester = createRealmDigester();
     
        try {
            StringReader sr = new StringReader(xml);
            realmDigester.push(this);
            realmDigester.parse(sr);
            sr.close();
        } catch (org.xml.sax.SAXException saxe) {
            logger.log(BasicLevel.ERROR, "Error when parsing the XML of the file " + xml);
            throw new Exception(saxe.getMessage());
        } catch (Exception e) {
            logger.log(BasicLevel.ERROR, "Error when reading config file from the xml file " + xml);
            throw new Exception("Error when reading config file from the xml file " + xml + " : " + e.getMessage());
        }
        
    }


    /**
     * Add a Memory resource
     * @param name the name of the JResourceMemory to create
     * @throws Exception if the resource can't be added
     */
    public void addJResourceMemory(String name) throws Exception {

        // Build a new JResourceMemory
        JResourceMemory jResourceMemory = new JResourceMemory();
        jResourceMemory.setName(name);

        // Build xml
        StringBuffer xml = new StringBuffer(HEADER_XML);
        xml.append("<jonas-realm>");
        xml.append("<jonas-memoryrealm>");
        xml.append(jResourceMemory.toXML());
        xml.append("</jonas-memoryrealm>");
        xml.append("</jonas-realm>");
           
        // Add the resource
        addResources(xml.toString());
        
    }


    /**
     * Add a DS resource
     * @param name the name of the JResourceDS to create
     * @param dsName Name of the datasource resource to use.
     * @param userTable Name of table which have the username/password
     * @param userTableUsernameCol Column of the username of the user table
     * @param userTablePasswordCol Column of the password of the user table
     * @param roleTable Name of table which have the username/role
     * @param roleTableUsernameCol Column of the username of the role table
     * @param roleTableRolenameCol Column of the role of the role table
     * @param algorithm Default algorithm. If specified, the default is not 'clear' password
     * @throws Exception if the resource can't be added
     */
    public void addJResourceDS(String name,
                               String dsName,
                               String userTable,
                               String userTableUsernameCol,
                               String userTablePasswordCol,
                               String roleTable,
                               String roleTableUsernameCol,
                               String roleTableRolenameCol,
                               String algorithm) throws Exception {

        // Build a new JResourceDS
        JResourceDS jResourceDS = new JResourceDS();
        jResourceDS.setName(name);
        jResourceDS.setDsName(dsName);
        jResourceDS.setUserTable(userTable);
        jResourceDS.setUserTableUsernameCol(userTableUsernameCol);
        jResourceDS.setUserTablePasswordCol(userTablePasswordCol);
        jResourceDS.setRoleTable(roleTable);
        jResourceDS.setRoleTableUsernameCol(roleTableUsernameCol);
        jResourceDS.setRoleTableRolenameCol(roleTableRolenameCol);
        jResourceDS.setAlgorithm(algorithm);

        // Build xml
        StringBuffer xml = new StringBuffer(HEADER_XML);
        xml.append("<jonas-realm>");
        xml.append("<jonas-dsrealm>");
        xml.append(jResourceDS.toXML());
        xml.append("</jonas-dsrealm>");
        xml.append("</jonas-realm>");
           
        // Add the resource
        addResources(xml.toString());
        
    }


    /**
     * Add a LDAP resource
     * @param name the name of the JResourceLDAP to create
     * @param initialContextFactory Initial context factory for the LDAp server
     * @param providerUrl Url of the ldap server
     * @param securityAuthentication Type of the authentication used during the authentication to the LDAP server
     * @param securityPrincipal DN of the Principal(username). He can retrieve the information from the user
     * @param securityCredentials Credential(password) of the principal
     * @param securityProtocol Constant that holds the name of the environment property for specifying the security protocol to use.
     * @param language Constant that holds the name of the environment property for specifying the preferred language to use with the service.
     * @param referral Constant that holds the name of the environment property for specifying how referrals encountered by the service provider are to be processed.
     * @param stateFactories Constant that holds the name of the environment property for specifying the list of state factories to use.
     * @param authenticationMode Mode for validate the authentication (BIND_AUTHENTICATION_MODE or COMPARE_AUTHENTICATION_MODE)
     * @param userPasswordAttribute Attribute in order to get the password from the ldap server
     * @param userRolesAttribute Attribute in order to get the user role from the ldap server
     * @param roleNameAttribute Attribute for the role name when performing a lookup on a role
     * @param baseDN DN used for the lookup
     * @param userDN DN used when searching the user DN. Override the baseDN if it is defined
     * @param userSearchFilter Filter used when searching the user
     * @param roleDN DN used when searching the role DN. Override the baseDN if it is defined
     * @param roleSearchFilter Filter used when searching the role
     * @param algorithm Default algorithm. If specified, the default is not 'clear' password
     * @throws Exception if the resource can't be added
     */
    public void addJResourceLDAP(String name,
                                 String initialContextFactory,
                                 String providerUrl,
                                 String securityAuthentication,
                                 String securityPrincipal,
                                 String securityCredentials,
                                 String securityProtocol,
                                 String language,
                                 String referral,
                                 String stateFactories,
                                 String authenticationMode,
                                 String userPasswordAttribute,
                                 String userRolesAttribute,
                                 String roleNameAttribute,
                                 String baseDN,
                                 String userDN,
                                 String userSearchFilter,
                                 String roleDN,
                                 String roleSearchFilter,
                                 String algorithm) throws Exception {

        // Build a new JResourceLDAP
        JResourceLDAP jResourceLDAP = new JResourceLDAP();
        jResourceLDAP.setName(name);
        jResourceLDAP.setInitialContextFactory(initialContextFactory);
        jResourceLDAP.setProviderUrl(providerUrl);
        jResourceLDAP.setSecurityAuthentication(securityAuthentication);
        jResourceLDAP.setSecurityPrincipal(securityPrincipal);
        jResourceLDAP.setSecurityCredentials(securityCredentials);
        jResourceLDAP.setSecurityProtocol(securityProtocol);
        jResourceLDAP.setLanguage(language);
        jResourceLDAP.setReferral(referral);
        jResourceLDAP.setStateFactories(stateFactories);
        jResourceLDAP.setAuthenticationMode(authenticationMode);
        jResourceLDAP.setUserPasswordAttribute(userPasswordAttribute);
        jResourceLDAP.setUserRolesAttribute(userRolesAttribute);
        jResourceLDAP.setRoleNameAttribute(roleNameAttribute);
        jResourceLDAP.setBaseDN(baseDN);
        jResourceLDAP.setUserDN(userDN);
        jResourceLDAP.setUserSearchFilter(userSearchFilter);
        jResourceLDAP.setRoleDN(roleDN);
        jResourceLDAP.setRoleSearchFilter(roleSearchFilter);
        jResourceLDAP.setAlgorithm(algorithm);

        // Build xml
        StringBuffer xml = new StringBuffer(HEADER_XML);
        xml.append("<jonas-realm>");
        xml.append("<jonas-ldaprealm>");
        xml.append(jResourceLDAP.toXML());
        xml.append("</jonas-ldaprealm>");
        xml.append("</jonas-realm>");
           
        // Add the resource
        addResources(xml.toString());
        
    }
}
