/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Greg Wilkins for the HashUserRealm
 * Copyright (c) 1996 Mort Bay Consulting Pty. Ltd. All rights reserved.
 * --------------------------------------------------------------------------
 * $Id: JRealmJAASJetty42.java,v 1.1 2003/07/18 11:14:52 benoitf Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.security.realm;

import java.security.Principal;
import java.security.acl.Group;
import java.security.cert.X509Certificate;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Iterator;

import javax.security.auth.Subject;
import javax.security.auth.login.AccountExpiredException;
import javax.security.auth.login.CredentialExpiredException;
import javax.security.auth.login.FailedLoginException;
import javax.security.auth.login.LoginContext;
import javax.security.auth.login.LoginException;

import org.mortbay.http.HttpRequest;
import org.mortbay.http.UserPrincipal;
import org.mortbay.http.UserRealm;

//import monolog
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

import org.objectweb.security.context.SecurityContext;
import org.objectweb.security.context.SecurityCurrent;

import org.objectweb.jonas.common.Log;
import org.objectweb.jonas.security.auth.callback.NoInputCallbackHandler;



/**
 * <p>Implementation of a JAAS Realm. (by a wrapper)
 * It uses the entry JAAS_CONFIG_NAME from the JAAS config file
 *
 * @author Greg Wilkins for the HashUserRealm
 * @author Alexandre Thaveau (JAAS support with JOnAS)
 * @author Marc-Antoine Bourgeot (JAAS support with JOnAS)
 */
public class JRealmJAASJetty42 implements UserRealm {

    /**
     * Name used in the JAAS config file
     */
    private static final String JAAS_CONFIG_NAME = "jetty";


    /**
     * The logger used in JOnAS
     */
    private static Logger logger = null;

    /**
     * Name of this realm
     */
    private String name;


    /**
     * Constructor
     * @throws Exception if the resource can't be retrieved
     */
    public JRealmJAASJetty42() throws Exception {

        if (logger == null) {
            logger = Log.getLogger(Log.JONAS_SECURITY_PREFIX);
        }

    }

    /**
     * Constructor
     * @param name name of the realm
     * @throws Exception if the resource can't be retrieved
     */
    public JRealmJAASJetty42(String name) throws Exception {
        this();
        this.name = name;
    }
     
    /**
     * Authenticate a user with a specific username and credentials
     * @param username name of the user
     * @param credentials credential of the user
     * @param request httprequest
     * @return a Jetty principal
     */
    public UserPrincipal authenticate(String username, Object credentials, HttpRequest request) {
        // No authentication can be made with a null username
        if (username == null) {
            return null;
        }

        NoInputCallbackHandler noInputCH = null;
        LoginContext loginContext = null;
        if (credentials instanceof X509Certificate[]) {
            // Format the DN as a special username
            String headerCertificate = "##DN##";
            X509Certificate [] certs = (X509Certificate[]) credentials;
            
            username = certs[0].getSubjectDN().getName();
            String usernameCert = headerCertificate.concat(username.replace('=', '#').replace(',', '%').replace(' ', '$'));
            // Fill the callback handler for the login module with DN and certificate
            noInputCH = new NoInputCallbackHandler(usernameCert, "", certs[0]);
        } else {
            // Fill the callback handler for the login module with username and password
            noInputCH = new NoInputCallbackHandler(username, (String) credentials, null);
        }
 
        //Establish a LoginContext to use for authentication
        try {
            loginContext = new LoginContext(JAAS_CONFIG_NAME, noInputCH);
        } catch (LoginException e) {
            logger.log(BasicLevel.WARN, "loginException : unable to create a LoginContext for : " + username);
            return null;
        }

        // Negotiate a login via this LoginContext
        Subject subject = null;
        try {
            loginContext.login();
            subject = loginContext.getSubject();
            if (subject == null) {
                logger.log(BasicLevel.ERROR, "failedLogin for user :" + username);
                return null;
            }
        } catch (AccountExpiredException e) {
            if (logger.isLoggable(BasicLevel.ERROR)) {
                logger.log(BasicLevel.ERROR, "accountExpired for user :" + username);
            }
            return null;
        } catch (CredentialExpiredException e) {
            if (logger.isLoggable(BasicLevel.ERROR)) {
                logger.log(BasicLevel.ERROR, "credentialExpired for user :" + username);
            }
            return null;
        } catch (FailedLoginException e) {
            if (logger.isLoggable(BasicLevel.ERROR)) {
                logger.log(BasicLevel.ERROR, "failedLogin for user :" + username);
            }
            return null;
        } catch (LoginException e) {
            if (logger.isLoggable(BasicLevel.ERROR)) {
                logger.log(BasicLevel.ERROR, "loginException for user :" + username);
            }
            return null;
        }
  
  
        // Get credentials iterators from the subject (first found)
        Iterator credentialsIterator = subject.getPrivateCredentials().iterator();
        String credential = (String) credentialsIterator.next();

        // Retrieve first principal name found (without groups)
        Iterator iterator = subject.getPrincipals(Principal.class).iterator();
        String userName = null;
        while (iterator.hasNext() && (userName == null)) {
            Principal principal = (Principal) iterator.next();
            if (!(principal instanceof Group)) {
               userName = principal.getName();
            }
        }

        // No name --> error
        if (userName == null) {
            logger.log(BasicLevel.ERROR, "No Username found in the subject");
            return null;
        }

        // Retrieve all roles of the user (Roles are members of the Group class)
        iterator = subject.getPrincipals(Group.class).iterator();
        ArrayList roles = new ArrayList();
        while (iterator.hasNext()) {
            Group group = (Group) iterator.next();
            Enumeration e = group.members();
            while (e.hasMoreElements()) {
                Principal p = (Principal) e.nextElement();
                roles.add(p.getName());
            }
        }

        // Create a JettyPrincipal for Jetty
        JettyPrincipal principal = new JettyPrincipal(userName, credential, roles);
        
        // Register the subject in the security context
        //SecurityContext ctx = new SecurityContext(subject);
        SecurityContext ctx = new SecurityContext(userName, roles);
        SecurityCurrent current = SecurityCurrent.getCurrent();
        current.setSecurityContext(ctx);
        return principal;
    }

    /**
     * Disassociate a user
     * Not implemented
     * @param user the given user
     */
    public void disassociate(UserPrincipal user) {
        // Not implemented
    }


    /** 
     * @return The realm name. 
     */
    public String getName() {
        return name;
    }

    /**
     * Pop a role to a user
     * Not implemented
     * @param user the given user
     * @return the new principal
     */
    public UserPrincipal popRole(UserPrincipal user) {
        return user;
    }
    
    /**
     * Push a role to a user
     * Not implemented
     * @param user the given user
     * @param role the role to push
     * @return the new principal
     */
    public UserPrincipal pushRole(UserPrincipal user, String role) {
        return user;
    }

    /**
     * Define a principal which can be used by Jetty
     */
    private class JettyPrincipal implements UserPrincipal {
        
        /**
         * Credential of the principal
         */
        private Object credential;
       
        /**
         * Name of the principal
         */
        private String name;

        /**
         * Roles for this principal
         */
        private ArrayList roles;

        /**
         * Constructor
         * @param name name of this principal
         * @param credential credential of the principal
         * @param roles roles for this principal
         */
        public JettyPrincipal(String name, Object credential, ArrayList roles) {
            this.name = name;
            this.credential = credential;
            this.roles = roles;
        }
  
        /**
         * Get the name of this principal
         * @return the name of this principal
         */
        public String getName() {
            return name;
        }

        /**
         * This user is authenticated ?
         * @return true if it is authenticated
         */
        public boolean isAuthenticated() {
            return true;
        }
        
        /**
         * Check if the given role is in the user's roles
         * @param role the given role
         * @return true if the role is in the user's roles
         */
        public boolean isUserInRole(String role) {
            if (roles == null) {
                return (false);
            }
            return roles.contains(role);
        }
       
    }

}



