/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: JonasAdminJmx.java,v 1.22 2003/09/19 09:23:50 antonma Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.webapp.jonasadmin;

import java.io.File;
import java.io.FileInputStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.ArrayList;
import java.util.Properties;
import java.util.Set;

// JMX
import javax.management.ObjectName;

// naming imports
import javax.naming.NamingEnumeration;
import javax.naming.NameClassPair;
import javax.naming.NamingException;
import javax.naming.Context;

import org.objectweb.jonas.common.JProp;
import org.objectweb.jonas.jmx.JonasManagementRepr;
import org.objectweb.jonas.jmx.JonasObjectName;
import org.objectweb.jonas.webapp.jonasadmin.logging.LoggerItem;
import org.objectweb.jonas.webapp.jonasadmin.common.ItemByNameComparator;

import org.apache.struts.util.MessageResources;

/**
 * Helper class supporting manipulation of MBeans (accessing MBeans).
 * @author Michel-Ange ANTON<p>
 * Contributors: Adriana Danes
 */
public class JonasAdminJmx {

// ----------------------------------------------------------- Constants

    private static Object s_Synchro = new Object();

// ----------------------------------------------------------- Constructors

    /**
     * Protected constructor to prevent instantiation.
     */
    protected JonasAdminJmx() {
    }

// --------------------------------------------------------- Public Methods

    /**
     * Replace any occurrence of the specified placeholder in the specified
     * template string with the specified replacement value.
     *
     * @param template Pattern string possibly containing the placeholder
     * @param placeholder Placeholder expression to be replaced
     * @param value Replacement value for the placeholder
     * @return A complete string
     */
    public static String replace(String template, String placeholder, String value) {
        if (template == null) {
            return (null);
        }
        if ((placeholder == null) || (value == null)) {
            return (template);
        }
        String sRet = new String(template);
        while (true) {
            int index = sRet.indexOf(placeholder);
            if (index < 0) {
                break;
            }
            StringBuffer temp = new StringBuffer(sRet.substring(0, index));
            temp.append(value);
            temp.append(sRet.substring(index + placeholder.length()));
            sRet = temp.toString();
        }
        return (sRet);

    }

    /**
     * Return a list of JonasServerItem connected to the same registry.
     *
     * @return List of JonasServerItem
     * @throws Exception
     */
    public static ArrayList getListRemoteJonasServerItem()
        throws Exception {
        synchronized (s_Synchro) {
            ArrayList oServers = new ArrayList();
            // Save current RmiConnector
            String sCurrentRMIConnectorName = JonasManagementRepr.getCurrentRMIConnectorName();

            // Find all RmiConnectors
            Set setRmiConnector = JonasManagementRepr.getRMIConnectorsNames();

            if (setRmiConnector.size() > 0) {
                Iterator oIt = setRmiConnector.iterator();
                while (oIt.hasNext()) {
                    JonasServerItem oItem = new JonasServerItem();
                    oItem.setNameRmiConnector(oIt.next().toString());
                    // Find Jonas server name
                    JonasManagementRepr.setCurrentRMIConnectorName(oItem.getNameRmiConnector());
                    oItem.setNameServer(getJonasServerName());
                    oServers.add(oItem);
                }
                try {
                    // Restore current RmiConnector
                    JonasManagementRepr.setCurrentRMIConnectorName(sCurrentRMIConnectorName);
                }
                catch (Exception e) {
                    // May be the current server is shutdown
                    if (oServers.size() > 0) {
                        JonasServerItem oItem = (JonasServerItem) oServers.get(0);
                        JonasManagementRepr.setCurrentRMIConnectorName(oItem.getNameRmiConnector());
                    }
                }
            }
            return oServers;
        }
    }

    /**
     * Return the current Jonas server name.
     *
     * @return The name
     * @throws Exception
     */
    public static String getJonasServerName()
        throws Exception {
        synchronized (s_Synchro) {
            ObjectName onSearch = JonasObjectName.ejbServer();
            return (String) (JonasManagementRepr.getAttribute(onSearch, "JonasName"));
        }
    }

    /**
     * Verify if the Mbean gotten by the query in the current MbeanServer exists.
     *
     * @param p_On Query Mbean name to search
     * @return true if MBean exists
     * @throws Exception
     */
    public static boolean hasMBeanName(ObjectName p_On)
        throws Exception {
        synchronized (s_Synchro) {
            ArrayList al = new ArrayList();
            Iterator itNames = JonasManagementRepr.queryNames(p_On).iterator();
            if (itNames.hasNext()) {
                return true;
            }
            return false;
        }
    }

    /**
     * Return the first Mbean name gotten by the query in the current MbeanServer.
     *
     * @param p_On Query Mbean name to search
     * @return The first MBean name or null if not found
     * @throws Exception
     */
    public static String getFirstMBeanName(ObjectName p_On)
        throws Exception {
        synchronized (s_Synchro) {
            ArrayList al = new ArrayList();
            Iterator itNames = JonasManagementRepr.queryNames(p_On).iterator();
            if (itNames.hasNext()) {
                return itNames.next().toString();
            }
            return null;
        }
    }

    /**
     * Return the list of Mbean name gotten by the query in the current MbeanServer.
     *
     * @param p_On Query Mbean name to search
     * @return A list of string Mbean name
     * @throws Exception
     */
    public static List getListMBeanName(ObjectName p_On)
        throws Exception {
        synchronized (s_Synchro) {
            ArrayList al = new ArrayList();
            Iterator itNames = JonasManagementRepr.queryNames(p_On).iterator();
            while (itNames.hasNext()) {
                al.add(itNames.next().toString());
            }
            Collections.sort(al);
            return al;
        }
    }

    /**
     * Return the first <code>ObjectName</code> Mbean gotten by the query
     * in the current MbeanServer.
     *
     * @param p_On Query Mbean name to search
     * @return The first <code>ObjectName</code> or null if not found
     * @throws Exception
     */
    public static ObjectName getFirstMbean(ObjectName p_On)
        throws Exception {
        synchronized (s_Synchro) {
            ArrayList al = new ArrayList();
            Iterator itNames = JonasManagementRepr.queryNames(p_On).iterator();
            if (itNames.hasNext()) {
                return (ObjectName) itNames.next();
            }
            return null;
        }
    }

    /**
     * Return the list of <code>ObjectName</code> Mbean gotten by the query
     * in the current MbeanServer.
     *
     * @param p_On Query Mbeans to search
     * @return The list of <code>ObjectName</code>
     * @throws Exception
     */
    public static List getListMbean(ObjectName p_On)
        throws Exception {
        synchronized (s_Synchro) {
            ArrayList al = new ArrayList();
            Iterator itNames = JonasManagementRepr.queryNames(p_On).iterator();
            while (itNames.hasNext()) {
                al.add(itNames.next());
            }
            return al;
        }
    }

    /**
     * Extract a value of a Mbean name.
     *
     * @param p_Name Name of property
     * @param p_MBeanName Complete Mbean name
     * @return The value or null if not found
     */
    public static String extractValueMbeanName(String p_Name, String p_MBeanName) {
        String sValue = null;
        try {
            String sSearch = p_Name.trim() + "=";
            int iPos = p_MBeanName.indexOf(sSearch);
            if (iPos > -1) {
                sValue = p_MBeanName.substring(iPos + sSearch.length());
                iPos = sValue.indexOf(",");
                if (iPos > -1) {
                    sValue = sValue.substring(0, iPos);
                }
            }
        }
        catch (NullPointerException e) {
            // none
            sValue = null;
        }
        return sValue;
    }

    /**
     * Extract the filename of complete path.
     *
     * @param p_Path Complete path (directory and filename)
     * @return The filename or null
     */
    public static String extractFilename(String p_Path) {
        String sFilename = null;
        try {
            int iPosSeparator = p_Path.lastIndexOf("/");
            if (iPosSeparator < 0) {
                iPosSeparator = p_Path.lastIndexOf("\\");
                if (iPosSeparator < 0) {
                    sFilename = new String(p_Path);
                }
                else {
                    sFilename = p_Path.substring(iPosSeparator + 1);
                }
            }
            else {
                sFilename = p_Path.substring(iPosSeparator + 1);
            }
            if (sFilename.length() > 0) {
                int iPos_1 = p_Path.indexOf(DIR_AUTOLOAD + "/" + sFilename);
                int iPos_2 = p_Path.indexOf(DIR_AUTOLOAD + "\\" + sFilename);
                if (iPos_1 > -1) {
                    sFilename = DIR_AUTOLOAD + "/" + sFilename;
                }
                else if (iPos_2 > -1) {
                    sFilename = DIR_AUTOLOAD + "\\" + sFilename;
                }
            }
            else {
                sFilename = null;
            }
        }
        catch (NullPointerException e) {
            // none action
            sFilename = null;
        }
        return sFilename;
    }

    /**
     *
     */
    private static String DIR_RARS = "rars";
    private static String DIR_AUTOLOAD = "autoload";
    private static String DIR_CONF = "conf";

    private static ArrayList getFilenames(String p_Directory, String p_Extension) {
        ArrayList al = new ArrayList();
        String sExt = "." + p_Extension.toLowerCase();
        String sFile;

        File oFile = new File(p_Directory);
        String[] asFiles = oFile.list();
        int iPos;
        if (asFiles != null) {
            for (int i = 0; i < asFiles.length; i++) {
                oFile = new File(p_Directory, asFiles[i]);
                if (oFile.isFile() == true) {
                    sFile = oFile.getName().toLowerCase();
                    iPos = sFile.lastIndexOf(sExt);
                    if (iPos > -1) {
                        if (iPos == (sFile.length() - sExt.length())) {
                            al.add(oFile.getName());
                        }
                    }
                }
            }
        }
        Collections.sort(al);
        return al;
    }

    private static ArrayList getDirectories(String p_Directory) {
        ArrayList al = new ArrayList();
        String sFile;

        File oFile = new File(p_Directory);
        String[] asFiles = oFile.list();

        if (asFiles != null) {
            for (int i = 0; i < asFiles.length; i++) {
                oFile = new File(p_Directory, asFiles[i]);
                if (oFile.isDirectory() == true) {
                    al.add(oFile.getName());
                }
            }
        }
        Collections.sort(al);
        return al;
    }

    private static void appendDirectory(ArrayList p_List, String p_Dir) {
        String sDir = p_Dir + "/";
        for (int i = 0; i < p_List.size(); i++) {
            p_List.set(i, sDir + p_List.get(i));
        }
    }

    /**
     * Return the list of JAR filename ready to deploy in the current server.
     *
     * @return The list of JAR filename.
     * @throws Exception
     */
    public static ArrayList getJarFilesDeployable()
        throws Exception {
        ObjectName on = JonasObjectName.ejbService();
        ArrayList al = (ArrayList) JonasManagementRepr.getAttribute(on, "DeployableJars");
        String sDir = (String) JonasManagementRepr.getAttribute(on, "EjbjarsDirectory");

        String sEarDir = null;
        try {
            on = JonasObjectName.earService();
            sEarDir = (String) JonasManagementRepr.getAttribute(on, "AppsDirectory");
        }
        catch (Exception e) {
            // nothing, Ear service not found
        }
        return prepareContainersToDisplay(al, sDir, sEarDir);
    }

    /**
     * Return the list of RAR filename ready to deploy in the current server.
     *
     * @return The list of RAR filename.
     * @throws Exception
     */
    public static ArrayList getRarFilesDeployable()
        throws Exception {
        ObjectName on = JonasObjectName.resourceService();
        ArrayList al = (ArrayList) JonasManagementRepr.getAttribute(on, "DeployableRars");
        String sDir = (String) JonasManagementRepr.getAttribute(on, "RarsDirectory");

        String sEarDir = null;
        try {
            on = JonasObjectName.earService();
            sEarDir = (String) JonasManagementRepr.getAttribute(on, "AppsDirectory");
        }
        catch (Exception e) {
            // nothing, Ear service not found
        }
        return prepareContainersToDisplay(al, sDir, sEarDir);
    }

    /**
     * Return the list of EAR filename ready to deploy in the current server.
     *
     * @return The list of EAR filename.
     * @throws Exception
     */
    public static ArrayList getEarFilesDeployable()
        throws Exception {
        ObjectName on = JonasObjectName.earService();
        ArrayList al = (ArrayList) JonasManagementRepr.getAttribute(on, "DeployableEars");
        String sDir = (String) JonasManagementRepr.getAttribute(on, "AppsDirectory");
        return prepareContainersToDisplay(al, sDir, null);
    }

    /**
     * Return the list of WAR filename ready to deploy in the current server.
     *
     * @return The list of WAR filename.
     * @throws Exception
     */
    public static ArrayList getWarFilesDeployable()
        throws Exception {
        ObjectName on = JonasObjectName.webContainerService();
        ArrayList al = (ArrayList) JonasManagementRepr.getAttribute(on, "DeployableWars");
        String sDir = (String) JonasManagementRepr.getAttribute(on, "WebappsDirectory");

        String sEarDir = null;
        try {
            on = JonasObjectName.earService();
            sEarDir = (String) JonasManagementRepr.getAttribute(on, "AppsDirectory");
        }
        catch (Exception e) {
            // nothing, Ear service not found
        }
        return prepareContainersToDisplay(al, sDir, sEarDir);
    }

    /**
     * Return the list of deployed JAR filename in the current server.
     *
     * @return The list of JAR filename.
     * @throws Exception
     */
    public static ArrayList getJarFilesDeployed()
        throws Exception {
        ObjectName on = JonasObjectName.ejbService();
        ArrayList al = (ArrayList) JonasManagementRepr.getAttribute(on, "DeployedJars");
        String sDir = (String) JonasManagementRepr.getAttribute(on, "EjbjarsDirectory");

        String sEarDir = null;
        try {
            on = JonasObjectName.earService();
            sEarDir = (String) JonasManagementRepr.getAttribute(on, "AppsDirectory");
        }
        catch (Exception e) {
            // nothing, Ear service not found
        }
        return prepareContainersToDisplay(al, sDir, sEarDir);
    }

    /**
     * Return the list of deployed EAR filename in the current server.
     *
     * @return The list of EAR filename.
     * @throws Exception
     */
    public static ArrayList getEarFilesDeployed()
        throws Exception {
        ObjectName on = JonasObjectName.earService();
        ArrayList al = (ArrayList) JonasManagementRepr.getAttribute(on, "DeployedEars");
        String sDir = (String) JonasManagementRepr.getAttribute(on, "AppsDirectory");
        return prepareContainersToDisplay(al, sDir, null);
    }

    /**
     * Return the list of deployed RAR filename in the current server.
     *
     * @return The list of RAR filename.
     * @throws Exception
     */
    public static ArrayList getRarFilesDeployed()
        throws Exception {
        ObjectName on = JonasObjectName.resourceService();
        ArrayList al = (ArrayList) JonasManagementRepr.getAttribute(on, "DeployedRars");
        String sDir = (String) JonasManagementRepr.getAttribute(on, "RarsDirectory");

        String sEarDir = null;
        try {
            on = JonasObjectName.earService();
            sEarDir = (String) JonasManagementRepr.getAttribute(on, "AppsDirectory");
        }
        catch (Exception e) {
            // nothing, Ear service not found
        }
        return prepareContainersToDisplay(al, sDir, sEarDir);
    }

    /**
     * Return the list of deployed WAR filename in the current server.
     *
     * @return The list of WAR filename.
     * @throws Exception
     */
    public static ArrayList getWarFilesDeployed()
        throws Exception {
        ObjectName on = JonasObjectName.webContainerService();
        ArrayList al = (ArrayList) JonasManagementRepr.getAttribute(on, "DeployedWars");
        String sDir = (String) JonasManagementRepr.getAttribute(on, "WebappsDirectory");

        String sEarDir = null;
        try {
            on = JonasObjectName.earService();
            sEarDir = (String) JonasManagementRepr.getAttribute(on, "AppsDirectory");
        }
        catch (Exception e) {
            // nothing, Ear service not found
        }
        return prepareContainersToDisplay(al, sDir, sEarDir);
    }

    /**
     * Prepare a list of containers to diplay.
     * Deleting prefix root container directory, sorting the list.
     * @param p_Containers The list of containers
     * @param p_ContainerDir The root container directory
     * @param p_EarDir The root ear directory or null
     * @return The list of containers ready to display
     */
    public static ArrayList prepareContainersToDisplay(ArrayList p_Containers
        , String p_ContainerDir, String p_EarDir) {
        int iPos;
        String sPath;
        boolean bAdd;
        ArrayList al = new ArrayList();
        for (int i = 0; i < p_Containers.size(); i++) {
            // Delete prefix root container directory
            sPath = p_Containers.get(i).toString();
            iPos = sPath.indexOf(p_ContainerDir);
            if (iPos > -1) {
                sPath = sPath.substring(p_ContainerDir.length());
            }
            // Delete suffix slash
            if (sPath.endsWith("/") == true) {
                sPath = sPath.substring(0, sPath.length() - 1);
            }
            // Flag to filter
            bAdd = true;
            // Filter Jar or War include in Ear
            if (p_EarDir != null) {
                iPos = sPath.indexOf(p_EarDir);
                if (iPos > -1) {
                    bAdd = false;
                }

            }
            if (bAdd == true) {
                al.add(sPath);
            }
        }
        // Sort
        Collections.sort(al);
        return al;
    }

    private static ArrayList getFilesDeployed(ObjectName on)
        throws Exception {
        ArrayList alDeployed = new ArrayList();
        String sPath;
        String sFile;
        Iterator itNames = JonasAdminJmx.getListMBeanName(on).iterator();
        while (itNames.hasNext()) {
            sPath = JonasAdminJmx.extractValueMbeanName("fname", itNames.next().toString());
            sFile = JonasAdminJmx.extractFilename(sPath);
            if (sFile != null) {
                alDeployed.add(sFile);
            }
        }
        Collections.sort(alDeployed);
        return alDeployed;

    }

    /**
     * Return the list of Mail factory files ready to deploy in the current server.
     *
     * @return The list of Mail factory properties filenames.
     * @throws Exception
     */
    public static ArrayList getMailFilesDeployable()
        throws Exception {
        ObjectName on = JonasObjectName.mailService();
        return (ArrayList) JonasManagementRepr.getAttribute(on, "MailFactoryPropertiesFiles");
    }

    /**
     * Return the list of MimePartDataSource Mail factory files ready to deploy in the current server.
     *
     * @return The list of MimePartDataSource Mail factory properties filenames.
     * @throws Exception
     */
    public static ArrayList getMimePartMailFilesDeployable()
        throws Exception {
        ObjectName on = JonasObjectName.mailService();
        return (ArrayList) JonasManagementRepr.getAttribute(on, "MimePartMailFactoryPropertiesFiles");
    }

    /**
     * Return the list of Session Mail factory files ready to deploy in the current server.
     *
     * @return The list of Session Mail factory properties filenames.
     * @throws Exception
     */
    public static ArrayList getSessionMailFilesDeployable()
        throws Exception {
        ObjectName on = JonasObjectName.mailService();
        return (ArrayList) JonasManagementRepr.getAttribute(on, "SessionMailFactoryPropertiesFiles");
    }

    /**
     * Return the list of deployed MimePartDataSource Mail factories in the current server.
     *
     * @return The list of MimePartDataSource Mail factory filename.
     * @throws Exception
     */
    public static ArrayList getMimePartMailFilesDeployed()
        throws Exception {
        ArrayList alDeployed = new ArrayList();
        String sName;
        Iterator itNames = getListMBeanName(JonasObjectName.allMimeMailFactories()).iterator();
        while (itNames.hasNext()) {
            sName = JonasAdminJmx.extractValueMbeanName("name", itNames.next().toString());
            if (sName != null) {
                alDeployed.add(sName);
            }
        }
        Collections.sort(alDeployed);
        return alDeployed;
    }

    /**
     * Return the list of deployed Session Mail factories in the current server.
     *
     * @return The list of Session Mail factory filename.
     * @throws Exception
     */
    public static ArrayList getSessionMailFilesDeployed()
        throws Exception {
        ArrayList alDeployed = new ArrayList();
        String sName;
        Iterator itNames = getListMBeanName(JonasObjectName.allSessionMailFactories()).iterator();
        while (itNames.hasNext()) {
            sName = JonasAdminJmx.extractValueMbeanName("name", itNames.next().toString());
            if (sName != null) {
                alDeployed.add(sName);
            }
        }
        Collections.sort(alDeployed);
        return alDeployed;
    }

    /**
     * Return the list of deployed Mail factories in the current server.
     *
     * @return The list of Mail factory filename.
     * @throws Exception
     */
    public static ArrayList getMailFilesDeployed()
        throws Exception {
        ArrayList alDeployed = getSessionMailFilesDeployed();
        alDeployed.addAll(getMimePartMailFilesDeployed());
        Collections.sort(alDeployed);
        return alDeployed;
    }

    /**
     * Return the list of Datasource properties files ready to deploy in the current server.
     *
     * @return The list of Datasource properties filenames.
     * @throws Exception
     */
    public static ArrayList getDatasourceFilesDeployable()
        throws Exception {
        ObjectName on = JonasObjectName.databaseService();
        return (ArrayList) JonasManagementRepr.getAttribute(on, "DataSourcePropertiesFiles");
    }

    /**
     * Return the list of deployed Datasources in the current server.
     *
     * @return The list of Datasource filename.
     * @throws Exception
     */
    public static ArrayList getDatasourceFilesDeployed()
        throws Exception {
        ArrayList alDeployed = new ArrayList();
        String sName;
        Iterator itNames = getListMBeanName(JonasObjectName.allDatasources()).iterator();
        while (itNames.hasNext()) {
            sName = JonasAdminJmx.extractValueMbeanName("name", itNames.next().toString());
            if (sName != null) {
                alDeployed.add(sName);
            }
        }
        Collections.sort(alDeployed);
        return alDeployed;
    }

    /**
     * Return the list of Datasource dependences for a given datasource name in the current server.
     *
     * @param p_DatasourceName The name of the datasource
     * @return The list of Datasource dependence (a list of names corresponding to EJBs using this datasource).
     * @throws Exception
     */
    public static ArrayList getDatasourceDependences(String p_DatasourceName)
        throws Exception {
        String[] asParam = new String[1];
        String[] asSignature = new String[1];
        asSignature[0] = "java.lang.String";
        ArrayList al = new ArrayList();
        asParam[0] = (String) JonasManagementRepr.getAttribute(JonasObjectName.datasource(
            p_DatasourceName), "DSName");
        if (JonasManagementRepr.isRegistered(JonasObjectName.ejbService()) == true) {
            String sName;
            Iterator it = ((java.util.Set) JonasManagementRepr.invoke(JonasObjectName.ejbService()
                , "getDataSourceDependence", asParam, asSignature)).iterator();
            while (it.hasNext()) {
                sName = JonasAdminJmx.extractValueMbeanName(",name", it.next().toString());
                if (sName != null) {
                    al.add(sName);
                }
            }
            Collections.sort(al);
        }
        return al;
    }

    /**
     * Return the list of Mail factory dependences in the current server for a given factory name.
     *
     * @param p_MailFactoryName The name of the mail factory
     * @return The list of Mail factory dependences (a list of names corresponding to EJBs using this mail factory).
     * @throws Exception
     */
    public static ArrayList getMailFactoryDependences(String p_MailFactoryName)
        throws Exception {
        String jndiName = null;
        // create an MBean name with the given MailFactory name
        // - try with a session Mail Factory type
        ObjectName on = JonasObjectName.sessionMailFactory(p_MailFactoryName);
        if (JonasManagementRepr.isRegistered(on) != true) {
            // in this case, this is a MimePartDataSource mail factory
            on = JonasObjectName.mimeMailFactory(p_MailFactoryName);
        }
        jndiName = (String) JonasManagementRepr.getAttribute(on, "Name");

        String[] asParam = new String[1];
        String[] asSignature = new String[1];
        asSignature[0] = "java.lang.String";
        ArrayList al = new ArrayList();
        asParam[0] = jndiName;
        if (JonasManagementRepr.isRegistered(JonasObjectName.ejbService()) == true) {
            String sName;
            Iterator it = ((java.util.Set) JonasManagementRepr.invoke(JonasObjectName.ejbService()
                , "getMailFactoryDependence", asParam, asSignature)).iterator();
            while (it.hasNext()) {
                sName = JonasAdminJmx.extractValueMbeanName(",name", it.next().toString());
                if (sName != null) {
                    al.add(sName);
                }
            }
            Collections.sort(al);
        }
        return al;
    }

    /**
     * Return the list of deployed Session Mail Factories in the current server.
     *
     * @return
     * @throws Exception
     */
    public static ArrayList getSessionMailFactoriesDeployed() throws Exception {
        ArrayList alDeployed = new ArrayList();
        String sName;
        Iterator itNames = getListMBeanName(JonasObjectName.allSessionMailFactories()).iterator();
        while (itNames.hasNext()) {
            sName = JonasAdminJmx.extractValueMbeanName("name", itNames.next().toString());
            if (sName != null) {
                alDeployed.add(sName);
            }
        }
        Collections.sort(alDeployed);
        return alDeployed;
    }

    /**
     * Return the list of deployed MimePartDatasource Mail Factories in the current server.
     *
     * @return
     * @throws Exception
     */
    public static ArrayList getMimeMailPartFactoriesDeployed() throws Exception {
        ArrayList alDeployed = new ArrayList();
        String sName;
        Iterator itNames = getListMBeanName(JonasObjectName.allMimeMailFactories()).iterator();
        while (itNames.hasNext()) {
            sName = JonasAdminJmx.extractValueMbeanName("name", itNames.next().toString());
            if (sName != null) {
                alDeployed.add(sName);
            }
        }
        Collections.sort(alDeployed);
        return alDeployed;
    }

    /**
     * Return the list of Security Memory Factories in the current server.
     *
     * @return The list of Security Memory Factories
     * @throws Exception
     */
    public static ArrayList getSecurityMemoryFactories()
        throws Exception {
        String sName;
        ArrayList al = new ArrayList();
        Iterator itNames = getListMBeanName(JonasObjectName.allSecurityMemoryFactories()).iterator();
        while (itNames.hasNext()) {
            sName = JonasAdminJmx.extractValueMbeanName(",name", itNames.next().toString());
            if (sName != null) {
                al.add(sName);
            }
        }
        Collections.sort(al);
        return al;
    }

    /**
     * Return the list of Security Datasource Factories in the current server.
     *
     * @return The list of Security Datasource Factories
     * @throws Exception
     */
    public static ArrayList getSecurityDatasourceFactories()
        throws Exception {
        String sName;
        ArrayList al = new ArrayList();
        Iterator itNames = getListMBeanName(JonasObjectName.allSecurityDatasourceFactories()).
            iterator();
        while (itNames.hasNext()) {
            sName = JonasAdminJmx.extractValueMbeanName(",name", itNames.next().toString());
            if (sName != null) {
                al.add(sName);
            }
        }
        Collections.sort(al);
        return al;
    }

    /**
     * Return the list of Security Ldap Factories in the current server.
     *
     * @return The list of Security Ldap Factories
     * @throws Exception
     */
    public static ArrayList getSecurityLdapFactories()
        throws Exception {
        String sName;
        ArrayList al = new ArrayList();
        Iterator itNames = getListMBeanName(JonasObjectName.allSecurityLdapFactories()).iterator();
        while (itNames.hasNext()) {
            sName = JonasAdminJmx.extractValueMbeanName(",name", itNames.next().toString());
            if (sName != null) {
                al.add(sName);
            }
        }
        Collections.sort(al);
        return al;
    }

    /**
     * Return the SubType of a Security Factory in the current server.
     *
     * @param p_NameFactory The factory name to find
     * @return The SubType or null if not found
     * @throws Exception
     */
    public static String findSecurityFactorySubType(String p_NameFactory)
        throws Exception {
        String sName;
        String sSubType = null;
        ObjectName on = null;

        Iterator itNames = getListMBeanName(JonasObjectName.allSecurityFactories()).iterator();
        while (itNames.hasNext()) {
            on = new ObjectName(itNames.next().toString());
            sName = on.getKeyProperty("name");
            if (p_NameFactory.equals(sName) == true) {
                sSubType = on.getKeyProperty("subtype");
                break;
            }
        }
        return sSubType;
    }

    /**
     * Return the list of all users in a resource.
     *
     * @param p_Resource The resource
     * @return The list of users
     * @throws Exception
     */
    public static ArrayList getUsers(String p_Resource)
        throws Exception {
        String sName;
        ArrayList al = new ArrayList();
        Iterator itNames = getListMBeanName(JonasObjectName.allUsers(p_Resource)).iterator();
        while (itNames.hasNext()) {
            sName = JonasAdminJmx.extractValueMbeanName(",name", itNames.next().toString());
            if (sName != null) {
                al.add(sName);
            }
        }
        Collections.sort(al);
        return al;
    }

    /**
     * Return the list of all roles in a resource.
     *
     * @param p_Resource The resource
     * @return The list of roles
     * @throws Exception
     */
    public static ArrayList getRoles(String p_Resource)
        throws Exception {
        String sName;
        ArrayList al = new ArrayList();
        Iterator itNames = getListMBeanName(JonasObjectName.allRoles(p_Resource)).iterator();
        while (itNames.hasNext()) {
            sName = JonasAdminJmx.extractValueMbeanName(",name", itNames.next().toString());
            if (sName != null) {
                al.add(sName);
            }
        }
        Collections.sort(al);
        return al;
    }

    /**
     * Return the list of all groups in a resource.
     *
     * @param p_Resource The resource
     * @return The list of groups
     * @throws Exception
     */
    public static ArrayList getGroups(String p_Resource)
        throws Exception {
        String sName;
        ArrayList al = new ArrayList();
        Iterator itNames = getListMBeanName(JonasObjectName.allGroups(p_Resource)).iterator();
        while (itNames.hasNext()) {
            sName = JonasAdminJmx.extractValueMbeanName(",name", itNames.next().toString());
            if (sName != null) {
                al.add(sName);
            }
        }
        Collections.sort(al);
        return al;
    }

    /**
     * Return a list of names bounded in the registry
     */
    public static ArrayList getRegistryList()
        throws NamingException {
        synchronized (s_Synchro) {
            ArrayList oNames = new ArrayList();
            // get the naming context
            Context ctx = JonasManagementRepr.getContext();
            // get all names bounded in the contexr
            for (NamingEnumeration e = ctx.list(""); e.hasMoreElements(); ) {
                oNames.add(((NameClassPair) e.nextElement()).getName());
            }
            return oNames;
        }
    }

    /**
     * Return the Queue destinations list
     */
    public static ArrayList getQueuesList() {
        synchronized (s_Synchro) {
            ObjectName jmsServMB = JonasObjectName.jmsService();
            Set queues = (Set) JonasManagementRepr.getAttribute(jmsServMB
                , "AllJmsQueueDestinationNames");
            ArrayList al = new ArrayList();
            Iterator itNames = queues.iterator();
            while (itNames.hasNext()) {
                al.add(itNames.next());
            }
            return al;
        }
    }

    /**
     * Return the Topic destinations list
     */
    public static ArrayList getTopicsList() {
        synchronized (s_Synchro) {
            ObjectName jmsServMB = JonasObjectName.jmsService();
            Set topics = (Set) JonasManagementRepr.getAttribute(jmsServMB
                , "AllJmsTopicDestinationNames");
            ArrayList al = new ArrayList();
            Iterator itNames = topics.iterator();
            while (itNames.hasNext()) {
                al.add(itNames.next());
            }
            return al;
        }
    }

    /**
     * Return the default Connection Factories
     */
    public static ArrayList getConnectionFactoriesList(MessageResources p_Resources) {
        synchronized (s_Synchro) {
            ObjectName jmsServMB = JonasObjectName.jmsService();
            String cfName = (String) JonasManagementRepr.getAttribute(jmsServMB
                , "DefaultConnectionFactoryName");
            String tcfName = (String) JonasManagementRepr.getAttribute(jmsServMB
                , "DefaultTopicConnectionFactoryName");
            String qcfName = (String) JonasManagementRepr.getAttribute(jmsServMB
                , "DefaultQueueConnectionFactoryName");
            ArrayList al = new ArrayList();
            String comment = p_Resources.getMessage("tab.connfact.defaultconnfact");
            JmsConnFact cf = new JmsConnFact(cfName, comment);
            al.add(cf);
            comment = p_Resources.getMessage("tab.connfact.defaulttopicconnfact");
            cf = new JmsConnFact(tcfName, comment);
            al.add(cf);
            comment = p_Resources.getMessage("tab.connfact.defaultqueueconnfact");
            cf = new JmsConnFact(qcfName, comment);
            al.add(cf);
            return al;
        }
    }

    /**
     * Return the list of all used loggers in this JOnAS server.
     * @param p_Resources The used message resource
     * @param p_WhereAreYou The used WhereAreYou instance
     * @param p_Action True to get the action (to use directly in jsp) or False to get the same action but in forward (write in the file struts-config.xml)
     * @return The list of loggers
     * @throws Exception
     */
    public static ArrayList getLoggers(MessageResources p_Resources, WhereAreYou p_WhereAreYou
        , boolean p_Action)
        throws Exception {
        ArrayList al = new ArrayList();
        String sActionForward;
        // Logger JOnAS
        if (p_Action == true) {
            sActionForward = "EditLoggingJonas.do";
        }
        else {
            sActionForward = "ActionEditLoggingJonas";
        }
        al.add(new LoggerItem(p_Resources.getMessage("logger.jonas.name"), LoggerItem.LOGGER_JONAS
            , sActionForward));
        // Logger Catalina
        if (p_WhereAreYou.isCatalinaServer() == true) {
            ObjectName on;
            Iterator it = getListMbean(CatalinaObjectName.catalinaValves(p_WhereAreYou.
                getCurrentCatalinaServiceName()
                , p_WhereAreYou.getCurrentCatalinaDefaultHostName())).iterator();
            String sClassName = null;
            while (it.hasNext()) {
                on = (ObjectName) it.next();
                try {
                    sClassName = (String) JonasManagementRepr.invoke(on, "getClassName", null, null);
                }
                catch (Exception e) {
                    sClassName = e.getMessage();
                }
                if (sClassName.indexOf("AccessLogValve") > -1) {
                    if (p_Action == true) {
                        sActionForward = "EditCatalinaAccessLogger.do?select=" + on.toString();
                    }
                    else {
                        sActionForward = "ActionEditCatalinaAccessLogger";
                    }
                    al.add(new LoggerItem(p_Resources.getMessage("logger.catalina.access.name")
                        , LoggerItem.LOGGER_CATALINA_ACCESS, sActionForward, on.toString()));
                }
            }
        }
        // sort
        Collections.sort(al, new ItemByNameComparator());
        return al;
    }
}
