/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial Developer : Delplanque Xavier & Sauthier Guillaume
 * --------------------------------------------------------------------------
 * $Id: JAxisServiceFactory.java,v 1.2 2003/10/03 09:55:41 sauthieg Exp $
 * --------------------------------------------------------------------------
*/

package org.objectweb.jonas.ws.axis;

import org.apache.axis.AxisProperties;
import org.apache.axis.EngineConfiguration;
import org.apache.axis.EngineConfigurationFactory;
import org.apache.axis.client.AxisClient;
import org.apache.axis.configuration.EngineConfigurationFactoryFinder;

import org.objectweb.jonas.common.Log;
import org.objectweb.jonas.naming.CompNamingContext;
import org.objectweb.jonas.ws.JServiceFactory;
import org.objectweb.jonas.ws.WSServiceException;

import org.objectweb.jonas_ws.deployment.api.ServiceRef;

import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

import javax.naming.Context;
import javax.naming.NamingException;

import javax.xml.rpc.Service;

/**
 * Axis specific JServiceFactory. Create an Axis Service instance.
 *
 * @author Guillaume Sauthier
 */
public class JAxisServiceFactory implements JServiceFactory {
    /**
     * logger
     */
    private static Logger logger = Log.getLogger(Log.JONAS_WS_PREFIX);

    /** Used to tell Axis engine to use a new EngineConfigurationFactory */
    private static final String AXIS_CONFIG_FACTORY_PROP =
        "org.apache.axis.EngineConfigurationFactory";

    /** The class name of the new Factory */
    private static final String AXIS_MODULE_CONFIG_CLASS =
        "org.objectweb.jonas.ws.axis.ModuleEngineConfigurationFactory";


    /**
     * Construct a new JAxisServiceFactory.
     */
    public JAxisServiceFactory() {
        // Add our Config Factory in Axis
        AxisProperties.setProperty(AXIS_CONFIG_FACTORY_PROP,
                                   AXIS_MODULE_CONFIG_CLASS);        
    }

    /**
     * Create the JaxRpx Service Object associated with the service-ref
     *
     * @param sr The service-ref description
     * @param cl The classloader used to load Service class
     *
     * @return The Service object
     *
     * @throws WSServiceException When instantiation fails
     */
    public Service getService(
        ServiceRef  sr,
        ClassLoader cl) throws WSServiceException {
        // Try to instanciate the specified Service class
        Class serviceClass = sr.getServiceClass();
        org.apache.axis.client.Service s = null;

        try {
            String serviceName = serviceClass.getName();
            Class clazz = cl.loadClass(serviceName);
            s = (org.apache.axis.client.Service) clazz.newInstance();
        } catch (InstantiationException ie) {
            String err =
                "Cannot instantiate Service class '" + serviceClass.getName()
                + "' for service-ref '" + sr.getServiceRefName() + "'";

            if (logger.isLoggable(BasicLevel.ERROR)) {
                logger.log(BasicLevel.ERROR, err);
            }

            throw new WSServiceException(err, ie);
        } catch (ClassNotFoundException cnfe) {
            String err =
                "ClassNotFound '" + serviceClass.getName()
                + "' for service-ref '" + sr.getServiceRefName() + "'";

            if (logger.isLoggable(BasicLevel.ERROR)) {
                logger.log(BasicLevel.ERROR, err);
            }

            throw new WSServiceException(err, cnfe);
        } catch (IllegalAccessException iae) {
            String err =
                "Cannot instantiate Service class '" + serviceClass.getName()
                + "' for service-ref '" + sr.getServiceRefName() + "'";

            if (logger.isLoggable(BasicLevel.ERROR)) {
                logger.log(BasicLevel.ERROR, err);
            }

            throw new WSServiceException(err, iae);
        }

        // Prepare the Factory Call for Service configuration
        Context ctx = null;

        try {
            ctx = new CompNamingContext(sr.getServiceRefName());
            ctx.rebind("serviceRef", sr);
            ctx.rebind("parentCL", cl);
        } catch (NamingException ne) {
            String err =
                "Cannot bind informations for service-ref '"
                + sr.getServiceRefName() + "'";

            if (logger.isLoggable(BasicLevel.ERROR)) {
                logger.log(BasicLevel.ERROR, err);
            }

            throw new WSServiceException(err, ne);
        }

        // Get the Configuration from the service-ref
        // 1. Get our Factory (ModuleEngineConfigurtaionFactory)
        //  -> our factory react to a CompNamingContext
        // 2. Create the EngineConfiguration with the factory
        // 3. Configure the Axis Service instance with the config.
        EngineConfigurationFactory f =
            EngineConfigurationFactoryFinder.newFactory(ctx);
        EngineConfiguration ec = f.getClientEngineConfig();
        s.setEngine(new AxisClient(ec));

        return s;
    }
}
