/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial Developer : Delplanque Xavier & Sauthier Guillaume
 * --------------------------------------------------------------------------
 * $Id: ModuleEngineConfigurationFactory.java,v 1.2 2003/09/16 14:15:55 sauthieg Exp $
 * --------------------------------------------------------------------------
*/

package org.objectweb.jonas.ws.axis;

import org.apache.axis.AxisProperties;
import org.apache.axis.EngineConfiguration;
import org.apache.axis.EngineConfigurationFactory;
import org.apache.axis.configuration.EngineConfigurationFactoryFinder;
import org.apache.axis.configuration.FileProvider;

import org.objectweb.jonas.common.Log;
import org.objectweb.jonas.naming.CompNamingContext;

import org.objectweb.jonas_ws.deployment.api.ServiceRef;

import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

import java.io.File;
import java.io.InputStream;

import javax.naming.NamingException;


/**
 * This is the JOnAS implementation of EngineConfigurationFactory. It override
 * Axis default one.
 *
 * @author Guillaume Sauthier
 */
public class ModuleEngineConfigurationFactory
    implements EngineConfigurationFactory {
    /**
     * Property name for client config file
     */
    public static final String OPTION_CLIENT_CONFIG_FILE =
        "axis.ClientConfigFile";

    /**
     * Default client config file name
     */
    protected static final String CLIENT_CONFIG_FILE = "client-config.wsdd";

    /**
     * client config parameter name in jonas-init-param
     */
    public static final String AXIS_CLIENT_CONFIG_PARAM = "axis.client-config";

    /** The file name to use for research */
    private String clientConfigFile;

    /** The context used for this ModuleEngineConfigurationFactory */
    private CompNamingContext ctx;

    /** The classLoader used to load client-config file */
    private ClassLoader cl;

    /** The logger to use */
    private Logger logger = Log.getLogger(Log.JONAS_WS_PREFIX);

    /**
     * Create the default engine configuration and detect whether the user has
     * overridden this with their own.
     *
     * @param ctx a CompNamingContext where we can find : <BR>
     *        - serviceRef : the service-ref where we can find config information
     *        - parentCL   : the classLoader used to read config file
     */
    protected ModuleEngineConfigurationFactory(CompNamingContext ctx) {

        this.ctx = ctx;

        ServiceRef sr = null;

        try {
            sr = (ServiceRef) ctx.lookup("serviceRef");
            cl = (ClassLoader) ctx.lookup("parentCL");
        } catch (NamingException ne) {
            // If properties not found, just use Axis Default values
            // log an error
            if (logger.isLoggable(BasicLevel.ERROR)) {
                logger.log(BasicLevel.ERROR,
                    "Error when looking for variables in Context");
            }

            clientConfigFile =
                AxisProperties.getProperty(OPTION_CLIENT_CONFIG_FILE,
                    CLIENT_CONFIG_FILE);
        }

        clientConfigFile = sr.getParam(AXIS_CLIENT_CONFIG_PARAM);

        if (clientConfigFile == null) {
            if (logger.isLoggable(BasicLevel.DEBUG)) {
                logger.log(BasicLevel.DEBUG,
                    "option '" + AXIS_CLIENT_CONFIG_PARAM
                    + "' file not found in jonas-init-param"
                    + ". Continue with default configuration.");
            }

            clientConfigFile =
                AxisProperties.getProperty(OPTION_CLIENT_CONFIG_FILE,
                    CLIENT_CONFIG_FILE);
        }
    }

    /**
     * Creates and returns a new ModuleEngineConfigurationFactory. If a factory
     * cannot be created, return 'null'. The factory may return non-NULL only
     * if: <br>
     * - it knows what to do with the param (param != null) <br>
     * - it can find it's configuration information
     *
     * @param param The object used to retrieved the right Factory instance
     *
     * @return null if param is not a CompNamingContext, or return the JOnAS 
     *         EngineConfigurationFactory used for 1 service-ref element.
     */
    public static EngineConfigurationFactory newFactory(Object param) {
        if (param == null) {
            return null; // not for us.
        }

        if (param instanceof CompNamingContext) {
            return new ModuleEngineConfigurationFactory((CompNamingContext) param);
        } else {
            return null;
        }
    }

    /**
     * Get a client engine configuration. Try to load it from a File or as a
     * ClassLoader Resource. Return default CLient config if no resource
     * found.
     *
     * @return a client EngineConfiguration
     */
    public EngineConfiguration getClientEngineConfig() {
        if (logger.isLoggable(BasicLevel.DEBUG)) {
            logger.log(BasicLevel.DEBUG,
                "Loading client-config file '" + clientConfigFile + "'");
        }

        if (new File(clientConfigFile).exists()) {
            return new FileProvider(clientConfigFile);
        } else {
            InputStream is = cl.getResourceAsStream(clientConfigFile);

            if (is != null) {
                return new FileProvider(is);
            } else {
                if (logger.isLoggable(BasicLevel.WARN)) {
                    logger.log(BasicLevel.WARN,
                        "Cannot load client-config file '" + clientConfigFile
                        + "'. Continue with default configuration.");
                }

                return EngineConfigurationFactoryFinder.newFactory()
                                                       .getClientEngineConfig();
            }
        }
    }

    /**
     * Return null. (Not used for WebService clients)
     *
     * @return null
     */
    public EngineConfiguration getServerEngineConfig() {
        return null;
    }
}
