/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial Developer : Delplanque Xavier & Sauthier Guillaume
 * --------------------------------------------------------------------------
 * $Id: FileWSDLHandler.java,v 1.6 2003/09/09 11:57:33 sauthieg Exp $
 * --------------------------------------------------------------------------
*/

package org.objectweb.jonas.ws.handler;

import org.objectweb.jonas.ws.WSServiceException;

import org.objectweb.jonas_ws.deployment.api.WSDLFile;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;

import javax.wsdl.WSDLException;
import javax.wsdl.factory.WSDLFactory;
import javax.wsdl.xml.WSDLWriter;


/**
 * In charge of published a given WSDL in a directory.
 */
public class FileWSDLHandler implements WSDLHandler {
    /**
     * Directory name where WSDL file must be created.
     */
    private String dirName;

    /**
     * WSDL Directory
     */
    private File location;

    /**
     * Return a new instance of FileWSDLHandler.
     *
     * @param directory the directory to publish WSDLs
     *
     * @throws WSServiceException When Cannot create the directory 
     *         (if it don't exist).
     */
    public FileWSDLHandler(String directory) throws WSServiceException {
        try {
            location = new File(directory).getCanonicalFile();
            dirName = location.getCanonicalPath();

            if (!location.exists()) {
                // if the given file doesn't exist, create it.
                location.mkdirs();
            }
        } catch (IOException ioe) {
            throw new WSServiceException(
                "cannot find/create the publishing directory '" + directory + "'",
                ioe);
        }
    }

    /**
     * Publish the given WSDL to the directory location.
     *
     * @param wsdl the wsdl file to publish.
     *
     * @throws WSServiceException When publication fails.
     */
    public void publish(WSDLFile wsdl) throws WSServiceException {
        // build the fully qualified file name for the published wsdl file
        String filePath = wsdl.getName();
        int ind = filePath.lastIndexOf("/");
        String fileName;

        if (ind != -1) {
            fileName = dirName + File.separator + filePath.substring(ind);
        } else {
            fileName = dirName + File.separator + filePath;
        }

        try {
            // write the wsdl file in the directory
            FileWriter fwriter = new FileWriter(fileName);
            WSDLFactory factory = WSDLFactory.newInstance();
            WSDLWriter writer = factory.newWSDLWriter();
            writer.writeWSDL(wsdl.getDefinition(), fwriter);
            fwriter.flush();
            fwriter.close();
        } catch (IOException ioe) {
            throw new WSServiceException("Error with writer of file '"
                + fileName + "'", ioe);
        } catch (WSDLException we) {
            throw new WSServiceException("Error with wsdl file '" + filePath
                + "' publishing in directory '" + dirName + "'", we);
        }
    }
}
