/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer: Florent BENOIT
 * --------------------------------------------------------------------------
 * $Id: ClientContainerDeploymentDesc.java,v 1.1 2003/06/27 09:03:13 benoitf Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_client.deployment.api;


// import java
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.IOException;
import java.io.Reader;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;
import java.util.jar.JarFile;
import java.util.zip.ZipEntry;

// import jonas_lib
import org.objectweb.jonas_lib.deployment.AbsDeploymentDesc;
import org.objectweb.jonas_lib.deployment.SAXParserHelper;

// import jonas_client
import org.objectweb.jonas_client.deployment.xml.JonasResourceEnv;
import org.objectweb.jonas_client.deployment.xml.ResourceEnvRef;
import org.objectweb.jonas_client.deployment.xml.JonasResource;
import org.objectweb.jonas_client.deployment.xml.ResourceRef;
import org.objectweb.jonas_client.deployment.xml.JonasEjbRef;
import org.objectweb.jonas_client.deployment.xml.EjbRef;
import org.objectweb.jonas_client.deployment.xml.JonasSecurity;
import org.objectweb.jonas_client.deployment.xml.EnvEntry;
import org.objectweb.jonas_client.deployment.xml.ApplicationClientUnmarshaller;
import org.objectweb.jonas_client.deployment.xml.JonasClientUnmarshaller;
import org.objectweb.jonas_client.deployment.xml.ApplicationClient;
import org.objectweb.jonas_client.deployment.xml.JonasClient;
import org.objectweb.jonas_client.deployment.xml.JonasClientImpl;


// import SAX
import org.xml.sax.ErrorHandler;
import org.xml.sax.EntityResolver;

/**
 * This class do the parsing of the application-client.xml file and jonas-client.xml files and
 * contruct a data structure associated to these two files.
 * @author Florent Benoit
 */
public class ClientContainerDeploymentDesc extends AbsDeploymentDesc {

    /** 
     * The path to the application-client.xml file. 
     */
    public static final String CLIENT_FILE_NAME = "META-INF/application-client.xml";

    /** 
     * The path to the jonas-client.xml file. 
     */
    public static final String JONAS_CLIENT_FILE_NAME = "META-INF/jonas-client.xml";

    /** 
     * Associates a JNDI name to a name of a resource env ref.
     */
    private Vector resourceEnvRefs = new Vector();

    /**
     * Associates a JNDI name to a name of a resource ref.
     */
    private Vector resourceRefs = new Vector();

    /**
     * List of the env-entry
     */
    private Vector envEntries = new Vector();

    /**
     * Associates a JNDI name to a name of an EJB ref.
     */
    private Vector ejbRefs = new Vector();

    /**
     * The name of the callback handler
     */
    private String callbackHandler = null;

    /**
     * The name of the jaas config file
     */
    private String jaasFile = null;

    /**
     * The name of the jaas entry
     */
    private String jaasEntry = null;

    /**
     * The username to use for the callback handler
     */
    private String username = null;

    /**
     * The password to use for the callback handler
     */
    private String password = null;

    /** 
     *  The error handler to use
     */
    private static ErrorHandler errorHandler = null;

    /** 
     * The entity resolver to use
     */
    private static EntityResolver entityResolver = null;

    /**
     * Flag for parser validation
     */
    private static boolean parsingWithValidation = false;

    /**
     * Construct an instance of a ClientContainerDeploymentDesc.<BR>
     * Constructor is private, call one of the static getInstance 
     * method instead.
     * @param classLoader the classloader for the classes.
     * @param applicationClient the data structure of the application-client (application-client.xml)
     * @param jonasClient the data structure of the jonas-client (jonas-client.xml)
     * @throws ClientContainerDeploymentDescException if the deployment 
     * descriptors are corrupted.
     */
    private ClientContainerDeploymentDesc(ClassLoader classLoader,
                                          ApplicationClient applicationClient,
                                          JonasClient jonasClient) 
        throws ClientContainerDeploymentDescException {
 
        // test classloader
        if (classLoader == null) {
            throw new ClientContainerDeploymentDescException("Classloader is null");
        }

        // display name
        displayName = (applicationClient.getDisplayName() != null) ? applicationClient.getDisplayName().getValue() : null;

        // callbackHandler
        callbackHandler = (applicationClient.getCallbackHandler() != null) ? applicationClient.getCallbackHandler().getValue() : null;
        
        JonasSecurity jonasSecurity = jonasClient.getJonasSecurity();
        if (jonasSecurity != null) {
            // jaas config file
            jaasFile = (jonasSecurity.getJaasfile() != null) ? jonasSecurity.getJaasfile().getValue() : null;

            // jaas entry
            jaasEntry = (jonasSecurity.getJaasentry() != null) ? jonasSecurity.getJaasentry().getValue() : null;

            // username
            username = (jonasSecurity.getUsername() != null) ? jonasSecurity.getUsername().getValue() : null;

            // password
            password = (jonasSecurity.getPassword() != null) ? jonasSecurity.getPassword().getValue() : null;
        }

        // resource-env-ref
        List resourceEnvRefList = applicationClient.getResourceEnvRefList();
        List jonasResourceEnvList = jonasClient.getJonasResourceEnvList();
        Hashtable resourceEnv = new Hashtable();
        for (Iterator i = jonasResourceEnvList.iterator(); i.hasNext();) {
            JonasResourceEnv jonasResourceEnv = (JonasResourceEnv) i.next();
            String rName = jonasResourceEnv.getResourceEnvRefName().getValue();
            boolean findResRef = false;
            for (Iterator j = resourceEnvRefList.iterator(); j.hasNext() && !findResRef;) {
                ResourceEnvRef resourceEnvRef = (ResourceEnvRef) j.next();
                if (rName.equals(resourceEnvRef.getResourceEnvRefName().getValue())) {
                    findResRef = true;
                }
            }
            if (!findResRef) {
                throw new ClientContainerDeploymentDescException("resource-env-ref missing for jonas-resource-env " + rName);
            }
            resourceEnv.put(rName, jonasResourceEnv);
        }
        for (Iterator i = resourceEnvRefList.iterator(); i.hasNext();) {
            ResourceEnvRef resourceEnvRef = (ResourceEnvRef) i.next();
            String rName = resourceEnvRef.getResourceEnvRefName().getValue();
            if (!resourceEnv.containsKey(rName)) {
                throw new ClientContainerDeploymentDescException("jonas-resource-env missing for resource-env-ref-name " + rName);
            }
            resourceEnvRefs.addElement(new ResourceEnvRefDesc(classLoader, resourceEnvRef, (JonasResourceEnv) resourceEnv.get(rName)));
        }


        // resource-ref
        List resourceRefList = applicationClient.getResourceRefList();
        List jonasResourceList = jonasClient.getJonasResourceList();
        Hashtable resource = new Hashtable();
        for (Iterator i = jonasResourceList.iterator(); i.hasNext();) {
            JonasResource jonasResource = ((JonasResource) i.next());
            String rName = jonasResource.getResRefName().getValue();
            boolean findResRef = false;
            for (Iterator j = resourceRefList.iterator(); j.hasNext() && !findResRef;) {
                if (rName.equals(((ResourceRef) j.next()).getResRefName().getValue())) {
                    findResRef = true;
                }
            }
            if (!findResRef) {
                throw new ClientContainerDeploymentDescException("resource-ref missing for jonas-resource " + rName);
            }
            resource.put(rName, jonasResource);
        }

        for (Iterator i = resourceRefList.iterator(); i.hasNext();) {
            ResourceRef resourceRef = ((ResourceRef) i.next());
            String rName = resourceRef.getResRefName().getValue();
            if (!resource.containsKey(rName)) {
                throw new ClientContainerDeploymentDescException("jonas-resource missing for res-ref-name " + rName);
            }
            resourceRefs.addElement(new ResourceRefDesc(classLoader, resourceRef, (JonasResource) resource.get(rName)));
        }

        // env-entry
        for (Iterator i = applicationClient.getEnvEntryList().iterator(); i.hasNext();) {
            EnvEntryDesc envEntryDesc = new EnvEntryDesc((EnvEntry) i.next());
            envEntries.addElement(envEntryDesc);
        }

        // ejb-ref
        List ejbRefList = applicationClient.getEjbRefList();
        List jonasEjbRefList = jonasClient.getJonasEjbRefList();

        Hashtable ejbRef = new Hashtable();
        for (Iterator i = jonasEjbRefList.iterator(); i.hasNext();) {
            JonasEjbRef jonasEjbRef = (JonasEjbRef) i.next();
            String brName = jonasEjbRef.getEjbRefName().getValue();
            boolean findBRef = false;
            for (Iterator j = ejbRefList.iterator(); j.hasNext() && !findBRef;) {
                EjbRef eRef = (EjbRef) j.next();
                if (brName.equals(eRef.getEjbRefName().getValue())) {
                    findBRef = true;
                }
            }
            if (!findBRef) {
                throw new ClientContainerDeploymentDescException("ejb-ref missing for jonas-ejb-ref " + brName);
            }
            ejbRef.put(brName, jonasEjbRef);
        }
        for (Iterator i = ejbRefList.iterator(); i.hasNext();) {
            EjbRef eRef = (EjbRef) i.next();
            String brName = eRef.getEjbRefName().getValue();
            JonasEjbRef jonasEjbRef = (JonasEjbRef) ejbRef.get(brName);
            if (jonasEjbRef == null) {
                // The ejb-link must be specified when there is no jonas-ejb-ref
                if ((eRef.getEjbLink() == null)) {
                    // Ignore the ejb-link syntax "product.jar#ProductEJB"
                    throw new ClientContainerDeploymentDescException("ejb-link missing for ejb-ref-name " + brName);
                }
            }
            ejbRefs.addElement(new EjbRefDesc(eRef, jonasEjbRef));
        }

    }

    /**
     * Get an instance of a Client deployment descriptor by parsing the application-client.xml 
     * and jonas-client.xml deployment descriptors.
     * @param clientFileName the fileName of the client file for the deployment 
     * descriptors.
     * @param classLoaderForCls the classloader for the classes.
     * @return a Client deployment descriptor by parsing the application-client.xml 
     * and jonas-client.xml deployment descriptors.
     * @throws ClientContainerDeploymentDescException if the deployment 
     * descriptors are corrupted.
     */
    public static ClientContainerDeploymentDesc getInstance(String clientFileName,
                                                            ClassLoader classLoaderForCls)
        throws ClientContainerDeploymentDescException {
        
        return getInstance(clientFileName, classLoaderForCls, null);
    }

    /**
     * Get an instance of a Client deployment descriptor by parsing the application-client.xml 
     * and jonas-client.xml deployment descriptors.
     * @param clientFileName the fileName of the client file for the deployment 
     * descriptors.
     * @param classLoaderForCls the classloader for the classes.
     * @param altClientXmlFilename the fileName to the application-client.xml for the alt-dd
     * tag in the Ear Case. This is used for specify an alternate DDesc file.
     * @return a Client deployment descriptor by parsing the application-client.xml 
     * and jonas-client.xml deployment descriptors.
     * @throws ClientContainerDeploymentDescException if the deployment 
     * descriptors are corrupted.
     */
    public static ClientContainerDeploymentDesc getInstance(String clientFileName,
                                                            ClassLoader classLoaderForCls,
                                                            String altClientXmlFilename)

        throws ClientContainerDeploymentDescException {
        // clientjar file
        JarFile clientFile = null;

        // Input streams
        InputStream applicationClientInputStream = null;
        InputStream jonasClientInputStream = null;
        
        // ZipEntry
        ZipEntry applicationClientZipEntry = null;
        ZipEntry jonasClientZipEntry = null;

        // Clients
        ApplicationClient applicationClient;
        JonasClient jonasClient;

        // Build the file
        File fClient = new File(clientFileName);

        
        //Check if the file exists.
        if (!(fClient.exists())) {
            String err = "' " + clientFileName + "' was not found.";
            throw new ClientContainerDeploymentDescException(err);
        }

        //Check if the Alt deploymentDesc file exists.
        //But only if it's a non null value because it's optionnal.
        if ((altClientXmlFilename != null) && (!new File(altClientXmlFilename).exists())) {
            String err = "The file for the altdd tag for the EAR case '" + altClientXmlFilename + "' was not found.";
            throw new ClientContainerDeploymentDescException(err);
        }


        // load the application-client deployment descriptor data (META-INF/application-client.xml 
        // and META-INF/jonas-client.xml)
        //No alt-dd case
        if (altClientXmlFilename == null) {
            try {
                clientFile = new JarFile(clientFileName);
                
                //Lookup in the JAR
                //Check the client entry
                applicationClientZipEntry = clientFile.getEntry(CLIENT_FILE_NAME);
                if (applicationClientZipEntry == null) {
                    throw new ClientContainerDeploymentDescException("The entry '" + CLIENT_FILE_NAME + "' was not found in the file '" + clientFileName + "'.");
                }
                //Get the stream
                applicationClientInputStream = clientFile.getInputStream(applicationClientZipEntry);
            } catch (Exception e) {
                if (clientFile != null) {
                    try {
                        clientFile.close();
                    } catch (IOException ioe) {
                        //We can't close the file
                    }
                }
                throw new ClientContainerDeploymentDescException("Can not read the XML deployment descriptors of the client jar file '" + clientFileName + "'.", e);
            }
        } else {
            try {
                applicationClientInputStream = new FileInputStream(altClientXmlFilename);
            } catch (FileNotFoundException ioe) {
                throw new ClientContainerDeploymentDescException("The altDD file '" + altClientXmlFilename + "' was not found.");
            }
        }
        applicationClient = loadApplicationClient(applicationClientInputStream, CLIENT_FILE_NAME);
        try {
            applicationClientInputStream.close();
        } catch (IOException e) {
            // Nothing to do
        }

            try {
                clientFile = new JarFile(clientFileName);
                
                // Lookup in the JAR
                // Check the client entry
                jonasClientZipEntry = clientFile.getEntry(JONAS_CLIENT_FILE_NAME);
                if (jonasClientZipEntry != null) {
                    //Get the stream
                    jonasClientInputStream = clientFile.getInputStream(jonasClientZipEntry);
                }
            } catch (Exception e) {
                if (clientFile != null) {
                    try {
                        clientFile.close();
                    } catch (IOException ioe) {
                        //We can't close the file
                    }
                }
                throw new ClientContainerDeploymentDescException("Can not read the XML deployment descriptors of the client jar file '" + clientFileName + "'.", e);
            }

        // load jonas-client deployment descriptor data 
        // (META-INF/jonas-client.xml)
        if (jonasClientInputStream != null) {
            jonasClient = loadJonasClient(jonasClientInputStream, JONAS_CLIENT_FILE_NAME);
            try {
                jonasClientInputStream.close();
            } catch (IOException e) {
                // Nothing to do
            }
        } else {
            jonasClient = new JonasClientImpl();
        }


        // instantiate client deployment descriptor
        return new ClientContainerDeploymentDesc(classLoaderForCls, applicationClient, jonasClient);
    }


    /** 
     * Load the application-client.xml file.
     * @param is the InputStream of the XML file.
     * @param fileName the name of the file (jonas-client.xml).
     * @return a structure containing the result of the jonas-client.xml parsing.
     * @throws ClientContainerDeploymentDescException if the deployment descriptor 
     * is corrupted.
     */
    private static ApplicationClient loadApplicationClient(InputStream is, String fileName) 
        throws ClientContainerDeploymentDescException {
        
        Reader reader = new InputStreamReader(is);
        SAXParserHelper helper = new SAXParserHelper(errorHandler, entityResolver);
        try {
            ApplicationClientUnmarshaller.setErrorHandler(helper.getWrappedErrorHandler());
            ApplicationClientUnmarshaller.setEntityResolver(helper.getWrappedEntityResolver());
            // Unmarshal the XML, and validate as it is unmarshalled
            return ApplicationClientUnmarshaller.unmarshal(reader, parsingWithValidation);
        } catch (IOException e) {
            if (helper.getSAXParseException() != null) {
                String detailMsg = getSAXMsg(fileName, helper.getSAXParseException(), "Invalid XML");
                throw new ClientContainerDeploymentDescException(detailMsg, e);
            } else {
                throw new ClientContainerDeploymentDescException("Error reading XML document " + fileName, e);
            }
        } finally {
            if (reader != null) {
                try {
                    reader.close();
                } catch (IOException ignored) { }
            }
        } 
    }

    /**
     * Load the jonas-client.xml file.
     * @param is the InputStream of the XML file.
     * @param fileName the name of the file (jonas-web.xml).
     * @return a structure containing the result of the jonas-web.xml parsing.
     * @throws ClientContainerDeploymentDescException if the deployment 
     * descriptor is corrupted.
     */
    private static JonasClient loadJonasClient(InputStream is, 
                                               String fileName) 
        throws ClientContainerDeploymentDescException {

        Reader reader = new InputStreamReader(is);
        SAXParserHelper helper = new SAXParserHelper(errorHandler, entityResolver);
        try {
            JonasClientUnmarshaller.setErrorHandler(helper.getWrappedErrorHandler());
            JonasClientUnmarshaller.setEntityResolver(helper.getWrappedEntityResolver());
            // Unmarshal the XML, and validate as it is unmarshalled
            return JonasClientUnmarshaller.unmarshal(reader, parsingWithValidation);
        } catch (IOException e) {
            if (helper.getSAXParseException() != null) {
                String detailMsg = getSAXMsg(fileName, helper.getSAXParseException(), "Invalid XML");
                throw new ClientContainerDeploymentDescException(detailMsg, e);
            } else {
                throw new ClientContainerDeploymentDescException("Error reading XML document " + fileName, e);
            }
        } finally {
            if (reader != null) {
                try {
                    reader.close();
                } catch (IOException ignored) { }
            }
        }
    }

    /**
     * Get an array of mapping between a res-env-ref name and its JNDI name.
     * @return an array of mapping between a res-env-ref name and its 
     * JNDI name.
     */
    public ResourceEnvRefDesc[] getResourceEnvRefTags() {
        ResourceEnvRefDesc ret[] = new ResourceEnvRefDesc[resourceEnvRefs.size()];
        resourceEnvRefs.copyInto(ret);
        return ret;
    }

    /**
     * Get an array of mapping between an res-ref name and its JNDI name.
     * @return an array of mapping between an res-ref name and its JNDI name.
     */
    public ResourceRefDesc[] getResourceRefTags() {
        ResourceRefDesc ret[] = new ResourceRefDesc[resourceRefs.size()];
        resourceRefs.copyInto(ret);
        return ret;
    }

    /**
     * Get an array of mapping between an entry name and its value.
     * @return an array of mapping between an entry name and its value.
     */
    public EnvEntryDesc[] getEnvEntryTags() {
        EnvEntryDesc ret[] = new EnvEntryDesc[envEntries.size()];
        envEntries.copyInto(ret);
        return ret;
    }

    /**
     * Get an array of mapping between an ejb name and its JNDI name.
     * @return an array of mapping between an ejb name and its JNDI name.
     */
    public EjbRefDesc[] getEjbRefTags() {
        EjbRefDesc ret[] = new EjbRefDesc[ejbRefs.size()];
        ejbRefs.copyInto(ret);
        return ret;
    }

    /**
     * Get the name of the jaas configuration file
     * @return the name of the jaas configuration file
     */
    public String getJaasFile() {
        return jaasFile;
    }

    /**
     * Get the entry in the jaas configuration file
     * @return the entry in the jaas configuration file
     */
    public String getJaasEntry() {
        return jaasEntry;
    }

    /**
     * Get the username used for a callback handler
     * @return the username used for a callback handler
     */
    public String getUsername() {
        return username;
    }

    /**
     * Get the password used for a callback handler
     * @return the password used for a callback handler
     */
    public String getPassword() {
        return password;
    }

    /**
     * Get the callback handler of this client application.
     * @return the callback handler of this client application.
     */
    public String getCallbackHandler() {
        return callbackHandler;
    }

    /**
     * Return a String representation of the ClientContainerDeploymentDesc.
     * @return a String representation of the ClientContainerDeploymentDesc.
     */
    public String toString() {
        StringBuffer ret = new StringBuffer();
        
        // Return the displayName
        ret.append("\ngetDisplayName()=" + getDisplayName());
        
        // Return the resource-env-ref
        ResourceEnvRefDesc[] rer = getResourceEnvRefTags();
        for (int i = 0; i < rer.length; i++) {
            ret.append("\ngetResourceEnvRefTags(" + i + ")=" + rer[i].getClass().getName());
            ret.append(rer[i].toString());
        }

        // Return the resource-ref
        ResourceRefDesc[] resourceRefDesc = getResourceRefTags();
        for (int i = 0; i < resourceRefDesc.length; i++) {
            ret.append("\ngetResourceRefTags(" + i + ")=" + resourceRefDesc[i].getClass().getName());
            ret.append(resourceRefDesc[i].toString());
        }
 
        // Return the env-entry
        EnvEntryDesc[] envEntries = getEnvEntryTags();
        for (int i = 0; i < envEntries.length; i++) {
            ret.append("\ngetEnvEntryTags(" + i + ")=" + envEntries[i].getClass().getName());
            ret.append(envEntries[i].toString());
        }

        // Return the ejb-ref
        EjbRefDesc[] ejbRefDesc = getEjbRefTags();
        for (int i = 0; i < ejbRefDesc.length; i++) {
            ret.append("\ngetEjbRefTags(" + i + ")=" + ejbRefDesc[i].getClass().getName());
            ret.append(ejbRefDesc[i].toString());
        }
        
        // Return the callbackHandler
        ret.append("\ngetCallbackHandler()=" + getCallbackHandler());

        return ret.toString();
    }

    /**
     * Get the current ErrorHandler.
     * @return the current ErrorHandler.
     */
    public static ErrorHandler getErrorHandler() {
        return errorHandler;
    }
    
    /**
     * Get the current EntityResolver.
     * @return the current EntityResolver.
     */
    public static EntityResolver getEntityResolver() {
        return entityResolver;
    }

    /**
     * Controls whether the parser is reporting all validity errors.
     * @return if true, all external entities will be read.
     */
    public static boolean getParsingWithValidation() {
        return ClientContainerDeploymentDesc.parsingWithValidation;
    }

    /**
     * Set the specified ErrorHandler. If the application does not register
     * an ErrorHandler, all error events reported by the reader are ignored.
     * @param errorHandler the Error Handler to set.
     */
    public static void setErrorHandler(ErrorHandler errorHandler) {
        ClientContainerDeploymentDesc.errorHandler = errorHandler;
    }

    /**
     * Set the specified EntityResolver.
     * @param entityResolver the Entity Resolver to set.
     */
    public static void setEntityResolver(EntityResolver entityResolver) {
        ClientContainerDeploymentDesc.entityResolver = entityResolver;
    }

    /**
     * Controls whether the parser is reporting all validity errors.
     * @param validation if true, all external entities will be read.
     */
    public static void setParsingWithValidation(boolean validation) {
        ClientContainerDeploymentDesc.parsingWithValidation = validation;
    }
}
