/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Florent BENOIT
 * --------------------------------------------------------------------------
 * $Id: ClientDeploymentDescManager.java,v 1.1 2003/06/27 09:03:14 benoitf Exp $
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas_client.deployment.lib;

//import java
import java.net.URL;
import java.net.URLClassLoader;
import java.net.MalformedURLException;
import java.io.File;
import java.io.IOException;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.StringTokenizer;


//import jonas_client
import org.objectweb.jonas_client.deployment.api.ClientContainerDeploymentDesc;
import org.objectweb.jonas_client.deployment.api.ClientContainerDeploymentDescException;
import org.objectweb.jonas_client.deployment.api.EjbRefDesc;

//import jonas_ejb
import org.objectweb.jonas_ejb.deployment.api.BeanDesc;
import org.objectweb.jonas_ejb.deployment.api.DeploymentDescException;
import org.objectweb.jonas_ejb.deployment.api.DeploymentDesc;
import org.objectweb.jonas_ejb.deployment.api.SessionDesc;
import org.objectweb.jonas_ejb.deployment.api.EntityDesc;
import org.objectweb.jonas_ejb.deployment.lib.EjbDeploymentDescManager;

/**
 * This class provide a way for managing the ClientContainerDeploymentDesc.
 * @author Florent Benoit
 */
public class ClientDeploymentDescManager {

    /**
     * Char separator in an ejbLink.
     */
    private static final String EJB_LINK_SEPARATOR = "#";

    /**
     * The unique instance of the ClientDeploymentDescManager.
     */
    private static ClientDeploymentDescManager unique;

    /**
     * Reference on the EjbDeploymentDescManager.
     */
    private EjbDeploymentDescManager ejbDDManager = null;

    /**
     * Associate a ear classLoader to an hashtable which contains
     * association between Urls of wars and their optional alt-dd
     */
    private Hashtable earCLAltDDBindings = null;


    /**
     * Contructs a unique ClientDeploymentDescManager
     */
    private ClientDeploymentDescManager() {
        ejbDDManager = EjbDeploymentDescManager.getInstance();
        earCLAltDDBindings = new Hashtable();
    }

    /**
     * Get an instance of the ClientDeploymentDescManager.
     * @return the instance of the ClientDeploymentDescManager.
     */
    public static ClientDeploymentDescManager getInstance() {
        if (unique == null) {
            unique = new ClientDeploymentDescManager();
        }
        return unique;
    }
 
    /**
     * Get the specified web deployment descriptor.
     * @param url the url where to load xml deployment descriptors.
     * @param loaderForCls classloader used to load web classes.
     * @param earLoader the ear classloader.
     * @return ClientContainerDeploymentDesc the web deployment descriptor. 
     * @throws ClientContainerDeploymentDescException when 
     * ClientContainerDeploymentDesc cannot be created  with the given files.
     */
    public ClientContainerDeploymentDesc getDeploymentDesc(URL url, 
                                                   ClassLoader loaderForCls,
                                                   ClassLoader earLoader) 
        throws ClientContainerDeploymentDescException {

        // Check if the jar exists ...
        if (!new File(url.getFile()).exists()) {
            String err = "Cannot get the deployment descriptor for ";
            err = err + "'" + url.getFile() + "'. The file doesn't exist.";
            throw new ClientContainerDeploymentDescException(err);
        }

        //url used to load an alternate DDesc in the EAR case
        URL altDDUrl = null;

        //check if it's an Ear case or not
        Hashtable urlAltddBindings = null;
        if (earLoader != null) {
            //Mapping ?
            urlAltddBindings = (Hashtable) earCLAltDDBindings.get(earLoader);          
            if (urlAltddBindings == null) {
                //If there is no mapping, the setAltDD function was badly called
                String err = "Cannot find if there is alt-dd for '" + url.getFile() + "', the setAltDD function was badly called";
                throw new ClientContainerDeploymentDescException(err);
            }
            //Now we can get the optional alt-dd url file
            altDDUrl = (URL) urlAltddBindings.get(url);
        }


        // ... and get the instance of the ClientContainerDeploymentDesc.
        // If there is an alternate url for the application-client.xml, call the method with this param.
        ClientContainerDeploymentDesc clientDD = null;
        if (altDDUrl != null) {
            clientDD = ClientContainerDeploymentDesc.getInstance(url.getFile(), loaderForCls, altDDUrl.getFile());
        } else {
            clientDD = ClientContainerDeploymentDesc.getInstance(url.getFile(), loaderForCls);
        }

        // Resolve the ejb-link for ejb-ref
        EjbRefDesc[] ejbRef = clientDD.getEjbRefTags();
        for (int i = 0; i < ejbRef.length; i++) {
            if (ejbRef[i].getJndiName() == null) {
                String ejbLink = ejbRef[i].getEjbLink();
                String ejbRefType = ejbRef[i].getEjbRefType();
                if (ejbLink != null) {
                    if (earLoader == null) {
                        throw new ClientContainerDeploymentDescException("Ejb-link is not authorized from a single client jar. The client jar must be in an ear.");
                    } else {
                        String jndiName = getJndiName(url, ejbLink, earLoader, ejbRefType);
                        ejbRef[i].setJndiName(jndiName);
                    }
                }
            }
        }
        return clientDD;
    }

    /**
     * Return the JNDI name from the ejbLink string.
     * ejbLink format : filename.jar#beanName in the same Ear File
     * @param clientURL the url of the jar being parsed. This is needed 
     * because ejbLink is relative. With the url and the ejbLink, we can 
     * know where the file is locate.
     * @param ejbLink the ejbLink tag of an ejb-ref.
     * @param earLoader the classloader of the ear.
     * @param ejbType the type of the referenced ejb in the ejb-ref tag.
     * @return the JNDI name if found, null otherwise
     * @throws ClientContainerDeploymentDescException when it failed
     */
    private String getJndiName(URL clientURL, String ejbLink,
                               ClassLoader earLoader, String ejbType) 
        throws ClientContainerDeploymentDescException {

        // Extract from the ejb link
        //   - the name of the file
        //   - the name of the bean
        String ejbJarLink = null;
        String beanNameLink = null;

        // Check the format of the ejb-link. It must contains .jar#
        if (ejbLink.toLowerCase().indexOf(".jar#") == -1) {
            String err = "Ejb-link " + ejbLink + " has a bad format. Correct format :  filename.jar#beanName";
            throw new ClientContainerDeploymentDescException(err);
        }
        StringTokenizer st = new StringTokenizer(ejbLink, EJB_LINK_SEPARATOR);

        // We must have only two elements after this step, one for the fileName
        //   before the # and the name of the bean after the # char       
        if (st.countTokens() != 2 || ejbLink.startsWith("#") || ejbLink.endsWith("#")) {
            String err = "Ejb-link " + ejbLink + " has a bad format. Correct format :  filename.jar#beanName";
            throw new ClientContainerDeploymentDescException(err);
        }

        //Get the token
        ejbJarLink = (String) st.nextToken();
        beanNameLink = (String) st.nextToken();

        //Check if ejbJarLink is a jar or not
        if (!ejbJarLink.endsWith(".jar")) {
            String err = "Ejbjar filename " + ejbJarLink + " from the ejb-link " + ejbLink + " has a bad format. Correct format :  filename.jar";
            throw new ClientContainerDeploymentDescException(err);
        }


        // Now construct the URL from the absolute path from the url ejbJar and
        // the relative path from ejbJarLink
        URL ejbJarLinkUrl = null;
        try {
            ejbJarLinkUrl = new File(new File(clientURL.getFile()).getParent() + File.separator + ejbJarLink).getCanonicalFile().toURL();
        } catch (MalformedURLException mue) {
            String err = "Error when creating an url for the ejb jar filename. Error :" + mue.getMessage();
            throw new ClientContainerDeploymentDescException(err);
        } catch (IOException ioe) {
            String err = "Error when creating/accessing a file. Error :" + ioe.getMessage();
            throw new ClientContainerDeploymentDescException(err);
        }

        // Check if the jar exist.
        if (!new File(ejbJarLinkUrl.getFile()).exists()) {
            String err = "Cannot get the deployment descriptor for '" + ejbJarLinkUrl.getFile() + "'. The file doesn't exist.";
            throw new ClientContainerDeploymentDescException(err);
        }

        // We've got the url
        //   Now, We can ask the Deployment Descriptor of this url
        URL[] ddURL = new URL[1];
        ddURL[0] = ejbJarLinkUrl;
        URLClassLoader loaderForClsEjb = new URLClassLoader(ddURL, earLoader);
        DeploymentDesc ejbDD = null;
        try {
            ejbDD = ejbDDManager.getDeploymentDesc(ejbJarLinkUrl, loaderForClsEjb, earLoader);
        } catch (DeploymentDescException e) {
            String err = "Cannot get the deployment descriptor for '" + ejbJarLinkUrl.getFile() + "'.";
            throw new ClientContainerDeploymentDescException(err, e);
        }

        //JndiName resolve.
        BeanDesc bd = ejbDD.getBeanDesc(beanNameLink);

        if (bd == null) {
            String err = "Ejb-link " + ejbLink + " not found inside the file " + clientURL.getFile() + ". The bean doesn't exists";
            throw new ClientContainerDeploymentDescException(err);
        }

        //Check if the type of the ejb-ref is correct.
        if (bd instanceof SessionDesc) {
            if (!ejbType.equalsIgnoreCase("Session")) {
                String err = "Deployment desc '" + clientURL.getFile() + "' has an incompatible ejb-ref-type: Required Session but found " + ejbType;
                throw new ClientContainerDeploymentDescException(err);
            }
        } else if (bd instanceof EntityDesc) {
            if (!ejbType.equalsIgnoreCase("Entity")) {
                String err = "Deployment desc '" + clientURL.getFile() + "' has an incompatible ejb-ref-type: Required Entity but found " + ejbType;
                throw new ClientContainerDeploymentDescException(err);
            }
        } else {
            String err = "Deployment desc '" + clientURL.getFile() + "' has a bad ejb-ref-type.";
            throw new ClientContainerDeploymentDescException(err);
        }

        return bd.getJndiName();
    }

    /**
     * Make a cleanup of the cache of deployment descriptor. This method must 
     * be invoked after the ear deployment by the EAR service.
     * @param earClassLoader the ClassLoader of the ear application to
     * remove from the cache.
     */
    public void removeCache(ClassLoader earClassLoader) {
        //Remove the altdd mapping
        earCLAltDDBindings.remove(earClassLoader);

        //Then remove the cache of the ejb dd manager
        ejbDDManager.removeCache(earClassLoader);
    }


    /**
     * Set the alt deployment desc which are used instead of
     * the web.xml file which is in the war file.
     * The alt-dd tag is in the application.xml file of the ear files
     * and is used ony in the EAR case. ie : deployment of wars packaged into
     * EAR applications. alt-dd tag is optionnal
     * @param earClassLoader the ear classloader which is used for mapped
     * the URLs of the wars to the Alt dd.
     * @param urls the urls of the wars
     * @param altDDs the alt-dd name for the specified war URLs
     */
    public void setAltDD(ClassLoader earClassLoader, URL[] urls,
                         URL[] altDDs) {
        
        //Associate an url to a altDD url
        Hashtable urlAltddBindings = new Hashtable();

        //Fill the hashtable for each url
        for (int i = 0; i < urls.length; i++) {
            if (altDDs[i] != null) {
                urlAltddBindings.put(urls[i], altDDs[i]);
            }
        }
        
        //Bind the hashtable
        earCLAltDDBindings.put(earClassLoader, urlAltddBindings);

    }

    /**
     * Get the size of the cache (number of entries in the cache).
     * This method is used only for the tests.
     * @return the size of the cache (number of entries in the cache).
     */
    public int getCacheSize() {
        int bufferSize = 0;

        Enumeration enum = earCLAltDDBindings.keys();
        while (enum.hasMoreElements()) {
            ClassLoader loader = (ClassLoader) enum.nextElement();
            Hashtable hashtab = (Hashtable) earCLAltDDBindings.get(loader);
            bufferSize = bufferSize + hashtab.size();
        }

        return bufferSize;
    }



}
