/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: JEntityFactory.java,v 1.26 2003/10/22 16:07:45 durieuxp Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ejb.container;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;
import javax.ejb.EJBException;
import javax.ejb.EntityBean;
import javax.naming.Context;
import javax.naming.NamingException;
import javax.transaction.RollbackException;
import javax.transaction.SystemException;
import javax.transaction.Transaction;
import org.objectweb.jonas_ejb.container.jorm.JormFactory;
import org.objectweb.jonas_ejb.deployment.api.EntityDesc;
import org.objectweb.jonas_ejb.deployment.api.EntityJdbcCmp1Desc;
import org.objectweb.jonas_ejb.deployment.api.EntityJdbcCmp2Desc;
import org.objectweb.jorm.api.PMapper;
import org.objectweb.medor.eval.prefetch.api.PrefetchCache;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * This class is a factory for an Entity Bean. It is responsible for
 * - managing Home and LocalHome.
 * - managing a pool of instances/contexts
 * - keeping the list of PKs and the associated JEntitySwitch's
 * - keeping the JNDI context for this component (java:comp/env)
 * @author Philippe Coq, Philippe Durieux
 */
public class JEntityFactory extends JFactory {

    /**
     * optional home
     */
    protected JEntityHome home = null;

    /**
     * optional local home
     */
    protected JEntityLocalHome localhome = null;

    /**
     * reentrant if instance can be used concurrently
     */
    protected boolean reentrant = false;

    /**
     * shared if the EJB container is not the only one to modify
     * the bean state on the database, or if a cluster of container
     * access the bean concurrently.
     */
    protected boolean shared = false;

    /**
     * enable the prefetch for CMP2 bean
     */
    protected boolean prefetch = false;

    /**
     * instance pool management (list of available JEntityContext objects)
     */
    protected List bctxlist = Collections.synchronizedList(new ArrayList());

    /**
     * initial value for pool size
     */
    protected int minPoolSize;

    protected int instanceCount = 0;
    protected int maxCacheSize;

    /**
     * List of JEntitySwitch objects
     * At each PK is associated a JEntitySwitch object
     */
    protected HashMap pklist = new HashMap();

    /**
     * List of Transaction Listeners.
     * At each Transaction is associated a TxListener object
     */
    protected HashMap txlist = new HashMap();

    /**
     * dirty list of JEntitySwitch
     */
    private LinkedList dirtyList = new LinkedList();

    // Datasource in case of CMP
    protected Object datasource = null;


    /**
     * This field is used to generate an uid for unique automatic pk 
     */   
    private static int count = 1; 

    /**
     * constructor
     * must be without parameters (required by Jorm)
     */
    public JEntityFactory() {
        TraceEjb.interp.log(BasicLevel.DEBUG, "");
    }

    /**
     * Init this object
     * @param dd the deployment descriptor
     * @param cont the Container
     */
    public void init(EntityDesc dd, JContainer cont) {
        super.init(dd, cont);
        reentrant = dd.isReentrant();
        shared = dd.isShared();
        prefetch = dd.isPrefetch();
        maxCacheSize = dd.getCacheMax();
        minPoolSize = dd.getPoolMin();

        // Finds the DataSource associated to the bean
        // Useful for Entity with Container Managed Persistence.
        if (dd instanceof EntityJdbcCmp1Desc) {
            String dsname = ((EntityJdbcCmp1Desc) dd).getDatasourceJndiName();
            if (dsname != null) {
                try {
                    datasource = getInitialContext().lookup(dsname);
                } catch (NamingException e) {
                    TraceEjb.logger.log(BasicLevel.ERROR, dsname + " is not known by the EJB server.", e);
                    throw new EJBException(dsname + " unknown");
                }
            }
        }

        if (dd instanceof EntityJdbcCmp2Desc) {
            String dsname = ((EntityJdbcCmp2Desc) dd).getDatasourceJndiName();
            if (dsname != null) {
                try {
                    datasource = getInitialContext().lookup(dsname);
                } catch (NamingException e) {
                    TraceEjb.logger.log(BasicLevel.ERROR, dsname + " is not known by the EJB server.");
                    throw new EJBException(dsname + " unknown");
                }
            }
        }

        // Create the Home if defined in DD
        Class homeclass = null;
        String clname = dd.getFullWrpHomeName();
        if (clname != null) {
            try {
                homeclass = cont.getClassLoader().loadClass(clname);
            } catch (ClassNotFoundException e) {
                throw new EJBException(ejbname + "Cannot load " + clname);
            }
            if (TraceEjb.isDebugIc()) {
                TraceEjb.interp.log(BasicLevel.DEBUG, ejbname + ": " + clname + " loaded");
            }
            try {
                // new JEntityHome(dd, this)
                int nbp = 2;
                Class ptype[] = new Class[nbp];
                Object pobj[] = new Object[nbp];
                ptype[0] = org.objectweb.jonas_ejb.deployment.api.EntityDesc.class;
                pobj[0] = (Object) dd;
                ptype[1] = org.objectweb.jonas_ejb.container.JEntityFactory.class;
                pobj[1] = (Object) this;
                home = (JEntityHome) homeclass.getConstructor(ptype).newInstance(pobj);
            } catch (Exception e) {
                if (TraceEjb.isVerbose()) {
                    e.printStackTrace();
                }
                throw new EJBException(ejbname + " Cannot create home ", e);
            }

            // register it in JNDI
            try {
                home.register();
            } catch (Exception e) {
                throw new EJBException(ejbname + " Cannot register home ", e);
            }
        }

        // Create the LocalHome if defined in DD
        clname = dd.getFullWrpLocalHomeName();
        if (clname != null) {
            try {
                homeclass = cont.getClassLoader().loadClass(clname);
            } catch (ClassNotFoundException e) {
                TraceEjb.logger.log(BasicLevel.ERROR, ejbname + " cannot load " + clname);
                throw new EJBException("Cannot load " + clname);
            }
            if (TraceEjb.isDebugIc()) {
                TraceEjb.interp.log(BasicLevel.DEBUG, ejbname + ": " + clname + " loaded");
            }
            try {
                // new JEntityLocalHome(dd, this)
                int nbp = 2;
                Class ptype[] = new Class[nbp];
                Object pobj[] = new Object[nbp];
                ptype[0] = org.objectweb.jonas_ejb.deployment.api.EntityDesc.class;
                pobj[0] = (Object) dd;
                ptype[1] = org.objectweb.jonas_ejb.container.JEntityFactory.class;
                pobj[1] = (Object) this;
                localhome = (JEntityLocalHome) homeclass.getConstructor(ptype).newInstance(pobj);
            } catch (Exception e) {
                throw new EJBException(ejbname + " Cannot create localhome ", e);
            }
            // register it in JNDI
            try {
                localhome.register();
            } catch (Exception e) {
                throw new EJBException(ejbname + " Cannot register localhome ", e);
            }
        }

        // pre-allocate a set of JEntityContext (bean instances)
        for (int i = 0; i < minPoolSize; i++) {
            JEntityContext ctx = null;
            try {
                ctx = createNewInstance();
                bctxlist.add(ctx);
            } catch (Exception e) {
                TraceEjb.logger.log(BasicLevel.ERROR, ejbname + " cannot create new instance", e);
                throw new EJBException(ejbname + " Cannot init pool of instances ", e);
            }
        }
    }

    /**
     * get EJB by its PK
     * Creates if not exist yet.
     * @param pk The Primary Key Object
     * @return The JEntitySwitch matching the PK.
     */
    public synchronized JEntitySwitch getEJB(Object pk) {

        TraceEjb.interp.log(BasicLevel.DEBUG, "");

        if (pk == null) {
            TraceEjb.logger.log(BasicLevel.ERROR, "pk null ???");
            return null;
        }
        JEntitySwitch bs = (JEntitySwitch) pklist.get(pk);
        if (bs == null) {
            bs = getJEntitySwitch();
            bs.init(this, pk);
            pklist.put(pk, bs);
        }
        return bs;
    }

    /**
     * get EJB by its PK
     * @param pk The Primary Key Object
     * @return The JEntitySwitch matching the PK, or null if none exist.
     */
    public synchronized JEntitySwitch existEJB(Object pk) {
        if (pk == null) {
            TraceEjb.logger.log(BasicLevel.ERROR, "pk null ???");
            return null;
        }
        JEntitySwitch ret = (JEntitySwitch) pklist.get(pk);
        TraceEjb.interp.log(BasicLevel.DEBUG, ret == null ? "no" : "yes");
        return ret;
    }

    /**
     * Bind a PK with a JEntitySwitch
     * @param pk The Primary Key Object
     * @param bs The JEntitySwitch
     */
    public synchronized void bindEJB(Object pk, JEntitySwitch bs) {
        TraceEjb.interp.log(BasicLevel.DEBUG, "");
        if (pk == null) {
            TraceEjb.logger.log(BasicLevel.ERROR, "pk null ???");
            return;
        }
        bs.init(this, pk);
        pklist.put(pk, bs);
    }

    /**
     * This method allocates a new JEntitySwitch. But no association has been
     * done between the primary key and the new JEntitySwitch. Therefore the
     * initialisation is not done.
     * @return The JEntitySwitch.
     */
    public JEntitySwitch getJEntitySwitch() {
        return new JEntitySwitch();
    }

    /**
     * remove an EJB by its PK
     * @param pk The Primary Key Object
     */
    public synchronized void removeEJB(Object pk) {
        TraceEjb.interp.log(BasicLevel.DEBUG, "");
        if (pk == null) {
            TraceEjb.logger.log(BasicLevel.ERROR, "pk null ???");
            return;
        }
        pklist.remove(pk);
    }

    /**
     * Register an EntitySwitch in the dirty list.
     * should be called each time a new instance is modified outside transaction
     * @param ejb The Entity Switch to be registered
     */
    public synchronized void registerEJB(JEntitySwitch ejb) {
        TraceEjb.swapper.log(BasicLevel.DEBUG, "");
        if (dirtyList.contains(ejb)) {
            TraceEjb.swapper.log(BasicLevel.ERROR, ejbname + " Elt already in the dirty list");
            return;
        }
        dirtyList.addLast(ejb);
    }

    /**
     * Ask swapper to synchronize all dirty EntitySwitch
     */
    public void synchronizeEntities() {
        TraceEjb.swapper.log(BasicLevel.DEBUG, "");
        cont.registerBFS(this);
    }

    // ---------------------------------------------------------------
    // BeanFactory implementation
    // ---------------------------------------------------------------

    /**
     * @return the Instance pool size for this Ejb
     */
    public int getPoolSize() {
        return bctxlist.size();
    }

    /**
     * stop this EJB.
     * Mainly unregister it in JNDI.
     */
    public void stop() {
        TraceEjb.interp.log(BasicLevel.DEBUG, "");
        try {
            if (home != null) {
                home.unregister();
            }
            if (localhome != null) {
                localhome.unregister();
            }
        } catch (NamingException e) {
        }
    }

    /**
     * Synchronize all dirty instances
     */
    public void sync() {
        TraceEjb.swapper.log(BasicLevel.DEBUG, "");

        // Invalid prefetch cache for "no tx"
        if (this instanceof JormFactory) {
            PMapper mapper = ((JormFactory) this).getMapper();
            PrefetchCache pc = mapper.getPrefetchCache();
            if (pc != null) {
                pc.invalidatePrefetchBuffer("NO_TX");
            }
        }
        
        // Set the JNDI component context (for ejbStore)
        Context bnctx = setComponentContext();

        // try to passivate all instances and build a new list.
        LinkedList wsdone = new LinkedList();
        synchronized (this) {
            for (ListIterator i = dirtyList.listIterator(0); i.hasNext(); ) {
                JEntitySwitch es = (JEntitySwitch) i.next();
                if (es.passivateIH(false)) {
                    // passivation OK.
                    i.remove();
                    wsdone.addLast(es);
                } else {
                    TraceEjb.swapper.log(BasicLevel.DEBUG, "swapper could not passivate an instance");
                }
            }
        }

        // for each instance passivated, notify it's OK to continue.
        // must be done after all instances are passivated, in case of relations.
        for (ListIterator i = wsdone.listIterator(0); i.hasNext(); ) {
            JEntitySwitch es = (JEntitySwitch) i.next();
            es.endIH();
        }

        // Reset old value for component context
        resetComponentContext(bnctx);
    }

    /**
     * Reduce number of instances in memory
     */
    public void reduceCache() {
        TraceEjb.swapper.log(BasicLevel.DEBUG, "");

        // passivate all instances that are not busy
        // count is the number of busy instances (not freed)
        // other instances should be added to the pool (bctxlist)
        int count = 0;

        // Set the JNDI component context (for ejbStore)
        Context bnctx = setComponentContext();

        Collection coll = null;
        HashMap clone = null;
        synchronized (this) {
            clone = (HashMap) pklist.clone();
        }
        JEntitySwitch bs = null;
        for (Iterator i = clone.values().iterator(); i.hasNext(); ) {
            bs = (JEntitySwitch) i.next();
            if (!bs.passivateIH(true)) {
                count++;
            }
        }
        // Reset old value for component context
        resetComponentContext(bnctx);


        // compute the new size we want for the pool
        // (if more busy instances than min size, we want an empty pool)
        int poolsz = minPoolSize - count;
        if (poolsz < 0) {
            poolsz = 0;
        }

        // reduce the pool to this new value
        synchronized (bctxlist) {
            TraceEjb.swapper.log(BasicLevel.DEBUG, "try to reduce " + bctxlist.size() + " to " + poolsz);
            while (bctxlist.size() > poolsz) {
                ListIterator i = bctxlist.listIterator();
                if (i.hasNext()) {
                    i.next();
                    i.remove();
                    instanceCount--; // here ???
                }
            }
        }
        TraceEjb.swapper.log(BasicLevel.DEBUG, "cacheSize " + getCacheSize());
    }

    /**
     * @return the home if it exists
     */
    public JHome getHome() {
        return home;
    }

    /**
     * @return the local home if it exists
     */
    public JLocalHome getLocalHome() {
        return localhome;
    }

    // ---------------------------------------------------------------
    // Context Pool
    // ---------------------------------------------------------------

    /**
     * Get a Context from the pool, or create a new one if no more
     * available in the pool.
     * This JContext must be initialized then by the caller.
     * @return a JEntityContext, not initialized.
     */
    public JEntityContext getJContext() {
        TraceEjb.interp.log(BasicLevel.DEBUG, "");
        JEntityContext bctx = null;

        // try to find a free context in the pool
        synchronized (bctxlist) {
            try {
                ListIterator i = bctxlist.listIterator();
                if (i.hasNext()) {
                    bctx = (JEntityContext) i.next();
                    i.remove();
                }
            } catch (IndexOutOfBoundsException ex) {
                // pool is empty
            }
        }

        if (bctx == null) {
            // create a new one if pool empty
            try {
                bctx = createNewInstance();
            } catch (Exception e) {
                throw new EJBException("Cannot create a new instance: " + e.toString());
            }
        }
        return bctx;
    }

    /**
     * Release a Context
     * @param ctx - The JContext to release.
     */
    public void releaseJContext(JContext ctx) {
        TraceEjb.interp.log(BasicLevel.DEBUG, "");
        JEntityContext bctx = (JEntityContext) ctx;
        bctx.razEntityContext();
        bctxlist.add(bctx);
    }

    // ---------------------------------------------------------------
    // other public methods
    // ---------------------------------------------------------------

    /**
     * @return min pool size
     * for Jmx
     */
    public int getMinPoolSize() {
        return minPoolSize;
    }

    /**
     * @return max cache size
     * for Jmx
     */
    public int getMaxCacheSize() {
        return maxCacheSize;
    }

    /**
     * @return current cache size
     * for Jmx
     */
    public int getCacheSize() {
        return instanceCount;
    }

    /**
     * @return true if reentrant
     */
    public boolean isReentrant() {
        return reentrant;
    }

    /**
     * @return true if shared
     */
    public boolean isShared() {
        return shared;
    }

    /**
     * @return true if prefetch enable
     */
    public boolean isPrefetch() {
        return prefetch;
    }

    /**
     * @return the Datasource used for CMP
     */
    public Object getDataSource() {
        if (datasource == null) {
            TraceEjb.logger.log(BasicLevel.ERROR, "datasource not defined in JEntityFactory");
        }
        return datasource;
    }

    /**
     * Check Transaction before calling a method on a bean.
     * For Entity beans, the only possible case is "Container Managed Tx"
     * @param rctx The Request Context
     */
    public void checkTransaction(RequestCtx rctx) {
        checkTransactionContainer(rctx);
    }

    /**
     * synchronize data modified in this transaction.
     * this is necessary in case of finder method, because
     * ejb-ql looks for on disk.
     * @param tx the Transaction
     */
    public void syncForFind(Transaction tx) {
        TraceEjb.interp.log(BasicLevel.DEBUG, "");
        if (tx == null) {
            sync();
        } else {
            TxListener txl = (TxListener) txlist.get(tx);
            if (txl != null) {
                txl.storeInstances();
            }
        }
    }

    /**
     * Remove a Transaction Listener from the list.
     * @param tx the Transaction to remove
     */
    public void removeTxListener(Transaction tx) {
        TraceEjb.interp.log(BasicLevel.DEBUG, "");
        txlist.remove(tx);
    }

    /**
     * Register a Context/Instance in the transaction.
     * @param tx current Transaction
     * @param ctx JEntityContext to be registered
     */
    public void registerContext(Transaction tx, JEntityContext ec) throws IllegalStateException {
        TraceEjb.interp.log(BasicLevel.DEBUG, "");

        // Get the TXListener associated to this tx, or create it.
        TxListener txl = (TxListener) txlist.get(tx);
        if (txl == null) {
            txl = new TxListener(this, tx);
            txlist.put(tx, txl);
            // register it as Synchronization.
            try {
                tx.registerSynchronization(txl);
            } catch (RollbackException e) {
                TraceEjb.logger.log(BasicLevel.WARN, "transaction has been marked rollbackOnly");
            } catch (SystemException e) {
                TraceEjb.logger.log(BasicLevel.ERROR, "cannot register synchro ", e);
                throw new IllegalStateException("cannot register synchro");
            }
        }

        // Add the JEntityContext 
        txl.addInstance(ec);
    }

    /**
     * Take a dump of current entity counters and return them
     * @return EntityCounters
     */
    public synchronized EntityCounters getEntityCounters() {
        EntityCounters ec = new EntityCounters();
        Collection coll = pklist.values();
        for (Iterator i = coll.iterator(); i.hasNext(); ) {
            JEntitySwitch es = (JEntitySwitch) i.next();
            switch (es.getState()) {
            case 0:
                ec.inTx++;
                break;
            case 1:
                ec.outTx++;
                break;
            case 2:
                ec.idle++;
                break;
            case 3:
                ec.passive++;
                break;
            case 4:
                ec.removed++;
                break;
            }
        }
        return ec;
    }


    /** 
     * Calculate a new uid for automatic pk creation
     * Used by JEntityCmpJdbc.vm
     * @return int (unique pk)
     */
    public int calculateAutomaticPk() {
        TraceEjb.interp.log(BasicLevel.DEBUG, "");
        int uid;
        synchronized (getClass()) {
            uid = (int) (System.currentTimeMillis() % (java.lang.Integer.MAX_VALUE / 10));
            uid = uid * 10 + count++;
        }
        return uid;

    }

    // ---------------------------------------------------------------
    // private methods
    // ---------------------------------------------------------------

    /**
     * Create a new instance of the bean and its EntityContext
     * In case of CMP, the bean class is derived to manage entity persistence.
     * @return JEntityContext
     * @throws Exception cannot instantiate bean
     */
    protected JEntityContext createNewInstance() throws Exception  {
        TraceEjb.interp.log(BasicLevel.DEBUG, "");

        // create the bean instance
        EntityBean bean = null;
        try {
            bean = (EntityBean) beanclass.newInstance();
        } catch (InstantiationException e) {
            TraceEjb.logger.log(BasicLevel.ERROR, ejbname + " cannot instantiate bean");
            throw e;
        } catch (IllegalAccessException e) {
            TraceEjb.logger.log(BasicLevel.ERROR, ejbname + " cannot instantiate bean");
            throw e;
        }

        // create a new EntityContext and bind it to the instance
        JEntityContext ec = new JEntityContext(this, bean);
        bean.setEntityContext(ec);

        // If too much instances, ask swapper to clean memory.
        instanceCount++;
        if (maxCacheSize != 0 && instanceCount > maxCacheSize) {
            TraceEjb.swapper.log(BasicLevel.DEBUG, "too much instances :" + instanceCount + ", max=" + maxCacheSize);
            cont.registerBF(this);
        }

        return ec;
    }

}
