/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: JEntitySwitch.java,v 1.27 2003/09/16 15:43:39 durieuxp Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ejb.container;

import java.rmi.NoSuchObjectException;
import java.rmi.RemoteException;
import javax.ejb.EJBException;
import javax.ejb.EntityBean;
import javax.ejb.TimedObject;
import javax.ejb.Timer;
import javax.ejb.TimerService;
import javax.ejb.TransactionRolledbackLocalException;
import javax.transaction.RollbackException;
import javax.transaction.Status;
import javax.transaction.SystemException;
import javax.transaction.Transaction;
import org.objectweb.jonas_ejb.deployment.api.MethodDesc;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * JEntitySwitch is used internally to synchronize accesses to the entity context
 * and thus to the entity bean instance.
 * All parts common to EJBObject and EJBLocalObject should be here.
 * Different policies can be applied to manage context/instance pairs:
 * - only 1 pair (container manages the transaction isolation)
 * - 1 pair for each transaction (transaction isolation managed by DataBase)
 * - 2 pairs (1 for transactional accesses, 1 for non transaction accesses)
 * In the current version, we implement only the first. Others will be implemented later.
 * @author Philippe Durieux
 * @author Philippe Coq
 */
public class JEntitySwitch {

    protected JEntityFactory bf;
    protected Object pk = null;
    protected JEntityLocal local = null;
    protected JEntityRemote remote = null;
    private JContainer cont;
    protected boolean shared;
    protected boolean reentrant;
    private TimerService myTimerService = null;

    // nb of non transacted requests running
    protected int countIH = 0;

    // unique EntityContext
    protected JEntityContext iContext = null;

    // ident is used mainly for debugging
    private static int counter = 0;
    protected String ident;

    // nb of threads waiting for the Context
    private int waiters = 0;

    // True if ctx has been used outside transactions.
    private boolean dirty = false;

    /**
     * True if a TX need this instance currently used outside tx
     * When the last release is done, we must store this instance.
     */
    private boolean mustStore = false;

    // True if we must reload the bean state before use.
    private boolean mustLoad = false;

    // True when context/instance has been discarded (outside tx only)
    private boolean discarded = false;

    // current tx running on this instance and not yet committed.
    protected Transaction currtx = null;

    // max time to wait when in a transaction, before checking if
    // transaction has been set rollback only (deadlock detection)
    private static int maxtime = 5000; // 5 sec.

    /**
     * empty constructor. Object is initialized via init() because it is
     * implemented differently according to jorm mappers.
     */
    public JEntitySwitch() {
    }

    /**
     * constructor.
     * A new object is build when a new PK is known in the container,
     * either when a new bean is created, or when a find occurs.
     * For create(), PK is not known yet when this object is build.
     * @param bf The Entity Factory
     * @param pk The Primary Key
     */
    public void init(JEntityFactory bf, Object pk) {
        TraceEjb.interp.log(BasicLevel.DEBUG, "pk=" + pk);
        this.bf = bf;
        this.pk = pk;
        if (pk == null) {
            TraceEjb.logger.log(BasicLevel.ERROR, ident + "Init Entity Switch with a null PK!");
            throw new EJBException("Init Entity Switch with a null PK!");
        }

        // keep these in local, to be more efficient.
        cont = bf.getContainer();
        shared = bf.isShared();
        reentrant = bf.isReentrant();

        // generate new ident (debug)
        ident = "<" + counter + ">";
        counter++;

        iContext = null;
        countIH = 0;
        waiters = 0;
        dirty = false;
        mustStore = false;
        mustLoad = false;
        discarded = false;
        currtx = null;

        // Create EJBObject if bean has a Remote Interface
        if (bf.getHome() != null) {
            try {
                remote = ((JEntityHome) bf.getHome()).createRemoteObject();
                remote.setEntitySwitch(this);
            } catch (RemoteException e) {
                throw new EJBException("cannot create Remote Object", e);
            }
        }

        // Create EJBLocalObject if bean has a Local Interface
        if (bf.getLocalHome() != null) {
            local = ((JEntityLocalHome) bf.getLocalHome()).createLocalObject();
            local.setEntitySwitch(this);
        }
    }

    /**
     * @return the underlaying EJBLocalObject
     */
    public JEntityLocal getLocal() {
        return local;
    }

    /**
     * @return the underlaying EJBObject
     */
    public JEntityRemote getRemote() {
        return remote;
    }

    /**
     * Obtains the TimerService associated for this Entity Bean (one / pk)
     * @return a JTimerService instance.
     */
    public TimerService getEntityTimerService() {
        if (myTimerService == null) {
            // TODO : Check that instance implements TimedObject ?
            myTimerService = new JTimerService(this);
        }
        return myTimerService;
    }

    /**
     * Notify a timeout for this bean and this Pk
     * @param timer timer whose expiration caused this notification.
     */
    public void notifyTimeout(Timer timer) {
        TraceEjb.interp.log(BasicLevel.DEBUG, ident);

        // Making this method generic leads to some restrictions listed here after :
        // - Transactional attribute is enforced to TX_REQUIRES_NEW.
        // - security is not checked.
        // LATER: Make it non generic and move it in JEntityLocal.vm and JEntityRemote.vm

        RequestCtx rctx = bf.preInvoke(MethodDesc.TX_REQUIRES_NEW, "");
        try {
            JEntityContext bctx = getICtx(rctx.currTx);
            EntityBean eb = bctx.getInstance();
            if (eb instanceof TimedObject) {
                TimedObject instance = (TimedObject) eb;
                instance.ejbTimeout(timer);
            } else {
                throw new EJBException("The bean does not implement the `TimedObject` interface");
            }
        } catch (EJBException e) {
            rctx.sysExc = e;
            throw e;
        } catch (RuntimeException e) {
            rctx.sysExc = e;
            throw new EJBException("RuntimeException thrown by an enterprise Bean", e);
        } catch (Error e) {
            rctx.sysExc = e;
            throw new EJBException("Error thrown by an enterprise Bean" + e);
        } catch (RemoteException e) {
            rctx.sysExc = e;
            throw new EJBException("Remote Exception raised:", e);
        } finally {
            try {
                bf.postInvoke(rctx);
            } finally {
                if (rctx.sysExc != null) {
                    discardICtx(rctx.currTx);
                } else {
                    releaseICtx(rctx.currTx);
                }
            }
        }
	}

    /**
     * @return the Primary Key Object for this instance.
     */
    public Object getPrimaryKey() {
        if (pk == null) {
            throw new EJBException("No Primary Key yet");
        }
        return pk;
    }

    /**
     * bind a JEntityContext for a create method.
     * No need to synchronize since object was unknown before.
     * @param tx - the Transaction object
     * @param bctx - the JEntityContext to bind
     */
    public synchronized void bindICtx(Transaction tx, JEntityContext bctx) {

        // In some rare cases, a context can already exist.
        // Only if it is marked removed : just discard it before.
        if (iContext != null) {
            // For example, this instance has been removed (HT) just before...
            TraceEjb.interp.log(BasicLevel.DEBUG, ident + "IT create: release existing Context first.");
            bf.releaseJContext(iContext); 
            discarded = false;
            dirty = false;
        }

        // Bind the context to this object and this transaction.
        // Do not activate the bean instance since ejbCreate has been done.
        iContext = bctx;
        bctx.initEntityContext(this, tx);

        if (tx == null) {
            TraceEjb.interp.log(BasicLevel.DEBUG, ident + "IH create");
            // Seems that we must set dirty here, because for example, CMR fields
            // will be updated in postCreate and they must be written on disk when a new
            // transaction is started later.
            setDirty();
        } else {
            TraceEjb.interp.log(BasicLevel.DEBUG, ident + "IT create");
            if (currtx == null) {
                // register this context to the transaction, as a Synchronization.
                // this will be used later to store instance state when needed :
                // before a finder, or at beforeCompletion, and to release instance
                // at commit.
                try {
                    bf.registerContext(tx, bctx);
                } catch (IllegalStateException e) {
                    TraceEjb.logger.log(BasicLevel.ERROR, ident + "IT create: Transaction is in an illegal state");
                    throw e;
                }
                currtx = tx; // not if an exception is raised! (possible deadlock)
            }
        }
    }

    /**
     * Try to bind a JEntityContext if none already bound.
     * Called by finder methods
     * @param tx - the Transaction object
     * @param bctx The Entity Context
     */
    public synchronized void tryBindICtx(Transaction tx, JEntityContext bctx) {

        // Don't bind if already a context.
        if (iContext != null) {
            TraceEjb.interp.log(BasicLevel.DEBUG, ident + "context already mapped");
            bf.releaseJContext(bctx);
            return;
        }

        // No synchronization, since this is used for readonly methods (finder);
        iContext = bctx;
        bctx.initEntityContext(this, tx);
        bctx.activate(true);

        if (tx == null) {
            TraceEjb.interp.log(BasicLevel.DEBUG, ident + "IH find");
        } else {
            TraceEjb.interp.log(BasicLevel.DEBUG, ident + "IT find");
            if (currtx == null) {
                // register this context to the transaction, as a Synchronization.
                // this will be used later to store instance state when needed :
                // before a finder, or at beforeCompletion, and to release instance
                // at commit.
                try {
                    bf.registerContext(tx, bctx);
                } catch (IllegalStateException e) {
                    TraceEjb.logger.log(BasicLevel.ERROR, ident + "bindICtx IT: Transaction is in an illegal state");
                    throw e;
                }
                currtx = tx; // not if an exception is raised! (possible deadlock)
            }
        }
    }

    /**
     * bind a JEntityContext for a remove method.
     * called in case of remove(pk) or remove(handle)
     * @param tx - the Transaction object
     * @param newctx - the JEntityContext to bind
     * @return the BeanContext
     */
    public JEntityContext getICtx(Transaction tx, JEntityContext newctx) {
        return mapICtx(tx, newctx, false, false);
    }

    /**
     * Get a context/instance associated with this transaction
     * Called at each request on the bean (including remove)
     * @param tx - the Transaction object
     * @return the BeanContext
     */
    public JEntityContext getICtx(Transaction tx) {
        return mapICtx(tx, null, false, true);
    }

    /**
     * Map a context and its instance.
     * This version manages only 1 context (iContext).
     * @param tx - the Transaction object
     * @param bctx - the JEntityContext to bind if not null
     * @param forced - force to take this context. (case of create)
     * @param holdit - increment count to hold it, a release will be called later.
     * @return JEntityContext actually mapped
     */
    public synchronized JEntityContext mapICtx(Transaction tx, JEntityContext bctx, boolean forced, boolean holdit) {

        // Synchronization.
        if (tx == null) {
            // ------------- IH begin -------------
            TraceEjb.interp.log(BasicLevel.DEBUG, ident + "mapICtx IH");

            // Must wait in case of TX or if non reentrant and already busy
            while (currtx != null || (! reentrant && countIH > 0)) {
                if (currtx != null) {
                    TraceEjb.interp.log(BasicLevel.DEBUG, ident + "mapICtx IH: WAIT end IT");
                } else {
                    TraceEjb.interp.log(BasicLevel.DEBUG, ident + "mapICtx IH: WAIT end IH");
                }
                waiters++;
                try {
                    wait(maxtime);
                    TraceEjb.interp.log(BasicLevel.DEBUG, ident + "mapICtx IH: NOTIFIED");
                } catch (InterruptedException e) {
                    TraceEjb.interp.log(BasicLevel.DEBUG, ident + "mapICtx IH: INTERRUPTED");
                } catch (Exception e) {
                    throw new EJBException("JEntitySwitch synchronization pb", e);
                } finally {
                    waiters--;
                }
            }
        } else {
            // ------------- IT begin -------------
            TraceEjb.interp.log(BasicLevel.DEBUG, ident + "mapICtx IT");

            // Must wait in case of TX or if instance has been modified outside transaction.
            while (dirty || (currtx != null && !tx.equals(currtx))) {
                // instance is dirty : must write it before working on it again.
                // this is mandatory to be able to retrieve a good state in case of rollback.
                if (dirty) {
                    TraceEjb.interp.log(BasicLevel.DEBUG, ident + "mapICtx IT: WAIT end IH");
                    // Cannot passivate here, because must be done outside tx context
                    bf.synchronizeEntities();
                } else {
                    TraceEjb.interp.log(BasicLevel.DEBUG, ident + "mapICtx IT: WAIT end IT");
                }
                waiters++;
                try {
                    wait(maxtime);
                    TraceEjb.interp.log(BasicLevel.DEBUG, ident + "mapICtx IT: NOTIFIED");
                } catch (InterruptedException e) {
                    TraceEjb.interp.log(BasicLevel.DEBUG, ident + "mapICtx IT: INTERRUPTED");
                } catch (Exception e) {
                    throw new EJBException("JEntitySwitch synchronization pb", e);
                } finally {
                    waiters--;
                }
                // If transaction has been rolledback or set rollback only, give up.
                int status = Status.STATUS_ROLLEDBACK;
                try {
                    status = tx.getStatus();
                } catch (SystemException e) {
                    TraceEjb.logger.log(BasicLevel.ERROR, ident + "getICtx IT: unexpected exception getting transaction status");
                }
                switch (status) {
                case Status.STATUS_MARKED_ROLLBACK:
                case Status.STATUS_ROLLEDBACK:
                case Status.STATUS_ROLLING_BACK:
                    TraceEjb.logger.log(BasicLevel.WARN, ident + "getICtx IT: transaction rolled back");
                    throw new TransactionRolledbackLocalException("rollback occured while waiting");
                }
            }
        }

        // Choose the context to use.
        if (forced) {
            // If the new context is enforced, we must first release the older
            if (iContext != null) {
                bf.releaseJContext(iContext);
            }
            iContext = bctx;
            iContext.initEntityContext(this, tx);
        } else {
            if (iContext != null) {
                // Reuse the iContext for this transaction.
                // If a context was supplied, release it first.
                if (bctx != null) {
                    bf.releaseJContext(bctx);
                }
                iContext.reuseEntityContext(tx);
                if (mustLoad) {
                    iContext.activate(false);
                    mustLoad = false;
                }
            } else {
                if (bctx != null) {
                    iContext = bctx;
                } else {
                    // no Context available : get one from the pool.
                    iContext = (JEntityContext) bf.getJContext();
                }
                iContext.initEntityContext(this, tx);
                iContext.activate(true);
            }
        }

        if (tx != null) {
            if (currtx == null) {
                TraceEjb.interp.log(BasicLevel.DEBUG, ident + "mapICtx IT: new tx, registerSynchronization");
                // register this context to the transaction, as a Synchronization.
                // this will be used later to store instance state when needed :
                // before a finder, or at beforeCompletion, and to release instance
                // at commit.
                try {
                    bf.registerContext(tx, iContext);
                } catch (IllegalStateException e) {
                    TraceEjb.logger.log(BasicLevel.ERROR, ident + "mapICtx IT: Transaction is in an illegal state");
                    throw e;
                }
                currtx = tx; // not if an exception is raised! (possible deadlock)
            }
        } else {
            if (holdit) {
                countIH++;
                TraceEjb.interp.log(BasicLevel.DEBUG, ident + "mapICtx IH count=" + countIH);
            }
            if (!dirty) {
                setDirty();
            }
        }

        return iContext;
    }

    /**
     * Release a context/instance at end of request.
     * @param tx - transaction associated to this context
     */
    public synchronized void releaseICtx(Transaction tx) {

        JEntityContext bctx = null;

        if (tx == null) {
            // ------------- IH end -------------
            bctx = iContext;
            countIH--;
            TraceEjb.interp.log(BasicLevel.DEBUG, ident + "releaseICtx IH count=" + countIH);
            if (countIH == 0) {
                if (bctx.isMarkedRemoved()) {
                    bf.releaseJContext(bctx);
                    iContext = null;
                    noLongerUsed();
                } else if (mustStore) {
                    try {
                        bctx.storeIfModified();
                    } catch (EJBException e) {
                        TraceEjb.logger.log(BasicLevel.WARN, ident + " ejbexception:" + e);
                    }
                    mustStore = false;
                    //TraceEjb.swapper.log(BasicLevel.DEBUG, ident + " store");
                    dirty = false;
                }
                if (waiters > 0) {
                    TraceEjb.synchro.log(BasicLevel.DEBUG, ident + " notify");
                    notifyAll();
                }
            }
        } else {
            // ------------- IT end -------------
            // nothing to do now. Wait for tx completed.
            TraceEjb.interp.log(BasicLevel.DEBUG, ident + "releaseICtx IT");

        }
    }

    /**
     * Discard a context/instance at end of request. A problem occured
     * on this insance and it must be discarded.
     * @param tx - transaction associated to this context
     */
    public synchronized void discardICtx(Transaction tx) {
        if (tx == null) {
            TraceEjb.interp.log(BasicLevel.DEBUG, ident + "discardICtx IH");
            JEntityContext bctx = iContext;
            countIH--;
            if (countIH == 0) {
                bf.releaseJContext(bctx);
                iContext = null;
                discarded = false;
                dirty = false;
            } else {
                discarded = true;
            }
        } else {
            // nothing to do now. Wait for tx completed.
            TraceEjb.interp.log(BasicLevel.DEBUG, ident + "discardICtx IT");
        }
    }

    /**
     * This transaction is now over. We can dispose of the instance
     * for another transaction or discard it.
     * @param tx the transaction object
     * @param committed true if transaction was committed.
     */
    public synchronized void txCompleted(Transaction tx, boolean committed) {
        JEntityContext bctx = iContext;
        currtx = null;

        TraceEjb.interp.log(BasicLevel.DEBUG, ident + "txCompleted " + committed);

        if (waiters > 0) {
            // Somebody needs this instance : don't release it.
            // Just mark it to be reloaded if needed and notify waiters.
            if (shared || !committed) {
                mustLoad = true;
            }
            TraceEjb.synchro.log(BasicLevel.DEBUG, ident + " notify");
            notifyAll();
        } else {
            if (bctx.isMarkedRemoved()) {
                bf.releaseJContext(bctx);
                iContext = null;
                discarded = false; // ?
                dirty = false; // ?
                if (committed) {
                    noLongerUsed(); // remove is effective !
                } else {
                    // remove operation has been rolled back
                    // we must clean memory instance now
                    TraceEjb.synchro.log(BasicLevel.DEBUG, ident + " remove has been rolled back");
                }
            } else {
                // If bean is shared : passivate and release this instance
                // because it can be used outside this server.
                // If bean is not shared : keep a ready version in cache, until
                // the swapper decide that there are too many instances in memory.
                if (shared || !committed) {
                    bctx.passivate();
                    bf.releaseJContext(bctx);
                    iContext = null;
                    discarded = false;
                    dirty = false;
                }
            }
        }
    }

    /**
     * Instance is ready to use for new transaction.
     */
    public synchronized void endIH() {
        // If the instance is busy, don't touch it.
        // If no context mapped, nothing to do
        if (countIH == 0 && currtx == null && iContext != null) {
            TraceEjb.swapper.log(BasicLevel.DEBUG, ident + " ready again");
            dirty = false;
            if (waiters > 0) {
                TraceEjb.synchro.log(BasicLevel.DEBUG, ident + " notify");
                notifyAll();
            }
        } else {
            TraceEjb.swapper.log(BasicLevel.DEBUG, ident + " busy!");
        }
    }

    /**
     * try to passivate IH (called from swapper)
     * @param passivation false if only store bean state.
     * @return true if OK, false if busy.
     */
    public synchronized boolean passivateIH(boolean passivation) {

        JEntityContext bctx = iContext;

        // If no context mapped, nothing to do
        if (bctx == null) {
            return true;
        }
        // If the instance is busy, don't touch it.
        if (countIH > 0 || currtx != null) {
            TraceEjb.swapper.log(BasicLevel.DEBUG, ident + "busy");
            TraceEjb.swapper.log(BasicLevel.DEBUG, ident + "countIH=" + countIH + " currtx=" + currtx + " waiters=" + waiters);
            if (countIH > 0) {
                // will be stored later, when released by last thread (countIH = 0)
                // return true to remove from the dirty list, since it will be written anyway.
                mustStore = true;
                return true;
            }
            return false;
        }
        TraceEjb.swapper.log(BasicLevel.DEBUG, ident);
        // Store the instance state
        // -> only if instance not discarded and not marked mustLoad (mustLoad => bad memory value)
        if (!discarded && !mustLoad) {
            try {
                bctx.storeIfModified();
            } catch (Exception e) {
                TraceEjb.logger.log(BasicLevel.ERROR, ident, "error while storing bean state:", e);
            }
            mustStore = false;
            // don't set dirty false here, will be done in endIH.
            // this prevent duplicates in dirty list.
        }

        // passivate this instance if required.
        if (passivation) {
            bctx.passivate();
            bf.releaseJContext(bctx);
            iContext = null;
            discarded = false;
            dirty = false;
            if (waiters > 0) {
                TraceEjb.synchro.log(BasicLevel.DEBUG, ident + " notify");
                notifyAll();
            }
        }
        return true;
    }

    /**
     * Remove this EJB and
     * UnExport the remote object.
     */
    private void noLongerUsed() {
        TraceEjb.interp.log(BasicLevel.DEBUG, ident);
        if (remote != null) {
            try {
                remote.unexportObject();
            } catch (NoSuchObjectException e) {
                TraceEjb.logger.log(BasicLevel.ERROR, ident + " unexport entity failed: " + e);
            }
        }
        bf.removeEJB(getPrimaryKey());
    }

    /**
     * instance is dirty : possibly modified outside a transaction.
     */
    private void setDirty() {
        TraceEjb.swapper.log(BasicLevel.DEBUG, ident);
        dirty = true;
        bf.registerEJB(this);
    }

    /**
     * @return ejb instance's state
     * state values are 0=in-tx, 1=out-tx, 2=idle, 3=passive, 4=removed.
     * we don't synchronize this method to avoid jadmin blocks
     */
    public int getState() {
        if (iContext == null) {
            return 3;
        }
        if (currtx != null) {
            return 0;
        }
        if (iContext.isMarkedRemoved()) {
            return 4;
        }
        if (dirty) {
            return 1;
        } else {
            return 2;
        }
    }

}
