/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: JStatelessFactory.java,v 1.9 2003/10/13 15:12:56 coqp Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ejb.container;

import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.EmptyStackException;
import java.util.List;
import java.util.ListIterator;
import javax.ejb.EJBException;
import javax.ejb.SessionBean;
import javax.ejb.TimedObject;
import javax.ejb.Timer;
import javax.naming.NamingException;
import org.objectweb.jonas_ejb.deployment.api.MethodDesc;
import org.objectweb.jonas_ejb.deployment.api.SessionStatelessDesc;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * This class is a factory for a Session Stateless Bean.
 * @author Philippe Durieux
 */
public class JStatelessFactory extends JSessionFactory {

    // instance pool management (list of available JSessionContext objects)
    // Contexts are pooled only for Stateless Session beans because for Stateful
    // sessions a newInstance() is required by the spec.
    protected List bctxlist = new ArrayList();

    protected int instanceCount = 0;
    // initial value for pool size
    protected int minPoolSize = 0;
    // nb max of instances in pool
    protected int maxCacheSize = 0;

    /**
     * constructor
     * @param dd Session Stateless Deployment Descriptor
     * @param cnt Container where the bean is defined
     */
    public JStatelessFactory(SessionStatelessDesc dd, JContainer cont) {
        super(dd, cont);
        TraceEjb.interp.log(BasicLevel.DEBUG, "");
        isStateful = false;
        maxCacheSize = dd.getCacheMax();
        minPoolSize = dd.getPoolMin();
        if(TraceEjb.isDebugSwapper()) {
            TraceEjb.swapper.log(BasicLevel.DEBUG," maxCacheSize = "+ maxCacheSize +
                                 " minPoolSize = "+minPoolSize);
        }
        // pre-allocate a set of JSessionContext (bean instances)
        synchronized (bctxlist) {
            for (int i = 0; i < minPoolSize; i++) {
                JSessionContext ctx = null;
                try {
                    ctx = createNewInstance();
                    bctxlist.add(ctx);
                } catch (Exception e) {
                    TraceEjb.logger.log(BasicLevel.ERROR, ejbname + 
                                        " cannot create new instance", e);
                    throw new EJBException(ejbname + " Cannot init pool of instances ", e);
                }
            }
        }
        if ( minPoolSize != 0) {
           TraceEjb.interp.log(BasicLevel.INFO, "pre-allocate a set of "+minPoolSize+
                               " stateless session instances"); 
        }
    }

    // ---------------------------------------------------------------
    // BeanFactory implementation
    // ---------------------------------------------------------------

    /**
     * @return the Instance pool size for this Ejb
     */
    public int getPoolSize() {
        return bctxlist.size();
    }

    /**
     * Reduce number of instances in memory in the free list
     * we reduce to the minPoolSize
     */
    public void reduceCache() {
        TraceEjb.swapper.log(BasicLevel.DEBUG, "");
        // reduce the pool to the minPoolSize
        int poolsz = minPoolSize;
        synchronized (bctxlist) {
            if(TraceEjb.isDebugSwapper()) {
                TraceEjb.swapper.log(BasicLevel.DEBUG, "try to reduce " + bctxlist.size() + 
                                     " to " + poolsz);
            }
            while (bctxlist.size() > poolsz) {
                ListIterator i = bctxlist.listIterator();
                if (i.hasNext()) {
                    i.next();
                    i.remove();
                    instanceCount--;
                }
            }
        }
        if(TraceEjb.isDebugSwapper()) {
            TraceEjb.swapper.log(BasicLevel.DEBUG, "cacheSize= " + getCacheSize());
        }

    }

    // ---------------------------------------------------------------
    // other public methods
    // ---------------------------------------------------------------

    /**
     * Creates a new Session Stateless
     * @return the new JSessionSwitch
     */
    public JSessionSwitch createNewSession() throws RemoteException {
        TraceEjb.interp.log(BasicLevel.DEBUG, "");
        JStatelessSwitch bs = new JStatelessSwitch(this);
        return bs;
    }

    /**
     * @return a SessionContext for Stateless Session Bean
     */
    public JSessionContext getJContext() {
        TraceEjb.interp.log(BasicLevel.DEBUG, "");
        JStatelessContext bctx = null;

        // try to find a free context in the pool
        synchronized(bctxlist) {
            try {
                ListIterator i = bctxlist.listIterator();
                if (i.hasNext()) {
                    bctx = (JStatelessContext) i.next();
                    i.remove();
                }
            } catch(IndexOutOfBoundsException ex) {
                // pool is empty
            }
        }

        if (bctx == null) {
            // create a new one if pool empty
            try {
                bctx = createNewInstance();
            } catch (Exception e) {
                TraceEjb.logger.log(BasicLevel.ERROR, "exception:"+e);
                throw new EJBException("Cannot create a new instance");
            }
        }
        if(TraceEjb.isDebugSwapper()) {
            TraceEjb.swapper.log(BasicLevel.DEBUG, "nb instances " + getCacheSize());
        }
        return bctx;
    }

    /**
     * Called after each method call
     * @param ctx the Session Context
     */
    public  void releaseJContext(JContext ctx) {
        TraceEjb.interp.log(BasicLevel.DEBUG, "");
        JStatelessContext bctx = (JStatelessContext) ctx;

        // if the max number of instances is reached 
        // we can reduce the number of instances 
        synchronized (bctxlist) {
            if (maxCacheSize != 0 && instanceCount > maxCacheSize) {
                if(TraceEjb.isDebugSwapper()) {
                    TraceEjb.swapper.log(BasicLevel.DEBUG, "too much instances :" + 
                                         instanceCount + ", max=" + maxCacheSize);
                }
                instanceCount--;
            } else {
                bctxlist.add(bctx);
            }
        }
        if(TraceEjb.isDebugSwapper()) {
            TraceEjb.swapper.log(BasicLevel.DEBUG, "nb instances " + getCacheSize());
        }
    }

    /**
     * Notify a timeout for this bean
     * @param timer timer whose expiration caused this notification.
     */
    public void notifyTimeout(Timer timer) {
        TraceEjb.interp.log(BasicLevel.DEBUG, "");

        // Get a JStatelessSwitch (similar to createEJB)
        // In fact, we don't need the Local and Remote objects here.
        // This object is just needed to initilize the SessionContext.
        JSessionSwitch bs = null;
        if (sessionList.size() > 0) {
            bs = (JSessionSwitch) sessionList.remove(0);
        } else {
            try {
                bs = new JStatelessSwitch(this);
            } catch (RemoteException e) {
                TraceEjb.logger.log(BasicLevel.ERROR, "Notify Timeout - Unexpected : " + e);
                return;
            }
        }

        // Making this method generic leads to some restrictions listed here after :
        // - Transactional attribute is enforced to TX_REQUIRES_NEW.
        // - security is not checked.
        // LATER: Make it non generic and move it in JSessionLocal.vm and JSessionRemote.vm

        RequestCtx rctx = preInvoke(MethodDesc.TX_REQUIRES_NEW, "");
        JSessionContext bctx = null;
        try {
            bctx = bs.getICtx(rctx.currTx);
            TimedObject instance = (TimedObject) bctx.getInstance();
            instance.ejbTimeout(timer);
        } catch (EJBException e) {
            rctx.sysExc = e;
            throw e;
        } catch (RuntimeException e) {
            rctx.sysExc = e;
            throw new EJBException("RuntimeException thrown by an enterprise Bean", e);
        } catch (Error e) {
            rctx.sysExc = e;
            throw new EJBException("Error thrown by an enterprise Bean" + e);
        } catch (RemoteException e) {
            rctx.sysExc = e;
            throw new EJBException("Remote Exception raised:", e);
        } finally {
            try {
                postInvoke(rctx);
            } finally {
                // release everything - Exactly as if the session timeout had expired.
                bs.timeoutExpired(null);
            }
        }
	}

    /**
     * @return min pool size
     * for Jmx
     */
    public int getMinPoolSize() {
        return minPoolSize;
    }

    /**
     * @return max cache size
     * for Jmx
     */
    public int getMaxCacheSize() {
        return maxCacheSize;
    }

    /**
     * @return current cache size ( = nb of instance created)
     * for Jmx
     */
    public int getCacheSize() {
        return instanceCount;
    }

    // ---------------------------------------------------------------
    // private methods
    // ---------------------------------------------------------------

    /**
     * Create a new instance of the bean and its StatelessContext
     */
    private JStatelessContext createNewInstance() throws Exception {
        TraceEjb.interp.log(BasicLevel.DEBUG, "");
        // create the bean instance
        SessionBean bean = null;
        try {
            bean = (SessionBean) beanclass.newInstance();
        } catch (InstantiationException e) {
            TraceEjb.logger.log(BasicLevel.ERROR, ejbname+" cannot instantiate session bean");
            throw e;
        } catch (IllegalAccessException e) {
            TraceEjb.logger.log(BasicLevel.ERROR, ejbname+" Cannot instantiate SessionBean");
            throw e;
        }

        // create a new StatelessContext and bind it to the instance
        JStatelessContext bctx = new JStatelessContext(this, bean);
        bean.setSessionContext(bctx);

        // call ejbCreate on the instance
        TraceEjb.interp.log(BasicLevel.DEBUG, "call ejbCreate on the instance");
        try {
            beanclass.getMethod("ejbCreate", null).invoke(bean, null);
        } catch (Exception e) {
            TraceEjb.logger.log(BasicLevel.ERROR, ejbname+" cannot call ejbCreate on Stateless Session "+e);
        }

        // We always create an instance even if the max is reached
        // see releaseJContext
        synchronized (bctxlist) {
            instanceCount++;
        }
        return bctx;
    }

}
