/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: JTimer.java,v 1.1 2003/09/11 15:44:44 durieuxp Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ejb.container;

import java.io.Serializable;
import java.util.Date;
import javax.ejb.EJBException;
import javax.ejb.NoSuchObjectLocalException;
import javax.ejb.Timer;
import javax.ejb.TimerHandle;
import org.objectweb.jonas_timer.TimerEvent;
import org.objectweb.jonas_timer.TimerEventListener;
import org.objectweb.jonas_timer.TimerManager;

/**
 * JOnAS Implementation of the Timer interface (from EJB 2.1)
 * This is a basic implementation based on jonas_timer.
 * A later (and better ?) implementation could be based on Quartz.
 * @author Philippe Durieux
 */
public class JTimer implements Timer, TimerEventListener {

    /**
     * true if "one-shot" timer
     */
    private boolean oneshot;

    /**
     * period in case of repeatable timer (millisec)
     */
    private long period;

    /**
     * initial duration (millisec)
     */
    private long initial;

    /**
     * The TimerService that manages this Timer.
     */
    private JTimerService timerservice;

    /**
     * Info provided at Timer Creation time.
     */
    private Serializable info;

    /**
     * Jonas Timer Manager
     */
    private TimerManager tim = TimerManager.getInstance();

    /**
     * jonas timer initial
     */
    private TimerEvent te1 = null;

    /**
     * repeatable jonas timer
     */
    private TimerEvent te2 = null;

    /**
     * Start Time of this Timer (millisec)
     * used for Timer identification.
     */
    private long starttime;

    /**
     * Next expiration time
     * Used to compute the reminding time.
     */
    private long endtime;

    /**
     * constructor
     */
    public JTimer(JTimerService timerservice, long initial, long period, Serializable info) {
        this.timerservice = timerservice;
        this.info = info;
        this.period = period;
        this.initial = initial;
        oneshot = (period == 0);
        starttime = System.currentTimeMillis();
        endtime = starttime + initial;
        // Create the jonas timer
        te1 = tim.addTimer(this, (initial + 500) / 1000, null, false);
    }

    /**
     * @return the start time in millisec.
     */
    public long getStartTime() {
        return starttime;
    }

    // ------------------------------------------------------------------------------------
    // The container must implement suitable equals() and hashCode() methods
    // ------------------------------------------------------------------------------------

    /**
     * Indicates whether some other object is "equal to" this one. 
     * @param obj - the reference object with which to compare.
     * @return true if this object is the same as the obj argument; false otherwise.
     */
    public boolean equals(Object obj) {
        if (obj instanceof JTimer) {
            JTimer timer2 = (JTimer) obj;
            if (timer2.getStartTime() == getStartTime()) {
                return true;
            }
        }
        return false;
    }

    /**
     * Returns a hash code value for the object.
     * @return a hash code value for this object.
     */
    public int hashCode() {
        return 1;
    }

    // ------------------------------------------------------------------------------------
    // TimerEventListener implemetation
    // ------------------------------------------------------------------------------------

    /**
     * The timer has just expired.
     */
    public void timeoutExpired(Object arg) {
        // propagate the timeout
        timerservice.notify(this);
        if (te2 == null) {
            te1 = null;
            if (oneshot) {
                timerservice.remove(this);
            } else {
                // start a new jonas timer (periodic)
                endtime = System.currentTimeMillis() + period;
                te2 = tim.addTimer(this, (period + 500) / 1000, null, true);
            }
        } else {
            // recompute time of next expiration
            endtime = System.currentTimeMillis() + period;
        }
    }

    // ------------------------------------------------------------------------------------
    // Timer implemetation
    // ------------------------------------------------------------------------------------

    /**
     * Cause the timer and all its associated expiration notifications to be cancelled. 
     * @throws IllegalStateException the instance is in a state that does not allow access to this method. 
     * @throws NoSuchObjectLocalException If invoked on a timer that has expired or has been cancelled. 
     * @throws EJBException If this method could not complete due to a system-level failure.
     */
    public void cancel() throws IllegalStateException, NoSuchObjectLocalException, EJBException {
        if (te1 != null) {
            te1.unset();
        }
        if (te2 != null) {
            te2.unset();
        }
        timerservice.remove(this);
    }

    /**
     * Get the number of milliseconds that will elapse before the next scheduled timer expiration. 
     * @return the number of milliseconds that will elapse before the next scheduled timer expiration. 
     * @throws IllegalStateException the instance is in a state that does not allow access to this method. 
     * @throws NoSuchObjectLocalException If invoked on a timer that has expired or has been cancelled. 
     * @throws EJBException If this method could not complete due to a system-level failure.
     */
    public long getTimeRemaining() throws IllegalStateException, NoSuchObjectLocalException, EJBException {
        return endtime - System.currentTimeMillis();
    }

    /**
     * Get the point in time at which the next timer expiration is scheduled to occur. 
     * @return the point in time at which the next timer expiration is scheduled to occur. 
     * @throws IllegalStateException the instance is in a state that does not allow access to this method. 
     * @throws NoSuchObjectLocalException If invoked on a timer that has expired or has been cancelled. 
     * @throws EJBException If this method could not complete due to a system-level failure.
     */
    public Date getNextTimeout() throws IllegalStateException, NoSuchObjectLocalException, EJBException {
        return new Date(endtime);
    }

    /**
     * Get the information associated with the timer at the time of creation. 
     * @return The Serializable object that was passed in at timer creation,
     * or null if the info argument passed in at timer creation was null. 
     * @throws IllegalStateException the instance is in a state that does not allow access to this method. 
     * @throws NoSuchObjectLocalException If invoked on a timer that has expired or has been cancelled. 
     * @throws EJBException If this method could not complete due to a system-level failure.
     */
    public Serializable getInfo() throws IllegalStateException, NoSuchObjectLocalException, EJBException {
        return info;
    }

    /**
     * Get a serializable handle to the timer. 
     * This handle can be used at a later time to re-obtain the timer reference. 
     * @return a serializable handle to the timer. 
     * @throws IllegalStateException the instance is in a state that does not allow access to this method.
     * @throws NoSuchObjectLocalException If invoked on a timer that has expired or has been cancelled. 
     * @throws EJBException If this method could not complete due to a system-level failure.
     */
    public TimerHandle getHandle() throws IllegalStateException, NoSuchObjectLocalException, EJBException {
        // TODO
        throw new NoSuchObjectLocalException("Not Implemented");
    }

}
