/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: JTimerService.java,v 1.1 2003/09/11 15:44:44 durieuxp Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ejb.container;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import javax.ejb.EJBException;
import javax.ejb.Timer;
import javax.ejb.TimerService;
import javax.ejb.TransactionRolledbackLocalException;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * JOnAS Implementation of the TimerService interface (from EJB 2.1)
 * One such object is created the first time a bean calls getTimerService.
 * Basically manages the list of the Timers for that bean.
 * @author Philippe Durieux
 */
public class JTimerService implements TimerService {

    private ArrayList mytimers = new ArrayList();
    private JFactory bf = null;
    private JEntitySwitch es = null;

    /**
     * constructor used for MDB or Session beans
     */
    public JTimerService(JFactory bf) {
        this.bf = bf;
    }

    /**
     * constructor used for Entity beans
     */
    public JTimerService(JEntitySwitch es) {
        this.es = es;
    }

    /**
     * Notify the timer to the listener
     * @param timer The Timer object that will be notified
     */
    public void notify(Timer timer) {
        try {
            if (es != null) {
                // Entity Bean
                es.notifyTimeout(timer);
            } else {
                if (bf instanceof JMdbFactory) {
                    ((JMdbFactory)bf).notifyTimeout(timer);
                } else if (bf instanceof JStatelessFactory) {
                    ((JStatelessFactory)bf).notifyTimeout(timer);
                } else {
                    TraceEjb.logger.log(BasicLevel.ERROR, "Cannot notify this type of bean");
                }
            }
        } catch (TransactionRolledbackLocalException e) {
            // TODO: Container should retry the timeout.
        }
    }

    /**
     * Remove the Timer
     * @param timer The Timer object that will be removed
     */
    public void remove(Timer timer) {
        synchronized(this) {
            mytimers.remove(mytimers.lastIndexOf(timer));
        }
    }

    // ------------------------------------------------------------------------------------
    // TimerService implemetation
    // ------------------------------------------------------------------------------------

    /**
     * Create an interval timer whose first expiration occurs at a given point in time and whose subsequent
     * expirations occur after a specified interval. 
     * @param initialExpiration The point in time at which the first timer expiration must occur.
     * @param intervalDuration The number of milliseconds that must elapse between timer expiration notifications.
     * @param info Application information to be delivered along with the timer expiration. This can be null. 
     * @return the newly created Timer.
     * @throws IllegalArgumentException  initialExpiration = 0, or intervalDuration < 0 or initialExpiration.getTime() < 0.
     * @throws IllegalStateException the instance is in a state that does not allow access to this method. 
     * @throws EJBException If this method could not complete due to a system-level failure.
     */
    public Timer createTimer(Date initialExpiration, long intervalDuration, Serializable info)
        throws IllegalArgumentException, IllegalStateException, EJBException {
        long initialDuration = initialExpiration.getTime() - System.currentTimeMillis();
        return createTimer(initialDuration, intervalDuration, info);
    }

    /**
     * Create a single-action timer that expires at a given point in time. 
     * @param expiration The point in time at which the timer expiration must occur.
     * @param info Application information to be delivered along with the timer expiration. This can be null.  
     * @return the newly created Timer.
     * @throws IllegalArgumentException expiration = 0, or expiration.getTime() < 0.
     * @throws IllegalStateException the instance is in a state that does not allow access to this method. 
     * @throws EJBException If this method could not complete due to a system-level failure.
     */
    public Timer createTimer(Date expiration, Serializable info)
        throws IllegalArgumentException, IllegalStateException, EJBException {
        return createTimer(expiration, 0, info);
    }

    /**
     * Create an interval timer whose first expiration occurs after a specified duration, and whose subsequent
     * expirations occur after a specified interval. 
     * @param initialDuration The number of milliseconds that must elapse before the first timer expiration notification.
     * @param intervalDuration The number of milliseconds that must elapse between timer expiration notifications.
     * @param info Application information to be delivered along with the timer expiration. This can be null. 
     * @return the newly created Timer. 
     * @throws IllegalArgumentException  initialExpiration = 0, or intervalDuration < 0.
     * @throws IllegalStateException the instance is in a state that does not allow access to this method. 
     * @throws EJBException If this method could not complete due to a system-level failure.
     */
    public Timer createTimer(long initialDuration, long intervalDuration, Serializable info)
        throws IllegalArgumentException, IllegalStateException, EJBException {
        Timer timer = new JTimer(this, initialDuration, intervalDuration, info);
        synchronized(this) {
            mytimers.add(timer);
        }
        return timer;
    }

    /**
     * Create a single-action timer that expires after a specified duration. 
     * @param duration The number of milliseconds that must elapse before the timer expires.
     * @param info Application information to be delivered along with the timer expiration. This can be null. 
     * @return the newly created Timer. 
     * @throws IllegalArgumentException  initialExpiration = 0, or intervalDuration < 0.
     * @throws IllegalStateException the instance is in a state that does not allow access to this method. 
     * @throws EJBException If this method could not complete due to a system-level failure.
     */
    public Timer createTimer(long duration, Serializable info)
        throws IllegalArgumentException, IllegalStateException, EJBException {
        return createTimer(duration, 0, info);
    }

    /**
     * Get all the active timers associated with this bean. 
     * @return A collection of javax.ejb.Timer objects.
     * @throws IllegalStateException the instance is in a state that does not allow access to this method.
     * @throws EJBException If this method could not complete due to a system-level failure.
     */
    public Collection getTimers() throws IllegalStateException, EJBException {
        return (ArrayList) mytimers.clone();
    }
}
