/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: TxListener.java,v 1.2 2003/10/22 14:39:59 durieuxp Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ejb.container;

import java.util.Iterator;
import java.util.LinkedList;
import javax.naming.Context;
import javax.transaction.Synchronization;
import javax.transaction.Transaction;

import org.objectweb.jonas_ejb.container.jorm.JormFactory;
import org.objectweb.jorm.api.PMapper;
import org.objectweb.medor.eval.prefetch.api.PrefetchCache;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * This class is a listener on the transaction. It is registered as
 * a Synchronization on the current transaction. It holds the list of all
 * instances that are possibly modified during the transaction.
 * @author Philippe Durieux
 */
public class TxListener implements Synchronization {

    private JEntityFactory bf;
    private Transaction tx;

    /**
     * list of contexts to be stored on disk.
     */
    private LinkedList ctxList = new LinkedList();

    /**
     * constructor
     */
    public TxListener(JEntityFactory bf, Transaction tx) {
        TraceEjb.interp.log(BasicLevel.DEBUG, "");
        this.bf = bf;
        this.tx = tx;
    }

    /**
     * Synchronizes all instances on disk. 
     * This must be called before every SQL request (finder) or at beforeCompletion().
     * We must be in the good transaction context here, and with the correct
     * component environment (java:comp).
     */
    public synchronized void storeInstances() {
        TraceEjb.interp.log(BasicLevel.DEBUG, "");
        for (Iterator i = ctxList.iterator(); i.hasNext(); ) {
            JEntityContext ec = (JEntityContext) i.next();
            ec.beforeCompletion();
        }
    }

    /**
     * Add a new Context in the list.
     * @param ctx the JEntityContext to synchronize
     */
    public synchronized void addInstance(JEntityContext ec) {
        TraceEjb.interp.log(BasicLevel.DEBUG, "");
        ctxList.addLast(ec);
    }

    // -------------------------------------------------------------------
    // Synchronization implementation
    // -------------------------------------------------------------------

    /**
     * This beforeCompletion method is called by the transaction
     * manager prior to the start of the transaction completion process.
     * This method executes in the transaction context of the calling
     * thread.
     * The Bean's state must be stored on the persistent storage before
     * the completion of the transaction.
     */
    public void beforeCompletion() {
        TraceEjb.interp.log(BasicLevel.DEBUG, "");

        Context bnctx = bf.setComponentContext();
        storeInstances();
        bf.resetComponentContext(bnctx);
    }

    /**
     * The afterCompletion method is called by the transaction
     * manager after the transaction is committed or rolled back.
     * This method executes without a transaction context.
     *
     * @param status The status of the transaction completion.
     */
    public void afterCompletion(int status) {

        // Send afterCompletion on each Context.
        synchronized (this) {
            for (Iterator i = ctxList.iterator(); i.hasNext(); ) {
                JEntityContext ec = (JEntityContext) i.next();
                ec.afterCompletion(status);
            }
        }

        // Invalid the prefetch cache
        if (bf instanceof JormFactory) {
            PMapper mapper = ((JormFactory) bf).getMapper();
            PrefetchCache pc = mapper.getPrefetchCache();
            if (pc != null) {
                pc.invalidatePrefetchBuffer(tx);
            }
        }

        // remove this in the tx HashMap.
        bf.removeTxListener(tx);
    }
}
