/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): 
 * Contributor(s):
 *
 * --------------------------------------------------------------------------
 * $Id: MedorFactory.java,v 1.22 2003/10/22 14:39:59 durieuxp Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ejb.container.jorm;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import javax.ejb.EJBException;

import org.objectweb.jonas_ejb.container.JContainer;
import org.objectweb.jonas_ejb.container.TraceEjb;
import org.objectweb.jonas_ejb.deployment.api.DeploymentCmp2Desc;
import org.objectweb.jonas_ejb.deployment.api.DeploymentDescException;
import org.objectweb.jonas_ejb.deployment.api.EntityCmp2Desc;
import org.objectweb.jonas_ejb.deployment.api.EntityDesc;
import org.objectweb.jonas_ejb.deployment.api.MethodCmp2Desc;
import org.objectweb.jonas_ejb.lib.EjbqlQueryTreeHolder;
import org.objectweb.jorm.metainfo.api.Manager;
import org.objectweb.jorm.api.PException;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.api.EvaluationException;
import org.objectweb.medor.eval.api.ConnectionResources;
import org.objectweb.medor.eval.api.QueryEvaluator;
import org.objectweb.medor.eval.prefetch.api.PrefetchBuffer;
import org.objectweb.medor.filter.api.ParameterOperand;
import org.objectweb.medor.optim.api.ExecPlanGenerator;
import org.objectweb.medor.optim.api.QueryTransformer;
import org.objectweb.medor.optim.lib.BasicQueryRewriter;
import org.objectweb.medor.optim.lib.IndexesGenerator;
import org.objectweb.medor.optim.jorm.JormFlatten2Rdb;
import org.objectweb.medor.optim.jorm.RdbJormLeafRewriter;
import org.objectweb.medor.optim.lib.FlattenQueryTreeRule;
import org.objectweb.medor.query.api.QueryLeaf;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.tuple.api.TupleCollection;
import org.objectweb.medor.lib.Log;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * This class does the initialisation of Medor and permits to access to the
 * query. The optimisation of the query tree is done during the first time that
 * it is used.
 * @author S.Chassande-Barrioz
 */
public abstract class MedorFactory extends JormFactory {
    /**
     * This field references the query transformer which must be used to
     * optimize the medor requests.
     */
    protected QueryTransformer queryTransformer = null;

    protected Manager miManager = null;

    protected ExecPlanGenerator indexesGenerator;

    protected boolean optimizeAtInit = false;

    protected static final Collection leafRewriters =
        Collections.singleton(new RdbJormLeafRewriter());

    // This is the prefetch index for the query method CURRENTLY evaluated
    private int prefetchIndex;

    public MedorFactory() {
        super();
        if (Log.loggerFactory != TraceEjb.loggerFactory) {
            Log.loggerFactory = TraceEjb.loggerFactory;
        }
    }

    /**
     * It retrieves a medor request which is evaluable and optimized. The method
     * index is translate into a request index, then the found request is
     * optimized if it is not already.
     *
     * @param methodDesc is the MethodCmp2Desc of the finder or select method.
     * @return the QueryTree optimized which is associated to the method index
     */
    public synchronized QueryEvaluator getOptimizedRequest(MethodCmp2Desc methodDesc)
        throws MedorException {
        TraceEjb.logger.log(BasicLevel.DEBUG, "getOptimizedRequest(method name: "
                            + methodDesc.getMethod().getName() + ")");
        try {
            EjbqlQueryTreeHolder qth = methodDesc.getQueryTreeHolder(mapper);
            // set the prefetch index for this method
            setPrefetchIndex(qth.getPrefetchIndex());
            // set the optimizer if not set yet
            if (qth.getQueryOptimizer() == null) {
                if (queryTransformer == null) {
                    ArrayList rules = new ArrayList(2);
                    rules.add(new FlattenQueryTreeRule());
                    rules.add(new JormFlatten2Rdb());
                    queryTransformer = new BasicQueryRewriter(rules);
                }
                qth.setQueryOptimizer(new QueryTransformer() {
                        public QueryTree transform(QueryTree qt) throws MedorException {
                            QueryTree qt1 = queryTransformer.transform(qt);
                            if (indexesGenerator == null) {
                                throw new Error("getOptimizedRequest: indexesGenerator == null");
                            }
                            return indexesGenerator.transform(qt1);
                        }
                    });
            }
            return qth.getOptimizedQueryTree();
        } catch (Exception e) {
            throw new MedorException("Impossible to optimize the query " + methodDesc.getQuery(), e);
        }
    }

    /**
     * It optimized all medor requests which are already not.
     */
    /* NO MORE USED ???
    public void optimizeAll() throws MedorException {
        for (Iterator it = dd.getMethodDescIterator(); it.hasNext();) {
            MethodCmp2Desc mcd = (MethodCmp2Desc) it.next();
            if (!mcd.isFindByPrimaryKey() && (mcd.isFinder() || mcd.isEjbSelect())) {
                getOptimizedRequest(mcd.getIndex());
            }
        }
    }
    */

    /**
     * It evaluate an optimized medor request according to the specified parameters
     * To evalute the medor request the query evaluator is used.
     * @param methodIndex is method index which permits to find a medor request.
     * @param parameters is the parameters
     * (key=parameter name  /  value=parameter value)
     * @return TupleCollection is the result of the request
     */
    public TupleCollection evaluate(int methodIndex, ParameterOperand[] parameters)
        throws MedorException {
        
        MethodCmp2Desc mcd = (MethodCmp2Desc) dd.getMethodDesc(methodIndex);
        QueryEvaluator evaluator = getOptimizedRequest(mcd);

        // Calculates and gets the required connection resources for this query
        ConnectionResources cr = evaluator.getRequiredConnectionResources();

        // Gets the QueryLeafs that requierd connections
        QueryLeaf[] leaves = cr.getRequiredQueryLeafConnection();

        // Setting QueryLeaf's appropriated connection Object
        // TODO: All jorm classes are not mapped on the same mapper 
        // than the current class - should use connections from different datasources.
        Object conn = null;
        try {
            conn = getConnection(null);
        } catch (PException e) {
            throw new MedorException("impossible to obtain connection to the data support", e);
        }
        for (int i = 0; (i < leaves.length); i++) {
            cr.setConnection(leaves[i], conn);
        }
        PrefetchBuffer prefetchBuffer = null;
        if (mcd.getPrefetch()) {
            Object tx = null;
            try {
                tx = this.getTransactionManager().getTransaction();
                if (tx == null) {
                    tx = "NO_TX";
                }
                prefetchBuffer = mapper.getPrefetchCache().createPrefetchBuffer(this, tx, getPrefetchIndex());
            } catch (javax.transaction.SystemException e) {
                // TODO: Que fait-on ???
                TraceEjb.logger.log(BasicLevel.ERROR, "Cannot get the current transaction to create the Prefetch buffer:", e);
            }
        }
        try {
            if (prefetchBuffer == null) {
                return evaluator.evaluate(parameters, cr);
            } else {
                return evaluator.evaluate(parameters, cr, prefetchBuffer);
            }
        } catch (EvaluationException e) {
            throw e;
        } finally {
            try {
                if (conn != null) {
                    releaseConnection(conn);
                }
            } catch (PException e) {
                TraceEjb.logger.log(BasicLevel.ERROR, "impossible to close a connection", e);
            }
        }
    }


    public void init(EntityDesc ed, JContainer c, String mapperName) {
        super.init(ed, c, mapperName);

        ecd = (EntityCmp2Desc) ed;

        //------------------- JORM META INFORMATION BUILDING ----------------//
        TraceEjb.logger.log(BasicLevel.DEBUG, "Jorm Meta information building");
        DeploymentCmp2Desc dd = ecd.getDeploymentCMP2Desc();
        try {
            miManager = dd.getJormManager();
        } catch (DeploymentDescException e) {
            TraceEjb.logger.log(BasicLevel.ERROR,
                                "impossible to load the jorm meta information into the manager", e);
            throw new EJBException("impossible to load the jorm meta information into the manager", e);
        }

        //---------------------- MEDOR INITIALIZATION -----------------------//
        TraceEjb.logger.log(BasicLevel.DEBUG, "Medor initialisation");
        indexesGenerator = new IndexesGenerator();

    }


    private void setPrefetchIndex(int i) {
        prefetchIndex = i;
    }

    private int getPrefetchIndex() {
        return prefetchIndex;
    }

}
