/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Helene Joanin
 * Contributors: Philippe Durieux, Sebastien Chassande
 *
 * --------------------------------------------------------------------------
 * $Id: BeanSources.java,v 1.22 2003/10/17 13:33:13 joaninh Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ejb.genic;

import java.io.File;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.Properties;
import java.util.StringTokenizer;
import org.apache.velocity.app.VelocityEngine;
import org.objectweb.jonas.common.Log;
import org.objectweb.jonas_ejb.deployment.api.BeanDesc;
import org.objectweb.jonas_ejb.deployment.api.EntityCmp2Desc;
import org.objectweb.jonas_ejb.deployment.api.EntityDesc;
import org.objectweb.jonas_ejb.deployment.api.EntityJdbcCmp1Desc;
import org.objectweb.jonas_ejb.lib.BeanNaming;
import org.objectweb.jonas_ejb.lib.RdbMappingBuilder;
import org.objectweb.jorm.metainfo.api.Manager;
import org.objectweb.jorm.compiler.lib.JormCompiler;
import org.objectweb.jorm.compiler.lib.JormCompilerParameter;
import org.objectweb.jorm.lib.JormConfigurator;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.wrapper.printwriter.LoggerImpl;
import org.objectweb.util.monolog.wrapper.velocity.VelocityLogger;


/**
 *  This class allows to generate the sources of:
 *  the class that implements the Enterprise bean's remote interface,
 *  the class that implements the Enterprise bean's home interface,
 *  the class that implements the Enterprise bean's local interface,
 *  the class that implements the Enterprise bean's localhome interface,
 *  the class of the Entity Handle in case of entity,
 *  and the extended class of the Bean for persistence in case of entity with CMP,
 * of a given Enterprise Java Bean.
 */
public class BeanSources {

    // EJBObject (Remote). May be null if no Remote interface.
    private String wrpRemoteFileName = null;

    // EJBLocalObject (Local). May be null if no Local interface.
    private String wrpLocalFileName = null;

    // Home. May be null if no Home interface.
    private String wrpHomeFileName = null;

    // Home cluster configuration file. May be null if no Home interface.
    private String wrpHomeClusterConfFileName  = null;

    // LocalHome. May be null if no LocalHome interface.
    private String wrpLocalHomeFileName = null;

    // Handle (for Entity only)
    private String wrpHandleFileName = null;

    // Bean (for CMP Entity only)
    private String wrpBeanFileName = null;

    private String itfCohCmp2Entity = null;

    private ArrayList mapperList;

    // Source objects used to generate classes
    private ArrayList sources;
    private ArrayList noRemoteJava;

    private String outdir = null; // directory where to place the generated files
    private BeanDesc dd = null; // Bean Deployment Description
    private Manager manager = null;
    private Logger logger = null;

    /**
     * GenIC Constructor
     *
     * @param     beanDesc      deployment descriptor of the bean
     * @param     dirOutputName  path of the directory where to place the generated files
     *                           (empty string "" if the output directory is the current directory)
     * @exception GenICException In error case
     */
    public BeanSources(BeanDesc beanDesc, String dirOutputName, Manager manager)
        throws GenICException {
        this(beanDesc, dirOutputName);
        this.manager = manager;
    }

    /**
     * GenIC Constructor
     *
     * @param     beanDesc      deployment descriptor of the bean
     * @param     dirOutputName  path of the directory where to place the generated files
     *                           (empty string "" if the output directory is the current directory)
     * @exception GenICException In error case
     */
    public BeanSources(BeanDesc beanDesc, String dirOutputName) throws GenICException {

        logger = Log.getLogger(Log.JONAS_GENIC_PREFIX);
        dd = beanDesc;
        outdir = dirOutputName;
        sources = new ArrayList();
        noRemoteJava = new ArrayList();

        // Create and initialize the Velocity Engine
        String jonasRoot = System.getProperty("install.root");
        if (jonasRoot == null) {
            throw new GenICException("System property 'install.root' not set");
        }
        String path2Tmpl = new String(jonasRoot + File.separatorChar + "templates"
                                      + File.separatorChar + "genic");

        VelocityEngine ve = new VelocityEngine();
        ve.setProperty(VelocityEngine.VM_LIBRARY, "");
        ve.setProperty(VelocityEngine.RESOURCE_LOADER, "file");
        ve.setProperty(VelocityEngine.FILE_RESOURCE_LOADER_PATH, path2Tmpl);
        // Velocity logs
        Logger vLogger = Log.getLogger(Log.JONAS_GENIC_VELOCITY_PREFIX);
        if (vLogger.getCurrentIntLevel() != BasicLevel.DEBUG) {
            // No more velocity logs to avoid the creation of an empty file "velocity.log"
            ve.setProperty(VelocityEngine.RUNTIME_LOG_LOGSYSTEM_CLASS,
                           "org.apache.velocity.runtime.log.NullLogSystem");
        } else {
            // Connect the velocity log system to monolog (no more needed)
            ve.setProperty(VelocityEngine.RUNTIME_LOG_LOGSYSTEM, new VelocityLogger(vLogger));
        }
        try {
            ve.init();
        } catch (Exception e) {
            throw new GenICException("Cannot initialize the Velocity engine", e);
        }

        // generated Home
        if (dd.getHomeClass() != null) {
            if (outdir.length() > 0) {
                wrpHomeFileName = outdir + File.separatorChar + BeanNaming.getPath(dd.getFullWrpHomeName());
            } else {
                wrpHomeFileName = dd.getWrpHomeName();
            }
            wrpHomeClusterConfFileName = wrpHomeFileName.concat("_Cluster.xml");
            sources.add(new Source(dd, wrpHomeClusterConfFileName, Source.CLUSTER_HOME, ve));
            wrpHomeFileName = wrpHomeFileName.concat(".java");
            sources.add(new Source(dd, wrpHomeFileName, Source.HOME, ve));
            // generated Handle for Remote Entity bean
            if (dd instanceof EntityDesc) {
                if (outdir.length() > 0) {
                    wrpHandleFileName = outdir + File.separatorChar + BeanNaming.getPath(dd.getFullWrpHandleName());
                } else {
                    wrpHandleFileName = dd.getWrpHandleName();
                }
                wrpHandleFileName = wrpHandleFileName.concat(".java");
                noRemoteJava.add(wrpHandleFileName);
                sources.add(new Source(dd, wrpHandleFileName, Source.ENTITY_HANDLE, ve));
            }
        }

        // generated EJBObject (Remote)
        if (dd.getRemoteClass() != null) {
            if (outdir.length() > 0) {
                wrpRemoteFileName = outdir + File.separatorChar + BeanNaming.getPath(dd.getFullWrpRemoteName());
            } else {
                wrpRemoteFileName = dd.getWrpRemoteName();
            }
            wrpRemoteFileName = wrpRemoteFileName.concat(".java");
            sources.add(new Source(dd, wrpRemoteFileName, Source.REMOTE, ve));
        }

        // generated LocalHome
        if (dd.getLocalHomeClass() != null) {
            if (outdir.length() > 0) {
                wrpLocalHomeFileName = outdir + File.separatorChar + BeanNaming.getPath(dd.getFullWrpLocalHomeName());
            } else {
                wrpLocalHomeFileName = dd.getWrpLocalHomeName();
            }
            wrpLocalHomeFileName = wrpLocalHomeFileName.concat(".java");
            noRemoteJava.add(wrpLocalHomeFileName);
            sources.add(new Source(dd, wrpLocalHomeFileName, Source.LOCAL_HOME, ve));
        }

        // generated EJBLocalObject (Local)
        if (dd.getLocalClass() != null) {
            if (outdir.length() > 0) {
                wrpLocalFileName = outdir + File.separatorChar + BeanNaming.getPath(dd.getFullWrpLocalName());
            } else {
                wrpLocalFileName = dd.getWrpLocalName();
            }
            wrpLocalFileName = wrpLocalFileName.concat(".java");
            noRemoteJava.add(wrpLocalFileName);
            sources.add(new Source(dd, wrpLocalFileName, Source.LOCAL, ve));
        }

        // generated derived bean class for Entity Bean CMP
        if (dd instanceof EntityJdbcCmp1Desc) {
            if (outdir.length() > 0) {
                wrpBeanFileName = outdir + File.separatorChar + BeanNaming.getPath(dd.getFullDerivedBeanName());
            } else {
                wrpBeanFileName = dd.getDerivedBeanName();
            }
            wrpBeanFileName = wrpBeanFileName.concat(".java");
            noRemoteJava.add(wrpBeanFileName);
            sources.add(new Source(dd, wrpBeanFileName, Source.ENTITY_CMP_JDBC, ve));
        }
        if (dd instanceof EntityCmp2Desc) {
            EntityCmp2Desc ecd = (EntityCmp2Desc) dd;
            getAllMapperNames();
            if (outdir.length() > 0) {
                wrpBeanFileName = outdir + File.separatorChar
                    + BeanNaming.getPath(dd.getFullDerivedBeanName()) + ".java";
                if (ecd.needJormCoherenceHelper()) {
                    itfCohCmp2Entity = outdir + File.separatorChar
                        + BeanNaming.getPath(ecd.getJormCoherenceHelperFQItfName()) + ".java";
                }
            } else {
                wrpBeanFileName = BeanNaming.getPath(dd.getDerivedBeanName()) + ".java";
                if (ecd.needJormCoherenceHelper()) {
                    itfCohCmp2Entity = BeanNaming.getPath(ecd.getJormCoherenceHelperItfName()) + ".java";
                }
            }
            if (ecd.needJormCoherenceHelper()) {
                sources.add(new Source(dd, itfCohCmp2Entity, Source.ITF_COH_CMP2_ENTITY, ve));
                noRemoteJava.add(itfCohCmp2Entity);
            }
            noRemoteJava.add(wrpBeanFileName);
            sources.add(new Source(dd, wrpBeanFileName, Source.ENTITY_CMP_JDBC, ve));
        }
    }

    /**
     * TODO: do not use a system property for this ???
     */
    private void getAllMapperNames() {
        String mns = System.getProperty("JONAS_CMP_MAPPERNAMES");
        mapperList = new ArrayList();
        if (mns != null && mns.length() > 0) {
            StringTokenizer st = new StringTokenizer(mns, ",", false);
            while (st.hasMoreElements()) {
                String mapper = st.nextToken().trim();
                if (mapper.startsWith("rdb.")) {
                    mapperList.add(mapper.substring(4));
                } else if ("rdb".equals(mapper)) {
                    mapperList.add("generic");
                }
            }
        }
    }

    /**
     * Generates the java sources
     */
    public void generate() throws GenICException {

        // Generates the java sources
        for (Iterator it = sources.iterator(); it.hasNext();) {
            Source src = (Source) it.next();
            src.generate();
        }

        // In case of Cmp2, call jorm compiler.
        // The list of mappers is given to the Jorm Compiler Configurator.
        if (dd instanceof EntityCmp2Desc) {
            EntityCmp2Desc ecd = (EntityCmp2Desc) dd;

            // Configure Jorm Compiler
            JormCompiler jc = new JormCompiler();
            JormConfigurator jcc = jc.getCompilerConfigurator();
            Properties prop = new Properties();
            prop.put("jorm.generator", "org.objectweb.jorm.generator.lib.JormGenerator");
            prop.put("jorm.mimanager","org.objectweb.jorm.metainfo.lib.JormManager");
            prop.put("jorm.writer","org.objectweb.jorm.mi2xml.lib.BasicDomWriter");
            prop.put("jorm.mapper.list", "rdb");
            prop.put("jorm.mapper.mifactory.rdb", "org.objectweb.jorm.mapper.rdb.metainfo.RdbMappingFactory");
            prop.put("jorm.mapper.mopfactory.rdb", "org.objectweb.jorm.mapper.rdb.generator.RdbMOPFactory");
            prop.put("jorm.mapper.gcmapping.rdb", "org.objectweb.jorm.mapper.rdb.genclass.RdbGenClassMapping");
            prop.put("jorm.mapper.schmgr.rdb", "org.objectweb.jorm.mapper.rdb.lib.RdbPMappingStructuresManager");

            StringBuffer ml = new StringBuffer();
            if (!mapperList.isEmpty()) {
                Iterator it = mapperList.iterator();
                ml = ml.append((String)it.next());
                while (it.hasNext()) {
                    ml = ml.append("," + (String)it.next());
                }
            }
            prop.put("jorm.mapper.submappers.rdb", ml.toString());
            jcc.configure(prop);
            jcc.setLoggerFactory(new LoggerImpl());

            // Run the Jorm Compiler
            JormCompilerParameter cp = jc.getCompilerParameter();
            cp.setLoggerFactory(new LoggerImpl());
            cp.logger = Log.getLogger(Log.JONAS_GENIC_JORMC_PREFIX);
            cp.projectName = RdbMappingBuilder.getProjectName();
            cp.keepSrc = true;
            cp.output = outdir;
            cp.bindingInheritance = "org.objectweb.jonas_ejb.container.JEntitySwitch";
            cp.classMappingInheritance = "org.objectweb.jonas_ejb.container.jorm.RdbFactory";

            Collection jormflist = null;
            try {
                jc.setMIManager(manager);
                jormflist = jc.generateFiles(ecd.getJormList());
            } catch (Exception e) {
                throw new GenICException("Problem during jorm generation", e);
            }
            // Save the list of Jorm generated files
            for (Iterator i = jormflist.iterator(); i.hasNext(); ) {
                String file = (String) i.next();
                logger.log(BasicLevel.DEBUG, "Jorm generated file: " + file);
                noRemoteJava.add(file);
            }
        }
    }

    /**
     * Return the bean's name
     */
    public String getEjbName() {
        return dd.getEjbName();
    }

    /**
     * Return the file name of the generated source for the Home
     * (null if none)
     */
    public String getWrpHomeFileName() {
        return wrpHomeFileName;
    }

    /**
     * Return the file name of the cluster configuration for the Home
     * (null if none)
     */
    public String getWrpHomeClusterConfFileName() {
        return wrpHomeClusterConfFileName;
    }

    /**
     * Return the file name of the generated source for the Remote
     * (null if none)
     */
    public String getWrpRemoteFileName() {
        return wrpRemoteFileName;
    }

    /**
     * Return the class name of the generated source for the Remote (package included)
     */
    public String getWrpRemoteClassName() {
        return dd.getFullWrpRemoteName();
    }

    /**
     * Return the class name of the generated source for the Home (package included)
     */
    public String getWrpHomeClassName() {
        return dd.getFullWrpHomeName();
    }

    public Collection getNoRemoteJavas() {
        return noRemoteJava;
    }
}
