/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial Developer : Delplanque Xavier & Sauthier Guillaume
 * --------------------------------------------------------------------------
 * $Id: HandlerDesc.java,v 1.4 2003/09/09 11:57:33 sauthieg Exp $
 * --------------------------------------------------------------------------
*/

package org.objectweb.jonas_ws.deployment.api;

import org.objectweb.jonas_ws.deployment.xml.Handler;
import org.objectweb.jonas_ws.deployment.xml.InitParam;
import org.objectweb.jonas_ws.deployment.xml.SoapHeader;
import org.objectweb.jonas_ws.deployment.xml.SoapRole;

import java.util.Hashtable;
import java.util.List;
import java.util.Vector;

import javax.xml.namespace.QName;


/**
 * This class provides methodes to manipulate handler's attributes.
 *
 * @author Guillaume Sauthier
 * @author Xavier Delplanque
 */
public class HandlerDesc {
    /** The name of the Handler (must be unique) */
    private String name;

    /** The class of the Handler */
    private Class clazz;

    /** The headers the handlers will access */
    private Vector headers = new Vector();

    /** The init params for Handler instanciation */
    private Hashtable params = new Hashtable();

    /** The list of roles the Handlers will play */
    private Vector roles = new Vector();

    /**
     * Constructor : creates a HandlerDesc object.
     *
     * @param jarCL ejbjar or war classloader.
     * @param handler contains informations defined in webservices  deployment
     *        descriptor (handler).
     *
     * @throws WSDeploymentDescException When error occurs.
     */
    public HandlerDesc(ClassLoader jarCL, Handler handler)
        throws WSDeploymentDescException {
        // set name
        if (handler.getHandlerName() == null) {
            throw new WSDeploymentDescException("handler name undefined");
        }

        name = handler.getHandlerName().getValue();

        // set class
        String className = handler.getHandlerClass().getValue().trim();

        try {
            clazz = jarCL.loadClass(className);
        } catch (ClassNotFoundException e) {
            throw new WSDeploymentDescException("handler class not found", e);
        }

        // set init parameters
        List ipl = handler.getInitParamList();

        try {
            for (int i = 0; i < ipl.size(); i++) {
                // add in params table each init parameter name and associated value
                if (ipl.get(i) != null) {
                    InitParam p = (InitParam) ipl.get(i);
                    params.put(p.getParamName().getValue(),
                        p.getParamValue().getValue());
                }
            }
        } catch (NullPointerException e) {
            throw new WSDeploymentDescException("parameter name missing", e);
        }

        // set soap headers list
        List shl = handler.getSoapHeaderList();

        for (int i = 0; i < shl.size(); i++) {
            // build qnames and add it in the table
            if (shl.get(i) != null) {
                SoapHeader sh = (SoapHeader) shl.get(i);
                QName qn = new QName(sh.getNamespaceURI().getValue(),
                        sh.getLocalpart().getValue());
                headers.add(qn);
            }
        }

        // fill roles a list containing soap role names
        List srl = handler.getSoapRoleList();

        for (int i = 0; i < srl.size(); i++) {
            if (srl.get(i) != null) {
                String role = ((SoapRole) srl.get(i)).getValue();
                roles.add(role);
            }
        }
    }

    /**
     * Return the list of SOAP Headers the Handler will access.
     *
     * @return the list of SOAP Headers the Handler will access.
     */
    public List getSOAPHeaders() {
        return headers;
    }

    /**
     * Return the name of the Handler.
     *
     * @return the name of the Handler.
     */
    public String getName() {
        return name;
    }

    /**
     * Return the class of the Handler.
     *
     * @return the class of the Handler.
     */
    public Class getHandlerClass() {
        return clazz;
    }

    /**
     * Return the table of init-param.
     *
     * @return the table of init-param.
     */
    public Hashtable getInitParams() {
        return params;
    }

    /**
     * Return the value of specified init-param.
     *
     * @param name The key of init-param map.
     *
     * @return the value of specified init-param.
     */
    public String getInitParam(String name) {
        return (String) params.get(name);
    }

    /**
     * Return the list of roles the Handler will play.
     *
     * @return the list of roles the Handler will play.
     */
    public List getSOAPRoles() {
        return roles;
    }
}
