/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial Developer : Delplanque Xavier & Sauthier Guillaume
 * --------------------------------------------------------------------------
 * $Id: HandlerRef.java,v 1.4 2003/09/09 11:57:33 sauthieg Exp $
 * --------------------------------------------------------------------------
*/

package org.objectweb.jonas_ws.deployment.api;

import java.util.Hashtable;
import java.util.List;
import java.util.Vector;

import javax.xml.namespace.QName;


/**
 * The HandlerRef class describe a Handler to use on the client side of a Web
 * Service.
 *
 * @author Guillaume Sauthier
 * @author Xavier Delplanque
 */
public class HandlerRef {
    /** The name of the handler (must be unique) */
    private String name;

    /** The handler class */
    private Class clazz;

    /** Params needed to initialize the handler */
    private Hashtable params;

    /** The list of SOAP Headers the handler will access */
    private Vector headers;

    /** The list of SOAP actor the handler will play as a role */
    private Vector roles;

    /** List of port names the handler is associated with */
    private Vector portNames;

    /**
     * Creates a new HandlerRef object.
     *
     * @param classLoader ejbjar classLoader
     * @param handler contains informations defined in ejb deployment
     *        descriptor (service-ref)
     *
     * @throws WSDeploymentDescException When Construction fails.
     */
    public HandlerRef(ClassLoader classLoader,
        org.objectweb.jonas_ejb.deployment.xml.Handler handler)
        throws WSDeploymentDescException {
        name = handler.getHandlerName().getValue();

        String className = handler.getHandlerClass().getValue().trim();

        try {
            clazz = classLoader.loadClass(className);
        } catch (ClassNotFoundException e) {
            throw new WSDeploymentDescException("handler class not found", e);
        }

        // fill params table
        params = new Hashtable();

        List iparams = handler.getInitParamList();

        try {
            for (int i = 0; i < iparams.size(); i++) {
                // add in params table each init parameter name and associated value
                org.objectweb.jonas_ejb.deployment.xml.InitParam p = (org.objectweb.jonas_ejb.deployment.xml.InitParam) iparams.get(i);

                if (p != null) {
                    params.put(p.getParamName().getValue(),
                        p.getParamValue().getValue());
                }
            }
        } catch (NullPointerException e) {
            throw new WSDeploymentDescException("parameter name missing", e);
        }

        // fill headers a list containing soap header QNames
        headers = new Vector();

        List shl = handler.getSoapHeaderList();

        for (int i = 0; i < shl.size(); i++) {
            // build qnames and add it in the table
            org.objectweb.jonas_ejb.deployment.xml.SoapHeader sh = (org.objectweb.jonas_ejb.deployment.xml.SoapHeader) shl.get(i);

            if (sh != null) {
                QName qn = new QName(sh.getNamespaceURI().getValue(),
                        sh.getLocalpart().getValue());
                headers.add(qn);
            }
        }

        // fill roles a list containing soap role names
        roles = new Vector();

        List srl = handler.getSoapRoleList();

        for (int i = 0; i < srl.size(); i++) {
            String role = ((org.objectweb.jonas_ejb.deployment.xml.SoapRole) srl.get(i)).getValue();

            if (role != null) {
                roles.add(role);
            }
        }

        // fill portNames a list containing ports names
        portNames = new Vector();

        List pnl = handler.getPortNameList();

        for (int i = 0; i < pnl.size(); i++) {
            String pn = ((org.objectweb.jonas_ejb.deployment.xml.PortName) pnl.get(i)).getValue();

            if (pn != null) {
                portNames.add(pn);
            }
        }
    }

    /**
     * Creates a new HandlerRef object.
     *
     * @param classLoader ejbjar classLoader
     * @param handler contains informations defined in web deployment
     *        descriptor (service-ref)
     *
     * @throws WSDeploymentDescException When Construction fails.
     */
    public HandlerRef(ClassLoader classLoader,
        org.objectweb.jonas_web.deployment.xml.Handler handler)
        throws WSDeploymentDescException {
        name = handler.getHandlerName().getValue();

        String className = handler.getHandlerClass().getValue().trim();

        try {
            clazz = classLoader.loadClass(className);
        } catch (ClassNotFoundException e) {
            throw new WSDeploymentDescException("handler class not found", e);
        }

        // fill params table
        params = new Hashtable();

        List iparams = handler.getInitParamList();

        try {
            for (int i = 0; i < iparams.size(); i++) {
                // add in params table each init parameter name and associated value
                org.objectweb.jonas_web.deployment.xml.InitParam p = (org.objectweb.jonas_web.deployment.xml.InitParam) iparams.get(i);

                if (p != null) {
                    params.put(p.getParamName().getValue(),
                        p.getParamValue().getValue());
                }
            }
        } catch (NullPointerException e) {
            throw new WSDeploymentDescException("parameter name missing", e);
        }

        // fill headers a list containing soap header QNames
        headers = new Vector();

        List shl = handler.getSoapHeaderList();

        for (int i = 0; i < shl.size(); i++) {
            // build qnames and add it in the table
            org.objectweb.jonas_web.deployment.xml.SoapHeader sh = (org.objectweb.jonas_web.deployment.xml.SoapHeader) shl.get(i);

            if (sh != null) {
                QName qn = new QName(sh.getNamespaceURI().getValue(),
                        sh.getLocalpart().getValue());
                headers.add(qn);
            }
        }

        // fill roles a list containing soap role names
        roles = new Vector();

        List srl = handler.getSoapRoleList();

        for (int i = 0; i < srl.size(); i++) {
            String role = ((org.objectweb.jonas_web.deployment.xml.SoapRole) srl.get(i)).getValue();

            if (role != null) {
                roles.add(role);
            }
        }

        // fill portNames a list containing ports names
        portNames = new Vector();

        List pnl = handler.getPortNameList();

        for (int i = 0; i < pnl.size(); i++) {
            String pn = ((org.objectweb.jonas_web.deployment.xml.PortName) pnl.get(i)).getValue();

            if (pn != null) {
                portNames.add(pn);
            }
        }
    }

    /**
     * Return the name of the Handler.
     *
     * @return the name of the Handler.
     */
    public String getName() {
        return name;
    }

    /**
     * Return the Handler implementation class.
     *
     * @return the Handler class
     */
    public Class getHandlerClass() {
        return clazz;
    }

    /**
     * Return all the init-params of the Handler.
     *
     * @return the init-params of the Handler
     */
    public Hashtable getInitParams() {
        return params;
    }

    /**
     * Return the value of an init-param.
     *
     * @param pname The key of init-param map.
     *
     * @return the value of an init-param
     */
    public String getInitParam(String pname) {
        return (String) params.get(pname);
    }

    /**
     * Return the list of Headers the Handlers will access.
     *
     * @return the list of Headers the Handlers will access.
     */
    public List getSOAPHeaders() {
        return headers;
    }

    /**
     * Return the list of SOAP Actor Definitions the Handler will play as a
     * role.
     *
     * @return the list of Role the Handler will play
     */
    public List getSOAPRoles() {
        return roles;
    }

    /**
     * Return the list of port name the Handler is associated with. The names
     * match the localPart of the Port QName.
     *
     * @return the list of port name the Handler is associated with.
     */
    public List getPortNames() {
        return portNames;
    }

    /**
     * Test Equality between 2 Objects.
     *
     * @param other The object to compare.
     *
     * @return true if the objects are equals in value, else false.
     */
    public boolean equals(Object other) {
        if (other == null) {
            return false;
        }

        if (!(other instanceof HandlerRef)) {
            return false;
        }

        HandlerRef ref = (HandlerRef) other;

        if (!name.equals(ref.getName())) {
            return false;
        }

        if (!clazz.getName().equals(ref.getHandlerClass().getName())) {
            return false;
        }

        if (!params.equals(ref.getInitParams())) {
            return false;
        }

        if (!headers.equals(ref.getSOAPHeaders())) {
            return false;
        }

        if (!roles.equals(ref.getSOAPRoles())) {
            return false;
        }

        if (!portNames.equals(ref.getPortNames())) {
            return false;
        }

        // After all theses tests, the 2 objects are equals in value
        return true;
    }
}
