/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial Developer : Delplanque Xavier & Sauthier Guillaume
 * --------------------------------------------------------------------------
 * $Id: MappingFile.java,v 1.4 2003/09/09 11:57:33 sauthieg Exp $
 * --------------------------------------------------------------------------
*/

package org.objectweb.jonas_ws.deployment.api;

import org.objectweb.jonas_ws.deployment.xml.JavaWsdlMapping;
import org.objectweb.jonas_ws.deployment.xml.JavaWsdlMappingUnmarshaller;
import org.objectweb.jonas_ws.deployment.xml.NamespaceURI;
import org.objectweb.jonas_ws.deployment.xml.PackageMapping;
import org.objectweb.jonas_ws.deployment.xml.PackageType;

import org.xml.sax.EntityResolver;
import org.xml.sax.ErrorHandler;
import org.xml.sax.SAXParseException;

import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;

import java.util.Hashtable;
import java.util.List;
import java.util.Map;


/**
 * this Class is used to manipulate jaxrpc-mapping-file. This file contains
 * informations for mapping between XML namespaces and java packages. We
 * actually support just a few part of this file. According with JSR 921, this
 * file must contain class mapping information (exceptions/faults,
 * types/classes, portTypes/interfaces ...).
 *
 * @author Guillaume Sauthier
 * @author Xavier Delplanque
 */
public class MappingFile {
    /** The error handler to use (null = default) */
    private static ErrorHandler errorHandler = null;

    /** The entity resolver to use (null = default) */
    private static EntityResolver entityResolver = null;

    /** Flag for parser validation */
    private static boolean parsingWithValidation = true;

    /** SAXParserHelper used */
    private static SAXParserHelper helper = null;

    /** Associates a namespace to its package. It's the cache. */
    private Hashtable namespacePackageMapping;

    /**
     * Constructor : creates a MappingFile object.
     *
     * @param is mapping file InputStream.
     * @param name file name (only for messages).
     *
     * @throws WSDeploymentDescException When reading/parsing fails.
     */
    public MappingFile(InputStream is, String name)
        throws WSDeploymentDescException {
        namespacePackageMapping = new Hashtable();

        JavaWsdlMapping mapping = loadMappingFile(is, name);
        fillPackageMapping(mapping);
    }

    /**
     * Fill up the package mapping hashtable
     *
     * @param mapping jaxrpc-mapping-file root element.
     */
    private void fillPackageMapping(JavaWsdlMapping mapping) {
        List pml = mapping.getPackageMappingList();
        PackageMapping pm = null;
        PackageType pt = null;
        NamespaceURI nuri = null;

        for (int i = 0; i < pml.size(); i++) {
            // get each mapping file package mappings
            pm = (PackageMapping) pml.get(i);

            if (pm != null) {
                pt = pm.getPackageType();
                nuri = pm.getNamespaceURI();

                if ((nuri != null) && (pt != null)) {
                    //add in the hashtable the namespace to package relation (cache)
                    namespacePackageMapping.put(nuri.getValue().trim(),
                        pt.getValue().trim());
                }
            }
        }
    }

    /**
     * Load the MappingFile file
     *
     * @param is The mapping file InputStream
     * @param name mapping filename
     *
     * @return Zeus Root Element of the xml file.
     *
     * @throws WSDeploymentDescException When parsing fails.
     */
    private JavaWsdlMapping loadMappingFile(InputStream is, String name)
        throws WSDeploymentDescException {
        Reader reader = new InputStreamReader(is);
        helper = new SAXParserHelper(errorHandler, entityResolver);

        try {
            JavaWsdlMappingUnmarshaller.setErrorHandler(helper.getWrappedErrorHandler());
            JavaWsdlMappingUnmarshaller.setEntityResolver(helper.getWrappedEntityResolver());

            // Unmarshal the XML, and validate as it is unmarshalled
            JavaWsdlMapping m = JavaWsdlMappingUnmarshaller.unmarshal(reader,
                    parsingWithValidation);

            return m;
        } catch (IOException e) {
            if (helper.getSAXParseException() != null) {
                String detailMsg = getSAXMsg(name,
                        helper.getSAXParseException(), "Invalid XML");
                throw new WSDeploymentDescException(detailMsg, e);
            } else {
                throw new WSDeploymentDescException(
                    "Error reading XML document " + name, e);
            }
        } finally {
            if (reader != null) {
                try {
                    reader.close();
                } catch (IOException ignored) {
                    ;
                }
            }
        }
    }

    /**
     * return the mapping between XML namespaces and Java packages defined in
     * the jaxrpc-mapping file.
     *
     * @return the mapping between XML namespaces and Java packages
     */
    public Map getMappings() {
        return namespacePackageMapping;
    }

    /**
     * Return the package associated with the specified namespaceURI (can be
     * null).
     *
     * @param namespaceURI the namespace key to retrieve the package name
     *
     * @return the package associated with the specified namespaceURI. (null if
     *         namespace not present).
     */
    public String getMapping(String namespaceURI) {
        return (String) namespacePackageMapping.get(namespaceURI);
    }

    /**
     * Build a string representation of MappingFile
     *
     * @return String representation of the mapping file.
     */
    public String toString() {
        StringBuffer ret = new StringBuffer();

        ret.append("\ngetMappings()=" + namespacePackageMapping);

        return ret.toString();
    }

    /**
     * build a message from SAX Exception in a consistent style, consistent
     * with emacs compile mode (same as grep, cc, javac, etc).
     *
     * @param fileName name of the parsed file.
     * @param exception The SAX Exception thrown during parsing.
     * @param msg message to be displayed.
     *
     * @return String representation of the error.
     */
    private String getSAXMsg(String fileName, SAXParseException exception,
        String msg) {
        return fileName + ":" + exception.getLineNumber() + ":" +
        exception.getColumnNumber() + ": " + msg;
    }

    /**
     * Return true if the 2 objects are equals in value.
     *
     * @param other the object to compare.
     *
     * @return true if objects are equals in value, else false.
     */
    public boolean equals(Object other) {
        if (other == null) {
            return false;
        }

        if (!(other instanceof MappingFile)) {
            return false;
        }

        MappingFile ref = (MappingFile) other;

        if (!namespacePackageMapping.equals(ref.getMappings())) {
            return false;
        }

        // After all theses tests, the 2 objects are equals in value
        return true;
    }
}
