/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial Developer : Delplanque Xavier & Sauthier Guillaume
 * --------------------------------------------------------------------------
 * $Id: PortComponentDesc.java,v 1.4 2003/09/09 11:57:33 sauthieg Exp $
 * --------------------------------------------------------------------------
*/

package org.objectweb.jonas_ws.deployment.api;

import org.objectweb.jonas_ejb.deployment.api.SessionStatelessDesc;

import org.objectweb.jonas_web.deployment.api.WebContainerDeploymentDesc;

import org.objectweb.jonas_ws.deployment.xml.Handler;
import org.objectweb.jonas_ws.deployment.xml.Localpart;
import org.objectweb.jonas_ws.deployment.xml.NamespaceURI;
import org.objectweb.jonas_ws.deployment.xml.PortComponent;
import org.objectweb.jonas_ws.deployment.xml.ServiceImplBean;
import org.objectweb.jonas_ws.deployment.xml.WsdlPort;

import java.net.URL;

import java.util.List;
import java.util.Vector;

import javax.xml.namespace.QName;


/**
 * This class provides methodes to manipulate portComponent's attributes.
 *
 * @author Guillaume Sauthier
 * @author Xavier Delplanque
 */
public class PortComponentDesc {
    /**
     * SSB Endpoint : Bean
     */
    private static final int BEAN_TYPE = 0;

    /**
     * JaxRpc Endpoint : Class
     */
    private static final int JAXRPC_TYPE = 1;

    /** The name of the Port (must be unique) */
    private String name;

    /** The interface of the endpoint/port */
    private Class sei;

    /** the sib link, can be ejb or servlet link */
    private String sibLink;

    /** The class of bean implementing the service */
    private Class sib;

    /** If the PortComponent is linked to a SSB */
    private SessionStatelessDesc beanDesc = null;

    /** If the PortComponent is linked to a servlet */
    private WebContainerDeploymentDesc webDesc = null;

    /** The list of Handlers the Port will run */
    private Vector handlers = new Vector();

    /** The QName of the WSDL port */
    private QName portQName;

    /** The PortComponent implmentation type : Bean/JaxRpc */
    private int implType;

    /** The URL where to access the port-component (null before resolving) */
    private URL endpoint = null;

    /**
     * Creates a new PortComponentDesc object.
     *
     * @param jarCL the module (ejbjar or war) classloader.
     * @param pc webservices portcomponent element
     *
     * @throws WSDeploymentDescException When construction fails.
     */
    public PortComponentDesc(ClassLoader jarCL, PortComponent pc)
        throws WSDeploymentDescException {
        // set name
        if (pc.getPortComponentName() == null) {
            throw new WSDeploymentDescException(
                "PortComponent's name undefined");
        }

        name = pc.getPortComponentName().getValue();

        // set sei
        String seiClass = null;

        try {
            seiClass = pc.getServiceEndpointInterface().getValue().trim();
            sei = jarCL.loadClass(seiClass);
        } catch (ClassNotFoundException e) {
            throw new WSDeploymentDescException("SEI class not found " +
                seiClass + " in port component :" + name, e);
        }

        // set sibLink
        ServiceImplBean pcsib = pc.getServiceImplBean();

        if (pcsib == null) {
            throw new WSDeploymentDescException(
                "SIB missing in port component :" + name);
        }

        if (pcsib.getEjbLink() != null) {
            // we are in ssb case
            sibLink = pcsib.getEjbLink().getValue().trim();

            // set the port component type
            implType = BEAN_TYPE;
        } else {
            // we are in jaxrpc endpoint case
            sibLink = pcsib.getServletLink().getValue().trim();

            // set the port component type
            implType = JAXRPC_TYPE;
        }

        // set and init handlers
        List hl = pc.getHandlerList();
        Handler h = null;

        for (int i = 0; i < hl.size(); i++) {
            // for each reference, build and add a HandlerDesc Object
            if (hl.get(i) != null) {
                // get the standard dd handler
                h = (Handler) hl.get(i);

                // build and add a new handler
                handlers.add(new HandlerDesc(jarCL, h));
            }
        }

        // set portQName
        WsdlPort wp = pc.getWsdlPort();

        if (wp == null) {
            throw new WSDeploymentDescException(
                "Wsdl port undefined in port component :" + name);
        }

        NamespaceURI nuri = wp.getNamespaceURI();

        if (nuri == null) {
            throw new WSDeploymentDescException(
                "WsdlPort's NamespaceURI undefined in port component :" + name);
        }

        Localpart lp = wp.getLocalpart();

        if (lp == null) {
            throw new WSDeploymentDescException(
                "WsdlPort's Localpart undefined in port component :" + name);
        }

        portQName = new QName(nuri.getValue().trim(), lp.getValue().trim());
    }

    /**
     * Return the name of the PortComponent.
     *
     * @return the name of the PortComponent.
     */
    public String getName() {
        return name;
    }

    /**
     * Return the Service Endpoint Interface.
     *
     * @return the Service Endpoint Interface.
     */
    public Class getServiceEndpointInterface() {
        return sei;
    }

    /**
     * Return the implementation bean. It's the bean class which do the real
     * work.
     *
     * @return the implementation bean.
     */
    public Class getServiceImplementationBean() {
        return sib;
    }

    /**
     * Return the WSDL's Port QName, the PortComponent is asssociated with.
     *
     * @return the port's QName
     */
    public QName getQName() {
        return portQName;
    }

    /**
     * Return the list of Handlers the PortComponent is associated with.
     *
     * @return the list of Handlers the PortComponent is associated with.
     */
    public List getHandlers() {
        return handlers;
    }

    /**
     * Return the sib name. the sib can be an ejb or a servlet.
     *
     * @return the sib name.
     */
    public String getSibLink() {
        return sibLink;
    }

    /**
     * Set the sib class.
     *
     * @param sib the sib class.
     *
     * @deprecated sib class set with the setBeanDesc or setWebDesc methods.
     */
    public void setSibClass(Class sib) {
        this.sib = sib;
    }

    /**
     * Return true if the Service Impl Bean is an EJB.
     *
     * @return true if the Service Impl Bean is an EJB.
     */
    public boolean hasBeanImpl() {
        return (implType == BEAN_TYPE);
    }

    /**
     * Return true if the Service Impl Bean is a JaxRpc component.
     *
     * @return true if the Service Impl Bean is a JaxRpc component.
     */
    public boolean hasJaxRpcImpl() {
        return (implType == JAXRPC_TYPE);
    }

    /**
     * Return the StatelessSessionDesc object linked with this
     * portComponentDesc or null if the portComp is linked to a servlet.
     *
     * @return the StatelessSessionDesc object linked with this
     *         portComponentDesc or null if the portComp is linked to a
     *         servlet.
     */
    public SessionStatelessDesc getBeanDesc() {
        return beanDesc;
    }

    /**
     * Return the WebContainerDeploymentDesc object linked with this
     * portComponentDesc or null if the portComp is linked to a stateless
     * session bean.
     *
     * @return the WebContainerDeploymentDesc object linked with this
     *         portComponentDesc or null if the portComp is linked to a
     *         stateless session bean.
     */
    public WebContainerDeploymentDesc getWebDesc() {
        return webDesc;
    }

    /**
     * Set the beanDesc for this endpoint.
     *
     * @param bean The SSB Object declaring the endpoint.
     */
    public void setBeanDesc(SessionStatelessDesc bean) {
        beanDesc = bean;
        sib = beanDesc.getEjbClass();

        // Ensure that port type is ssb
        implType = BEAN_TYPE;
    }

    /**
     * Set the webDesc for this endpoint.
     *
     * @param web The web DD declaring the JaxRpc Endpoint.
     */
    public void setWebDesc(WebContainerDeploymentDesc web) {
        webDesc = web;
        sib = webDesc.getServletClass(sibLink);

        // Ensure that port type is jaxrpc
        implType = JAXRPC_TYPE;
    }

    /**
     * Return the URL where the port-component can be accessed.
     *
     * @return the URL where the port-component can be accessed.
     */
    public URL getEndpointURL() {
        return endpoint;
    }

    /**
     * Set the Endpoint URL of the port-component.
     *
     * @param url the resolved endpoint URL.
     */
    public void setEndpointURL(URL url) {
        endpoint = url;
    }
}
