/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial Developer : Delplanque Xavier & Sauthier Guillaume
 * --------------------------------------------------------------------------
 * $Id: ServiceDesc.java,v 1.5 2003/10/20 13:50:10 sauthieg Exp $
 * --------------------------------------------------------------------------
*/

package org.objectweb.jonas_ws.deployment.api;

import org.objectweb.jonas_ws.deployment.xml.JaxrpcMappingFile;
import org.objectweb.jonas_ws.deployment.xml.PortComponent;
import org.objectweb.jonas_ws.deployment.xml.WebserviceDescription;
import org.objectweb.jonas_ws.deployment.xml.WsdlFile;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;

import java.util.Hashtable;
import java.util.List;
import java.util.Vector;
import java.util.jar.JarFile;

import javax.xml.namespace.QName;


/**
 * This class corresponds to 1 <code>webservices-description</code> XML
 * element. It's used to get a MappingFile instance if it's defined, idem for
 * WSDLFile instance and give access to the Web service Port component list.
 *
 * @author Guillaume Sauthier
 * @author Xavier Delplanque
 */
public class ServiceDesc {

    /** The webservice-description-name */
    private String name;

    /** The list of PortComponentDesc */
    private Hashtable namePCDescBindings = new Hashtable();

    /** The WSDL of the WebService */
    private WSDLFile wsdl = null;

    /** The JaxRpc Mapping of the WebService */
    private MappingFile mapping = null;

    /**
     * Constructor : creates a ServiceDesc object.
     *
     * @param jarCL ejbjar or war classLoader
     * @param wsd Zeus object containing Webservices informations
     * @param fileName artifact jar or directory name
     *
     * @throws WSDeploymentDescException When contruction fails.
     */
    public ServiceDesc(
        ClassLoader           jarCL,
        WebserviceDescription wsd,
        String                fileName) throws WSDeploymentDescException {
        // set name
        if (wsd.getWebserviceDescriptionName().getValue() == null) {
            throw new WSDeploymentDescException("Service Desc name missing");
        }

        name = wsd.getWebserviceDescriptionName().getValue().trim();

        // set wsdl
        File moduleFile = new File(fileName);

        // get ServiceRef WSDLFile
        WsdlFile wf = wsd.getWsdlFile();

        String wsdlFileName = "";

        // If there is one WSDL to load
        if (wf != null) {
            wsdlFileName = wf.getValue().trim();

            InputStream isWsdl = null;

            // If we have file stored in Jar
            if (moduleFile.exists()) {
                if (moduleFile.isFile()) {
                    JarFile jf = null;

                    // Get the WSDL file of the jar as InputStream
                    try {
                        jf = new JarFile(moduleFile);
                        isWsdl = jf.getInputStream(jf.getEntry(wsdlFileName));
                    } catch (Exception e) {
                        if (jf != null) {
                            try {
                                jf.close();
                            } catch (IOException i) {
                                ;
                            }
                        }

                        String svc =
                            (name == null) ? "" : (" for Service " + name);
                        throw new WSDeploymentDescException(
                            "Cannot read the wsdl-file " + wsdlFileName
                            + " in " + fileName + svc,
                            e);
                    }
                }
                // filename is a Directory
                else {
                    try {
                        isWsdl =
                            new FileInputStream(new File(moduleFile,
                                    wsdlFileName));
                    } catch (IOException ioe) {
                        String svc =
                            (name == null) ? "" : (" for Service " + name);
                        throw new WSDeploymentDescException(
                            "Cannot read the wsdl-file " + wsdlFileName
                            + " in " + fileName + svc,
                            ioe);
                    }
                }

                // build the WSDLFile
                wsdl = new WSDLFile(isWsdl, wsdlFileName);
            } else { // the moduleFile doesn't exist
                throw new WSDeploymentDescException(fileName
                    + " doesn't exists");
            }
        }

        // set mapping file
        JaxrpcMappingFile mf = wsd.getJaxrpcMappingFile();

        String mappingFileName = "";

        // If there is no Mapping to load
        if (mf != null) {
            mappingFileName = mf.getValue().trim();

            InputStream isMapping = null;

            // If we have file stored in Jar
            if (moduleFile.exists()) {
                // If we have file stored in Jar
                if (moduleFile.isFile()) {
                    JarFile jf = null;

                    // Get the Mapping file of the jar as InputStream
                    try {
                        jf = new JarFile(fileName);
                        isMapping =
                            jf.getInputStream(jf.getEntry(mappingFileName));
                    } catch (Exception e) {
                        if (jf != null) {
                            try {
                                jf.close();
                            } catch (IOException i) {
                                ;
                            }
                        }

                        String svc =
                            (name == null) ? "" : (" for Service " + name);
                        throw new WSDeploymentDescException(
                            "Cannot read the jaxrpc-mapping-file "
                            + mappingFileName + " in " + fileName + svc,
                            e);
                    }
                }
                // filename is a Directory
                else {
                    try {
                        isMapping =
                            new FileInputStream(new File(moduleFile,
                                    mappingFileName));
                    } catch (IOException ioe) {
                        String svc =
                            (name == null) ? "" : (" for Service " + name);
                        throw new WSDeploymentDescException(
                            "Cannot read the jaxrpc-mapping-file "
                            + mappingFileName + " in " + fileName + svc,
                            ioe);
                    }
                }

                // build the MappingFile
                mapping = new MappingFile(isMapping, mappingFileName);
            } else { // the moduleFile doesn't exist
                throw new WSDeploymentDescException(fileName
                    + " doesn't exists");
            }
        }

        // set namePCDescBindings
        List pcl = wsd.getPortComponentList();

        for (int i = 0; i < pcl.size(); i++) {
            // for each port component, build and add a PortComponentDesc Object
            if (pcl.get(i) != null) {
                PortComponent pc = (PortComponent) pcl.get(i);
                PortComponentDesc pcd = new PortComponentDesc(jarCL, pc);

                if (!wsdl.hasPort(pcd.getQName())) {
                    throw new WSDeploymentDescException("port-component '"
                        + pcd.getName()
                        + "' is linked with an unknown WSDL port '"
                        + pcd.getQName() + "'");
                }

                if (namePCDescBindings.put(pcd.getName(), pcd) != null) {
                    throw new WSDeploymentDescException("port-component '"
                        + pcd.getName()
                        + "' has same name than a previously recorded port-component '");
                }
            }
        }

        // validate informations : 
        List ports = getPortComponents();

        // SoapHeader validity between handlers and wsdl
        for (int i = 0; i < ports.size(); i++) {
            // for each port component desc
            if (ports.get(i) != null) {
                List hl = ((PortComponentDesc) ports.get(i)).getHandlers();

                for (int j = 0; j < hl.size(); j++) {
                    // for each handler
                    if (hl.get(j) != null) {
                        HandlerDesc hd = (HandlerDesc) hl.get(j);
                        List shl = hd.getSOAPHeaders();

                        for (int k = 0; k < shl.size(); k++) {
                            if (!wsdl.hasSOAPHeader((QName) shl.get(k))) {
                                // the handler use a soap header undefined in the wsdl
                                throw new WSDeploymentDescException("handler '"
                                    + hd.getName()
                                    + "' use an undefined soap header : '"
                                    + (QName) shl.get(k) + "' in "
                                    + wsdlFileName);
                            }
                        }
                    }
                }
            }
        }

        // Ports use Soap bindings
        for (int i = 0; i < ports.size(); i++) {
            if (ports.get(i) != null) {
                PortComponentDesc pcd = (PortComponentDesc) ports.get(i);

                if (!wsdl.hasSOAPBinding(pcd.getQName())) {
                    throw new WSDeploymentDescException("port-component '"
                        + pcd.getName() + "' associated with port '"
                        + pcd.getQName() + "' doesn't use a SOAP binding");
                }
            }
        }
    }

    /**
     * Return the name of the WebService.
     *
     * @return the name of the WebService.
     */
    public String getName() {
        return name;
    }

    /**
     * Return the Mapping File.
     *
     * @return the MappingFile.
     */
    public MappingFile getMapping() {
        return mapping;
    }

    /**
     * Return the WSDL File.
     *
     * @return the WSDL File.
     */
    public WSDLFile getWSDL() {
        return wsdl;
    }

    /**
     * Return the port component list of this service desc.
     *
     * @return the port component list of this service desc.
     */
    public List getPortComponents() {
        return (List) new Vector(namePCDescBindings.values());
    }

    /**
     * Return the port component desc with the given name if it exists, null
     * else.
     *
     * @param pcName the name of the portComponent to retrieve.
     *
     * @return the port component desc with the given name if it exists, null
     *         else.
     */
    public PortComponentDesc getPortComponent(String pcName) {
        return (PortComponentDesc) namePCDescBindings.get(pcName);
    }
}
