/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial Developer : Delplanque Xavier & Sauthier Guillaume
 * --------------------------------------------------------------------------
 * $Id: ServiceRef.java,v 1.4 2003/09/09 11:57:33 sauthieg Exp $
 * --------------------------------------------------------------------------
*/

package org.objectweb.jonas_ws.deployment.api;

import org.objectweb.jonas_ejb.lib.BeanNaming;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;

import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;
import java.util.jar.JarFile;

import javax.xml.namespace.QName;


/**
 * A ServiceRef describe a dependancy from a J2EE component onto a WebService
 * (it is a Web Service Client). The component will have the hability to
 * lookup the Service Implementation and use it.
 *
 * @author Guillaume Sauthier
 * @author Xavier Delplanque
 */
public class ServiceRef {
    /** jonas-init-param property name for service implementation. */
    public static final String SERVICE_CLASS_NAME = "serviceClassName";

    /** All the params of the ServiceRef */
    private Hashtable params = new Hashtable();

    /** The list of PortComponentRefs */
    private Vector pcRefs = new Vector();

    /** The list of Handlers */
    private Vector hRefs = new Vector();

    /** The WSDLFile of the ServiceRef */
    private WSDLFile wsdl;

    /** The MappingFile of the ServiceRef */
    private MappingFile mapping;

    /** The name of the ServiceRef (used in lookup) */
    private String name;

    /** The jndi name of the ServiceRef */
    private String jndiName;

    /** The service class */
    private Class serviceClass;

    /** The service interface */
    private Class serviceInterface;

    /** The Service QName */
    private QName serviceQName = null;

    /** The Module file (directory or jar) */
    private File moduleFile = null;

    /** filename */
    private String filename = null;

    /** WSDL FileName */
    private String wsdlFileName = null;

    /** DOCUMENT ME! */
    private String mappingFileName = null;

    /**
     * Creates a new ServiceRef object.
     *
     * @param classLoader ejb class loader
     * @param sref generated object containing service-ref informations.
     * @param jsref JOnAS specific service-ref informations.
     * @param filename the name of the archive where retrieve WSDL and mapping
     *        files
     *
     * @throws WSDeploymentDescException if the wsdl file is undefined  or if
     *         informations in wsdl and handler doesn't match.
     */
    public ServiceRef(
        ClassLoader                                            classLoader,
        org.objectweb.jonas_ejb.deployment.xml.ServiceRef      sref,
        org.objectweb.jonas_ejb.deployment.xml.JonasServiceRef jsref,
        String                                                 filename)
        throws WSDeploymentDescException {
        this.filename = filename;
        moduleFile = new File(filename);

        // set ServiceRef name
        name = sref.getServiceRefName().getValue().trim();

        // set jndiName
        jndiName = jsref.getJndiName().getValue().trim();

        // set Service QName (if exists)
        if (sref.getServiceQname() != null) {
            serviceQName =
                new QName(sref.getServiceQname().getNamespaceURI().getValue()
                              .trim(),
                    sref.getServiceQname().getLocalpart().getValue().trim());
        }

        // set Service Interface
        if (sref.getServiceInterface() != null) {
            String si = sref.getServiceInterface().getValue().trim();

            try {
                serviceInterface = classLoader.loadClass(si);
            } catch (ClassNotFoundException cnf) {
                throw new WSDeploymentDescException("service-interface : " + si
                    + " not found !", cnf);
            }
        }

        // this interface must implements javax.xml.rpc.Service
        if (!javax.xml.rpc.Service.class.isAssignableFrom(serviceInterface)) {
            String err =
                "The service-interface : " + serviceInterface.getName()
                + " must extends the javax.xml.rpc.Service interface.";
            throw new WSDeploymentDescException(err);
        }

        // ServiceRef init parameters : jonas specific deployment desc
        List jipl = jsref.getJonasInitParamList();

        try {
            for (int i = 0; i < jipl.size(); i++) {
                // add in params table each init parameter name and associated value
                org.objectweb.jonas_ejb.deployment.xml.JonasInitParam p =
                    (org.objectweb.jonas_ejb.deployment.xml.JonasInitParam) jipl
                    .get(i);

                if (p != null) {
                    params.put(p.getParamName().getValue().trim(),
                        p.getParamValue().getValue().trim());
                }
            }
        } catch (NullPointerException e) {
            throw new WSDeploymentDescException("parameter name missing", e);
        }

        // set Service Class
        String sc = getParam(SERVICE_CLASS_NAME);

        if (sc != null) {
            // the name is defined, the service class is already generated.
            try {
                serviceClass = classLoader.loadClass(sc);
            } catch (ClassNotFoundException cnf) {
                throw new WSDeploymentDescException("jonas-init-param/"
                    + SERVICE_CLASS_NAME + " : " + sc + " not found !", cnf);
            }

            // we test that classes are valids
            if (!serviceInterface.isAssignableFrom(serviceClass)) {
                String err =
                    "The service-class must implement the service-interface."
                    + serviceClass.getName() + " must be assignable from "
                    + serviceInterface.getName();
                throw new WSDeploymentDescException(err);
            }
        }

        // fill portComponentRef list	
        List pcrl = sref.getPortComponentRefList();

        for (int i = 0; i < pcrl.size(); i++) {
            // for each reference, build and add a PortComponentRef Object
            if (pcrl.get(i) != null) {
                org.objectweb.jonas_ejb.deployment.xml.PortComponentRef ref =
                    (org.objectweb.jonas_ejb.deployment.xml.PortComponentRef) pcrl
                    .get(i);
                pcRefs.add(new PortComponentRef(classLoader, ref));
            }
        }

        // fill handlers list
        List hl = sref.getHandlerList();
        org.objectweb.jonas_ejb.deployment.xml.Handler h = null;

        for (int i = 0; i < hl.size(); i++) {
            // for each reference, build and add a HandlerRef Object
            if (hl.get(i) != null) {
                // get the standard dd handler
                h = (org.objectweb.jonas_ejb.deployment.xml.Handler) hl.get(i);

                // build and add a new handler
                hRefs.add(new HandlerRef(classLoader, h));
            }
        }

        // get ServiceRef WSDLFile name
        org.objectweb.jonas_ejb.deployment.xml.WsdlFile wf = sref.getWsdlFile();

        if (wf != null) {
            this.wsdlFileName = wf.getValue().trim();
            setWSDLFile();

            if ((wsdl.getNbServices() > 1) && (serviceQName == null)) {
                String err =
                    "the wsdl : " + this.wsdlFileName
                    + " have more than one service,"
                    + " you must define a QName for service-ref : " + name;
                throw new WSDeploymentDescException(err);
            }
            // When the WSDL defines only 1 service, we can retrieve
            // automatically the service QName for service-ref
            else if ((wsdl.getNbServices() == 1) && (serviceQName == null)) {
                serviceQName = wsdl.getServiceQname();
            }
            // serviceQName not null
            else {
                // chech service QName existence in WSDL
                if (!wsdl.hasService(serviceQName)) {
                    throw new WSDeploymentDescException("service-qname '"
                        + serviceQName + "' must be defined in WSDL : '"
                        + wsdlFileName + "'");
                }
            }
        }

        // get ServiceRef MappingFile name
        org.objectweb.jonas_ejb.deployment.xml.JaxrpcMappingFile mf =
            sref.getJaxrpcMappingFile();

        if (mf != null) {
            this.mappingFileName = mf.getValue().trim();
            setMappingFile();
        }

        // mapping is required when generated service interface is specified
        if (!serviceInterface.getName().equals("javax.xml.rpc.Service")
                && (mapping == null)) {
            throw new WSDeploymentDescException(
                "jaxrpc-mapping-file is required when service-interface "
                + "('" + serviceInterface.getName() + "') is a a Generated one");
        }

        validate();
    }

    /**
     * Creates a new ServiceRef object.
     *
     * @param classLoader web class loader
     * @param sref generated object containing service-ref informations.
     * @param jsref JOnAS specific service-ref informations.
     * @param filename the name of the archive where retrieve WSDL and mapping
     *        files
     *
     * @throws WSDeploymentDescException if the wsdl file is undefined  or if
     *         informations in wsdl and handler doesn't match.
     */
    public ServiceRef(
        ClassLoader                                            classLoader,
        org.objectweb.jonas_web.deployment.xml.ServiceRef      sref,
        org.objectweb.jonas_web.deployment.xml.JonasServiceRef jsref,
        String                                                 filename)
        throws WSDeploymentDescException {
        this.filename = filename;
        moduleFile = new File(filename);

        // set ServiceRef name
        name = sref.getServiceRefName().getValue().trim();

        // set jndiName
        jndiName = jsref.getJndiName().getValue().trim();

        // set Service QName (if exists)
        if (sref.getServiceQname() != null) {
            serviceQName =
                new QName(sref.getServiceQname().getNamespaceURI().getValue()
                              .trim(),
                    sref.getServiceQname().getLocalpart().getValue().trim());
        }

        // set Service Interface
        if (sref.getServiceInterface() != null) {
            String si = sref.getServiceInterface().getValue().trim();

            try {
                serviceInterface = classLoader.loadClass(si);
            } catch (ClassNotFoundException cnf) {
                throw new WSDeploymentDescException("service-interface : " + si
                    + " not found !", cnf);
            }
        }

        // this interface must implements javax.xml.rpc.Service
        if (!javax.xml.rpc.Service.class.isAssignableFrom(serviceInterface)) {
            String err =
                "The service-interface : " + serviceInterface.getName()
                + " must extends the javax.xml.rpc.Service interface.";
            throw new WSDeploymentDescException(err);
        }

        // ServiceRef init parameters : jonas specific deployment desc
        List jipl = jsref.getJonasInitParamList();

        try {
            for (int i = 0; i < jipl.size(); i++) {
                // add in params table each init parameter name and associated value
                org.objectweb.jonas_web.deployment.xml.JonasInitParam p =
                    (org.objectweb.jonas_web.deployment.xml.JonasInitParam) jipl
                    .get(i);

                if (p != null) {
                    params.put(p.getParamName().getValue().trim(),
                        p.getParamValue().getValue().trim());
                }
            }
        } catch (NullPointerException e) {
            throw new WSDeploymentDescException("parameter name missing", e);
        }

        // set Service Class
        String sc = getParam(SERVICE_CLASS_NAME);

        if (sc != null) {
            // the name is defined, the service class is already generated.
            try {
                serviceClass = classLoader.loadClass(sc);
            } catch (ClassNotFoundException cnf) {
                throw new WSDeploymentDescException("jonas-init-param/"
                    + SERVICE_CLASS_NAME + " : " + sc + " not found !", cnf);
            }

            // we test that classes are valids
            if (!serviceInterface.isAssignableFrom(serviceClass)) {
                String err =
                    "The service-class must implement the service-interface."
                    + serviceClass.getName() + " must be assignable from "
                    + serviceInterface.getName();
                throw new WSDeploymentDescException(err);
            }
        }

        // fill portComponentRef list	
        List pcrl = sref.getPortComponentRefList();

        for (int i = 0; i < pcrl.size(); i++) {
            // for each reference, build and add a PortComponentRef Object
            if (pcrl.get(i) != null) {
                org.objectweb.jonas_web.deployment.xml.PortComponentRef ref =
                    (org.objectweb.jonas_web.deployment.xml.PortComponentRef) pcrl
                    .get(i);
                pcRefs.add(new PortComponentRef(classLoader, ref));
            }
        }

        // fill handlers list
        List hl = sref.getHandlerList();
        org.objectweb.jonas_web.deployment.xml.Handler h = null;

        for (int i = 0; i < hl.size(); i++) {
            // for each reference, build and add a HandlerRef Object
            if (hl.get(i) != null) {
                // get the standard dd handler
                h = (org.objectweb.jonas_web.deployment.xml.Handler) hl.get(i);

                // build and add a new handler
                hRefs.add(new HandlerRef(classLoader, h));
            }
        }

        // get ServiceRef WSDLFile name
        org.objectweb.jonas_web.deployment.xml.WsdlFile wf = sref.getWsdlFile();

        if (wf != null) {
            this.wsdlFileName = wf.getValue().trim();
            setWSDLFile();

            if ((wsdl.getNbServices() > 1) && (serviceQName == null)) {
                String err =
                    "the wsdl : " + this.wsdlFileName
                    + " have more than one service,"
                    + " you must define a QName for service-ref : " + name;
                throw new WSDeploymentDescException(err);
            }
            // When the WSDL defines only 1 service, we can retrieve
            // automatically the service QName for service-ref
            else if ((wsdl.getNbServices() == 1) && (serviceQName == null)) {
                serviceQName = wsdl.getServiceQname();
            }
            // serviceQName not null
            else {
                // chech service QName existence in WSDL
                if (!wsdl.hasService(serviceQName)) {
                    throw new WSDeploymentDescException("service-qname '"
                        + serviceQName + "' must be defined in WSDL : '"
                        + wsdlFileName + "'");
                }
            }
        }

        // get ServiceRef MappingFile name
        org.objectweb.jonas_web.deployment.xml.JaxrpcMappingFile mf =
            sref.getJaxrpcMappingFile();

        if (mf != null) {
            this.mappingFileName = mf.getValue().trim();
            setMappingFile();
        }

        // mapping is required when generated service interface is specified
        if (!serviceInterface.getName().equals("javax.xml.rpc.Service")
                && (mapping == null)) {
            throw new WSDeploymentDescException(
                "jaxrpc-mapping-file is required when service-interface "
                + "('" + serviceInterface.getName() + "') is a a Generated one");
        }

        validate();
    }

    /**
     * Return the list of PortComponentRef.
     *
     * @return the list of PortComponentRef
     */
    public List getPortComponentRefs() {
        return pcRefs;
    }

    /**
     * Return the list of Handler.
     *
     * @return the list of Handler
     */
    public List getHandlerRefs() {
        return hRefs;
    }

    /**
     * Return the name used for Service interface lookup.
     *
     * @return the service-ref-name value
     */
    public String getServiceRefName() {
        return name;
    }

    /**
     * Return the Class object representing the service-interface.
     *
     * @return the Class object representing the service-interface.
     */
    public Class getServiceInterface() {
        return serviceInterface;
    }

    /**
     * Return the WSDLFile object describing the WebService.
     *
     * @return the WSDLFile object describing the WebService.
     */
    public WSDLFile getWSDLFile() {
        return wsdl;
    }

    /**
     * Return the MappingFile object.
     *
     * @return the MappingFile object.
     */
    public MappingFile getMappingFile() {
        return mapping;
    }

    /**
     * Return the implementaion class of the Service Interface. Can be null if
     * serviceClassName has not been set (GenIC/WsGen case) in
     * jonas-init-param because the class has not been generated at this time.
     *
     * @return the implementaion class of the Service Interface.
     */
    public Class getServiceClass() {
        return serviceClass;
    }

    /**
     * Return all the params of the ServiceRef as an Hashtable.
     *
     * @return all the params of the ServiceRef as an Hashtable.
     */
    public Hashtable getParams() {
        return params;
    }

    /**
     * Return the value of the specified parameter
     *
     * @param name the parameter to retrieve
     *
     * @return the value of the specified parameter
     */
    public String getParam(String name) {
        return (String) params.get(name);
    }

    /**
     * Return the serviceRef jndi name
     *
     * @return the value of the specified parameter
     */
    public String getJndiName() {
        return jndiName;
    }

    /**
     * Return the name of WSDL inside of the module.
     *
     * @return the name of WSDL inside of the module
     */
    public String getWsdlFileName() {
        return wsdlFileName;
    }

    /**
     * Return the QName identifying the service in the WSDL. can return null if
     * WSDL not defined.
     *
     * @return Return the QName identifying the service in the WSDL (can be
     *         null).
     */
    public QName getServiceQName() {
        return serviceQName;
    }

    /**
     * Return <code>true</code> if the parameter is a ServiceRef and if it
     * equals this object. Return <code>false</code> else.
     *
     * @param other the object to compare
     *
     * @return true if objects are equals
     */
    public boolean equals(Object other) {
        if (other == null) {
            return false;
        }

        if (!(other instanceof ServiceRef)) {
            return false;
        }

        ServiceRef sr = (ServiceRef) other;

        if (!mapping.equals(sr.getMappingFile())) {
            return false;
        }

        if (!wsdl.equals(sr.getWSDLFile())) {
            return false;
        }

        if (!params.equals(sr.getParams())) {
            return false;
        }

        for (Iterator i = sr.getPortComponentRefs().iterator(); i.hasNext();) {
            PortComponentRef pcr = (PortComponentRef) i.next();

            if (!pcRefs.contains(pcr)) {
                return false;
            }
        }

        for (Iterator i = sr.getHandlerRefs().iterator(); i.hasNext();) {
            HandlerRef hr = (HandlerRef) i.next();

            if (!hRefs.contains(hr)) {
                return false;
            }
        }

        return true;
    }

    /**
     * Create the MappingFile.
     *
     * @throws WSDeploymentDescException When MappingFile creation fails.
     */
    private void setMappingFile() throws WSDeploymentDescException {
        InputStream isMapping = null;

        // If we have file stored in Jar
        if (moduleFile.exists()) {
            // If we have file stored in Jar
            if (moduleFile.isFile()) {
                JarFile jf = null;

                // Get the Mapping file of the jar as InputStream
                try {
                    jf = new JarFile(filename);
                    isMapping = jf.getInputStream(jf.getEntry(mappingFileName));
                } catch (Exception e) {
                    if (jf != null) {
                        try {
                            jf.close();
                        } catch (IOException i) {
                            ;
                        }
                    }

                    String svc =
                        (serviceQName == null) ? ""
                                               : (
                            " for Service " + serviceQName
                        );
                    throw new WSDeploymentDescException(
                        "Cannot read the jaxrpc-mapping-file "
                        + mappingFileName + " in " + filename + svc,
                        e);
                }
            }
            // filename is a Directory
            else {
                try {
                    isMapping =
                        new FileInputStream(new File(moduleFile, mappingFileName));
                } catch (IOException ioe) {
                    String svc =
                        (serviceQName == null) ? ""
                                               : (
                            " for Service " + serviceQName
                        );
                    throw new WSDeploymentDescException(
                        "Cannot read the jaxrpc-mapping-file "
                        + mappingFileName + " in " + filename + svc,
                        ioe);
                }
            }

            // build the MappingFile
            mapping = new MappingFile(isMapping, mappingFileName);
        } else { // the moduleFile doesn't exist
            throw new WSDeploymentDescException(filename + " doesn't exists");
        }
    }

    /**
     * Create the WSDLFile object.
     *
     * @throws WSDeploymentDescException When WSDL Creation fails.
     */
    private void setWSDLFile() throws WSDeploymentDescException {
        InputStream isWsdl = null;

        // If we have file stored in Jar
        if (moduleFile.exists()) {
            if (moduleFile.isFile()) {
                JarFile jf = null;

                // Get the WSDL file of the jar as InputStream
                try {
                    jf = new JarFile(moduleFile);
                    isWsdl = jf.getInputStream(jf.getEntry(wsdlFileName));
                } catch (Exception e) {
                    if (jf != null) {
                        try {
                            jf.close();
                        } catch (IOException i) {
                            ;
                        }
                    }

                    throw new WSDeploymentDescException(
                        "Cannot get wsdl file: " + wsdlFileName + " in "
                        + filename);
                }
            }
            // filename is a Directory
            else {
                try {
                    isWsdl =
                        new FileInputStream(new File(moduleFile, wsdlFileName));
                } catch (IOException ioe) {
                    throw new WSDeploymentDescException(
                        "Cannot get wsdl file: " + wsdlFileName + " in "
                        + filename);
                }
            }

            // build the WSDLFile
            wsdl = new WSDLFile(isWsdl, wsdlFileName);
        } else { // the moduleFile doesn't exist
            throw new WSDeploymentDescException(filename + " doesn't exists");
        }
    }

    /**
     * Validate the ServiceRef
     *
     * @throws WSDeploymentDescException DOCUMENT ME!
     */
    private void validate() throws WSDeploymentDescException {
        // validate informations : 
        // todo : pcLinkExist
        // SoapHeader validity between handlers and wsdl
        for (int i = 0; i < hRefs.size(); i++) {
            HandlerRef myHRef = (HandlerRef) hRefs.get(i);
            List shl = myHRef.getSOAPHeaders();

            for (int j = 0; j < shl.size(); j++) {
                if (!wsdl.hasSOAPHeader((QName) shl.get(j))) {
                    // the handler use a soap header undefined in the wsdl
                    throw new WSDeploymentDescException("handler "
                        + myHRef.getName() + " use an undefined soap header : "
                        + (QName) shl.get(j) + " in " + wsdlFileName);
                }
            }
        }

        // Port validity between handlers and wsdl
        for (int i = 0; i < hRefs.size(); i++) {
            HandlerRef myHRef = (HandlerRef) hRefs.get(i);
            List pnl = myHRef.getPortNames();

            for (int j = 0; j < pnl.size(); j++) {
                if (!wsdl.hasPort((String) pnl.get(j))) {
                    // the handler use a port undefined in the wsdl
                    throw new WSDeploymentDescException("handler "
                        + myHRef.getName() + " use an undefined port : "
                        + (String) pnl.get(j) + " in " + wsdlFileName);
                }
            }
        }

        // Check service-interface
        // if service interface is not generic, Full WSDL Knowledge is required
        if (!serviceInterface.equals(javax.xml.rpc.Service.class)) {
            if ((wsdl == null)
                    || (
                        (wsdl != null)
                        && (
                            wsdl.getDefinition().getServices().values().size() == 0
                        )
                    )) {
                throw new WSDeploymentDescException(
                    "generated service-interface '"
                    + serviceInterface.getName() + "' from service-ref '"
                    + name + "' require Full WSDL Knowledge");
            }

            // Check mapping for GeneratedServiceInterface	
            String namespaceURI = serviceQName.getNamespaceURI();
            String realPackage =
                BeanNaming.getPackageName(serviceInterface.getName());

            // real != mapping provided package
            if (!realPackage.equals(mapping.getMapping(namespaceURI))) {
                throw new WSDeploymentDescException("jaxrpc-mapping-file '"
                    + mappingFileName + "' need a mapping for package '"
                    + realPackage + "'");
            }

            /*
               // TODO : Uncomment this part will require substancial modification to number of tests (integ + unit)
               // -> We need to create generated clients interfaces for each endpoint ...
               // keep service QName namespaceURI (targetNamespace)
               // check mapping for ServiceEndpointInterface when having a generated service-interface
               for (Iterator p = pcRefs.iterator() ; p.hasNext() ; ) {
                   PortComponentRef pcr = (PortComponentRef) p.next();
                   String sei = pcr.getSei().getName();
                   realPackage = BeanNaming.getPackageName(sei);
                   if (!realPackage.equals(mapping.getMapping(namespaceURI)))
                       throw new WSDeploymentDescException("jaxrpc-mapping-file '" + mappingFileName +
                                                           "' need a mapping for package '" + realPackage + "'");
               }
             */
        }
    }
}
