/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial Developer : Delplanque Xavier & Sauthier Guillaume
 * --------------------------------------------------------------------------
 * $Id: WSDeploymentDesc.java,v 1.5 2003/09/09 11:57:33 sauthieg Exp $
 * --------------------------------------------------------------------------
*/

package org.objectweb.jonas_ws.deployment.api;

import org.objectweb.jonas_lib.deployment.AbsDeploymentDesc;

import org.objectweb.jonas_ws.deployment.xml.JonasWebservices;
import org.objectweb.jonas_ws.deployment.xml.WebserviceDescription;
import org.objectweb.jonas_ws.deployment.xml.Webservices;

import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

import java.util.List;
import java.util.Vector;


/**
 * This Classes is a data structure that encapsulate informations contained in
 * <code>webservices.xml</code> deployment descriptor. It provides methods to
 * manipulate these informations.
 *
 * @author Guillaume Sauthier
 * @author Xavier Delplanque
 */
public class WSDeploymentDesc extends AbsDeploymentDesc {
    /** The list of ServiceDesc described in webservices.xml */
    private Vector services = new Vector();

    /** The Webservices display name */
    private String displayName;

    /** The WebApp dispatching SOAP request to the PortComponents */
    private String warFile = null;

    /** The Webservices logger */
    private Logger logger;

    /**
     * Constructor : creates a WSDeploymentDesc object
     *
     * @param jarCL module (war or ejbjar) class loader.
     * @param logger the logger to use.
     * @param ws Zeus object containing webservices informations
     * @param jws Zeus object containing jonas-webservices informations
     * @param fileName module (ejbjar or war) jar or directory
     *
     * @throws WSDeploymentDescException if in the Webservices file : - each
     *         service haven't got an unique name. - each port component
     *         haven't got an unique name. - each handler haven't got an
     *         unique name. - wsdl ports aren't inclued in portComponents.
     */
    public WSDeploymentDesc(ClassLoader jarCL, Logger logger, Webservices ws,
        JonasWebservices jws, String fileName) throws WSDeploymentDescException {
        this.logger = logger;

        // set displayName
        displayName = ws.getDisplayName().getValue();

        // Store the WebApp filename
        if ((jws != null) && (jws.getWar() != null)) {
            warFile = jws.getWar().getValue();
        }

        // set services
        List wdl = ws.getWebserviceDescriptionList();

        for (int i = 0; i < wdl.size(); i++) {
            if (wdl.get(i) != null) {
                WebserviceDescription wd = (WebserviceDescription) wdl.get(i);
                services.add(new ServiceDesc(jarCL, wd, fileName));
            }
        }

        // validation 
        // services names are unique
        // fill the list of service names
        Vector serviceNames = new Vector();

        for (int i = 0; i < services.size(); i++) {
            String sn = ((ServiceDesc) services.get(i)).getName();

            if (serviceNames.contains(sn)) {
                // if the service name is ever contained, it isn't unique
                String err = "service '" + sn + "' haven't got a unique name";

                if (logger.isLoggable(BasicLevel.ERROR)) {
                    logger.log(BasicLevel.ERROR, err);
                }

                throw new WSDeploymentDescException(err);
            }

            serviceNames.add(sn);
        }

        // port components names are unique
        // fill the list of port component names
        Vector pcNames = new Vector();

        for (int i = 0; i < services.size(); i++) {
            ServiceDesc s = (ServiceDesc) services.get(i);
            List pcl = s.getPortComponents();

            for (int j = 0; j < pcl.size(); j++) {
                PortComponentDesc pcd = (PortComponentDesc) pcl.get(j);
                String pcn = pcd.getName();

                if (pcNames.contains(pcn)) {
                    // if the port component name is already contained, it isn't unique
                    String err = "port-component '" + pcn +
                        "' haven't got a unique name";

                    if (logger.isLoggable(BasicLevel.ERROR)) {
                        logger.log(BasicLevel.ERROR, err);
                    }

                    throw new WSDeploymentDescException(err);
                }

                pcNames.add(pcn);
            }
        }

        // handlers names are unique
        // fill the list of handler names
        Vector handlerNames = new Vector();

        for (int i = 0; i < services.size(); i++) {
            List pcl = ((ServiceDesc) services.get(i)).getPortComponents();

            for (int j = 0; j < pcl.size(); j++) {
                List pchl = ((PortComponentDesc) pcl.get(j)).getHandlers();

                for (int k = 0; k < pchl.size(); k++) {
                    String hn = ((HandlerDesc) pchl.get(k)).getName();

                    if (handlerNames.contains(hn)) {
                        // if the handler name is already contained, it isn't unique
                        String err = "handler '" + hn +
                            "' haven't got a unique name";

                        if (logger.isLoggable(BasicLevel.ERROR)) {
                            logger.log(BasicLevel.ERROR, err);
                        }

                        throw new WSDeploymentDescException(err);
                    }

                    handlerNames.add(hn);
                }
            }
        }

        // verify that all wsdl ports are included inside portComponents
        // build the list of port component QNames
        Vector pcQNames = new Vector();

        for (int i = 0; i < services.size(); i++) {
            List pcl = ((ServiceDesc) services.get(i)).getPortComponents();

            for (int j = 0; j < pcl.size(); j++) {
                PortComponentDesc pc = (PortComponentDesc) pcl.get(j);
                pcQNames.add(pc.getQName());
            }
        }

        // verify for each WSDLs that port are inside pcQNames
        for (int i = 0; i < services.size(); i++) {
            WSDLFile wsdlf = ((ServiceDesc) services.get(i)).getWSDL();

            if (!wsdlf.hasPortsIncludedIn(pcQNames)) {
                String err = "WSDL '" + wsdlf.getName() +
                    "' uses a port undefined as portComponent";

                if (logger.isLoggable(BasicLevel.ERROR)) {
                    logger.log(BasicLevel.ERROR, err);
                }

                throw new WSDeploymentDescException(err);
            }
        }
    }

    /**
     * Return the list of ServiceDesc.
     *
     * @return the list of ServiceDesc.
     */
    public List getServiceDescs() {
        return services;
    }

    /**
     * Return the Webservices displayName.
     *
     * @return the Webservices displayName.
     */
    public String getDisplayName() {
        return displayName;
    }

    /**
     * Return a String representation of the WSDeploymentDesc.
     *
     * @return a String representation of the WSDeploymentDesc.
     */
    public String toString() {
        return "WSDeploymentDesc";
    }

    /**
     * Return the filename of the WebApp dispatching SOAP requests to the
     * components (can be null if no filename specified  in
     * jonas-webservices.xml).
     *
     * @return the filename of the WebApp dispatching SOAP requests to the
     *         components.
     */
    public String getWarFile() {
        return warFile;
    }
}
