/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial Developer : Delplanque Xavier & Sauthier Guillaume
 * --------------------------------------------------------------------------
 * $Id: WSDeploymentDescManager.java,v 1.5 2003/09/09 11:57:33 sauthieg Exp $
 * --------------------------------------------------------------------------
*/

package org.objectweb.jonas_ws.deployment.lib;


// import monolog
import org.objectweb.jonas.common.Log;

// import ejb
import org.objectweb.jonas_ejb.deployment.api.BeanDesc;
import org.objectweb.jonas_ejb.deployment.api.DeploymentDesc;
import org.objectweb.jonas_ejb.deployment.api.DeploymentDescException;
import org.objectweb.jonas_ejb.deployment.api.SessionStatelessDesc;
import org.objectweb.jonas_ejb.deployment.lib.EjbDeploymentDescManager;
import org.objectweb.jonas_ejb.lib.BeanNaming;

import org.objectweb.jonas_web.deployment.api.WebContainerDeploymentDesc;
import org.objectweb.jonas_web.deployment.api.WebContainerDeploymentDescException;

// import web
import org.objectweb.jonas_web.deployment.lib.WebDeploymentDescManager;

import org.objectweb.jonas_ws.deployment.api.PortComponentDesc;
import org.objectweb.jonas_ws.deployment.api.SAXParserHelper;
import org.objectweb.jonas_ws.deployment.api.ServiceDesc;

// import ws
import org.objectweb.jonas_ws.deployment.api.WSDeploymentDesc;
import org.objectweb.jonas_ws.deployment.api.WSDeploymentDescException;
import org.objectweb.jonas_ws.deployment.xml.JonasWebservices;
import org.objectweb.jonas_ws.deployment.xml.JonasWebservicesUnmarshaller;

// import xml
import org.objectweb.jonas_ws.deployment.xml.Webservices;
import org.objectweb.jonas_ws.deployment.xml.WebservicesUnmarshaller;

import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

// import sax
import org.xml.sax.EntityResolver;
import org.xml.sax.ErrorHandler;
import org.xml.sax.SAXParseException;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;

// import io
import java.io.InputStreamReader;
import java.io.Reader;

import java.net.MalformedURLException;
import java.net.URL;

import java.util.HashMap;

// import util
import java.util.List;
import java.util.jar.JarFile;
import java.util.zip.ZipEntry;


/**
 * This class provide a way for managing the WSDeploymentDesc. Note that there
 * is 1 instance of the WSDeploymentDescManager on each JOnAS server.
 *
 * @author Guillaume Sauthier
 * @author Xavier Delplanque
 */
public class WSDeploymentDescManager {
    /** META-INF/webservices.xml */
    public static final String WS_EJB_JAR_FILE_NAME =
        "META-INF" + File.separator + "webservices.xml";

    /** WEB-INF/webservices.xml */
    public static final String WS_WAR_FILE_NAME =
        "WEB-INF" + File.separator + "webservices.xml";

    /** META-INF/jonas-webservices.xml */
    public static final String JONAS_WS_EJB_JAR =
        "META-INF" + File.separator + "jonas-webservices.xml";

    /** WEB-INF/jonas-webservices.xml */
    public static final String JONAS_WS_WAR =
        "WEB-INF" + File.separator + "jonas-webservices.xml";

    /** The error handler to use (null = default) */
    private static ErrorHandler errorHandler = null;

    /** The entity resolver to use (null = default) */
    private static EntityResolver entityResolver = null;

    /** Flag for parser validation */
    private static boolean parsingWithValidation = true;

    /** The unique instance of the WSDeploymentDescManager */
    private static WSDeploymentDescManager unique = null;

    /** is the manager instanciated ? */
    private static boolean isInstanciated = false;

    /** Logger for the deployment desc manager. */
    private static Logger logger = Log.getLogger(Log.JONAS_WS_PREFIX);

    /** Modes for DD opening : XML case */
    private static final int XML_MODE = 0;

    /** DModes for DD opening : Jar case */
    private static final int JAR_MODE = 1;

    /** DModes for DD opening : Directory case */
    private static final int DIR_MODE = 2;

    /** Associates module or webservices dd URL  and a WSDeploymentDesc */
    private HashMap urlWsddBindings;

    /** Reference on the EjbDeploymentDescManager. */
    private EjbDeploymentDescManager ejbDDManager = null;

    /** Reference on the WebDeploymentDescManager. */
    private WebDeploymentDescManager wcDDManager = null;

    /**
     * Constructor : creates a WSDeploymentDescManager object. Used only when
     * loading files in the running JOnAS instance.
     */
    private WSDeploymentDescManager() {
        urlWsddBindings = new HashMap();

        ejbDDManager = EjbDeploymentDescManager.getInstance();
        wcDDManager = WebDeploymentDescManager.getInstance();

    }

    /**
     * Get the unique instance of the WSDeploymentDescManager.
     *
     * @return the instance of the WSDeploymentDescManager.
     */
    public static WSDeploymentDescManager getInstance() {
        if (!isInstanciated) {
            isInstanciated = true;
            unique = new WSDeploymentDescManager();

        }

        return unique;

    }

    /**
     * Get the specified WebService deployment descriptor. Used by WsGen.
     *
     * @param file module (ejbjar or war). It is required that file is not a
     *        directory.
     * @param jarCL the classloader where classes are stored.
     *
     * @return the module webservices deployment descriptor if it exists, null
     *         else.
     *
     * @throws WSDeploymentDescException when WSDeploymentDesc cannot be
     *         created with the given files.
     */
    public static WSDeploymentDesc getDeploymentDesc(
        String      file,
        ClassLoader jarCL) throws WSDeploymentDescException {
        WSDeploymentDesc wsdd = loadDeploymentDesc(file, jarCL);

        // webservices.xml not present in file, no resolution needed
        if (wsdd == null) {
            return null;

        }

        // Resolve ejb-link and servlet-link elements
        List sdl = wsdd.getServiceDescs();

        for (int i = 0; i < sdl.size(); i++) {
            List pcdl = ((ServiceDesc) sdl.get(i)).getPortComponents();

            for (int j = 0; j < pcdl.size(); j++) {
                // Resolve the ejb-link and servlet-link of each port components
                PortComponentDesc pcd = (PortComponentDesc) pcdl.get(j);
                String sibLink = pcd.getSibLink();

                // a link is defined, we resolve it
                if (pcd.hasBeanImpl()) {
                    SessionStatelessDesc desc =
                        getBeanDesc(file, sibLink, jarCL);
                    pcd.setBeanDesc(desc);

                } else if (pcd.hasJaxRpcImpl()) {
                    WebContainerDeploymentDesc desc =
                        getWebDesc(file, sibLink, jarCL);
                    pcd.setWebDesc(desc);

                }
            }
        }

        return wsdd;

    }

    /**
     * Get the specified ws deployment descriptor and put it in the cache if it
     * is not in.
     *
     * @param url module (ejbjar or war). It can be a directory or a jar file.
     * @param jarCL classloader used to load bean classes.
     * @param earCL the parent classloader (the ear classloader). Null when not
     *        in the case of an ear application.
     *
     * @return the module webservices deployment descriptor if it exists, null
     *         else.
     *
     * @throws WSDeploymentDescException when WSDeploymentDesc cannot be
     *         created with the given files.
     */
    public WSDeploymentDesc getDeploymentDesc(
        URL         url,
        ClassLoader jarCL,
        ClassLoader earCL) throws WSDeploymentDescException {
        WSDeploymentDesc wsdd = null;
        File moduleFile = new File(url.getFile());

        if (moduleFile.exists()) {
            // if the module has yet been loaded
            if (!urlWsddBindings.containsKey(url)) {
                wsdd = getDeploymentDescriptor(url, jarCL, earCL);

            }
            // module already loaded
            else {
                // Can be null if no webservices.xml has been found in specified module
                wsdd = (WSDeploymentDesc) urlWsddBindings.get(url);

            }

            return wsdd;

        } else {
            throw new WSDeploymentDescException("'" + moduleFile.getName()
                + "' doesn't exists");

        }
    }

    /**
     * return the specified webservices deployment descriptor and put it in
     * cache.
     *
     * @param url an existing module (ejbjar or war). It can be a directory or
     *        a jar file
     * @param jarCL classloader used to load war or ejbjar classes.
     * @param earCL the parent classloader (the ear classloader). Null when not
     *        in the case of an ear application.
     *
     * @return the module webservices deployment descriptor if it exists, null
     *         else.
     *
     * @throws WSDeploymentDescException when WSDeploymentDesc cannot be
     *         created with the given files.
     */
    private WSDeploymentDesc getDeploymentDescriptor(
        URL         url,
        ClassLoader jarCL,
        ClassLoader earCL) throws WSDeploymentDescException {
        WSDeploymentDesc wsdd = loadDeploymentDesc(url.getFile(), jarCL);

        // add in cache the deployment descriptor and return it
        try {
            urlWsddBindings.put(url, wsdd);

        } catch (NullPointerException npe) {
            throw new WSDeploymentDescException(npe);

        }

        // No WSDeploymentDesc, return null
        if (wsdd == null) {
            return null;

        }

        // Resolve ejb-link and servlet-link elements
        List sdl = wsdd.getServiceDescs();

        for (int i = 0; i < sdl.size(); i++) {
            List pcdl = ((ServiceDesc) sdl.get(i)).getPortComponents();

            for (int j = 0; j < pcdl.size(); j++) {
                // Resolve the ejb-link and servlet-link of each port components
                PortComponentDesc pcd = (PortComponentDesc) pcdl.get(j);
                String sibLink = pcd.getSibLink();

                // a link is defined, we resolve it
                if (pcd.hasBeanImpl()) {
                    SessionStatelessDesc desc =
                        getBeanDesc(url.getFile(), sibLink, jarCL, earCL);
                    pcd.setBeanDesc(desc);

                } else if (pcd.hasJaxRpcImpl()) {
                    WebContainerDeploymentDesc desc =
                        getWebDesc(url.getFile(), sibLink, jarCL, earCL);
                    pcd.setWebDesc(desc);

                }
            }
        }

        return wsdd;

    }

    /**
     * Load webservices.xml in the WSDeploymentDesc structure (null if
     * webservices.xml not found). Can take as parameter a directory (like an
     * unpacked jar root directory), a XML file name (path to
     * (META-INF|WEB-INF)/webservices.xml) or a packed jar.
     *
     * @param file the filename to load (can be a directory, xml, jar).
     * @param cl the classLoader to use for class loading.
     *
     * @return WSDeploymentDesc object.
     *
     * @throws WSDeploymentDescException when parsing error.
     */
    private static WSDeploymentDesc loadDeploymentDesc(
        String      file,
        ClassLoader cl) throws WSDeploymentDescException {
        File module = new File(file);
        String moduleRoot = "";
        int fileMode = JAR_MODE;

        // get the Webservices object for this module
        String wsFileName = "";
        InputStream isWebservices = null;

        if (module.exists()) {
            // If we have file
            if (module.isFile()) {
                // Just a webservices.xml
                if (module.getPath().endsWith(".xml")) {
                    if (logger.isLoggable(BasicLevel.DEBUG)) {
                        logger.log(BasicLevel.DEBUG,
                            "Loading from an XML file : '" + module + "'");

                    }

                    fileMode = XML_MODE;
                    moduleRoot =
                        module.getParentFile().getParentFile().getAbsolutePath();
                    wsFileName = moduleRoot;

                    try {
                        isWebservices = new FileInputStream(module);

                    } catch (IOException ignored) {
                        ;

                    }
                }
                // Just a jar file
                else {
                    JarFile jf = null;
                    moduleRoot = module.getAbsolutePath();

                    fileMode = JAR_MODE;

                    // Get the Webservices file of the jar as InputStream
                    try {
                        if (logger.isLoggable(BasicLevel.DEBUG)) {
                            logger.log(BasicLevel.DEBUG,
                                "Loading from a Jar File : '" + module + "'");

                        }

                        jf = new JarFile(module);

                        ZipEntry wsEntry = jf.getEntry(WS_EJB_JAR_FILE_NAME);

                        // EJB-JAR
                        if (wsEntry != null) {
                            // the module is an ejbjar
                            wsFileName = WS_EJB_JAR_FILE_NAME;
                            isWebservices = jf.getInputStream(wsEntry);

                        } else {
                            wsEntry = jf.getEntry(WS_WAR_FILE_NAME);

                            // WEBAPP
                            if (wsEntry != null) {
                                // the module is a war
                                wsFileName = WS_WAR_FILE_NAME;
                                isWebservices = jf.getInputStream(wsEntry);

                            }
                        }
                    } catch (Exception e) {
                        if (jf != null) {
                            try {
                                jf.close();

                            } catch (IOException i) {
                                ;

                            }
                        }

                        String err =
                            "Cannot read webservices '" + wsFileName + "' in '"
                            + file + "'";

                        if (logger.isLoggable(BasicLevel.ERROR)) {
                            logger.log(BasicLevel.ERROR, err);

                        }

                        throw new WSDeploymentDescException(err, e);

                    }
                }
            }
            // filename is a Directory
            else {
                if (logger.isLoggable(BasicLevel.DEBUG)) {
                    logger.log(BasicLevel.DEBUG,
                        "Loading from a Directory : '" + module + "'");

                }

                fileMode = DIR_MODE;

                try {
                    File wsFile = new File(module, WS_EJB_JAR_FILE_NAME);
                    moduleRoot = module.getAbsolutePath();

                    if (wsFile.exists()) {
                        // the module is an exploded ejbjar
                        wsFileName = WS_EJB_JAR_FILE_NAME;
                        isWebservices = new FileInputStream(wsFile);

                    } else {
                        wsFile = new File(module, WS_WAR_FILE_NAME);

                        if (wsFile.exists()) {
                            // the module is an exploded war
                            wsFileName = WS_WAR_FILE_NAME;
                            isWebservices = new FileInputStream(wsFile);

                        }
                        // webservices.xml not found in the filename
                        else {
                            return null;

                        }
                    }
                } catch (IOException ioe) {
                    throw new WSDeploymentDescException(
                        "Cannot read webservices.xml file '" + wsFileName
                        + "' in '" + file + "'",
                        ioe);

                }
            }
        }

        // webservices.xml not found in module, return null
        if (isWebservices == null) {
            return null;

        }

        // get the Webservices object for this module
        String jwsFileName = "";
        InputStream isJonasWebservices = null;

        // try to load jonas-webservices.xml
        switch (fileMode) {
        // Try to load (WEB-INF|META-INF)/jonas-webservices.xml
        case DIR_MODE:

            try {
                File jwsFile = new File(module, JONAS_WS_EJB_JAR);

                if (jwsFile.exists()) {
                    // the module is an exploded ejbjar
                    jwsFileName = JONAS_WS_EJB_JAR;
                    isJonasWebservices = new FileInputStream(jwsFile);

                } else {
                    jwsFile = new File(module, JONAS_WS_WAR);

                    if (jwsFile.exists()) {
                        // the module is an exploded war
                        jwsFileName = JONAS_WS_WAR;
                        isJonasWebservices = new FileInputStream(jwsFile);

                    } else {
                        // jonas-webservices.xml not found in the filename
                        // -> must be in webapp case
                        // -> or ejbjar with default naming convention 
                        //    (<ejbjar-name>.jar -> <ejbjar-name>.war)
                    }
                }
            } catch (IOException ioe) {
                throw new WSDeploymentDescException(
                    "Cannot read jonas-webservices.xml file '" + jwsFileName
                    + "' in '" + file + "'",
                    ioe);

            }

            break;

        // Try to load jonas-webservices.xml colocated with the webservices.xml
        case XML_MODE:
            jwsFileName = BeanNaming.getJonasXmlName(module.getPath());

            try {
                if (new File(jwsFileName).exists()) {
                    isJonasWebservices = new FileInputStream(jwsFileName);

                }
            } catch (IOException ignored) {
                ;

            }

            break;

        // Try to load jonas-webservices.xml as JarInputStream
        case JAR_MODE:

            JarFile jf = null;

            // Get the JonasWebservices file of the jar as InputStream
            try {
                if (logger.isLoggable(BasicLevel.DEBUG)) {
                    logger.log(BasicLevel.DEBUG,
                        "Loading from a Jar File : '" + module + "'");

                }

                jf = new JarFile(module);

                ZipEntry jwsEntry = jf.getEntry(JONAS_WS_EJB_JAR);

                // EJB-JAR
                if (jwsEntry != null) {
                    // the module is an ejbjar
                    jwsFileName = JONAS_WS_EJB_JAR;
                    isJonasWebservices = jf.getInputStream(jwsEntry);

                } else {
                    jwsEntry = jf.getEntry(JONAS_WS_WAR);

                    // WEBAPP
                    if (jwsEntry != null) {
                        // the module is a war
                        jwsFileName = JONAS_WS_WAR;
                        isJonasWebservices = jf.getInputStream(jwsEntry);

                    }
                }
            } catch (Exception e) {
                if (jf != null) {
                    try {
                        jf.close();

                    } catch (IOException i) {
                        ;

                    }
                }

                String err =
                    "Cannot read webservices '" + jwsFileName + "' in '" + file
                    + "'";

                if (logger.isLoggable(BasicLevel.ERROR)) {
                    logger.log(BasicLevel.ERROR, err);

                }

                // neber thrown ?
                throw new WSDeploymentDescException(err, e);

            }

            break;
        }

        // Load and Parse files
        Webservices webservices = loadWebservices(isWebservices, wsFileName);
        JonasWebservices jonasWebservices = null;

        if (isJonasWebservices != null) {
            jonasWebservices =
                loadJonasWebservices(isJonasWebservices, jwsFileName);

        }

        // Create DeploymentDesc
        WSDeploymentDesc wsdd =
            new WSDeploymentDesc(cl, logger, webservices, jonasWebservices,
                moduleRoot);

        return wsdd;

    }

    /**
     * return a Zeus object containing webservices.xml informations
     *
     * @param is webservices.xml file input stream
     * @param name webservices.xml file name
     *
     * @return a Zeus object containing webservices.xml informations
     *
     * @throws WSDeploymentDescException when Webservices cannot be created
     *         with the given files.
     */
    private static Webservices loadWebservices(
        InputStream is,
        String      name) throws WSDeploymentDescException {
        Reader reader = new InputStreamReader(is);
        SAXParserHelper helper =
            new SAXParserHelper(errorHandler, entityResolver);

        try {
            WebservicesUnmarshaller.setErrorHandler(helper
                .getWrappedErrorHandler());
            WebservicesUnmarshaller.setEntityResolver(helper
                .getWrappedEntityResolver());

            // Unmarshal the XML, and validate as it is unmarshalled
            Webservices ws =
                WebservicesUnmarshaller.unmarshal(reader, parsingWithValidation);

            return ws;

        } catch (IOException ioe) {
            if (helper.getSAXParseException() != null) {
                String detailMsg =
                    getSAXMsg(name, helper.getSAXParseException(), "Invalid XML");
                throw new WSDeploymentDescException(detailMsg, ioe);

            } else {
                throw new WSDeploymentDescException(
                    "Error reading XML document " + name,
                    ioe);

            }
        } finally {
            if (reader != null) {
                try {
                    reader.close();

                } catch (IOException ignored) {
                    ;

                }
            }
        }
    }

    /**
     * return a Zeus object containing jonas-webservices.xml informations
     *
     * @param is jonas-webservices.xml file input stream
     * @param name jonas-webservices.xml file name
     *
     * @return a Zeus object containing jonas-webservices.xml informations
     *
     * @throws WSDeploymentDescException when JonasWebservices cannot be
     *         created with the given files.
     */
    private static JonasWebservices loadJonasWebservices(
        InputStream is,
        String      name) throws WSDeploymentDescException {
        Reader reader = new InputStreamReader(is);
        SAXParserHelper helper =
            new SAXParserHelper(errorHandler, entityResolver);

        try {
            JonasWebservicesUnmarshaller.setErrorHandler(helper
                .getWrappedErrorHandler());
            JonasWebservicesUnmarshaller.setEntityResolver(helper
                .getWrappedEntityResolver());

            // Unmarshal the XML, and validate as it is unmarshalled
            JonasWebservices jws =
                JonasWebservicesUnmarshaller.unmarshal(reader,
                    parsingWithValidation);

            return jws;

        } catch (IOException ioe) {
            if (helper.getSAXParseException() != null) {
                String detailMsg =
                    getSAXMsg(name, helper.getSAXParseException(), "Invalid XML");
                throw new WSDeploymentDescException(detailMsg, ioe);

            } else {
                throw new WSDeploymentDescException(
                    "Error reading XML document " + name,
                    ioe);

            }
        } finally {
            if (reader != null) {
                try {
                    reader.close();

                } catch (IOException ignored) {
                    ;

                }
            }
        }
    }

    /**
     * build a message from SAX Exception in a consistent style, consistent
     * with emacs compile mode (same as grep, cc, javac, etc).
     *
     * @param fileName DOCUMENT ME!
     * @param exception DOCUMENT ME!
     * @param msg DOCUMENT ME!
     *
     * @return DOCUMENT ME!
     */
    private static String getSAXMsg(
        String            fileName,
        SAXParseException exception,
        String            msg) {
        return fileName + ":" + exception.getLineNumber() + ":"
        + exception.getColumnNumber() + ": " + msg;

    }

    /**
     * Get the size of the cache (number of entries in the cache). Used only
     * for debugging.
     *
     * @return the size of the cache (number of entries in the cache).
     */
    public int getCacheSize() {
        return urlWsddBindings.size();

    }

    /**
     * Clear the cache.
     */
    public void clearCache() {
        urlWsddBindings = new HashMap();

    }

    /**
     * Return a string representation of the cache. (Used only for debugging).
     *
     * @return a string representation of the cache.
     */
    public String toString() {
        return "" + getCacheSize();

    }

    /**
     * DOCUMENT ME!
     *
     * @param v DOCUMENT ME!
     */
    public static void setParsingWithValidation(boolean v) {
        parsingWithValidation = v;

    }

    /**
     * DOCUMENT ME!
     *
     * @return DOCUMENT ME!
     */
    public static boolean getParsingWithValidation() {
        return (parsingWithValidation);

    }

    /**
     * Return a StatelessSessionDesc found in filename with the name link. Used
     * by WsGen only. (No JOnAS running).
     *
     * @param filename the ejb-jar file to load.
     * @param link the ejb-name to retrieve.
     * @param cl the classLoader to use.
     *
     * @return he StatelessSessionDesc associated with this port component
     *
     * @throws WSDeploymentDescException if ejb-link not found or something
     *         went wrong when parsing ejb deployment desc.
     */
    private static SessionStatelessDesc getBeanDesc(
        String      filename,
        String      link,
        ClassLoader cl) throws WSDeploymentDescException {
        BeanDesc bd = null;

        try {
            DeploymentDesc dd =
                EjbDeploymentDescManager.getDeploymentDesc(filename, cl);
            bd = dd.getBeanDesc(link);

        } catch (DeploymentDescException dde) {
            throw new WSDeploymentDescException("Unable to find the ejb-link '"
                + link + "' in '" + filename + "'", dde);

        }

        if (!(bd instanceof SessionStatelessDesc)) {
            throw new WSDeploymentDescException("ejb-link '" + link
                + "' must be a Stateless Session Bean");

        }

        return (SessionStatelessDesc) bd;

    }

    /**
     * Return a StatelessSessionDesc found in filename with the name link. Used
     * when JOnAS is running.
     *
     * @param filename the ejb-jar file to load.
     * @param link the ejb-name to retrieve.
     * @param jarCL the loader for classes.
     * @param earCL the EAR classLoader .
     *
     * @return the StatelessSessionDesc associated with this port component
     *
     * @throws WSDeploymentDescException if ejb-link not found or something
     *         went wrong when parsing ejb deployment desc.
     */
    private SessionStatelessDesc getBeanDesc(
        String      filename,
        String      link,
        ClassLoader jarCL,
        ClassLoader earCL) throws WSDeploymentDescException {
        URL url;

        try {
            url = new File(filename).toURL();

        } catch (MalformedURLException mue) {
            throw new WSDeploymentDescException("Url Error with '" + filename
                + "'", mue);

        }

        BeanDesc bd = null;

        try {
            DeploymentDesc dd =
                ejbDDManager.getDeploymentDesc(url, jarCL, earCL);
            bd = dd.getBeanDesc(link);

        } catch (DeploymentDescException dde) {
            throw new WSDeploymentDescException("Unable to load EjbJar '" + url
                + "'", dde);

        }

        // Ejb Link not found
        if (bd == null) {
            throw new WSDeploymentDescException("Unable to find the ejb-link '"
                + link + "' in '" + filename + "'");

        }

        if (!(bd instanceof SessionStatelessDesc)) {
            throw new WSDeploymentDescException("ejb-link '" + link
                + "' must be a Stateless Session Bean");

        }

        return (SessionStatelessDesc) bd;

    }

    /**
     * Return the Web Deployment Desc found in filename with the name link.
     *
     * @param filename the war file to load.
     * @param link the servlet-name to retrieve.
     * @param cl the classLoader to use.
     *
     * @return the Web Deployment Desc associated with this port component
     *
     * @throws WSDeploymentDescException if servlet-link not found or something
     *         went wrong when parsing web deployment desc.
     */
    private static WebContainerDeploymentDesc getWebDesc(
        String      filename,
        String      link,
        ClassLoader cl) throws WSDeploymentDescException {
        WebContainerDeploymentDesc wd = null;

        try {
            wd = WebDeploymentDescManager.getDeploymentDesc(filename, cl);

        } catch (WebContainerDeploymentDescException wcdde) {
            throw new WSDeploymentDescException("Error while reading/parsing "
                + filename, wcdde);

        }

        // link not found in this deployment desc
        if (wd.getServletClass(link) == null) {
            throw new WSDeploymentDescException(
                "Unable to find the servlet-link '" + link + "' in '"
                + filename + "'");

        }

        return wd;

    }

    /**
     * Return the Web Deployment Desc found in filename with the name link.
     * Used when JOnAS is up and running.
     *
     * @param filename the war file to load.
     * @param link the servlet-name to retrieve.
     * @param jarCL the classLoader to us.
     * @param earCL the EAR classLoader.
     *
     * @return the Web Deployment Desc associated with this port component
     *
     * @throws WSDeploymentDescException if servlet-link not found or something
     *         went wrong when parsing web deployment desc.
     */
    private WebContainerDeploymentDesc getWebDesc(
        String      filename,
        String      link,
        ClassLoader jarCL,
        ClassLoader earCL) throws WSDeploymentDescException {
        URL url;

        try {
            url = new File(filename).toURL();

        } catch (MalformedURLException mue) {
            throw new WSDeploymentDescException("Url Error with '" + filename
                + "'", mue);

        }

        WebContainerDeploymentDesc wd = null;

        try {
            wd = wcDDManager.getDeploymentDesc(url, jarCL, earCL);

        } catch (WebContainerDeploymentDescException wcdde) {
            throw new WSDeploymentDescException("Error while reading/parsing "
                + filename, wcdde);

        }

        // link not found in this deployment desc
        if (wd.getServletClass(link) == null) {
            throw new WSDeploymentDescException(
                "Unable to find the servlet-link '" + link + "' in '"
                + filename + "'");

        }

        return wd;

    }
}
