/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Xavier Delplanque & Guillaume Sauthier
 * --------------------------------------------------------------------------
 * $Id: AbsConfigGenerator.java,v 1.9 2003/10/17 15:55:59 sauthieg Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ws.wsgen;

import java.io.File;
import java.io.FileFilter;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.Vector;

import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

import org.objectweb.jonas.common.Log;

import org.objectweb.jonas_ejb.deployment.api.BeanDesc;

import org.objectweb.jonas_ws.deployment.api.PortComponentDesc;
import org.objectweb.jonas_ws.deployment.api.ServiceDesc;
import org.objectweb.jonas_ws.deployment.api.ServiceRef;
import org.objectweb.jonas_ws.deployment.api.WSDeploymentDesc;
import org.objectweb.jonas_ws.wsgen.jarfile.EjbJarManager;
import org.objectweb.jonas_ws.wsgen.jarfile.WarManager;

import org.objectweb.common.Cmd;


/**
 * This class realizes configuration files generation.
 * This class implements genEndpointConfig & genClientConfig methods;
 * configuration methods stay abstract (WS implementation specific).<br>
 * A class is created for each WS Engine for example, a new class
 * called GlueConfigGenerator for Glue integration.
 *
 * @author Xavier Delplanque & Guillaume Sauthier
 */
public abstract class AbsConfigGenerator implements ConfigGenerator {

    /** WsGen logger */
    protected static Logger logger = Log.getLogger(Log.JONAS_WSGEN_PREFIX);

    /** GenerationContext */
    protected GenerationContext gCtx;
 
    /**
     * Constructor : creates a AbsConfigGenerator object.
     * @param ctx GenerationContext storing properties
     */
    public AbsConfigGenerator(GenerationContext ctx) {

        logger.log(BasicLevel.DEBUG, "Create AbsConfigGenerator with param : " + ctx);

        gCtx = ctx;

        // get JAVA_HOME environment property.
        // used for compilation and file generation.
        if (gCtx.getJavaHomeBin() == null) {
            String javaHomeBin = System.getProperty("java.home", "");
            if (!("".equals(javaHomeBin))) {
                javaHomeBin = javaHomeBin + File.separator + ".."
                    + File.separator + "bin" + File.separator;
            }
            gCtx.setProperty(GenerationContext.JAVA_HOME, javaHomeBin);
        }

        // transform current class loader in classpath.
        gCtx.setProperty(GenerationContext.CLASSPATH, getContextClassPath());

    }

    
    /**
     * Generate endpoint configuration files for the set of ServiceDesc
     * included in WSDeploymentDesc contained in a WebApp. Common method 
     * for all Web Services implementation (Glue/Axis).
     *
     * @param wm the managed war file.
     * @param dd The DeploymentDesc to process (image of a webservices.xml)
     *
     * @throws WsGenException if generation fails
     */
    public void genWebAppEndpointConfig(WarManager wm,
                                        WSDeploymentDesc dd)
        throws WsGenException {

        logger.log(BasicLevel.DEBUG,
                   "Generating WebApp Endpoint Config in WarManager '"
                   + wm.getFilename() + "'");

        // Create an ID for process
        String id = dd.getDisplayName();

        File gen = null;
        File dest = null;
        // create a temp directory for file generation
        try {
            gen = WsGen.createTempDir();
            dest = WsGen.createTempDir();
        } catch (java.io.IOException ie) {
            throw new WsGenException("Cannot create the temporary directory for "
                                     + "webapp endpoint config file generation", ie);
        }

        logger.log(BasicLevel.DEBUG, "GenDir for '" + id + "' is '" + gen + "'");
        logger.log(BasicLevel.DEBUG, "DestDir for '" + id + "' is '" + dest + "'");

        // Prepare the Context
        gCtx.setProperty(id + "-abs.webapp.endpoint.gen.dir", gen);
        gCtx.setProperty("processing.key", id);
        gCtx.setProperty("processing.war-manager", wm);

        // generate files in the temp directory
        // for each ServiceDesc
        for (Iterator i = dd.getServiceDescs().iterator(); i.hasNext();) {
            ServiceDesc sd = (ServiceDesc) i.next();

            // for each Port Component Desc
            for (Iterator p = sd.getPortComponents().iterator(); p.hasNext();) {
                PortComponentDesc pcd = (PortComponentDesc) p.next();
                gCtx.setProperty("processing.port-component-desc", pcd);
                configureWebAppEndpoint();
            }
        }

        // Unset Context
        gCtx.removeProperty("processing.key");
        gCtx.removeProperty("processing.war-manager");
        gCtx.removeProperty("processing.port-component-desc");
        
        // compile sources
        gCtx.setProperty(id + "-abs.webapp.endpoint.out.gen.dir", dest);
        compileDir(gen, dest);

        // add generated classes in War
        wm.addDir("WEB-INF/classes", dest, ".class");

        // add sources ?
        if (isKeepGenerated()) {
            wm.addDir("WEB-INF/src", gen, ".java");
        }
        
    }

    /**
     * Generate endpoint configuration files for the set of ServiceDesc
     * included in WSDeploymentDesc contained in an EjbJar. Common method 
     * for all Web Services implementation (Glue/Axis).
     *
     * @param wm the war file dispatching request on EJBs endpoint.
     * @param dd The DeploymentDesc to process (image of a webservices.xml)
     *
     * @throws WsGenException if generation fails
     */
    public void genEjbJarEndpointConfig(WarManager wm,
                                        WSDeploymentDesc dd)
        throws WsGenException {

        logger.log(BasicLevel.DEBUG,
                   "Generating EjbJar Endpoint Config for WarManager '" + wm.getFilename() + "'");

        // Create an ID for process
        String id = dd.getDisplayName();

        File gen = null;
        File dest = null;
        // create a temp directory for file generation
        try {
            gen = WsGen.createTempDir();
            dest = WsGen.createTempDir();
        } catch (java.io.IOException ie) {
            throw new WsGenException("Cannot create the temporary directory for "
                                     + "webapp endpoint config file generation", ie);
        }

        logger.log(BasicLevel.DEBUG, "GenDir for '" + id + "' is '" + gen + "'");
        logger.log(BasicLevel.DEBUG, "DestDir for '" + id + "' is '" + dest + "'");

        // Prepare the Context
        gCtx.setProperty(id + "-abs.ejbjar.endpoint.gen.dir", gen);
        gCtx.setProperty("processing.key", id);
        gCtx.setProperty("processing.war-manager", wm);
        gCtx.setProperty("processing.wsdd", dd);

        // generate files in the temp directory
        configureEjbJarEndpoint();

        // Unset Context
        gCtx.removeProperty("processing.key");
        gCtx.removeProperty("processing.war-manager");
        gCtx.removeProperty("processing.wsdd");
        
        // compile sources
        gCtx.setProperty(id + "-abs.ejbjar.endpoint.out.gen.dir", dest);
        compileDir(gen, dest);

        // add generated classes in War
        wm.addDir("WEB-INF/classes", dest, ".class");

        // add sources ?
        if (isKeepGenerated()) {
            wm.addDir("WEB-INF/src", gen, ".java");
        }

    }

    /**
     * Inner FileFilter class : select only the java sources.
     */
    private static class JavaFileFilter implements FileFilter {
        public boolean accept(File file) {
            return file.getName().endsWith(".java");
        }
    }
    /**
     * Inner FileFilter class : select only the directories.
     */
    private static class DirFileFilter implements FileFilter {
        public boolean accept(File file) {
            return file.isDirectory();
        }
    }

    /**
     * Add all *.java files from directory into Vector.
     * @param dir directory containing source files.
     * @param srcs Vector updates with source files.
     *
     * @throws WsGenException when IOException occurs.
     */
    protected static void addJavaSources(File dir, Vector srcs)
        throws WsGenException {

        try {

            // add each java sources contained in the directory
            File[] files = dir.listFiles(new JavaFileFilter());
            for (int i = 0; i < files.length; i++) {
                srcs.add(files[i]); 
            }

            // reapply on subdirectories
            files = dir.listFiles(new DirFileFilter());
            for (int i = 0; i < files.length; i++) {
                addJavaSources(files[i], srcs);
            }

        } catch (Exception e) {
            throw new WsGenException("Error when adding files for directory '" 
                                     + dir.getName() + "'", e);
        }

    }

    
    /**
     * Compile Java sources files of directory.
     *
     * @param srcDir File representing source directory.
     * @param destDir File representing destination directory.
     *
     * @throws WsGenException when compilation fails.
     */
    protected void compileDir(File srcDir, File destDir)
        throws WsGenException  {

        logger.log(BasicLevel.DEBUG, "Compile Directory '" + srcDir + "' into '" + destDir + "'");
        

        try {
            // get all java files contained in the gen dir
            Vector srcs = new Vector();
            addJavaSources(srcDir, srcs);

            String javaC = (String) gCtx.getProperty(JAVAC_NAME, JAVAC_NAME_DEFAULT);
            Cmd cmd = new Cmd(gCtx.getJavaHomeBin() + javaC);
            cmd.addArgument("-classpath");
            cmd.addArgument(gCtx.getClasspath());
            cmd.addArgument("-d");
            cmd.addArgument(destDir.getCanonicalPath());
            cmd.addArguments((Vector) gCtx.getProperty(JAVAC_OPTS, new Vector()));

            if (srcs.size() != 0) {
                for (Enumeration e = srcs.elements(); e.hasMoreElements();) {
                    String srcName = ((File) e.nextElement()).getCanonicalPath();
                    cmd.addArgument(srcName);
                }
                
                trace("Running '" + cmd.toString() + "'");
                if (cmd.run()) {
                    WsGen.info("WebServices Classes successfully compiled.");
                } else {
                    String err = "Error when running javac for WebServices clients compilation.";
                    throw new WsGenException(err);
                }
            }
        } catch (Exception e) {
            throw new WsGenException("Error when compiling sources", e);
        }


    }

    /**
     * Generate client files for a set of ServiceRef.
     *
     * @param servRefs an Array of ServiceRef.
     *
     * @throws WsGenException when configureClient fails.
     */
    protected void genClientFiles(ServiceRef[] servRefs)
        throws WsGenException  {

        logger.log(BasicLevel.DEBUG, "Gen Client Files");

        for (int i = 0; i < servRefs.length; i++) {

            // finish preparing Context
            ServiceRef sr = servRefs[i];
            gCtx.setProperty("processing.service-ref", sr);
            configureClient();

            // remove property
            gCtx.removeProperty("processing.service-ref");
        }
    }


    /**
     * Generate client configuration files for each ServiceRef of bean descs.
     * Common method for all Web Services implementation (Glue/Axis).
     *
     * @param jm the jar file containing clients.
     * @param bdIt an iterator on a bean desc list
     *
     * @throws WsGenException if generation fails
     */
    public void genEjbJarClientConfig(EjbJarManager jm,
                                      Iterator bdIt)
        throws WsGenException  {

        if (logger.isLoggable(BasicLevel.DEBUG)) {
            logger.log(BasicLevel.DEBUG,
                       "Clients generation in JarManager '" + jm.getFilename() + "'");
        }

        // Create an ID for process
        String id = jm.getFilename();

        File gen = null;
        File dest = null;
        // create a temp directory for file generation
        try {
            gen = WsGen.createTempDir();
            dest = WsGen.createTempDir();
        } catch (java.io.IOException ie) {
            throw new WsGenException("Cannot create the temporary directory for "
                                     + "ejbjar client config file generation", ie);
        }

        logger.log(BasicLevel.DEBUG, "GenDir for '" + id + "' is '" + gen + "'");
        logger.log(BasicLevel.DEBUG, "DestDir for '" + id + "' is '" + dest + "'");

        // Prepare the Context
        gCtx.setProperty(id + "-abs.client.gen.dir", gen);
        gCtx.setProperty("processing.key", id);
        gCtx.setProperty("processing.ws-manager", jm);

        while (bdIt.hasNext()) {
            BeanDesc bd = (BeanDesc) bdIt.next();
            if (bd.getServiceRefDesc().length != 0) {
                // the ejb may have clients
                // add clients files in temporary file
                gCtx.setProperty("processing.bean-name", bd.getEjbName());
                genClientFiles(bd.getServiceRefDesc());
                gCtx.removeProperty("processing.bean-name");
            }
        }

        // Unset Context
        gCtx.removeProperty("processing.key");
        gCtx.removeProperty("processing.ws-manager");
        
        // compile sources
        gCtx.setProperty(id + "-abs.client.out.gen.dir", dest);
        compileDir(gen, dest);

        // add generated classes in EjbJar
        jm.addDir("", dest, ".class");

        // add sources ?
        if (isKeepGenerated()) {
            jm.addDir("META-INF/src/", gen, ".java");
        }

    }
 
    /**
     * Generate client configuration files for the list of ServiceRef of a War.
     * Common method for all Web Services implementation (Glue/Axis).
     *
     * @param wm the managed war file.
     * @param servRefs The list of ServiceRef to process.
     *
     * @throws WsGenException if generation fails
     */
    public void genWebAppClientConfig(WarManager wm,
                                      ServiceRef[] servRefs)
        throws WsGenException  {
        
        if (logger.isLoggable(BasicLevel.DEBUG)) {
            logger.log(BasicLevel.DEBUG,
                       "WebApp Clients Generation in '" + wm.getFilename() + "'");
        }

        // Create an ID for process
        String id = wm.getFilename();

        File gen = null;
        File dest = null;
        // create a temp directory for file generation
        try {
            gen = WsGen.createTempDir();
            dest = WsGen.createTempDir();
        } catch (java.io.IOException ie) {
            throw new WsGenException("Cannot create the temporary directory for "
                                     + "webapp client config file generation", ie);
        }

        logger.log(BasicLevel.DEBUG, "GenDir for '" + id + "' is '" + gen + "'");
        logger.log(BasicLevel.DEBUG, "DestDir for '" + id + "' is '" + dest + "'");

        // Prepare the Context
        gCtx.setProperty(id + "-abs.client.gen.dir", gen);
        gCtx.setProperty("processing.key", id);
        gCtx.setProperty("processing.ws-manager", wm);

        genClientFiles(servRefs);

        // Unset Context
        gCtx.removeProperty("processing.key");
        gCtx.removeProperty("processing.ws-manager");
        
        // compile sources
        gCtx.setProperty(id + "-abs.client.out.gen.dir", dest);
        compileDir(gen, dest);

        // add generated classes in War
        wm.addDir("WEB-INF/classes/", dest, ".class");

        // add sources ?
        if (isKeepGenerated()) {
            wm.addDir("WEB-INF/src/", gen, ".java");
        }

    }
 
    /**
     * Recursively destroy the given file.
     * 
     * @param file the file to destroy.
     * 
     * @return true if deletion OK, otherwise false.
     */
    protected static boolean destroyFile(File file) {
        boolean isDestroy = true;
        if (file.isDirectory()) {
            // destroy all childs
            File[] childs = file.listFiles();
            for (int i = 0; i < childs.length; i++) {
                isDestroy &= destroyFile(childs[i]);
            }
        }
        return isDestroy && file.delete();
    }
 
    /**
     * Destroy the temporary directory used for generation.
     *
     */
    protected void destroyTempDirectories() {

        // retrieve properties with name finishing with ".gen.dir"
        for (Iterator p = gCtx.getPropertyNames(); p.hasNext();) {
            String name = (String) p.next();

            // if gen directory
            if (name.endsWith(".gen.dir")) {
                File genDir = (File) gCtx.getProperty(name);

                logger.log(BasicLevel.DEBUG, "Trying to delete '" + genDir.getName() + "'");

                // try to delete the directory
                if (!destroyFile(genDir)) {
                    WsGen.warning("Cannot delete temporary dir '" + genDir.getName() + "'");
                }
            }
        }
    }

    /**
     * Return the processing ID from GenerationContext.
     * @return the processing ID from GenerationContext.
     */
    protected String getProcessingID() {
        return (String) gCtx.getProperty("processing.key");
    }

    /**
     * Return the generation directory from ID
     *
     * @param id processing id
     *
     * @return the File for the directory
     */
    protected File getWebAppEndpointGenDir(String id) {
        return (File) gCtx.getProperty(id + "-abs.webapp.endpoint.gen.dir");
    }

    /**
     * Return the generation directory from ID
     *
     * @param id processing id
     *
     * @return the File for the directory
     */
    protected File getEjbJarEndpointGenDir(String id) {
        return (File) gCtx.getProperty(id + "-abs.ejbjar.endpoint.gen.dir");
    }

    /**
     * Return the generation directory from ID
     *
     * @param id processing id
     *
     * @return the File for the directory
     */
    protected File getClientGenDir(String id) {
        return (File) gCtx.getProperty(id + "-abs.client.gen.dir");
    }
 
    /**
     * configureWebAppEndpoint has to be implemented in subClass.
     * hold specific process for endpoint with specific WebService engine.
     *
     * @throws WsGenException if generation fails if the method isn't implemented
     */
    protected void configureWebAppEndpoint() throws WsGenException {
        throw new WsGenException("configureWebAppEndpoint needs to be overridden " 
                                 + "in subclasses of " + this.getClass().getName());
    }

    /**
     * configureEjbJarEndpoint has to be implemented in subClass.
     * hold specific process for endpoint with specific WebService engine.
     *
     * @throws WsGenException if generation fails if the method isn't implemented
     */
    protected void configureEjbJarEndpoint() throws WsGenException {
        throw new WsGenException("configureEjbJarEndpoint needs to be overridden " 
                                 + "in subclasses of " + this.getClass().getName());
    }

    /**
     * configureClient has to be implemented in subClass.
     * hold specific process for client with specific WebService engine.
     *
     * @throws WsGenException if generation fails if the method isn't implemented
     */
    protected void configureClient() throws WsGenException {
        throw new WsGenException("configureClient needs to be overridden " 
                                 + "in subclasses of " + this.getClass().getName());
    }
 
    /**
     * Return a String containing the path representation
     * of the ContextClassLoader.
     *
     * @return Return a String containing the path representation
     * of the ContextClassLoader.
     */
    private static String getContextClassPath() {
        URLClassLoader ucl = (URLClassLoader) (Thread.currentThread().getContextClassLoader());
        String ps = System.getProperty("path.separator");
        String cp = ".";
        URL[] urls = ucl.getURLs();
        for (int i = 0; i < urls.length; i++) {
            cp = cp + ps + urls[i].getFile();
        }
        return cp;
    }

    /**
     * Display the specified message only if verbose.
     * @param msg the message to display
     */
    protected void trace(String msg) {
        logger.log(isVerbose() ? BasicLevel.INFO : BasicLevel.DEBUG, msg);
    }

    /**
     * return true if verbose mode is set.
     * @return true if verbose mode is set.
     */
    protected boolean isVerbose() {
        return ((Boolean) gCtx.getProperty(VERBOSE, VERBOSE_DEFAULT)).booleanValue();
    }

    /**
     * return true if debug mode is set.
     * @return true if debug mode is set.
     */
    protected boolean isDebug() {
        return ((Boolean) gCtx.getProperty(DEBUG, DEBUG_DEFAULT)).booleanValue();
    }

    /**
     * return true if noConfig mode is set.
     * @return true if noConfig mode is set.
     */
    protected boolean isNoConfig() {
        return ((Boolean) gCtx.getProperty(NO_CONFIG, NO_CONFIG_DEFAULT)).booleanValue();
    }

    /**
     * return true if keep generated files mode is set.
     * @return true if keep generated files mode is set.
     */
    protected boolean isKeepGenerated() {
        return ((Boolean) gCtx.getProperty(KEEP_GEN, KEEP_GEN_DEFAULT)).booleanValue();
    }

    /**
     * Remove temp directories.
     */
    public void clean() {
        destroyTempDirectories();
    }
}
