/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Xavier Delplanque & Guillaume Sauthier
 * --------------------------------------------------------------------------
 * $Id: WsGen.java,v 1.11 2003/10/17 15:55:59 sauthieg Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ws.wsgen;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.StringTokenizer;
import java.util.Vector;

import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

import org.objectweb.jonas.common.JProp;
import org.objectweb.jonas.common.Log;
import org.objectweb.jonas.ear.lib.EarFileManager;
import org.objectweb.jonas.ear.lib.EarFileManagerException;


/**
 * 
 *
 * @author Xavier Delplanque & Guillaume Sauthier
 */
public abstract class WsGen {
 
    /** The Generator used for webservices classes and configuration files creation */
    protected ConfigGenerator cg;

    protected boolean keepGenerated;

    /** temp directory for archive creation */
    protected File tmpDir;

    /** Is the command on verbose ? */
    protected boolean verbose = false;

    /** Is the command is on debug mode ? */
    protected boolean debug = false;

    /** Is the command used for help message ? */
    protected boolean isHelp = false;

    /** Name of the directory where to place the generated file */
    protected String outputdir = null;

    /** input file name */
    protected String fileInputName = null;

    /** The logger to use */
    protected static Logger logger = null;

    private static WsGen wsg;
    /**
     * Constructor : creates a WsGen object.
     * Generate source files for client and configuration files
     * for both endpoints and clients (if needed).
     *
     * @param outputDir the directory where to store generated files
     * @param cGen The ConfigGenerator to use (Axis/Glue/...)
     * @param isVerbose Indicate if command is in verbose mode
     * @param isDebug Indicate if command is in debug mode
     * @param filename file to process.
     */
    protected WsGen(String outputDir,
                    ConfigGenerator cGen,
                    boolean isVerbose,
                    boolean isDebug,
                    String filename)
        throws WsGenException {
        
        verbose = isVerbose;
        cg = cGen;
        debug = isDebug;
        outputdir = outputDir;
        fileInputName = filename;

        if (logger == null) {
            logger = Log.getLogger(Log.JONAS_WSGEN_PREFIX);
            System.out.println("level : " + logger.getCurrentIntLevel());
            System.out.println(" debug level" + BasicLevel.DEBUG);
        }

        try {
            tmpDir = createTempDir();
        } catch (IOException e){
            String err = "Cannot create temp directory.";
            throw new WsGenException(err, e);
        }

        // Launch the configuration process
        configure();

        // Remove temp directories
        removeTempDirectories();
    }

    protected void removeTempDirectories() {
        cg.clean();
        removeDir(tmpDir);
    }

    /**
     * JOnAS Method
     */
    public static void main(String args[]) {
        boolean  error = false;

        boolean  isHelp              = false;
        boolean  isVerbose           = false;
        boolean  isDebug             = false;
        boolean  isKeepGenerated     = false;
        boolean  noConfig            = false;
        boolean  parseWithValidation = true;
        String   cGenClassName       = null;

        String   fileInputName = null;
        String   dirOutputName = null;
        String   nameJavac     = null;
        Vector   optionsJavaC  = new Vector();

        // Get args
        for (int argn = 0; argn < args.length; argn++) {
            String arg = args[argn];
            if (arg.equals("-help") || arg.equals("-?")) {
                isHelp = true;
                continue;
            }
            if (arg.equals("-verbose")) {
                isVerbose = true;
                continue;
            }
            if (arg.equals("-debug")) {
                isDebug = true;
                isVerbose = true;
                continue;
            }
            if (arg.equals("-keepgenerated")) {
                isKeepGenerated = true;
                continue;
            }
            if (arg.equals("-noconfig")) {
                noConfig = true;
                continue;
            }
            if (arg.equals("-novalidation")) {
                parseWithValidation = false;
                continue;
            }
            if (arg.equals("-javac")) {
                argn++;
                continue;
            }
            if (arg.equals("-javacopts")) {
                argn++;
                if (argn < args.length) {
                    StringTokenizer st = new StringTokenizer(args[argn]);
                    while (st.hasMoreTokens()) {
                        optionsJavaC.addElement(st.nextToken());
                    }
                } else {
                    error = true;
                }
                continue;
            }
            if (arg.equals("-d")) {
                argn++;
                if (argn < args.length) {
                    dirOutputName = args[argn];
                } else {
                    error = true;
                }
                continue;
            }
            fileInputName = args[argn];
        }

        Log.configure("trace");

        logger = Log.getLogger(Log.JONAS_WSGEN_PREFIX);
        
        // Usage ?
        if (isHelp) {
            usage();
            return;
        }

        // Check args
        if (error || (fileInputName == null) 
            || (dirOutputName == null)) {
            usage();
            throw new RuntimeException();
        }

        File inputFile = null;
        try {
            inputFile = new File(fileInputName);
            fileInputName = inputFile.getCanonicalPath();
        } catch (java.io.IOException ioe) {
            String err = "FileInputName : " + fileInputName 
                + " not found ! ";
            WsGen.fatalError(err);
        }

        File outputDir = null;
        try {
            outputDir = new File(dirOutputName);
            dirOutputName = outputDir.getCanonicalPath();
        } catch (java.io.IOException ioe) {
            String err = "dirOutputName : " + dirOutputName 
                + " not found ! ";
            WsGen.fatalError(err);
        }

        
        try {
            JProp jp = JProp.getInstance();
            cGenClassName = jp.getValue("jonas.wsgen.generator.class");
        } catch (Exception e) {
            String err = "Exception when trying to read jonas.properties";
            WsGen.fatalError(err);
        }

        if (cGenClassName == null) {
            String err = "property 'jonas.wsgen.generator.class' must be set"
                + "in jonas.properties";
            WsGen.fatalError(err);
        }

        // Get the ConfigGenerator class to use
        ConfigGenerator confGen = null;
        try {
            
            GenerationContext ctx = new GenerationContextImpl();

            ctx.setProperty(ConfigGenerator.VERBOSE,     new Boolean(isVerbose));
            ctx.setProperty(ConfigGenerator.DEBUG,       new Boolean(isDebug));
            ctx.setProperty(ConfigGenerator.NO_CONFIG,   new Boolean(noConfig));
            ctx.setProperty(ConfigGenerator.KEEP_GEN,    new Boolean(isKeepGenerated));
            ctx.setProperty(ConfigGenerator.JAVAC_NAME,  nameJavac);
            ctx.setProperty(ConfigGenerator.JAVAC_OPTS,  optionsJavaC);
            ctx.setProperty(ConfigGenerator.PARSING_VAL, new Boolean(parseWithValidation));
            
            Class cg = Class.forName(cGenClassName);
            Constructor c = cg.getConstructor(new Class[] {GenerationContext.class});
            confGen = (ConfigGenerator) c.newInstance(new Object[] {ctx});
            
        } catch (ClassNotFoundException cnf) {
            String err = "ConfigGenerator subClass : " + cGenClassName
                + " not found ! " + cnf;
            WsGen.fatalError(err);
        } catch (InstantiationException ie) {
            String err = "Error when instantiate : " + cGenClassName + " : " + ie;
            WsGen.fatalError(err);
        } catch (IllegalAccessException iae) {
            String err = "Error when instantiate : " + cGenClassName + " : " + iae;
            WsGen.fatalError(err);
        } catch (NoSuchMethodException nsme) {
            String err = "Error when instantiate : " + cGenClassName + " : " + nsme;
            WsGen.fatalError(err);
        } catch (InvocationTargetException ite) {
            String err = "Error when instantiate : " + cGenClassName + " : " + ite;
            WsGen.fatalError(err);
        }

        try {
            // EAR case : deflate the ear in tempo. directory
            if (fileInputName.endsWith(".ear")) {
                wsg = new EarWsGen(dirOutputName,
                                   confGen,
                                   isVerbose,
                                   isDebug,
                                   fileInputName);
                
            } else if (fileInputName.endsWith(".jar")) {
                wsg = new EjbJarWsGen(dirOutputName,
                                      confGen,
                                      isVerbose,
                                      isDebug,
                                      fileInputName);
                
            } else if (fileInputName.endsWith(".war")) {
                wsg = new WebAppWsGen(dirOutputName,
                                      confGen,
                                      isVerbose,
                                      isDebug,
                                      fileInputName);
                
            } else {
                // if inputfile not an EAR or EJBJAR or WAR
                usage();
                throw new RuntimeException();
            }
        } catch (WsGenException wge) {
            String err = "Exception during config generation";
            WsGen.fatalError(err);
        }
    }

    /**
     * Display the usage
     */
    public static void usage() {
        StringBuffer msg = new StringBuffer();
        msg.append("Usage: java org.objectweb.jonas_ws.wsgen.WsGen -help \n");
        msg.append("   to print this help message \n");
        msg.append(" or java org.objectweb.jonas_ws.wsgen.WsGen <Options> <Input_File> \n");
        msg.append("Options include: \n");
        msg.append("    -d <output_dir>  specify where to place the generated files \n");
        msg.append("    -novalidation parse the XML deployment descriptors without \n");
        msg.append("      validation \n");
        msg.append("    -javac  <opt> specify the java compiler to use \n");
        msg.append("    -javacopts <opt> specify the options to pass to the java compiler \n");
        msg.append("    -keepgenerated   do not delete intermediate generated files \n");
        msg.append("    -noconfig  do not generate configuration files (require \n");
        msg.append("      user privided files) \n");
        msg.append("    -verbose \n");
        msg.append("    -debug \n");
        msg.append(" \n");
        msg.append("    Input_File    the ejb-jar, war or ear filename\n");
        WsGen.info(msg.toString());
    }

    /**
     * Display the specified message only if verbose.
     * @param msg the message to display
     */
    void trace(String msg) {
        logger.log(verbose ? BasicLevel.INFO : BasicLevel.DEBUG, msg);
    }

    /**
     * Display the specified message.
     * @param msg the message to display
     */
    public static void info(String msg) {
        if (logger.isLoggable(BasicLevel.INFO)) {
            logger.log(BasicLevel.INFO, msg);
        }
    }

    /**
     * Display the specified message.
     * @param msg the message to display
     */
    public static void debug(String msg) {
        if (logger.isLoggable(BasicLevel.DEBUG)) {
            logger.log(BasicLevel.DEBUG, msg);
        }
    }

    /**
     * Display the specified warning message.
     * @param msg the warning message to display
     */
    public static void warning(String msg) {
        if (logger.isLoggable(BasicLevel.WARN)) {
            logger.log(BasicLevel.WARN, msg);
        }
    }

    /**
     * Display the specified error message.
     * @param msg the error message to display
     */
    public static void error(String msg) {
        if (logger.isLoggable(BasicLevel.ERROR)) {
            logger.log(BasicLevel.ERROR, msg);
        }
    }


    /**
     * Display the specified error message and exits with an
     * EXIT_FAILURE status.
     * @param msg the error message to display
     */
    public static void fatalError(String msg) {
        System.err.println("WsGen fatal error: " + msg);
        throw new RuntimeException(msg);
    }

    /**
     * Display the specified error message and exits with an
     * EXIT_FAILURE status.
     * @param e the error to display
     */
    public static void fatalError(Exception e) {
        System.err.println("WsGen fatal error: ");
        printException(e, System.err);
        e.printStackTrace(System.err);
        throw new RuntimeException(e.getMessage());
    }
    /**
     * Display the exception and its nested exception if exists, on the given printstream.
     * @param e the exception to display
     * @param out the PrintStream on which the exception has to displayed
     */ 
    static void printException(Exception e, PrintStream out) {
        Exception inner = null;
        if (e instanceof WsGenException
            && (inner = ((WsGenException) e).inner) != null) {
            out.println(e.getMessage());
            printException(inner, out);
        } else {
            e.printStackTrace();
        }
    }

    /**
     * Create a cleaned temporary directory.
     * @return the temp directory file
     * @throws IOException if a temp directory cannot be created.
     */
    public static File createTempDir() throws IOException {
        File tmpDir = File.createTempFile("wsgen", null, null); 
        tmpDir.delete();
        if (!tmpDir.mkdir()) {
            throw new IOException("Cannot create the temporary directory '" + tmpDir + "'.");
        }
        return tmpDir;
    }

    /**
     * Unpack the jar in a temporary directory.
     * @return the output directory URL
     */
    public static URL getExplodedJar(String fileInputName) {
        File tempDir = null;
        // unpack the file inside a temporary directory
        try {
            tempDir = createTempDir();
        } catch (IOException ioe) {
            String err = "Cannot create temporary Directory !";
            fatalError(err);
        }
        URL fileURL = null;
        URL temp = null;
        try {
            fileURL = new File(fileInputName).toURL();
            URL tempURL = tempDir.toURL();
            temp = EarFileManager.unpackEar(fileURL, tempURL);
        } catch (MalformedURLException mue) {
            error(mue.getMessage());
            fatalError(mue);
        } catch (EarFileManagerException fme) {
            error(fme.getMessage());
            fatalError(fme);
        }

        return temp;
    }

    /**
     * Copy a source file in a directory.
     * @param fileName the source filename.
     * @param dirOutputName the directory where file will be copied.
     * @return the path to the new file.
     */
    private static String copyFileTo(String fileName, String dirOutputName) {

        File inputFile = new File(fileName);
        File outputDir = new File(dirOutputName);

        String outFileName = null;

        // we build a copy of the given file in the given repertory
        File outFile;
        try {
            if (!outputDir.exists()) {
                // if the given directory doesn't exist, create it.
                outputDir.mkdirs();
            }
            outFile = new File(dirOutputName + File.separator + inputFile.getName());

            if (!outFile.equals(inputFile)) {
                FileOutputStream os = new FileOutputStream(outFile);
                FileInputStream is = new FileInputStream(inputFile);
                // copy the file
                byte[] buf = new byte[1024];
                int length;
                while ((length = is.read(buf)) > 0) {
                    os.write(buf, 0, length);
                }
                is.close();
                os.flush();
                os.close();
                outFileName = outFile.getCanonicalPath();
            }
        } catch (Exception e) {
            String err = "Error when copying '" + fileName
                + "' in '" + dirOutputName + "'";
            WsGen.fatalError(err);
        }

        return outFileName;
    }

    protected void removeDir(File dir) {

        if (!destroyFile(dir)) {
            warning("Cannot delete tmp directory '" + dir + "'");
        }
    }

    /**
     * Recursively destroy the given file/directory.
     * 
     * @param file the file to destroy.
     * 
     * @return true if deletion OK, otherwise false.
     */
    protected static boolean destroyFile(File file) {
        boolean isDestroy = true;
        if (file.isDirectory()) {
            // destroy all childs
            File[] childs = file.listFiles();
            for (int i = 0; i < childs.length; i++) {
                isDestroy &= destroyFile(childs[i]);
            }
        }
        return isDestroy && file.delete();
    }

    /**
     * Abstract method for configuration of the Object.
     * @throws WsGenException when configuration fails.
     */
    public abstract void configure() throws WsGenException;
    
}
