/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Xavier Delplanque & Guillaume Sauthier
 * --------------------------------------------------------------------------
 * $Id: AxisConfigGenerator.java,v 1.9 2003/10/22 12:39:56 sauthieg Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ws.wsgen.axis;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Iterator;

import javax.xml.namespace.QName;

import org.objectweb.util.monolog.api.BasicLevel;

import org.objectweb.jonas_ws.deployment.api.MappingFile;
import org.objectweb.jonas_ws.deployment.api.PortComponentDesc;
import org.objectweb.jonas_ws.deployment.api.ServiceRef;
import org.objectweb.jonas_ws.deployment.api.WSDeploymentDesc;
import org.objectweb.jonas_ws.wsgen.AbsConfigGenerator;
import org.objectweb.jonas_ws.wsgen.GenerationContext;
import org.objectweb.jonas_ws.wsgen.WsGen;
import org.objectweb.jonas_ws.wsgen.WsGenException;
import org.objectweb.jonas_ws.wsgen.jarfile.WarManager;
import org.objectweb.jonas_ws.wsgen.jarfile.WsJarManager;

import org.apache.velocity.Template;
import org.apache.velocity.VelocityContext;
import org.apache.velocity.app.VelocityEngine;

import org.apache.axis.wsdl.toJava.Utils;

import org.objectweb.common.Cmd;

/**
 * Implement methods generating AXIS specific configuration files and classes.
 *
 * @author Guillaume Sauthier & Xavier Delplanque
 */
public class AxisConfigGenerator extends AbsConfigGenerator {

    private VelocityEngine vEngine;
    private int fileCptr;

    private static final String AXIS_SERVLET_NAME    = "AxisServlet";
    private static final String AXIS_SERVLET_CLASS   = "org.apache.axis.transport.http.AxisServlet";
    private static final String AXIS_DEFAULT_MAPPING = "/services/*";
 
    /**
     * Create a new AxisConfigGenerator with the specified GenerationContext.
     * @param ctx the GenerationContext.
     */
    public AxisConfigGenerator(GenerationContext ctx)
        throws WsGenException {
        super(ctx);

        logger.log(BasicLevel.DEBUG, "Create AxisConfigGenerator");

        // set Default properties
        gCtx.setProperty(SERVLET_NAME   , AXIS_SERVLET_NAME);
        gCtx.setProperty(SERVLET_CLASS  , AXIS_SERVLET_CLASS);
        gCtx.setProperty(DEFAULT_MAPPING, AXIS_DEFAULT_MAPPING);
  
        String jonasRoot = System.getProperty("install.root");
        if (jonasRoot == null) {
            throw new WsGenException("System property 'install.root' not set");
        }
        String path2Tmpl = new String(jonasRoot + File.separatorChar + "templates"
                                      + File.separatorChar + "wsgen" 
                                      + File.separatorChar + "axis");
        vEngine = new VelocityEngine();
        vEngine.setProperty(VelocityEngine.VM_LIBRARY, "");
        vEngine.setProperty(VelocityEngine.RESOURCE_LOADER, "file");
        vEngine.setProperty(VelocityEngine.FILE_RESOURCE_LOADER_PATH, path2Tmpl);
        try {
            vEngine.init();
        } catch (Exception e) {
            throw new WsGenException("Cannot initialize the Velocity engine", e);
        }
        fileCptr = 0;
    }

    public void configureEjbJarEndpoint()
        throws WsGenException {

        logger.log(BasicLevel.DEBUG, "configureEjbJarEndpoint()");

        String id = getProcessingID();

        File genDir = getEjbJarEndpointGenDir(id);
        WSDeploymentDesc dd = (WSDeploymentDesc) gCtx.getProperty("processing.wsdd");
        WarManager wm = (WarManager) gCtx.getProperty("processing.war-manager");

        logger.log(BasicLevel.DEBUG, " -id=" + id);
        logger.log(BasicLevel.DEBUG, " -dd=" + dd);
        logger.log(BasicLevel.DEBUG, " -wm=" + wm.getFilename());

        // create deploy.wsdd
        // use velocity template
        String tmplName = "deploy_endpoint.vm";
        String srcFileName = genDir.getPath() + File.separator + "server-config.wsdd";

        logger.log(BasicLevel.DEBUG, "Creating '" + srcFileName + "'");

        VelocityContext vctx = VContextFactory.createEndpointCtx(dd);

        // generate server-config.wsdd file
        genFile(srcFileName, tmplName, vctx);

        // update servlets
        wm.addServlet(AXIS_SERVLET_NAME, AXIS_SERVLET_CLASS, AXIS_DEFAULT_MAPPING);

        wm.addFile("WEB-INF/server-config.wsdd", new File(srcFileName));
        
    }

    public void configureWebAppEndpoint()
        throws WsGenException {

        logger.log(BasicLevel.DEBUG, "configureWebAppEndpoint()");
        
        String id = getProcessingID();

        File genDir = getWebAppEndpointGenDir(id);
        PortComponentDesc pcd =
            (PortComponentDesc) gCtx.getProperty("processing.port-component-desc");
        WarManager wm = (WarManager) gCtx.getProperty("processing.war-manager");

        logger.log(BasicLevel.DEBUG, " -id=" + id);
        logger.log(BasicLevel.DEBUG, " -pcd=" + pcd);
        logger.log(BasicLevel.DEBUG, " -wm=" + wm.getFilename());

        // create deploy.wsdd
        // use velocity template
        String tmplName = "deploy_endpoint.vm";
        String srcFileName = genDir.getPath() + File.separator + pcd.getName() + "-server-config.wsdd";

        logger.log(BasicLevel.DEBUG, "Creating '" + srcFileName + "'");

        VelocityContext vctx = VContextFactory.createEndpointCtx(pcd);
        
        // generate *-server-config.wsdd
        genFile(srcFileName, tmplName, vctx);

        // complete the xml desc
        String sName = pcd.getSibLink();
        wm.setServletClass(sName, AXIS_SERVLET_CLASS);
        
        // add server-config file in init-param
        wm.addServletInitParam(sName,
                               "axis.serverConfigFile",
                               pcd.getName() + "-server-config.wsdd");

    }

 
    protected void configureClient()
        throws WsGenException {

        logger.log(BasicLevel.DEBUG, "configureClient()");

        String id = getProcessingID();

        File genDir = getClientGenDir(id);
        ServiceRef servRef = (ServiceRef) gCtx.getProperty("processing.service-ref");
        WsJarManager ws = (WsJarManager) gCtx.getProperty("processing.ws-manager");

        String beanName = (String) gCtx.getProperty("processing.bean-name");
        
        logger.log(BasicLevel.DEBUG, " -id=" + id);
        logger.log(BasicLevel.DEBUG, " -gen=" + genDir);
        logger.log(BasicLevel.DEBUG, " -sRef=" + servRef/*.getServiceRefName()*/);
        logger.log(BasicLevel.DEBUG, " -ws=" + ws/*.getFilename()*/);
        logger.log(BasicLevel.DEBUG, " -name=" + beanName);
        

        // the source generation is possible only when a
        // WSDL Definition is provided
        if (servRef.getWsdlFileName() != null 
            && !servRef.getServiceInterface().getName().equals("javax.xml.rpc.Service")) {
            // Source generation for the client
            String cmdJava = gCtx.getJavaHomeBin() + "java";
            String wsdl = null;

            // Command preparation
            Cmd axisCmd = new Cmd(cmdJava);
            axisCmd.addArgument("-classpath");
            axisCmd.addArgument(gCtx.getClasspath());
            axisCmd.addArgument("org.apache.axis.wsdl.WSDL2Java");

            // get informations from ServiceRef ...

            // WSDL file
            try {
                // todo : test under Win32 for URL
                String pathWSDL = new File(ws.getFilename()).getCanonicalFile().toURL().toExternalForm();
                wsdl = "jar:" + pathWSDL + "!/" + servRef.getWsdlFileName();
            } catch (IOException ioe) {
                String err = "Cannot retrieve file '" + ws.getFilename() + "'";
                throw new WsGenException(err, ioe);
            }

            // Mapping File
            MappingFile mf = servRef.getMappingFile();

            for (Iterator i = mf.getMappings().keySet().iterator() ; i.hasNext() ;) {
                String namespace = (String) i.next();
                String pkg = (String) mf.getMappings().get(namespace);
                axisCmd.addArgument("--NStoPkg");
                axisCmd.addArgument(namespace + "=" + pkg);
            }

            // Set the output directory
            axisCmd.addArgument("--output");
            axisCmd.addArgument(genDir.getPath());

            if (isVerbose()) {
                axisCmd.addArgument("--verbose");
            }

            if (isDebug()) {
                axisCmd.addArgument("--Debug");
            }

            axisCmd.addArgument(wsdl);

            trace("Running '" + axisCmd.toString() + "'");
            if (axisCmd.run()) {
                WsGen.info("Web Services Classes successfully generated by Axis.");
            } else {
                String err = "Error when running axis WSDL2Java tool.";
                throw new WsGenException(err);
            }

            // !!! update jonas-service-ref with serviceClassName
            // beanName can be null : not important
            ws.addJServiceRefInitParam(beanName,
                                       servRef.getServiceRefName(),
                                       "serviceClassName",
                                       getServiceClassname(servRef));
            

        } else { // End source generation
            logger.log(BasicLevel.DEBUG, "javax.xml.rpc.Service interface used");
            ws.addJServiceRefInitParam(beanName,
                                       servRef.getServiceRefName(),
                                       "serviceClassName",
                                       "org.apache.axis.client.Service");
        }

        // Client-config.wsdd ??
        // only if handlers are specified in ServiceRef
        // but if noConfig set to true, assume that
        // user configure himself the client. (expert)

        // Could we create configuration ?
        if (!isNoConfig() && (servRef.getHandlerRefs().size() != 0)) {
            // create deploy.wsdd
            // use velocity template
            String tmplName = "deploy_client.vm";
            // build a unique file name
            fileCptr++;
            String srcFileName = genDir.getPath() + File.separator 
                + "client-deploy-" + fileCptr + ".wsdd";
            VelocityContext vctx = VContextFactory.createClientCtx(servRef);

            logger.log(BasicLevel.DEBUG, "Creating '" + srcFileName + "'");

            genFile(srcFileName, tmplName, vctx);

            if (ws instanceof WarManager) {
                ws.addFile("WEB-INF/client-deploy-" + fileCptr + ".wsdd", new File(srcFileName));
            } else {
                ws.addFile("META-INF/client-deploy-" + fileCptr + ".wsdd", new File(srcFileName));
            }

            // !!! add init param
            // maybe add "WEB-INF" ??
            ws.addJServiceRefInitParam(beanName,
                                       servRef.getServiceRefName(),
                                       "axis.clientConfigFile",
                                       "client-deploy-" + fileCptr + ".wsdd");

        } // End !noConfig
    }

    /**
     * File generation using Velocity templates.
     */
    private void genFile(String srcFileName, String tmplName, VelocityContext vctx)
        throws WsGenException {

        Template tmpl = null;
        FileWriter fwriter = null;
        trace("create :" + srcFileName);
        try {
            tmpl = vEngine.getTemplate(tmplName);
        } catch (Exception e) {
            e.printStackTrace();
            throw new WsGenException("Cannot get the '" + tmplName + "' template file", e);
        }

        try {
            // Create before the parent directory
            File fs = new File(srcFileName);
            File fdir = fs.getParentFile();
            if (fdir != null) {
                if (!fdir.exists()) {
                    if (!fdir.mkdirs()) {
                        throw new IOException("Cannot create the directory '" + fdir.getPath() + "'");
                    }
                }
            }
            fwriter = new FileWriter(srcFileName);
        } catch (IOException e) {
            e.printStackTrace();
            throw new WsGenException("Cannot create the '" + srcFileName
                                     + "' file", e);
        }

        trace("Generate the file " + srcFileName);
        try {
            tmpl.merge(vctx, fwriter);
        } catch (Exception e) {
            throw new WsGenException("Cannot generate the '" + srcFileName
                                     + "' file from the '" + tmplName + "' template", e);
        }

        try {
            fwriter.flush();
            fwriter.close();
        } catch (IOException e) {
            throw new WsGenException("Cannot close the '" + srcFileName
                                     + "' file", e);
        }

    }


    /**
     * Return the Service Implementatin classname generated by Axis' WSDL2Java.
     *
     * @return the Service Implementatin classname generated by Axis' WSDL2Java.
     */
    private String getServiceClassname(ServiceRef sr) {
        QName qn = sr.getServiceQName();
        MappingFile mf = sr.getMappingFile();

        String p = (String) mf.getMappings().get(qn.getNamespaceURI());
        
        String classname = "";
        if (p != null) {
            classname = p + "." + Utils.xmlNameToJavaClass(qn.getLocalPart()) + "Locator";
        } else {
            classname = Utils.xmlNameToJavaClass(qn.getLocalPart()) + "Locator";
        }

        return classname;
    }

}
