/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Xavier Delplanque & Guillaume Sauthier
 * --------------------------------------------------------------------------
 * $Id: EarManager.java,v 1.2 2003/10/17 15:56:00 sauthieg Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ws.wsgen.jarfile;

//import java
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.util.jar.JarFile;

import org.objectweb.util.monolog.api.BasicLevel;

import org.objectweb.jonas_ws.wsgen.WsGenException;
import org.objectweb.jonas_ws.wsgen.xml.ApplicationXml;

/**
 * Use xml classes to get and write
 * informations inside an ear file.
 * 
 * @author Guillaume Sauthier
 */
public class EarManager extends JarManager {
 
    /** application.xml entry name */
    private static final String EAR = "META-INF/application.xml";

    /** application.xml instance */
    private ApplicationXml app = null;

    /**
     * Build a new instance of a EarManager.
     *
     * @param earFilename the name of the ear file
     * that we want to manage.
     *
     * @throws WsGenException when instanciation fails.
     */
    public EarManager(String earFilename) throws WsGenException {
        super(earFilename);
  
        logger.log(BasicLevel.DEBUG, "Creating EarManager for '" + earFilename + "'");

        // get the ear file DD.
        if (!isNewFile()) {
            try {
                JarFile jarFile = new JarFile(filename);
                InputStream is = jarFile.getInputStream(jarFile.getEntry(EAR));
                app = loadApplication(is);
                is.close();
            } catch (IOException ioe) {
                if (logger.isLoggable(BasicLevel.WARN)) {
                    logger.log(BasicLevel.WARN, "Cannot get '" + EAR 
                               + "' JarEntry from '" + filename + "'\n" 
                               + "will continue with default application.xml");
                }
                app = createDefaultApplicationDesc();
            }
        } else {
            logger.log(BasicLevel.INFO, "'" + filename + "' is empty, create a default application.xml");
            app = createDefaultApplicationDesc();
        }

        // add application.xml in save list
        addXml(EAR, app);
    }

    /**
     * Build a new instance of a EarManager.
     *
     * @param earFilename the name of the ear file
     * that we want to manage.
     * @param dir output directory
     *
     * @throws WsGenException if the ear file doesn't exist.
     */
    public EarManager(String earFilename, String dir) throws WsGenException {
        this(earFilename);
        setOutputDir(dir);
    }

    /**
     * Add a new War in the current ear file.
     *
     * @param warFileName the new war file name.
     *
     * @throws WsGenException
     */
    public void addWar (String warFileName)
        throws WsGenException {

        try {
            File war = new File(warFileName) ;
            String warName = war.getName();
            addFile(warName, war);
            app.addWeb(warName, warName.substring(0, warName.length() - 4));
        } catch (Exception e) {
            throw new WsGenException(e);
        }
    }

    /**
     * Add a new EjbJar in the current ear file.
     *
     * @param ejbjarFileName the new ejbjar file name.
     *
     * @throws WsGenException
     */
    public void addEjbjar (String ejbjarFileName)
        throws WsGenException {

        try {
            File jar = new File(ejbjarFileName) ;
            String ejbName = jar.getName();
            addFile(ejbName, jar);
            app.addEjb(ejbName);
        } catch (Exception e) {
            throw new WsGenException(e);
        }
    }

    /**
     * Create a Default application.xml object
     */
    private ApplicationXml createDefaultApplicationDesc() throws WsGenException {

        if (logger.isLoggable(BasicLevel.DEBUG)) {
            logger.log(BasicLevel.DEBUG, "Create Default application.xml for '" + filename + "'");
        }

        String jonasRoot = System.getProperty("install.root");
        if (jonasRoot == null) {
            throw new WsGenException("System property 'install.root' not set");
        }
        String path2Tmpl = new String(jonasRoot + File.separatorChar + "templates"
                                      + File.separatorChar + "wsgen");
        InputStream is = null;
        String filename = path2Tmpl + File.separatorChar + "application.xml";
        try {
            is = new FileInputStream(filename);
        } catch (FileNotFoundException ignored) {
            ;
        }

        ApplicationXml application = loadApplication(is);
        application.removeModules();
        return application;
    }

    /** 
     * Load the application.xml file.
     * @param is the InputStream of the XML file.
     * @return a structure containing the result of the application.xml parsing.
     * @throws WsGenException if the deployment descriptor 
     * is corrupted.
     */
    private static ApplicationXml loadApplication(InputStream is) 
        throws WsGenException {
        return new ApplicationXml(is);
    }

    /**
     * Omit application.xml file
     */
    protected boolean omit(String name) {
        return (name.equals(EAR));
    }

}
