/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Xavier Delplanque & Guillaume Sauthier
 * --------------------------------------------------------------------------
 * $Id: EjbJarManager.java,v 1.2 2003/10/17 15:56:00 sauthieg Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ws.wsgen.jarfile;

//import java
import java.io.IOException;
import java.io.InputStream;
import java.util.jar.JarFile;

import org.objectweb.util.monolog.api.BasicLevel;

import org.objectweb.jonas_ws.wsgen.WsGenException;
import org.objectweb.jonas_ws.wsgen.xml.JonasEjbJarXml;

/**
 * Use xml classes to get and write
 * informations inside an ejbjar file.
 *
 * @author Guillaume Sauthier
 */
public class EjbJarManager extends WsJarManager {

    // ejb-jar Depl. Desc. names
    private static final String EJBJAR = "META-INF/ejb-jar.xml";
    private static final String JONAS_EJBJAR = "META-INF/jonas-ejb-jar.xml";


    /** jonas-ejb-jar.xml */
    private JonasEjbJarXml jejbjar = null;


    /**
     * Build a new instance of a EjbJarManager.
     *
     * @param jarFilename the name of the ejbjar file
     * that we want to manage.
     *
     * @throws WsGenException if the jar file doesn't exist.
     */
    public EjbJarManager(String jarFilename) throws WsGenException {
        super(jarFilename);

        logger.log(BasicLevel.DEBUG, "Creating EjbJarManager for '" + jarFilename + "'");

        // get the ejbjar file DD.
        if (!isNewFile()) {
            
            // try to load META-INF/jonas-ejb-jar.xml
            try {
                JarFile jarFile = new JarFile(filename);
                InputStream is = jarFile.getInputStream(jarFile.getEntry(JONAS_EJBJAR));
                jejbjar = loadJonasEjbJar(is);
                is.close();
            } catch (IOException ioe) {
                String err =  "Cannot get '" + JONAS_EJBJAR 
                    + "' JarEntry from '" + filename + "'";
                if (logger.isLoggable(BasicLevel.ERROR)) {
                    logger.log(BasicLevel.ERROR, err);
                }

                throw new WsGenException(err, ioe);
            }
        } else {
            String err = "File '" + filename + "' does not exist." 
                + " Cannot create a default EjbJar file.";
            if (logger.isLoggable(BasicLevel.ERROR)) {
                logger.log(BasicLevel.ERROR, err);
            }
            
            throw new WsGenException(err);
        }

        // add xmls into save list
        addXml(JONAS_EJBJAR, jejbjar);
    }
 
    /**
     * Build a new instance of a EjbJarManager.
     *
     * @param jarFilename the name of the ejbjar file
     * that we want to manage.
     * @param dir output directory
     *
     * @throws WsGenException if the jar file doesn't exist.
     */
    public EjbJarManager(String jarFilename, String dir) throws WsGenException {
        this(jarFilename);
        setOutputDir(dir);
    }

    /**
     * Add a jonas-init-param for a specified service-ref
     *
     * @param name the bean name where put init-param
     * @param serviceRefName the service-ref-name (key)
     * @param initParamName the param name
     * @param initParamValue the param value
     *
     * @throws WsGenException
     */
    public void addJServiceRefInitParam(String beanName,
                                        String serviceRefName,
                                        String initParamName,
                                        String initParamValue)
        throws WsGenException {
        
        jejbjar.addJonasInitParam(beanName,
                                  serviceRefName,
                                  initParamName,
                                  initParamValue);
    }


    /**
     * Load the jonas-ejb-jar.xml file.
     * @param is the InputStream of the XML file.
     * @return a structure containing the result of the jonas-ejb-jar.xml parsing.
     * @throws WsGenException if the deployment descriptor is corrupted.
     */
    private static JonasEjbJarXml loadJonasEjbJar(InputStream is) 
        throws WsGenException { 
        return new JonasEjbJarXml(is); 
    }

    /**
     * Omit only jonas-ejb-jar.xml.
     * ejb-jar.xml is not omitted because not used by EjbJarManager.
     */
    protected boolean omit(String name) {
        return (name.equals(JONAS_EJBJAR));
    }
 
}
