/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public 
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Xavier Delplanque & Guillaume Sauthier
 * --------------------------------------------------------------------------
 * $Id: JarManager.java,v 1.3 2003/10/20 13:50:10 sauthieg Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ws.wsgen.jarfile;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;
import java.util.jar.JarOutputStream;
import java.util.zip.ZipEntry;
import java.util.zip.ZipException;

import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

import org.objectweb.jonas.common.Log;

import org.objectweb.jonas_ws.wsgen.WsGenException;
import org.objectweb.jonas_ws.wsgen.xml.Xml;
import org.objectweb.jonas_ws.wsgen.xml.XmlException;

import org.xml.sax.SAXParseException;

/**
 * Manage a jar file.
 * @author Xavier Delplanque
 */
public class JarManager {
 
    /** original jar filename */
    protected String filename;

    /** output directory */
    protected String outDir = ".";

    /** temporary file (copy of filename) */
    private JarFile copy;

    /** Map (Entry->File) */
    protected Map newEntries;

    /** xml objects to be marshalled (Entry->Xml) */
    private Map xmls;

    /** The logger to use */
    protected static Logger logger = Log.getLogger(Log.JONAS_WSGEN_PREFIX);

    /**
     * Build an instance of JarManager.
     * 
     * @param jarFilename the name of the jar file
     * that we want to manage.
     * @param destDir the name of the directory where final file will be created.
     *
     * @throws WsGenException if the  jar cannot be created
     */
    public JarManager(String jarFilename, String destDir) throws WsGenException {
        this(jarFilename);
        setOutputDir(destDir);
    }

    /**
     * Build a new instance of a JarManager.
     * Create the file if it doesn't exists.
     * By default, modify the file.
     *
     * @param jarFilename the name of the jar file
     * that we want to manage.
     *
     * @throws WsGenException if jar cannot be created
     */
    public JarManager(String jarFilename) throws WsGenException {

        logger.log(BasicLevel.DEBUG, "Creating JarManager for '" + jarFilename + "'");

        // init fields
        newEntries = new Hashtable();
        xmls = new Hashtable();
        copy = null;
        
        try {
            File jar = new File(jarFilename);
            filename = jar.getCanonicalPath();
            
            if (!jar.exists()) {
                createFile(jar);
            } else {
                makeCopy();
            }
            // set default out directory
            setOutputDir(new File(jarFilename).getCanonicalFile().getParent());

        } catch (IOException ioe) {
            throw new WsGenException("cannot find/open/create the jar file '" 
                                     + filename + "'", ioe);
        }
    }

    /**
     * Return the filename managed.
     * @return the filename managed.
     */
    public String getFilename() {
        return filename;
    }

    /**
     * Return the filename of the destination file.
     * @return the filename of the destination file.
     */
    public String getOutFilename() {
        // get the last part of the filename
        String name = new File(filename).getName();
        return outDir + File.separator + name;
    }

    public void setOutputDir(String dir) throws WsGenException {
        try {
            File d = new File(dir);
            outDir = d.getCanonicalPath();
        } catch (IOException ioe){
            throw new WsGenException(ioe);
        }
    }
 
    /**
     * Add all Files contained in a directory in the current jar file.
     * This adding keep the same file structure (directory & subdir).
     *
     * @param dir the directory to add
     *
     * @throws WsGenException when IOExceptio occurs?
     */
    public void addDir(File dir) 
        throws WsGenException {
        addDir ("", dir, "");
    }

    /**
     * Add all Files contained in a directory in the current jar file.
     * This adding keep the same file structure (directory & subdir).
     *
     * @param base entry base name.
     * @param dir the directory to add.
     *
     * @throws WsGenException when IOException occurs
     */
    public void addDir(String base, File dir) 
        throws WsGenException {
        addDir(base, dir, "");
    }

    /**
     * Add Files contained in a directory in the current jar file 
     * at the specified entryName location.
     * This adding keep the same file structure (directory & subdir).
     *
     * @param base where to put directory files in the jar
     * @param dir the directory to add
     * @param ends the filename accept pattern
     *
     * @throws WsGenException when IOException occurs
     */
    public void addDir(String base, File dir, String ends)
        throws WsGenException {

        try {
            File[] files = dir.listFiles();
            // add each repertory file and explore subdirectories
            for (int i = 0; i < files.length; i++) {
                File crtFile = files[i];
                if (crtFile.isFile()) {
                    // add the file in the jar if respect pattern
                    if (crtFile.getName().endsWith(ends)) {
                        addFile(base + crtFile.getName(), crtFile);
                    }
                } else {
                    // explore the subdirectory
                    addDir(base + crtFile.getName() + "/", crtFile, ends);
                }
            }
        } catch (Exception e) {
            throw new WsGenException("Error when adding files for directory '" + dir.getName() + "'", e);
        }
    }

    /**
     * Add a new File in the current jar file.
     *
     * @param entryName where to put the file in the jar
     * @param file the file to add
     *
     * @throws WsGenException when IOException occurs
     */
    public void addFile(String entryName, File file)
        throws WsGenException {

        logger.log(BasicLevel.DEBUG, "entry '" + entryName + "' -> file '" + file + "'");

        // get the zip entry for the given entry name
        String filename = file.getName();

        // add this entry in the entries
        try {
            newEntries.put(entryName, file);
        } catch (NullPointerException npe) {
            throw new WsGenException("Null Jar entry added '" + entryName + "'", npe);
        }
    }

    /**
     * Create the subdirectory structure of the File if needed.
     * @param file the file to create
     * @throws IOException if an IO error occurs
     * @throws WsGenException (needed when overridding)
     */
    private void createFile(File file) throws IOException, WsGenException {

        if (logger.isLoggable(BasicLevel.DEBUG)) {
            logger.log(BasicLevel.DEBUG, "Create File : '" + file.getPath() + "'");
        }
        File parent = file.getCanonicalFile().getParentFile();

        // if parent dir not existant
        // create the directories
        if (!parent.exists()) {
            parent.mkdirs();
        }

    }

    /**
     * build a message from SAX Exception in a consistent style, consistent
     * with emacs compile mode (same as grep, cc, javac, etc).
     * @param fileName name of the file
     * @param exception the SAX exception
     * @param msg the string message
     * @return a message which is in a consistent style
     */
    protected  static String getSAXMsg(String fileName,
                                       SAXParseException exception, String msg) {
        String ret = fileName + ":" + exception.getLineNumber() + ":" 
            + exception.getColumnNumber() + ": " + msg;
        return ret;
    }

    /**
     * Return true if a new jar file has been created.
     * @return true if a new jar file has been created.
     */
    protected boolean isNewFile() {
        return (copy == null);
    }

    /**
     * write a new entry in the jar output stream.
     * @param en the entry name
     * @param is the input stream to read.
     * @param jos the output stream to write.
     *
     * @throws WsGenException when writing entry fails.
     */
    protected static void writeEntry(String en, InputStream is, JarOutputStream jos) 
        throws WsGenException {

        try {
            jos.putNextEntry(new ZipEntry(en));
        } catch (ZipException zipe) {
            throw new WsGenException("Jar entry format error for '" + en + "'", zipe);
        } catch (IOException ioe) {
            throw new WsGenException("Error when adding jar entry '" + en + "'", ioe);
        }

        byte[] buf = new byte[1024];
        int length;

        // write the file in the jar
        try {
            while ((length = is.read(buf)) > 0) {
                jos.write(buf, 0, length);
            }
        } catch (Exception e) {
            throw new WsGenException("Error writing entry '" + en + "'", e);
        }

    }

    /**
     * Copy the jar into a tempo directory.
     *
     * @throws WsGenException when copy fails
     */
    private void makeCopy() throws WsGenException {
        // create a temporary jar file
        logger.log(BasicLevel.DEBUG, "make a copy of '" + filename + "'");

        // prepare the Streams
        FileOutputStream os;
        FileInputStream is;
        File temp = null;
        try {
            temp = File.createTempFile("wsgen", ".jar");
            os = new FileOutputStream(temp);
            is = new FileInputStream(filename);
        } catch (Exception e) {
            e.printStackTrace(System.out);
            throw new WsGenException("Error creating jar copy '" + copy.getName() + "'", e);
        }
        
        // copy the file
        byte[] buf = new byte[1024];
        int length;
        try {
            while ((length = is.read(buf)) > 0) {
                os.write(buf, 0, length);
            }
            is.close();
            os.flush();
            os.close();
            copy = new JarFile(temp);
        } catch (Exception e) {
            throw new WsGenException("Error writing file '" + copy.getName() + "'", e);
        }
    }

    /**
     * Add all jar entries in the jar file.
     * Save the file on the disk.
     * 
     * @throws WsGenException when save fails.
     */ 
    public void save() throws WsGenException {

        logger.log(BasicLevel.DEBUG, "saving '" + getOutFilename() + "'");

        // Create the OutputStream
        JarOutputStream os = null;
        try {
            os = new JarOutputStream(new FileOutputStream(getOutFilename()));
        } catch (IOException ioe) {
            throw new WsGenException(ioe);
        }

        logger.log(BasicLevel.DEBUG, "newEntries '" + newEntries + "'");

        // add file entries
        for (Iterator i = newEntries.keySet().iterator(); i.hasNext();) {
            String name = (String) i.next();
            File file = (File) (newEntries.get(name));
            try {
                logger.log(BasicLevel.DEBUG, "saving file '" + file + "'");
                writeEntry(name, new FileInputStream(file), os);
            } catch (Exception ex) {
                throw new WsGenException("Error when adding file '" + file.getName() + "' in jar", ex);
            }
        }
        
        logger.log(BasicLevel.DEBUG, "xmls '" + xmls + "'");
        // add xmls entries
        for (Iterator i = xmls.keySet().iterator(); i.hasNext();) {
            ZipEntry ze = (ZipEntry) (i.next());
            Xml xml = (Xml) (xmls.get(ze));

            // do not insert this xml entry if it does not have content
            if (!xml.isEmpty()) {
                try {
                    os.putNextEntry(ze);
                    xml.marshal(os);
                } catch (XmlException xe) {
                    throw new WsGenException("Error when adding xml '" + ze.getName() + "' in jar", xe);
                } catch (IOException ioe) {
                    throw new WsGenException("Error when adding xml '" + ze.getName() + "' in jar", ioe);
                }
            }
        }

        if (!isNewFile()) {
            
            // add temp jar entries
            try {
                // traverse all copy entries
                for (Enumeration e = copy.entries(); e.hasMoreElements();) {
                    ZipEntry ze = (ZipEntry) (e.nextElement());
                    
                    logger.log(BasicLevel.DEBUG, "entry '" + ze + "'");

                    // avoid twins entry
                    // priority for newEntries Entry
                    if (!newEntries.containsKey(ze.getName())) {
                        // if file not has to be omitted
                        if (!omit(ze.getName())) {
                            writeEntry(ze.getName(), copy.getInputStream(ze), os);
                        }
                    }
                }
            } catch (IOException ioe) {
                throw new WsGenException("Error when adding temporary jar '" 
                                         + copy.getName() + "' entries", ioe);
            }
        }

        // Close the Stream on filename
        try {
            os.close();
        } catch (IOException ioe) {
            throw new WsGenException("Error when closing '" + filename + "'", ioe);
        }

    }

    /**
     * Delete the copy
     */
    public void close () {
        // if deletion fails
        if (!isNewFile() && !(new File(copy.getName()).delete())) {
            logger.log(BasicLevel.INFO, "Error when delete temp file '" + copy.getName() + "'");
        }
    }


    /**
     * When instance destructed, close the Streams.
     */
    public void finalize() {
        if (copy != null) {
            close();
        }
    }

    /**
     * Omit filename yes/no ?.
     * Need to be overridden in sub classes.
     *
     * @param name name of entry to be tested
     * @return true if file must be omitted, otherwise false
     */
    protected boolean omit(String name) {
        return false;
    }

    /**
     * Add an Xml object to be marshalled when saving.
     * 
     * @param name entry name.
     * @param xml the Xml object.
     */
    protected void addXml(String name, Xml xml) {
        xmls.put(new ZipEntry(name), xml);
    }
}
