/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Xavier Delplanque & Guillaume Sauthier
 * --------------------------------------------------------------------------
 * $Id: WarManager.java,v 1.2 2003/10/17 15:56:00 sauthieg Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ws.wsgen.jarfile;

//import java
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.jar.JarFile;

import org.objectweb.util.monolog.api.BasicLevel;

import org.objectweb.jonas_ws.wsgen.WsGenException;
import org.objectweb.jonas_ws.wsgen.xml.JonasWebXml;
import org.objectweb.jonas_ws.wsgen.xml.WebXml;

/**
 * Use xml classes to get and write
 * informations inside a war file.
 *
 * @author Xavier Delplanque
 */
public class WarManager extends WsJarManager {

    // web Depl. Desc. filenames
    private static final String WEB = "WEB-INF/web.xml";
    private static final String JONAS_WEB = "WEB-INF/jonas-web.xml";


    /** jonas-web.xml */
    private JonasWebXml jweb = null;
    /** web.xml */
    private WebXml      web  = null;

    /**
     * Build a new instance of a WarManager.
     *
     * @param warFilename the name of the war file
     * that we want to manage.
     *
     * @throws WsGenException when instanciation fails.
     */
    public WarManager(String warFilename) throws WsGenException {
        super(warFilename);

        logger.log(BasicLevel.DEBUG, "Creating WarManager for '" + warFilename + "'");

        // get the war file DD.
        if (!isNewFile()) {

            // try to load WEB-INF/web.xml
            try {
                JarFile jarFile = new JarFile(filename);
                InputStream is = jarFile.getInputStream(jarFile.getEntry(WEB));
                web = loadWeb(is);
                is.close();
            } catch (IOException ioe) {
                if (logger.isLoggable(BasicLevel.WARN)) {
                    logger.log(BasicLevel.WARN, "Cannot get '" + WEB 
                               + "' JarEntry from '" + filename + "'\n"
                               + "will continue with default web.xml");
                }
                web = createDefaultWebDesc();
            }
            
            // try to load WEB-INF/jonas-web.xml
            try {
                JarFile jarFile = new JarFile(filename);
                InputStream is = jarFile.getInputStream(jarFile.getEntry(JONAS_WEB));
                jweb = loadJonasWeb(is);
                is.close();
            } catch (IOException ioe) {
                if (logger.isLoggable(BasicLevel.WARN)) {
                    logger.log(BasicLevel.WARN, "Cannot get '" + JONAS_WEB 
                               + "' JarEntry from '" + filename + "'\n"
                               + "will continue with default jonas-web.xml");
                }
                jweb = createDefaultJonasWebDesc();
            }
        } else {
            logger.log(BasicLevel.DEBUG, "'" + filename + "' is a new file :"
                       + "create defaults web descs (web / jonas-web).");
            web = createDefaultWebDesc();
            jweb = createDefaultJonasWebDesc();
        }
                   
        // add xmls into save list
        addXml(WEB, web);
        addXml(JONAS_WEB, jweb);
    }

    /**
     * Build a new instance of a WarManager.
     *
     * @param warFilename the name of the war file
     * that we want to manage.
     * @param dir output directory
     *
     * @throws WsGenException when instanciation fails.
     */
    public WarManager(String warFilename, String dir) throws WsGenException {
        this(warFilename);
        setOutputDir(dir);
    }


    /**
     * Add a new Servlet in the current war file.
     * Temporary (no tests name redundance : servlet-name).
     * @todo add ejb-ref / ejb-local-ref for minimum security
     *
     * @param sName the new servlet name
     * @param className the new servlet class name.
     * @param mapping mapping for the servlet
     *
     * @throws WsGenException
     */
    public void addServlet(String sName, String className, String mapping)
        throws WsGenException {

        // add the servlet in web.xml
        web.addServlet(sName, className);

        // add the servlet-mapping in web.xml
        web.addServletMapping(sName, mapping);
    }

    /**
     * Set a new Class for Servlet in the current war file.
     *
     * @param sName the new servlet name
     * @param className the new servlet class name.
     *
     * @throws WsGenException
     */
    public void setServletClass(String sName, String className)
        throws WsGenException {

        // add the servlet in web.xml
        web.setServletClass(sName, className);
    }

    /**
     * Add a jonas-init-param for a specified service-ref
     *
     * @param name the bean name where put init-param (unused)
     * @param serviceRefName the service-ref-name (key)
     * @param initParamName the param name
     * @param initParamValue the param value
     *
     * @throws WsGenException
     */
    public void addJServiceRefInitParam(String name,
                                        String serviceRefName,
                                        String initParamName,
                                        String initParamValue)
        throws WsGenException {

        logger.log(BasicLevel.DEBUG, "addJServiceRefInitParam (" + name + ", " 
                   + serviceRefName + ", " + initParamName + ", " + initParamValue + ")");
        
        jweb.addJonasInitParam(serviceRefName,
                               initParamName,
                               initParamValue);
    }

    /**
     * Add a new init-param for a specified servlet
     *
     * @param servletName the servlet-name (key)
     * @param initParamName the param name
     * @param initParamValue the param value
     *
     * @throws WsGenException
     */
    public void addServletInitParam(String servletName,
                                    String initParamName,
                                    String initParamValue)
        throws WsGenException {
        
        web.addServletInitParam(servletName,
                                initParamName,
                                initParamValue);
    }

    public void setServiceClass(String srn, String classname) {
        jweb.setServiceClass(srn, classname);
    }

    private WebXml createDefaultWebDesc() throws WsGenException {

        if (logger.isLoggable(BasicLevel.DEBUG)) {
            logger.log(BasicLevel.DEBUG, "Create Default web.xml for '" + filename + "'");
        }
  
        String jonasRoot = System.getProperty("install.root");
        if (jonasRoot == null) {
            throw new WsGenException("System property 'install.root' not set");
        }
        String path2Tmpl = new String(jonasRoot + File.separatorChar + "templates"
                                      + File.separatorChar + "wsgen");
        InputStream is = null;
        String filename = path2Tmpl + File.separatorChar + "web.xml";
        WebXml webapp = null;
        try {
            is = new FileInputStream(filename);
            webapp = loadWeb(is);
            is.close();
        } catch (Exception e) {
            throw new WsGenException("Error loading web dd '" + filename + "'", e);
        }

        return webapp;
    }
 
    private JonasWebXml createDefaultJonasWebDesc() throws WsGenException {
  
        if (logger.isLoggable(BasicLevel.DEBUG)) {
            logger.log(BasicLevel.DEBUG, "Create Default jonas-web.xml for '"
                       + filename + "'");
        }
  
        String jonasRoot = System.getProperty("install.root");
        if (jonasRoot == null) {
            throw new WsGenException("System property 'install.root' not set");
        }
        String path2Tmpl = new String(jonasRoot + File.separatorChar + "templates"
                                      + File.separatorChar + "wsgen");
        InputStream is = null;
        String filename = path2Tmpl + File.separatorChar + "jonas-web.xml";
        JonasWebXml jwebapp = null;
        try {
            is = new FileInputStream(filename);
            jwebapp = loadJonasWeb(is);
            is.close();
        } catch (Exception e) {
            throw new WsGenException("Error loading jonas-web dd '" + filename + "'", e);
        }
        return jwebapp;
    }
 
    /** 
     * Load the web.xml file.
     * @param is the InputStream of the XML file.
     * @return a structure containing the result of the web.xml parsing.
     * @throws WsGenException if the deployment descriptor 
     * is corrupted.
     */
    private static WebXml loadWeb(InputStream is) 
        throws WsGenException {
        return new WebXml(is);
    }

    /**
     * Load the jonas-web.xml file.
     * @param is the InputStream of the XML file.
     * @return a structure containing the result of the jonas-web.xml parsing.
     * @throws WsGenException if the deployment 
     * descriptor is corrupted.
     */
    private static JonasWebXml loadJonasWeb(InputStream is) 
        throws WsGenException { 
        return new JonasWebXml(is); 
    }

    /**
     * Omit web.xml and jonas-web.xml
     */
    protected boolean omit(String name) {
        return (name.equals(WEB) || name.equals(JONAS_WEB));
    }
}
