/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Xavier Delplanque & Guillaume Sauthier
 * --------------------------------------------------------------------------
 * $Id: JonasEjbJarXml.java,v 1.2 2003/10/20 13:50:10 sauthieg Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ws.wsgen.xml;

import org.w3c.dom.Document;
import org.w3c.dom.NodeList;
import org.w3c.dom.Node;
import org.w3c.dom.Element;

import java.io.InputStream;

/**
 * Wrapper class for managing jonas-ejb-jar.xml file.
 * @author Guillaume Sauthier
 */
public class JonasEjbJarXml extends JonasSpecificXml {

    private static final String JONAS_ENTITY_TAG  = "jonas-entity";
    private static final String JONAS_SESSION_TAG = "jonas-session";
    private static final String JONAS_MDB_TAG     = "jonas-message-driven-bean";

    /**
     * Build JonasEjbJarXml Object wrapping xml file of the InputStream.
     * @param is InputStream of a jonas-ejb-jar.xml file
     * @throws XmlException When creation fails.
     */
    public JonasEjbJarXml (InputStream is) throws XmlException {
        super(is);
    }

    /**
     * Set the implementing service class.
     * @param beanName the name of the EjbJar using service-ref
     * @param serviceRefName service-ref-name value
     * @param serviceClassName the service class implementing service-interface
     */
    public void setServiceClass(String beanName,
                                String serviceRefName,
                                String serviceClassName) {

        // find the jonas-service-ref
        Element jsr = findJonasServiceRef(beanName, serviceRefName);

        // set the service class for this element
        setServiceClass(jsr, serviceClassName);
    }

    /**
     * Add param name/value for the specified jonas-service-ref
     * @param beanName the name of the EjbJar using service-ref
     * @param serviceRefName name of the jonas-service-ref (key)
     * @param paramName name of the param
     * @param paramValue value of the param
     */
    public void addJonasInitParam(String beanName,
                                  String serviceRefName,
                                  String paramName,
                                  String paramValue) {
        
        // find the jonas-service-ref
        Element jsr = findJonasServiceRef(beanName, serviceRefName);
        
        // add the param in this element
        addJonasInitParam(jsr, paramName, paramValue);
    }

    /**
     * Return the specified jonas-service-ref element.
     * @param beanName name of the bean using WebServices (key)
     * @param serviceRefName name of the jonas-service-ref (key)
     * @return the jonas-service-ref element with specified name
     */
    private Element findJonasServiceRef(String beanName, String serviceRefName) {
        Document core = getDocument();
        Element root = core.getDocumentElement();

        Element jsr = null;

        // Search entities
        NodeList entities = getJonasEntityBeans();
        int size = entities.getLength();
        for (int i = 0; i < size; i++) {
            Element entity = (Element) entities.item(i);
            if (getBeanName(entity).equals(beanName)) {
                jsr = getJonasServiceRef(entity, serviceRefName);
            }
        }

        // Search sessions
        NodeList sessions = getJonasSessionBeans();
        size = sessions.getLength();
        for (int i = 0; i < size; i++) {
            Element session = (Element) sessions.item(i);
            if (getBeanName(session).equals(beanName)) {
                jsr = getJonasServiceRef(session, serviceRefName);
            }
        }

        // Search MessageDrivenBeans
        NodeList mdbs = getJonasMessageDrivenBeans();
        size = mdbs.getLength();
        for (int i = 0; i < size; i++) {
            Element mdb = (Element) mdbs.item(i);
            if (getBeanName(mdb).equals(beanName)) {
                jsr = getJonasServiceRef(mdb, serviceRefName);
            }
        }

        return jsr;

    }

    /**
     * Return the list of jonas-entity Element
     * @return the list of jonas-entity Element
     */
    private NodeList getJonasEntityBeans() {
        Document core = getDocument();
        Element root = core.getDocumentElement();
        return root.getElementsByTagName(JONAS_ENTITY_TAG);
    }

    /**
     * Return the list of jonas-session Element
     * @return the list of jonas-session Element
     */
    private NodeList getJonasSessionBeans() {
        Document core = getDocument();
        Element root = core.getDocumentElement();
        return root.getElementsByTagName(JONAS_SESSION_TAG);
    }

    /**
     * Return the list of jonas-message-driven-bean Element
     * @return the list of jonas-message-driven-bean Element
     */
    private NodeList getJonasMessageDrivenBeans() {
        Document core = getDocument();
        Element root = core.getDocumentElement();
        return root.getElementsByTagName(JONAS_MDB_TAG);
    }

    /**
     * Return the ejb-name of the element. Notice that Element must
     * be jonas-session/jonas-entity/jonas-message-driven-bean !
     * @param e the jonas-* element
     * @return the bean name
     */
    private String getBeanName(Element e) {
        // access ejb-name
        Node ejbName = e.getFirstChild();
        // access ejb-name TextNode value
        return ejbName.getFirstChild().getNodeValue();
    }

}
