/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Xavier Delplanque & Guillaume Sauthier
 * --------------------------------------------------------------------------
 * $Id: JonasSpecificXml.java,v 1.1 2003/10/16 14:55:52 sauthieg Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ws.wsgen.xml;

import org.w3c.dom.Document;
import org.w3c.dom.NodeList;
import org.w3c.dom.Node;
import org.w3c.dom.Element;

import java.io.InputStream;

/**
 * Abstract class for managing jonas specific DD file.
 * (jonas-web.xml / jonas-ejb-jar.xml)
 * @author Guillaume Sauthier
 */
public abstract class JonasSpecificXml extends Xml {

    private static final String JONAS_SERVICE_REF_TAG = "jonas-service-ref";
    private static final String SERVICE_REF_NAME_TAG  = "service-ref-name";
    private static final String JONAS_INIT_PARAM_TAG  = "jonas-init-param";
    private static final String JONAS_PARAM_VALUE_TAG = "param-value";
    private static final String JONAS_PARAM_NAME_TAG  = "param-name";

    private static final String JONAS_SERVICE_CLASS  = "serviceClassName";

    
    /**
     * Build JonasSpecificXml Object wrapping xml file of the InputStream.
     * @param is InputStream of a jonas-*.xml file
     * @throws XmlException When creation fails.
     */
    public JonasSpecificXml (InputStream is) throws XmlException {
        super(is);
    }

    /**
     * Helper method used to retrieve the Element 
     * jonas-service-ref from base Element.
     * @param base the Element where we search childs.
     * @param srn service-ref-name (key)
     * @return the Element representing the right jonas-service-ref.
     */
    protected static Element getJonasServiceRef(Element base, String srn) {

        // get base childs who are jonas-service-ref
        NodeList nl = base.getElementsByTagName(JONAS_SERVICE_REF_TAG);

        Element my_jsr = null;

        int size = nl.getLength();
        for (int i = 0; i < size && my_jsr == null; i++) {
            Element jsr = (Element) nl.item(i);
            NodeList childs = jsr.getElementsByTagName(SERVICE_REF_NAME_TAG);
            // get service-ref-name value
            if (childs.item(0).getFirstChild().getNodeValue().equals(srn)) {
                // found the right jonas-service-ref
                my_jsr = jsr;
            }
        }

        return my_jsr;
        
    }

    /**
     * Set the implementing service class.
     * @param jsr jonas-service-ref element
     * @param serviceClassName the service class implementing service-interface
     */
    protected void setServiceClass(Element jsr,
                                   String serviceClassName) {
        addJonasInitParam(jsr, JONAS_SERVICE_CLASS, serviceClassName);
    }

    
    /**
     * Add param name/value as childs of base Element
     * @param base "father" element of newly created jonas-init-param
     * @param name name of the param
     * @param value value of the param
     */
    protected void addJonasInitParam(Element base, String name, String value) {

        Document core = getDocument();
        
        // Build Nodes
        /*
          <jonas-init-param>
           <param-name>#name#</param-name>
           <param-value>#value#</param-value>
          </jonas-init-param>
         */
        Node pNameTxt      = core.createTextNode(name);
        Node pValueTxt     = core.createTextNode(value);
        Element jInitParam = core.createElement(JONAS_INIT_PARAM_TAG);
        Element pName      = core.createElement(JONAS_PARAM_NAME_TAG);
        Element pValue     = core.createElement(JONAS_PARAM_VALUE_TAG);

        pName.appendChild(pNameTxt);
        pValue.appendChild(pValueTxt);
        jInitParam.appendChild(pName);
        jInitParam.appendChild(pValue);

        // insert
        Element first = getFirstJonasInitParam(base);
        base.insertBefore(jInitParam, first);
    }


    /**
     * Return the first jonas-init-param of base element.
     * @param base the element where to perform the search.
     * @return the first jonas-init-param of base element.
     */
    private Element getFirstJonasInitParam(Element base) {
        NodeList nl = base.getElementsByTagName(JONAS_INIT_PARAM_TAG);
        return (Element) nl.item(0);
    }

}
