/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Xavier Delplanque & Guillaume Sauthier
 * --------------------------------------------------------------------------
 * $Id: WebXml.java,v 1.3 2003/10/16 14:55:52 sauthieg Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ws.wsgen.xml;

import org.w3c.dom.Document;
import org.w3c.dom.NodeList;
import org.w3c.dom.Node;
import org.w3c.dom.Element;

import java.io.InputStream;

/**
 * Wrapping class for managing a web.xml file.
 * Warning no Error Support (servlet not found, ...)
 * @author Guillaume Sauthier
 */
public class WebXml extends Xml {

    private static final String SERVLET_TAG       = "servlet";
    private static final String SERVLET_NAME_TAG  = "servlet-name";
    private static final String SERVLET_CLASS_TAG = "servlet-class";

    private static final String MAP_TAG      = "servlet-mapping";
    private static final String MAP_URL_TAG  = "url-pattern";
    private static final String MAP_NAME_TAG = "servlet-name";

    private static final String INIT_PARAM_TAG  = "init-param";
    private static final String PARAM_VALUE_TAG = "param-value";
    private static final String PARAM_NAME_TAG  = "param-name";
    
    /**
     * Build WebXml Object wrapping xml file of the InputStream.
     * @param is InputStream of a web.xml file
     * @throws XmlException When creation fails.
     */
    public WebXml (InputStream is) throws XmlException {
        super(is);
    }

    /**
     * Add servlet mapping in web.xml
     * @param servName servlet-name value
     * @param url url-pattern value
     */
    public void addServletMapping(String servName, String url) {

        Document core = getDocument();
        Element root = core.getDocumentElement();

        Element followingNode = findFirstElement(root, new String[] {MAP_TAG,
                                                                     "session-config",
                                                                     "mime-mapping",
                                                                     "welcome-file-list",
                                                                     "error-page",
                                                                     "taglib",
                                                                     "resource-env-ref",
                                                                     "resource-ref",
                                                                     "security-constraint",
                                                                     "login-config",
                                                                     "security-role",
                                                                     "env-entry",
                                                                     "ejb-ref",
                                                                     "ejb-local-ref"});
        // build nodes
        /*
          <servlet-mapping>
           <servlet-name>#servName#</servlet-name>
           <url-pattern>#url#</url-pattern>
          </servlet-mapping>
         */
        Node servNameTxt = core.createTextNode(servName);
        Node urlTxt = core.createTextNode(url);
        Element servNameE = core.createElement(MAP_NAME_TAG);
        Element urlE = core.createElement(MAP_URL_TAG);
        Element newNode = core.createElement(MAP_TAG);

        servNameE.appendChild(servNameTxt);
        urlE.appendChild(urlTxt);
        newNode.appendChild(servNameE);
        newNode.appendChild(urlE);
        root.insertBefore(newNode, followingNode);
    }

    /**
     * Add servlet in web.xml
     * @param servName servlet-name value
     * @param servClass servlet-class value
     */
    public void addServlet(String servName, String servClass) {

        Document core = getDocument();
        Element root = core.getDocumentElement();

        Element followingNode = findFirstElement(root, new String[] {SERVLET_TAG,
                                                                     MAP_TAG,
                                                                     "session-config",
                                                                     "mime-mapping",
                                                                     "welcome-file-list",
                                                                     "error-page",
                                                                     "taglib",
                                                                     "resource-env-ref",
                                                                     "resource-ref",
                                                                     "security-constraint",
                                                                     "login-config",
                                                                     "security-role",
                                                                     "env-entry",
                                                                     "ejb-ref",
                                                                     "ejb-local-ref"});

        // build nodes
        /*
          <servlet>
           <servlet-name>#servName#</servlet-name>
           <servlet-class>#servClass#</servlet-class>
          </servlet>
         */
        
        Node servNameTxt = core.createTextNode(servName);
        Node servClassTxt = core.createTextNode(servClass);
        Element servNameE = core.createElement(SERVLET_NAME_TAG);
        Element servClassE = core.createElement(SERVLET_CLASS_TAG);
        Element newNode = core.createElement(SERVLET_TAG);

        servNameE.appendChild(servNameTxt);
        servClassE.appendChild(servClassTxt);
        newNode.appendChild(servNameE);
        newNode.appendChild(servClassE);
        root.insertBefore(newNode, followingNode);

    }

    /**
     * Change the servlet-class value of the specified servlet.
     * @param sName servlet name
     * @param sClass new fully qualified classname
     */
    public void setServletClass(String sName, String sClass) throws XmlException {
        
        // find the servlet
        Element servlet = findServlet(sName);
        NodeList nl = servlet.getElementsByTagName(SERVLET_CLASS_TAG);

        Element servletClass = null;
        // no servlet-class found
        if (nl.getLength() == 0) {
            String err = "Servlet '" + sName + "' contains no servlet-class element";
            throw new XmlException(err);
        } else {
            servletClass = (Element) nl.item(0);
        }

        // set the new classname
        servletClass.getFirstChild().setNodeValue(sClass);
    }
    
    /**
     * Add servlet init-param in web.xml
     * @param servName servlet-name value
     * @param pName param-name value
     * @param pValue param-value value
     */
    public void addServletInitParam(String servName, 
                                    String pName,
                                    String pValue) throws XmlException {
        
        Document core = getDocument();

        // build nodes
        /*
          <init-param>
           <param-name>#pName#</param-name>
           <param-value>#pValue#</param-value>
          </init-param>
         */
        Element servlet = findServlet(servName);

        Node pNameTxt = core.createTextNode(pName);
        Node pValueTxt = core.createTextNode(pValue);
        Element pNameE = core.createElement(PARAM_NAME_TAG);
        Element pValueE = core.createElement(PARAM_VALUE_TAG);
        Element initParam = core.createElement(INIT_PARAM_TAG);

        pNameE.appendChild(pNameTxt);
        pValueE.appendChild(pValueTxt);
        initParam.appendChild(pNameE);
        initParam.appendChild(pValueE);

        Element followingNode = findFirstElement(servlet, new String[] {"init-param",
                                                                        "load-on-startup",
                                                                        "run-as",
                                                                        "security-role-ref"});

        servlet.insertBefore(initParam, followingNode);

    }

    /**
     * Return the servlet Element with the specified servlet-name 
     * (can be null if servlet-name not found).
     * @param sName the searched servlet name.
     * @return the servlet Element with the specified servlet-name.
     */
    private Element findServlet(String sName) throws XmlException {
        
        Document core = getDocument();
        Element root = core.getDocumentElement();
        NodeList nl = root.getElementsByTagName(SERVLET_TAG);
        
        Element servlet = null;
        
        // Search the servlet Element with right servlet-name value
        int size = nl.getLength();
        for (int i = 0; i < size && servlet == null; i++) {
            Element e = (Element) nl.item(i);
            // servlet-name is mandatory and unique in servlet
            Node sn = e.getElementsByTagName(SERVLET_NAME_TAG).item(0);
            String servletName = sn.getFirstChild().getNodeValue();
            if (servletName.equals(sName)) {
                servlet = e;
            }
        }

        if (servlet == null) {
            String err = "Servlet '" + sName + "' not found in web.xml";
            throw new XmlException(err);
        }
        
        return servlet;
    }

    private static Element findFirstElement(Element base, String[] names) {

        NodeList nl = null;
        Element first = null;

        // test each name
        for (int i = 0; i < names.length && first == null; i++) {
            // get the list of items
            nl = base.getElementsByTagName(names[i]);
            // found an element with name name
            if (nl.getLength() != 0) {
                first = (Element) nl.item(0);
            }
        }

        return first;
        
    }
    
}
