/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Jeff Mesnil.
 *
 * --------------------------------------------------------------------------
 * $Id: SecurityContext.java,v 1.5 2003/09/25 13:16:11 joaninh Exp $
 * --------------------------------------------------------------------------
 */



package org.objectweb.security.context;

import java.io.Serializable;
import java.security.Principal;
import java.util.ArrayList;
import java.util.Iterator;

/**
 * Implementation of the JOnAS Security Context
 * CAUTION: Don't forget to modify the marshalling and the unmarshalling in the
 * org.objectweb.jonas.security.interceptors.iiop.SecurityInterceptor class
 * when adding attributes in SecurityContext.
 * @author Jeff Mesnil (initial developer)
 * @author Florent Benoit 
 */

public class SecurityContext implements Serializable {

    /**
     * Name of the principal
     */
    private String principalName = null ;

    /**
     * List of roles of the principal
     */
    private final String[] roles;

    /**
     * run-as stack
     * It uses ArrayList instead of java.util.Stack class
     * This is for use ArrayList and not Vector as Stack implementation
     */
    private ArrayList runAsStack = null;

    /**
     * Constructor SecurityContext
     * use the default principal name
     *
     */
    public SecurityContext () {
        this.principalName = "JOnAS_client";
        this.roles = null;
    }

    /**
     * Constructor SecurityContext
     * @param principalName principal name
     * @param roles the roles of the principal
     */
    public SecurityContext (String principalName, String[] roles) {
        this.principalName = principalName;
        this.roles = roles;
    }

    /**
     * Constructor SecurityContext
     * @param principalName principal name
     * @param arrayRoles the list of the roles of the principal
     */
    public SecurityContext (String principalName, ArrayList arrayRoles) {
        this(principalName, arrayRoles, null);
    }

    /**
     * Constructor SecurityContext
     * @param principalName principal name
     * @param arrayRoles the list of the roles of the principal
     * @param arrayRunas the RunAs stack
     */
    public SecurityContext (String principalName, ArrayList arrayRoles, ArrayList arrayRunas) {
        this.principalName = principalName;
        if (arrayRoles != null) {
            //Convert list into array
            String[] r = new String[arrayRoles.size()];
            r = (String[]) arrayRoles.toArray(r);
            this.roles = r;
        } else {
            this.roles = null;
        }
        this.runAsStack = arrayRunas;
    }


    /**
     * Constructor SecurityContext
     * @param principalName principal name
     *
     */
    public SecurityContext (String principalName) {
        this.principalName = principalName;
        this.roles = null;
    }


    /**
     * Method getCallerPrincipal
     * @return the Principal in the Security Context
     *
     */
    public Principal getCallerPrincipal() {
        return new Principal () {
                public String getName() {
                    return principalName;
                }
                
                public boolean equals (Object o) {
                    if (o instanceof Principal) {
                        return principalName.equals(((Principal) o).getName());
                    }
                    return false;
                }

                public String toString () {
                    return "name = " + principalName;
                }
            };
    }

    /**
     * Return the roles of the principal
     * @return roles
     */
    public String[] getCallerPrincipalRoles() {
        return roles;
    }

    /**
     * Push : Pushes an item onto the top of this stack.
     * @param role the role to add on top of the stack
     */
    public void pushRunAs(String role) {
        getRunAsStack().add(role);
    }

    /**
     * Pop : Removes the object at the top of the run-as role stack
     */
    public void popRunAs() {
        if (!getRunAsStack().isEmpty()) {
            getRunAsStack().remove(getRunAsStack().size() - 1);
        }
    }

    /**
     * Peek : Looks at the object at the top of this stack without removing it from the stack.
     * @return the role at the top of the stack
     */
    public String peekRunAs() {
        if (getRunAsStack().isEmpty()) {
            return null;
        } else {
            return (String) getRunAsStack().get(getRunAsStack().size() - 1);
        }
    }


    /**
     * Gets the stack which manage the run-as
     * @return the stack which manage the run-as
     */
    public ArrayList getRunAsStack() {
        if (runAsStack == null) {
            runAsStack = new ArrayList();
        }
        return runAsStack;
    }
    
    /**
     * Method toString
     * @return String a string representation of the object
     */
    public String toString () {
        String txt =  "principal : name = " + principalName + "\n";
        if (roles != null) {
            for (int i = 0; i < roles.length; i++) {
                txt += "role[" + i + "] = " + roles[i] + "\n";
            }
        }
        if (runAsStack != null) {
            Iterator iRunas = runAsStack.iterator();
            int i = 0;
            while (iRunas.hasNext()) {
                txt += "runas[" + i + "] = " + ((String) iRunas.next()) + "\n";
            }
        }
        return txt;
    }
 
}

