/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or 1any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer: Florent BENOIT
 * --------------------------------------------------------------------------
 * $Id: F_JonasAdminDeploy.java,v 1.2 2003/09/26 14:27:43 joaninh Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.examples.clients.jonasadmin;

import com.meterware.httpunit.*;
import com.meterware.servletunit.*;
import junit.framework.*;

import org.objectweb.jonas.examples.util.JExampleTestCase;

/**
 * Define a class to test the JOnAS Admin console
 * Test authentication and deploy/undeploy ears
 * @author Florent Benoit
 */
public class F_JonasAdminDeploy extends JExampleTestCase {


    /**
     * URL of the login page
     */
    private static final String URL_JONASADMIN = "/jonasAdmin/";

    /**
     * URL of the deployment of EARs
     */
    private static final String URL_JONASADMIN_DEPLOYEAR = "/jonasAdmin/EditDeploy.do?type=ear";

    /**
     * Main method
     * @param args the arguments
     */
    public static void main(String[] args) {

        String testtorun = null;
        // Get args
        for (int argn = 0; argn < args.length; argn++) {
            String sArg = args[argn];
            if (sArg.equals("-n")) {
                testtorun = args[++argn];
            }
        }
        if (testtorun == null) {
            junit.textui.TestRunner.run(suite());
        } else {
            junit.textui.TestRunner.run(new F_JonasAdminDeploy(testtorun));
        }
    }

    /**
     * Get a new TestSuite for this class
     * @return a new TestSuite for this class
     */
    public static TestSuite suite() {
        return new TestSuite(F_JonasAdminDeploy.class);
    }

    /**
     * Setup need for these tests
     * jonasAdmin is required
     * @throws Exception if it fails
     */
    protected void setUp() throws Exception {
        super.setUp();
        useWar("jonasAdmin");
    }


    /**
     * Constructor with a specified name
     * @param s name
     */
    public F_JonasAdminDeploy(String s) {
        super(s, URL_JONASADMIN);
    }

    /**
     * Try to authenticate with a bad login/password
     * @throws Exception if an error occurs
     */
    public void testTryBadAuth() throws Exception {
        WebResponse wRes = doAuth("bad", "bad");
        String txt = wRes.getText();
        
        if (txt.indexOf("Invalid username or password") == -1) {
            fail("The login/password bad/bad can't be a valid l/p.");
        }
    }

    /**
     * Try to authenticate with a right login/password
     * @throws Exception if an error occurs
     */
    public void testTryWithGoodAuth() throws Exception {
        WebResponse wRes = doValidAuth();
        String txt = wRes.getText();
        
        if (txt.indexOf("Invalid username or password") != -1) {
            fail("The login/password jadmin/jonas must be valid.");
        }
    }

    /**
     * Return the text of the deployed table or the deployable table
     * @param wr the WebResponse to use
     * @param deployed if true return the text deployed table, if false text of deployable
     * @return the text of the deployed table or the deployable table
     * @throws Exception if an error occurs
     */
    private String getDeployTable(WebResponse wr, boolean deployed) throws Exception {

        //Extract right table
        WebTable bigTable = wr.getTables()[0];
        WebTable firstTable = bigTable.getTableCell(1, 0).getTables()[0];
        WebTable cellTable = firstTable.getTableCell(0, 0).getTables()[0];
        WebTable subTable = cellTable.getTableCell(0, 0).getTables()[0];

        TableCell deployableCell = subTable.getTableCell(1, 0);
        TableCell deployedCell = subTable.getTableCell(1, 2);
    
        if (deployed) {
            return deployedCell.asText();
        } else {
            return deployableCell.asText();
        }
    }


    /**
     * Deploy the ear with the testcase method and then, try to undeploy the ear with jonaAdmin
     * If it was undeployed, deploy it again with jonasAdmin
     * @param name of the ear (without extension) to test
     * @throws Exception if an error occurs
     */ 
    private void unDeployAndDeployEar(String name) throws Exception {
        String fileName = name + ".ear";
        String url = URL_JONASADMIN_DEPLOYEAR;

        //First load Ear
        useEar(name);

        WebResponse wr = wc.getResponse(getAbsoluteUrl(url));
        String txt = getDeployTable(wr, true);

        if (txt.indexOf(fileName) == -1) {
            fail("The file" + fileName + " was not found as deployed.");
        }

        WebForm[] webForms = wr.getForms();
        WebForm webForm = webForms[0];

        String params = webForm.getParameterValue("undeploy");
        WebForm.Scriptable script = webForm.getScriptableObject();

        if (params.length() == 0) {
            params += fileName;
        } else {
            params += "," + fileName;
        }

        script.setParameterValue("undeploy", params);

        WebResponse submitUndeploy = webForm.submit();

        //alarm is in the table
        webForms = submitUndeploy.getForms();
        webForm = webForms[0];

        WebResponse endResp = webForm.submit();

        wr = wc.getResponse(getAbsoluteUrl(url));
        txt = getDeployTable(wr, true);

        if (txt.indexOf(fileName) != -1) {
            fail("The jonasAdmin webApp has not remove " + fileName);
        }

        //now deploy
        webForms = wr.getForms();
        webForm = webForms[0];

        params = webForm.getParameterValue("deploy");
        script = webForm.getScriptableObject();

        if (params.length() == 0) {
            params += fileName;
        } else {
            params += "," + fileName;
        }

        script.setParameterValue("deploy", params);

        submitUndeploy = webForm.submit();

        //alarm is in the table
        webForms = submitUndeploy.getForms();
        webForm = webForms[0];

        endResp = webForm.submit();

        wr = wc.getResponse(getAbsoluteUrl(url));
        txt = getDeployTable(wr, true);

        if (txt.indexOf(fileName) == -1) {
            fail("The file " + fileName + " was not found as deployed.");
        }
    }

    /**
     * Deploy the ear with the testcase method and then, try to undeploy the ear with jonaAdmin
     * If it was undeployed, deploy it again with jonasAdmin
     * Test with ears alarm and cmp2
     * @throws Exception if an error occurs
     */ 
    public void testTryDeployUndeployEars() throws Exception {
        doValidAuth();
        unDeployAndDeployEar("alarm");
        unDeployAndDeployEar("cmp2");
    }


    /**
     * Authenticate with a valid login/password
     * @return the response
     * @throws Exception if an error occurs
     */ 
    private WebResponse doValidAuth() throws Exception {
        return doAuth("jonas", "jonas");
    }

    /**
     * Authenticate with the specified login/password
     * @param login login to use
     * @param password password to use
     * @return the response after authentication
     * @throws Exception if an error occurs
     */ 
    private WebResponse doAuth(String login, String password) throws Exception {
        WebResponse wr = wc.getResponse(url);
        WebForm[] webForms = wr.getForms();
        WebForm webForm = webForms[0];

        webForm.setParameter("j_username", login);
        webForm.setParameter("j_password", password);
        return webForm.submit();
        
    }


}
