/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or 1any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer: Florent BENOIT
 * --------------------------------------------------------------------------
 * $Id: A_WebServicesEndpoint.java,v 1.3 2003/10/23 12:18:22 benoitf Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.examples.clients.webservices;


import com.meterware.httpunit.*;
import com.meterware.servletunit.*;
import java.io.File;
import junit.framework.*;
import org.objectweb.jonas.common.JProp;


import javax.wsdl.Definition;
import javax.wsdl.factory.WSDLFactory;
import javax.wsdl.xml.WSDLReader;
import javax.wsdl.Port;
import javax.wsdl.Service;
import javax.wsdl.extensions.soap.SOAPAddress;

import javax.xml.namespace.QName;

/**
 * Define a class to test the webServices examples wsWarExample
 * Test WebServices deployment OK and test Call on deployed WebService
 * @author Guillaume Sauthier
 */
public abstract class A_WebServicesEndpoint extends A_WebServices {

    /**
     * property name => file1 WSDLHandler is mandatory for this test !
     */
    private static final String JPROP_WSDL_LOCATION = "jonas.service.ws.file1.location";

    public A_WebServicesEndpoint(String s, String urlPrefix) {
        super(s, urlPrefix);
    }

    /**
     * Check if Axis is Running
     * @param idServicesLink indicate where is the link to /services in the homepage
     * @param portName the name of the port we want to check
     * @throws Exception if an error occurs
     */
    public void checkAxisServicesAccessible(int idServicesLink, String portName) throws Exception {

        WebResponse wr = wc.getResponse(url + "index.html");
        WebLink services = wr.getLinks()[idServicesLink];

        // click the link
        WebResponse swr = services.click();

        // traverse the links and find the interresting one.
        boolean found = false;
        for (int i = 0; i < swr.getLinks().length; i++) {
            String ws_url = swr.getLinks()[i].getURLString();
            if (ws_url.equals(url + "services/" + portName + "?wsdl"))
                found = true;
            if (ws_url.equals(url + "services/" + portName + "?WSDL"))
                found = true;
        }
        if (!found)
            fail("endpoint '" + portName + "' not found in /services");

    }

    /**
     * Check if WSDL is generated
     * @param idWSDLServiceLink indicate where we can find the link 
     *        to WSDL of the port we want to test. (first, second, ...)
     * @throws Exception if an error occurs
     */
    public void checkAxisWSDL(int idWSDLServiceLink) throws Exception {

        WebResponse wr = wc.getResponse(url + "index.html");
        WebLink wsdl = wr.getLinks()[idWSDLServiceLink];

        // click the link
        WebResponse wr2 = wsdl.click();
        assertEquals("WSDL OK", "text/xml", wr2.getContentType());

    }

    /**
     * Check if deployed WebService is running
     * @param formPage the filename of the webpage containing the form
     * @param endpoint the URL where local service can be found
     * @param name the name to set as parameter
     * @param pageTitle title of the returned page
     * @throws Exception if an error occurs
     */
    public void checkDeployedWebService(String formPage,
                                        String endpoint,
                                        String name,
                                        String pageTitle) throws Exception {

        // Get the page containing the form
        WebResponse wr = wc.getResponse(url + formPage);

        // get the form named "prepare"
        WebForm form = wr.getFormWithName("prepare");
        // set all parameters
        form.setParameter("endpoint", endpoint);
        form.setParameter("name", name);
        // submit
        WebResponse wr2 = form.submit();

        String text = wr2.getText();

        String valTitle = "<title>" + pageTitle + "</title>";
        String valURL = "Working with URL : <i>" + endpoint + "</i><br/>";
        String valHello = "<b>result of sayHello(name) method :</b><i>Hello " + name + "</i><br/>";
        String valQuotes = "<b>result of getCotes() method :</b><i>12</i><br/>";
        boolean testOK = false;

        if ((text.indexOf(valTitle) != -1)
            && (text.indexOf(valURL) != -1)
            && (text.indexOf(valURL) != -1)
            && (text.indexOf(valURL) != -1)) {
            testOK = true;
        }

        if (!testOK) {
            fail("Call fails");
        }

    }


    /**
     * Check if WSDL has been published.
     * @param wsdlFilename the filename of the WSDL file
     * @param tns targetNameSpace of the Definition
     * @param serviceName the localpart of the Service QName
     * @param portName the name of the WSDL' Port
     * @param endpoint the expected URL
     * @throws Exception if an error occurs
     */
    public void checkWSDLPublication(String wsdlFilename,
                                     String tns,
                                     String serviceName,
                                     String portName,
                                     String endpoint) throws Exception {

        JProp jp = JProp.getInstance();

        String wsdlLoc = jp.getValue(JPROP_WSDL_LOCATION);

        File wsdl = new File(wsdlLoc, wsdlFilename);

        assertTrue("WSDL exists", wsdl.exists());
        assertTrue("WSDL is copied", wsdl.isFile());

        WSDLFactory f = WSDLFactory.newInstance();
        WSDLReader r = f.newWSDLReader();
        
        r.setFeature("javax.wsdl.verbose", false);
        r.setFeature("javax.wsdl.importDocuments", true);

        Definition def = r.readWSDL(wsdl.getPath());
        Service s = def.getService(new QName(tns, serviceName));
        Port p = s.getPort(portName);
        SOAPAddress sa = (SOAPAddress) p.getExtensibilityElements().get(0);

        assertEquals("URL updated", endpoint, sa.getLocationURI());

    }


}
