/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
*/



package org.objectweb.jonas_ejb.deployment.lib;

import junit.framework.TestCase;

import org.objectweb.jonas_ejb.deployment.api.BeanDesc;
import org.objectweb.jonas_ejb.deployment.api.DeploymentDesc;
import org.objectweb.jonas_ejb.deployment.api.DeploymentDescException;

import org.objectweb.jonas_ws.deployment.api.HandlerRef;
import org.objectweb.jonas_ws.deployment.api.MappingFile;
import org.objectweb.jonas_ws.deployment.api.PortComponentRef;
import org.objectweb.jonas_ws.deployment.api.ServiceRef;
import org.objectweb.jonas_ws.deployment.api.WSDLFile;

import java.io.File;
import java.io.FileInputStream;

import java.net.URL;
import java.net.URLClassLoader;

import javax.xml.namespace.QName;


/**
 * Integration tests for WebManager. (tests for ServiceRef objects)
 */
public class EjbDeploymentDescManagerTest extends TestCase {
    EjbDeploymentDescManager ejbDDManager = null;
    private String           path = null;

    /**
     * Constructor
     *
     * @param name ?
     */
    public EjbDeploymentDescManagerTest(String name) {
        super(name);
        path = System.getProperty("base.dir");

    }

    /**
     * return an instance of Web Manager
     *
     * @return DOCUMENT ME!
     *
     * @throws Exception ?
     */
    public EjbDeploymentDescManager createInstance() throws Exception {
        return EjbDeploymentDescManager.getInstance();

    }

    protected void setUp() throws Exception {
        super.setUp();
        ejbDDManager = createInstance();

    }

    protected void tearDown() throws Exception {
        ejbDDManager = null;
        super.tearDown();

    }

    /**
     * Tests if the right exception is thrown when a WSDL is not correctly
     * specified in the webDD.
     *
     * @throws Exception ?
     */
    public void testWsdlFileMissing() throws Exception {
        boolean isOk = false;

        try {
            DeploymentDesc dd = getDeplDesc("output/ejbjars/EjbError1.jar");

        } catch (DeploymentDescException wcdde) {
            String err = "Cannot get wsdl file:";
            isOk = wcdde.getMessage().startsWith(err);

        }

        assertTrue("DeploymentDescException thrown : EjbError1.jar", isOk);

    }

    /**
     * Tests if the right exception is thrown when a Mapping file is not
     * correctly specified in the webDD.
     *
     * @throws Exception ?
     */
    public void testMappingFileMissing() throws Exception {
        boolean isOk = false;

        try {
            DeploymentDesc dd = getDeplDesc("output/ejbjars/EjbError2.jar");

        } catch (DeploymentDescException wcdde) {
            String err = "Cannot read the jaxrpc-mapping-file ";
            isOk = wcdde.getMessage().startsWith(err);

        }

        assertTrue("DeploymentDescException thrown : EjbError2.jar", isOk);

    }

    /**
     * Tests if the right exception is thrown when the service-qname specified
     * in webDD do not match the WSDL
     *
     * @throws Exception ?
     */
    public void testQNameNotInWsdl() throws Exception {
        boolean isOk = false;

        try {
            DeploymentDesc dd = getDeplDesc("output/ejbjars/EjbError3.jar");

        } catch (DeploymentDescException wcdde) {
            String err = "service-qname '{http://ssbendpoint.beans.ws.objectweb.org}BadName' " + "must be defined in WSDL : 'META-INF/wsdl/jarendpoint.wsdl'";
            isOk = wcdde.getMessage().startsWith(err);

        }

        assertTrue("DeploymentDescException thrown : EjbError3.jar", isOk);

    }

    /**
     * Tests if the right exception is thrown when the unicity of the handler
     * name is not respected.
     *
     * @throws Exception ?
     */
    public void testHandlerNameNotUnique() throws Exception {
        boolean isOk = false;

        try {
            DeploymentDesc webDD = getDeplDesc("output/ejbjars/EjbError4.jar");

        } catch (DeploymentDescException wcdde) {
            String err = "handler-name 'SimpleHandler' must be unique in the module";
            isOk = wcdde.getMessage().startsWith(err);

        }

        assertTrue("DeploymentDescException thrown : EjbError4.jar", isOk);

    }

    // todo : add this test when we'll resolve the problem with soap:header QName

    /*
       public void testHandlerSoapHeaderNotInWsdl() throws Exception {
       boolean isOk = true;
       try {
       DeploymentDesc dd = getDeplDesc("output/ejbjars/EjbError5.war");
       }
       catch (DeploymentDescException wcdde) {
       isOk = false;
       }
       assertFalse("DeploymentDescException thrown : EjbError5.jar", isOk);
       }
     */

    /**
     * Tests if the right exception is thrown when the handler specifies a
     * port-name which one do not match the WSDL. port not present in WSDL.
     *
     * @throws Exception ?
     */
    public void testHandlerPortNotInWsdl() throws Exception {
        boolean isOk = false;

        try {
            DeploymentDesc dd = getDeplDesc("output/ejbjars/EjbError6.jar");

        } catch (DeploymentDescException wcdde) {
            String err = "handler SimpleHandler use an undefined port : wrongPortName in META-INF/wsdl/jarendpoint.wsdl";
            isOk = wcdde.getMessage().startsWith(err);

        }

        assertTrue("DeploymentDescException thrown : EjbError6.jar", isOk);

    }

    /**
     * Tests if the right exception is thrown when the
     * service-endpoint-interface is specified but the class not found.
     *
     * @throws Exception ?
     */
    public void testSiClassMissing() throws Exception {
        boolean isOk = false;

        try {
            DeploymentDesc dd = getDeplDesc("output/ejbjars/EjbError7.jar");

        } catch (DeploymentDescException wcdde) {
            String err = "service-interface : javax.xml.rpc.InexistantService not found !";
            isOk = wcdde.getMessage().startsWith(err);

        }

        assertTrue("DeploymentDescException thrown : EjbError7.jar", isOk);

    }

    /**
     * Tests if the right exception is thrown when a service-ref do not have a
     * corresponding jonas-service-ref.
     *
     * @throws Exception ?
     */
    public void testServRefNameNotInJonasWeb() throws Exception {
        boolean isOk = false;

        try {
            DeploymentDesc dd = getDeplDesc("output/ejbjars/EjbError8.jar");

        } catch (DeploymentDescException wcdde) {
            String err = "jonas-service-ref missing for service-ref-name service/jarendpoint";
            isOk = wcdde.getMessage().startsWith(err);

        }

        assertTrue("DeploymentDescException thrown : EjbError8.jar", isOk);

    }

    /**
     * Tests if the right exception is thrown when a generated
     * service-interface is specified but WSDL is not specified / is partial.
     *
     * @throws Exception ?
     */
    public void testServIntSpeNoWSDL() throws Exception {
        boolean isOk = false;

        try {
            DeploymentDesc dd = getDeplDesc("output/ejbjars/EjbError9.jar");

        } catch (DeploymentDescException wcdde) {
            String err = "generated service-interface 'org.objectweb.ws.services.SimpleService' " + "from service-ref 'service/jarendpoint' require Full WSDL Knowledge";
            isOk = wcdde.getMessage().startsWith(err);

        }

        assertTrue("DeploymentDescException thrown : EjbError9.jar", isOk);

    }

    /**
     * Tests if the right exception is thrown when a generated
     * service-interface is specified but mapping-file doesn't contain a
     * correct mapping
     *
     * @throws Exception ?
     */
    public void testServIntfSpecIncorrectMapping() throws Exception {
        boolean isOk = false;

        try {
            DeploymentDesc dd = getDeplDesc("output/ejbjars/EjbError10.jar");

        } catch (DeploymentDescException wcdde) {
            String err = "jaxrpc-mapping-file 'META-INF/incorrectMapping.xml' need a " + "mapping for package 'org.objectweb.ws.services'";
            isOk = wcdde.getMessage().startsWith(err);
	    if (!isOk) {
		System.err.println("testServIntfSpecIncorrectMapping");
		wcdde.printStackTrace(System.err);
	    }
        }

        assertTrue("DeploymentDescException thrown : EjbError10.jar", isOk);

    }

    /**
     * Tests if the right exception is thrown when multiple beans in EJBJARS
     * contains an error
     *
     * @throws Exception ?
     */
    public void testMultipleEjbJarServiceRefDontMatch() throws Exception {
        boolean isOk = false;

        try {
            DeploymentDesc dd = getDeplDesc("output/ejbjars/ejbjarsClientError.jar");

        } catch (DeploymentDescException wcdde) {
            String err = "service-ref missing for jonas-service-ref service/jarendpoint";
            isOk = wcdde.getMessage().startsWith(err);
	    if (!isOk) {
		System.err.println("testMultipleEjbJarServiceRefDontMatch");
		wcdde.printStackTrace(System.err);
	    }
        }

        assertTrue("DeploymentDescException thrown : ejbjarsClientError.jar", isOk);

    }

    /**
     * JOnAS Method
     *
     * @throws Exception DOCUMENT ME!
     */
    public void testOkSimpleSSB() throws Exception {
        DeploymentDesc dd = getDeplDesc("output/ejbjars/ssbclient.jar");

        BeanDesc       bd = dd.getBeanDesc("SSBClientSLL");

        // Check if service-ref have been correctly loaded
        assertEquals("bd ServiceRef(s)", 1, bd.getServiceRefDesc().length);

        ServiceRef sr = bd.getServiceRefDesc()[0];
        assertNull("jaxrpc-mapping-file", sr.getMappingFile());
        assertNull("wsdl-file", sr.getWSDLFile());
        assertEquals("service-ref-name", "service/jarendpoint", sr.getServiceRefName());
        assertEquals("jndi-name", "service_1", sr.getJndiName());
        assertEquals("service-interface", javax.xml.rpc.Service.class, sr.getServiceInterface());

    }

    /**
     * JOnAS Method
     *
     * @throws Exception DOCUMENT ME!
     */
    public void testOkSimpleFSB() throws Exception {
        DeploymentDesc dd = getDeplDesc("output/ejbjars/fsbclient.jar");

        BeanDesc       bd = dd.getBeanDesc("FsbClientSFL");

        // Check if service-ref have been correctly loaded
        assertEquals("bd ServiceRef(s)", 1, bd.getServiceRefDesc().length);

        ServiceRef sr = bd.getServiceRefDesc()[0];
        assertNull("jaxrpc-mapping-file", sr.getMappingFile());
        assertNotNull("wsdl-file", sr.getWSDLFile());
        assertEquals("service-ref-name", "service/jarendpoint", sr.getServiceRefName());
        assertEquals("jndi-name", "service_1", sr.getJndiName());
        assertEquals("service-interface", javax.xml.rpc.Service.class, sr.getServiceInterface());
        assertEquals("service-qname", new QName("http://ssbendpoint.beans.ws.objectweb.org", "SSBEndpointLocalService"), sr.getServiceQName());

    }

    /**
     * JOnAS Method
     *
     * @throws Exception DOCUMENT ME!
     */
    public void testOkSimpleCMP1() throws Exception {
        DeploymentDesc dd = getDeplDesc("output/ejbjars/cmp1client.jar");

        BeanDesc       bd = dd.getBeanDesc("Cmp1ClientECL");

        // Check if service-ref have been correctly loaded
        assertEquals("bd ServiceRef(s)", 2, bd.getServiceRefDesc().length);

        ServiceRef sr1 = bd.getServiceRefDesc()[0];
        assertNotNull("jaxrpc-mapping-file", sr1.getMappingFile());
        assertNotNull("wsdl-file", sr1.getWSDLFile());
        assertEquals("service-ref-name", "service/jarendpoint1", sr1.getServiceRefName());
        assertEquals("jndi-name", "service_1", sr1.getJndiName());
        assertEquals("service-interface", javax.xml.rpc.Service.class, sr1.getServiceInterface());
        assertEquals("serviceClassName", org.apache.axis.client.Service.class, sr1.getServiceClass());
        assertEquals("service-qname", new QName("http://ssbendpoint.beans.ws.objectweb.org", "SSBEndpointLocalService"), sr1.getServiceQName());
        assertEquals("port-component-ref length", 1, sr1.getPortComponentRefs().size());

        ServiceRef sr2 = bd.getServiceRefDesc()[1];
        assertNull("jaxrpc-mapping-file", sr2.getMappingFile());
        assertNotNull("wsdl-file", sr2.getWSDLFile());
        assertEquals("service-ref-name", "service/jarendpoint2", sr2.getServiceRefName());
        assertEquals("jndi-name", "service_2", sr2.getJndiName());
        assertEquals("service-interface", javax.xml.rpc.Service.class, sr2.getServiceInterface());
        assertEquals("service-qname unspecified but automatically retrieved", new QName("http://ssbendpoint.beans.ws.objectweb.org", "SSBEndpointLocalService"), sr2.getServiceQName());
        assertEquals("serviceClassName", org.apache.axis.client.Service.class, sr2.getServiceClass());
        assertEquals("port-component-ref length", 0, sr2.getPortComponentRefs().size());

    }

    /**
     * JOnAS Method
     *
     * @throws Exception DOCUMENT ME!
     */
    public void testOkSimpleCMP2() throws Exception {
        DeploymentDesc dd = getDeplDesc("output/ejbjars/cmp2client.jar");

        BeanDesc       bd = dd.getBeanDesc("Cmp2ClientEC2L");

        // Check if service-ref have been correctly loaded
        assertEquals("bd ServiceRef(s)", 1, bd.getServiceRefDesc().length);

        ServiceRef sr = bd.getServiceRefDesc()[0];
        assertNull("jaxrpc-mapping-file", sr.getMappingFile());
        assertNull("wsdl-file", sr.getWSDLFile());
        assertEquals("service-ref-name", "service/jarendpoint", sr.getServiceRefName());
        assertEquals("jndi-name", "service_1", sr.getJndiName());
        assertEquals("service-interface", javax.xml.rpc.Service.class, sr.getServiceInterface());
        assertNull("service-qname", sr.getServiceQName());
        assertEquals("serviceClassName", org.apache.axis.client.Service.class, sr.getServiceClass());

    }

    /**
     * JOnAS Method
     *
     * @throws Exception DOCUMENT ME!
     */
    public void testOkSimpleMDB() throws Exception {
        DeploymentDesc dd = getDeplDesc("output/ejbjars/mdbclient.jar");

        BeanDesc       bd = dd.getBeanDesc("MdbClientMD");

        // Check if service-ref have been correctly loaded
        assertEquals("bd ServiceRef(s)", 1, bd.getServiceRefDesc().length);

        ServiceRef sr = bd.getServiceRefDesc()[0];
        assertNull("jaxrpc-mapping-file", sr.getMappingFile());
        assertNotNull("wsdl-file", sr.getWSDLFile());
        assertEquals("service-ref-name", "service/jarendpoint", sr.getServiceRefName());
        assertEquals("jndi-name", "service_1", sr.getJndiName());
        assertEquals("service-interface", javax.xml.rpc.Service.class, sr.getServiceInterface());
        assertEquals("service-qname", new QName("http://ssbendpoint.beans.ws.objectweb.org", "SSBEndpointLocalService"), sr.getServiceQName());
        assertEquals("serviceClass", org.apache.axis.client.Service.class, sr.getServiceClass());

    }

    /**
     * JOnAS Method
     *
     * @throws Exception DOCUMENT ME!
     */
    public void testOkComplexStaticMethod() throws Exception {
        URL            url = new File(path + "output/ejbjars/ejbjarsClientOK.jar").toURL();
        URLClassLoader cl = new URLClassLoader(new URL[] { url }, Thread.currentThread().getContextClassLoader());

        DeploymentDesc dd = EjbDeploymentDescManager.getDeploymentDesc(path + "output/ejbjars/ejbjarsClientOK.jar", cl);

        BeanDesc       ssb = dd.getBeanDesc("SSBClientSLL");
        BeanDesc       fsb = dd.getBeanDesc("FsbClientSFL");
        BeanDesc       cmp1 = dd.getBeanDesc("Cmp1ClientECL");
        BeanDesc       cmp2 = dd.getBeanDesc("Cmp2ClientEC2L");
        BeanDesc       mdb = dd.getBeanDesc("MdbClientMD");

        // check servicerefs length
        assertEquals("ssb ServiceRef(s)", 2, ssb.getServiceRefDesc().length);
        assertEquals("fsb ServiceRef(s)", 0, fsb.getServiceRefDesc().length);
        assertEquals("cmp1 ServiceRef(s)", 1, cmp1.getServiceRefDesc().length);
        assertEquals("cmp2 ServiceRef(s)", 2, cmp2.getServiceRefDesc().length);
        assertEquals("mdb ServiceRef(s)", 1, mdb.getServiceRefDesc().length);

        // Ssb
        //----------------------------------------
        ServiceRef ssb_sr = ssb.getServiceRefDesc()[0];
        ServiceRef ssb_sr2 = ssb.getServiceRefDesc()[1];

        assertEquals("ssb service-ref-name 1", "service/jarendpoint", ssb_sr.getServiceRefName());
        assertEquals("ssb jndi-name 1", "service_1", ssb_sr.getJndiName());
        assertNotNull("ssb WSDL file 1", ssb_sr.getWSDLFile());
        assertNotNull("ssb Mapping file 1", ssb_sr.getMappingFile());
        assertEquals("ssb port-component-ref length", 1, ssb_sr.getPortComponentRefs().size());
        assertEquals("ssb port-component-ref/pcLink", "SSBEndpointLocalService", ((PortComponentRef) ssb_sr.getPortComponentRefs().get(0)).getPortComponentLink());

        assertEquals("ssb service-ref-name 2", "service/warendpoint", ssb_sr2.getServiceRefName());
        assertEquals("ssb jndi-name 2", "service_2", ssb_sr2.getJndiName());
        assertEquals("ssb service-interface", javax.xml.rpc.Service.class, ssb_sr2.getServiceInterface());
        assertNull("ssb WSDL file 2", ssb_sr2.getWSDLFile());
        assertNull("ssb Mapping file 2", ssb_sr2.getMappingFile());

        //-----------------------------------------
        // Fsb not tested, it doesn't contain any service-ref
        // Cmp1
        //----------------------------------------
        ServiceRef cmp1_sr = cmp1.getServiceRefDesc()[0];
        assertEquals("cmp1 service-ref-name 1", "service/jarendpoint", cmp1_sr.getServiceRefName());
        assertEquals("cmp1 jndi-name 1", "service_1", cmp1_sr.getJndiName());
        assertNotNull("cmp1 WSDL file 1", cmp1_sr.getWSDLFile());
        assertNotNull("cmp1 Mapping file 1", cmp1_sr.getMappingFile());
        assertEquals("cmp1 port-component-ref length", 2, cmp1_sr.getPortComponentRefs().size());
        assertEquals("cmp1 port-component-ref/pcLink", "SSBEndpointLocalService", ((PortComponentRef) cmp1_sr.getPortComponentRefs().get(0)).getPortComponentLink());
        assertNull("cmp1 port-component-ref/pcLink", ((PortComponentRef) cmp1_sr.getPortComponentRefs().get(1)).getPortComponentLink());
        assertNotNull("cmp1 handler", cmp1_sr.getHandlerRefs().get(0));
        assertEquals("cmp1 handler-name", "handlerCmp1", ((HandlerRef) cmp1_sr.getHandlerRefs().get(0)).getName());

        //-----------------------------------------
        // Cmp2
        //----------------------------------------
        ServiceRef cmp2_sr = cmp2.getServiceRefDesc()[0];
        ServiceRef cmp2_sr2 = cmp2.getServiceRefDesc()[1];

        assertEquals("cmp2 service-ref-name 1", "service/warendpoint1", cmp2_sr.getServiceRefName());
        assertEquals("cmp2 jndi-name 1", "service_1", cmp2_sr.getJndiName());
        assertNotNull("cmp2 WSDL file 1", cmp2_sr.getWSDLFile());
        assertNotNull("cmp2 Mapping file 1", cmp2_sr.getMappingFile());
        assertEquals("cmp2 port-component-ref length", 0, cmp2_sr.getPortComponentRefs().size());

        assertEquals("cmp2 service-ref-name 2", "service/warendpoint2", cmp2_sr2.getServiceRefName());
        assertEquals("cmp2 jndi-name 2", "service_2", cmp2_sr2.getJndiName());
        assertEquals("cmp2 service-interface", org.objectweb.ws.services.SimpleService.class, cmp2_sr2.getServiceInterface());
        assertNotNull("cmp2 WSDL file 2", cmp2_sr2.getWSDLFile());
        assertNotNull("cmp2 Mapping file 2", cmp2_sr2.getMappingFile());
        assertNotNull("cmp2 handler", cmp2_sr2.getHandlerRefs().get(0));
        assertEquals("cmp2 handler-name", "handlerCmp2", ((HandlerRef) cmp2_sr2.getHandlerRefs().get(0)).getName());

        //-----------------------------------------
        // Mdb
        //----------------------------------------
        ServiceRef mdb_sr = mdb.getServiceRefDesc()[0];

        assertEquals("mdb service-ref-name 1", "service/jarendpoint", mdb_sr.getServiceRefName());
        assertEquals("mdb jndi-name 1", "service_1", mdb_sr.getJndiName());
        assertNotNull("mdb WSDL file 1", mdb_sr.getWSDLFile());
        assertNull("mdb Mapping file 1", mdb_sr.getMappingFile());
        assertEquals("mdb port-component-ref length", 2, mdb_sr.getPortComponentRefs().size());
        assertEquals("mdb service-interface", javax.xml.rpc.Service.class, mdb_sr.getServiceInterface());
        assertEquals("mdb handler length", 2, mdb_sr.getHandlerRefs().size());
        assertEquals("mdb handler-name 1", "handler1Mdb", ((HandlerRef) mdb_sr.getHandlerRefs().get(0)).getName());
        assertEquals("mdb handler-name 2", "handler2Mdb", ((HandlerRef) mdb_sr.getHandlerRefs().get(1)).getName());

        //-----------------------------------------        
    }

    /**
     * Return the DeploymentDesc created from the filename.
     *
     * @param filename The name of the file containning a
     *        ejb-jar.xml/jonas-ejb-jar.xml
     *
     * @return the DeploymentDesc created from the filename
     *
     * @throws Exception when an error occurs (file not found, other exceptions
     *         ...).
     */
    private DeploymentDesc getDeplDesc(String filename) throws Exception {
        File f = new File(path + filename);
        URL  jar = null;

        try {
            jar = f.toURL();

        } catch (java.net.MalformedURLException neveruse) {
            ;
        }

        ClassLoader loader = new URLClassLoader(new URL[] { jar }, Thread.currentThread().getContextClassLoader());

        return ejbDDManager.getDeploymentDesc(jar, loader, null);

    }

    /**
     * Return the input stream from the filename.
     *
     * @param filename ?
     *
     * @return DOCUMENT ME!
     *
     * @throws Exception ?
     */
    private FileInputStream getInputStream(String filename) throws Exception {
        return new FileInputStream(path + filename);

    }

    /**
     * JOnAS Method
     *
     * @param args DOCUMENT ME!
     */
    public static void main(String[] args) {
        junit.textui.TestRunner.run(EjbDeploymentDescManagerTest.class);

    }

}
