package org.objectweb.jonas_web.deployment.lib;

import junit.framework.TestCase;

import org.objectweb.jonas_web.deployment.lib.WebDeploymentDescManager;

import org.objectweb.jonas_web.deployment.api.WebContainerDeploymentDesc;
import org.objectweb.jonas_web.deployment.api.WebContainerDeploymentDescException;
import org.objectweb.jonas_ws.deployment.api.ServiceRef;
import org.objectweb.jonas_ws.deployment.api.MappingFile;
import org.objectweb.jonas_ws.deployment.api.WSDLFile;

import java.net.URLClassLoader;
import java.net.URL;

import java.io.File;
import java.io.FileInputStream;

/**
 * Integration tests for WebManager. (tests for ServiceRef objects)
 */
public class WebDeploymentDescManagerTest extends TestCase {

    WebDeploymentDescManager webDDManager = null;

    private String path = null;
  
    /**
     * Constructor
     */
    public WebDeploymentDescManagerTest(String name) {
	super(name);
	path = System.getProperty("base.dir");
    }
  
    /**
     * return an instance of Web Manager
     */
    public WebDeploymentDescManager createInstance() throws Exception {
	return WebDeploymentDescManager.getInstance();
    }
  
    protected void setUp() throws Exception {
	super.setUp();
	webDDManager = createInstance();

    }
  
    protected void tearDown() throws Exception {
	webDDManager = null;
	super.tearDown();

    }

    /**
     * Tests if the right exception is thrown when a
     * WSDL is not correctly specified in the webDD.
     */
    public void testWsdlFileMissing() throws Exception {
	boolean isOk = false;
	try {
	    WebContainerDeploymentDesc webDD = getDeplDesc("output/webapps/clientErr1.war");
	}
	catch (WebContainerDeploymentDescException wcdde) {
	    String err = "Cannot get wsdl file:";
	    isOk = wcdde.getMessage().startsWith(err);	    
	}	
	assertTrue("WebContainerDeploymentDescException thrown : clientErr1.war", isOk);
    }

    /**
     * Tests if the right exception is thrown when a
     * Mapping file is not correctly specified in the webDD.
     */
    public void testMappingFileMissing() throws Exception {
	boolean isOk = false;
	try {
	    WebContainerDeploymentDesc webDD = getDeplDesc("output/webapps/clientErr2.war");
	}
	catch (WebContainerDeploymentDescException wcdde) {
	    String err = "Cannot read the jaxrpc-mapping-file ";
	    isOk = wcdde.getMessage().startsWith(err);
	}	
	assertTrue("WebContainerDeploymentDescException thrown : clientErr2.war", isOk);
    }

    /**
     * Tests if the right exception is thrown when the
     * service-qname specified in webDD do not match the WSDL
     */
    public void testQNameNotInWsdl() throws Exception {
	boolean isOk = false;
	try {
	    WebContainerDeploymentDesc webDD = getDeplDesc("output/webapps/clientErr3.war");
	}
	catch (WebContainerDeploymentDescException wcdde) {
	    String err = "service-qname '{http://wsendpoint.servlets.ws.objectweb.org}BadName' " + 
		"must be defined in WSDL : 'WEB-INF/wsdl/warendpoint.wsdl'";
	    isOk = wcdde.getMessage().startsWith(err);
	}	
	assertTrue("WebContainerDeploymentDescException thrown : clientErr3.war", isOk);
    }

    /**
     * Tests if the right exception is thrown when the
     * unicity of the handler name is not respected.
     */
    public void testHandlerNameNotUnique() throws Exception {
	boolean isOk = false;
	try {
	    WebContainerDeploymentDesc webDD = getDeplDesc("output/webapps/clientErr4.war");
	}
	catch (WebContainerDeploymentDescException wcdde) {
	    String err = "handler-name 'simpleHandler' must be unique in the module";
	    isOk = wcdde.getMessage().startsWith(err);
	    //wcdde.printStackTrace(System.err);
	}	
	assertTrue("WebContainerDeploymentDescException thrown : clientErr4.war", isOk);
    }

    // todo : add this test when we'll resolve the problem with soap:header QName
    /*
      public void testHandlerSoapHeaderNotInWsdl() throws Exception {
      boolean isOk = true;
      try {
      WebContainerDeploymentDesc webDD = getDeplDesc("output/webapps/clientErr5.war");
      }
      catch (WebContainerDeploymentDescException wcdde) {
      isOk = false;
      }	
      assertFalse("WebContainerDeploymentDescException thrown : clientErr5.war",isOk);
      }
    */    

    /**
     * Tests if the right exception is thrown when the
     * handler specifies a port-name which one do not match the WSDL.
     * port not present in WSDL.
     */
    public void testHandlerPortNotInWsdl() throws Exception {
	boolean isOk = false;
	try {
	    WebContainerDeploymentDesc webDD = getDeplDesc("output/webapps/clientErr6.war");
	}
	catch (WebContainerDeploymentDescException wcdde) {
	    String err = "handler simpleHandler use an undefined port : noendpoint in WEB-INF/wsdl/warendpoint.wsdl";
	    isOk = wcdde.getMessage().startsWith(err);
	}	
	assertTrue("WebContainerDeploymentDescException thrown : clientErr6.war", isOk);
    }

    /**
     * Tests if the right exception is thrown when the
     * service-endpoint-interface is specified but the class not found.
     */
    public void testSeiClassMissing() throws Exception {
	boolean isOk = false;
	try {
	    WebContainerDeploymentDesc webDD = getDeplDesc("output/webapps/clientErr7.war");
	}
	catch (WebContainerDeploymentDescException wcdde) {
	    String err = "class not found noSei: noSei";
	    isOk = wcdde.getMessage().startsWith(err);
	}	
	assertTrue("WebContainerDeploymentDescException thrown : clientErr7.war", isOk);
    }

    /**
     * Tests if the right exception is thrown when a
     * service-ref do not have a corresponding jonas-service-ref.
     */
    public void testServRefNameNotInJonasWeb() throws Exception {
	boolean isOk = false;
	try {
	    WebContainerDeploymentDesc webDD = getDeplDesc("output/webapps/clientErr8.war");
	}
	catch (WebContainerDeploymentDescException wcdde) {
	    String err = "service-ref missing for jonas-service-ref service/warendpoint";
	    isOk = wcdde.getMessage().startsWith(err);
	}	
	assertTrue("WebContainerDeploymentDescException thrown : clientErr8.war", isOk);
    }

    /**
     * Tests if the right exception is thrown when a
     * generated service-interface is specified but WSDL
     * is not specified / is partial.
     */
    public void testServIntSpeNoWSDL() throws Exception {
	boolean isOk = false;
	try {
	    WebContainerDeploymentDesc webDD = getDeplDesc("output/webapps/clientErr9.war");
	}
	catch (WebContainerDeploymentDescException wcdde) {
	    String err = "generated service-interface 'org.objectweb.ws.services.SimpleService' from service-ref 'service/warendpoint' require Full WSDL Knowledge: generated service-interface 'org.objectweb.ws.services.SimpleService' from service-ref 'service/warendpoint' require Full WSDL Knowledge";
	    isOk = wcdde.getMessage().startsWith(err);
	}	
	assertTrue("WebContainerDeploymentDescException thrown : clientErr9.war", isOk);
    }

    /**
     * Tests if the right exception is thrown when a
     * generated service-interface is specified but mapping-file
     * doesn't contain a correct mapping
     */
    public void testServIntSpeIncorrectMapping() throws Exception {
	boolean isOk = false;
	try {
	    WebContainerDeploymentDesc webDD = getDeplDesc("output/webapps/clientErr10.war");
	}
	catch (WebContainerDeploymentDescException wcdde) {
	    String err = "jaxrpc-mapping-file 'WEB-INF/incorrectMapping.xml' need a mapping " + 
		"for package 'org.objectweb.ws.services'";
	    isOk = wcdde.getMessage().startsWith(err);
	}	
	assertTrue("WebContainerDeploymentDescException thrown : clientErr10.war", isOk);
    }

    public void testOkSimple() throws Exception {

	WebContainerDeploymentDesc webDD = getDeplDesc("output/webapps/clientSimple.war");

	// Check if service-ref have been correctly loaded

	ServiceRef sr = webDD.getServiceRefTags()[0];

	assertEquals("webDD ServiceRef(s)", 1, webDD.getServiceRefTags().length);      
	assertNull("jaxrpc-mapping-file",sr.getMappingFile());
	assertNull("wsdl-file",sr.getWSDLFile());
	assertEquals("service-ref-name", "service/warendpoint", sr.getServiceRefName());
	assertEquals("jndi-name", "warendpoint_1", sr.getJndiName());

    }

    public void testOkComplex() throws Exception {

	WebContainerDeploymentDesc webDD = getDeplDesc("output/webapps/clientComplex.war");

	// Check if service-ref have been correctly loaded
	FileInputStream wsdlis = getInputStream("etc/warClientComplex/wsdl/warendpoint.wsdl");
	WSDLFile wsdlf = new WSDLFile(wsdlis, path + "etc/warClientComplex/wsdl/warendpoint.wsdl");

	FileInputStream mfis = getInputStream("etc/warClientComplex/warEndpointMapping.xml");
	MappingFile mf = new MappingFile(mfis, path + "etc/warClientComplex/warEndpointMapping.xml");
      
	wsdlis = getInputStream("etc/warClientComplex/wsdl/jarendpoint.wsdl");
	WSDLFile wsdlf2 = new WSDLFile(wsdlis, path + "etc/warClientComplex/wsdl/jarendpoint.wsdl");
      
	mfis = getInputStream("etc/warClientComplex/jarEndpointMapping.xml");
	MappingFile mf2 = new MappingFile(mfis, path + "etc/warClientComplex/jarEndpointMapping.xml");
      
	ServiceRef sr = webDD.getServiceRefTags()[0];
	ServiceRef sr2 = webDD.getServiceRefTags()[1];
      
	assertEquals("webDD ServiceRef(s)", 2, webDD.getServiceRefTags().length);
      
	assertEquals("service-ref-name 1", "service/warendpoint", sr.getServiceRefName());
	assertEquals("jndi-name 1", "warendpoint_1", sr.getJndiName());      
	assertTrue("WSDL file 1", wsdlf.equals(sr.getWSDLFile()));
	assertTrue("Mapping file 1", mf.equals(sr.getMappingFile()));

	assertEquals("service-ref-name 2", "service/jarendpoint", sr2.getServiceRefName());
	assertEquals("jndi-name 2", "service_1", sr2.getJndiName());
	assertTrue("WSDL file 2", wsdlf2.equals(sr2.getWSDLFile()));
	assertTrue("Mapping file 2", mf2.equals(sr2.getMappingFile()));

	// todo : ajouter des tests.

    }



    /**
     * Return the webDeploymentDesc created from the filename.
     * @param filename The name of the file containning a web.xml/jonas-web.xml
     * @return the webDeploymentDesc created from the filename
     * @throws Exception when an error occurs(file not found, other exceptions ...).
     */
    private WebContainerDeploymentDesc getDeplDesc(String filename) 
	throws Exception {	

	File f = new File(path + filename);
	URL war = null;
	try {
	    war = f.toURL();
	}
	catch(java.net.MalformedURLException neveruse){}
	ClassLoader loader = new URLClassLoader(new URL[] {war}, Thread.currentThread().getContextClassLoader());
	return webDDManager.getDeploymentDesc(war, loader, null);      
    }
  
    /**
     * Return the input stream from the filename.
     */
    private FileInputStream getInputStream(String filename) throws Exception {
	return new FileInputStream(path + filename);
    }

    public static void main(String[] args) {
	junit.textui.TestRunner.run(WebDeploymentDescManagerTest.class);
    }
}
