package org.objectweb.jonas_ws.deployment.lib;

import junit.framework.TestCase;
import org.objectweb.jonas_ws.deployment.api.WSDeploymentDesc;
import org.objectweb.jonas_ws.deployment.api.WSDeploymentDescException;
import org.objectweb.jonas_ws.deployment.api.ServiceDesc;
import org.objectweb.jonas_ws.deployment.api.PortComponentDesc;
import org.objectweb.jonas_ws.deployment.api.MappingFile;
import org.objectweb.jonas_ws.deployment.api.WSDLFile;

import java.net.URLClassLoader;
import java.net.URL;

import java.io.File;
import java.io.FileInputStream;

/**
 * Integration tests for WebServiceManager. (tests for ServiceDesc objects)
 */
public class WSDeploymentDescManagerTest extends TestCase {

    WSDeploymentDescManager wsDDManager = null;

    private String path = null;
  
    /**
     * Constructor
     */
    public WSDeploymentDescManagerTest(String name) {
	super(name);
	path = System.getProperty("base.dir");
    }
  
    /**
     * return an instance of WebService Manager
     */
    public WSDeploymentDescManager createInstance() throws Exception {
	WSDeploymentDescManager.getInstance().clearCache();
	return WSDeploymentDescManager.getInstance();
    }
  
    protected void setUp() throws Exception {
	super.setUp();
	wsDDManager = createInstance();
    }
  
    protected void tearDown() throws Exception {
	wsDDManager = null;
	super.tearDown();
    }

    /**
     * Tests if the right exception is thrown when a
     * WSDL is not correctly specified in the wsDD.
     */
    public void testWsdlFileMissing() throws Exception {
	boolean isOk = false;
	try {
	    WSDeploymentDesc wsDD = getDeploymentDesc("output/ejbjars/endpointError1.jar");
	}
	catch (WSDeploymentDescException wsdde) {
	    String err = "Cannot read the wsdl-file META-INF/wsdl/no.wsdl";
	    isOk = wsdde.getMessage().startsWith(err);
	    if (!isOk) {
		System.err.println("testWsdlFileMissing");
		wsdde.printStackTrace(System.err);
	    }
		
	}	
	assertTrue("WSDeploymentDescException thrown : endpointError1.jar", isOk);
    }

    /**
     * Tests if the right exception is thrown when a
     * Mapping file is not correctly specified in the wsDD.
     */
    public void testMappingFileMissing() throws Exception {
	boolean isOk = false;
	try {
	    WSDeploymentDesc wsDD = getDeploymentDesc("output/webapps/endpointError2.war");
	}
	catch (WSDeploymentDescException wsdde) {
	    String err = "Cannot read the jaxrpc-mapping-file ";
	    isOk = wsdde.getMessage().startsWith(err);
	    if (!isOk) {
		System.err.println("testMappingFileMissing");
		wsdde.printStackTrace(System.err);
	    }
	}	
	assertTrue("WSDeploymentDescException thrown : endpointError2.war", isOk);
    }

    /**
     * Tests if the right exception is thrown when the
     * webservice-description-name is not unique in the module
     */
    public void testServiceNameNotUnique() throws Exception {
	boolean isOk = false;
	try {
	    WSDeploymentDesc wsDD = getDeploymentDesc("output/ejbjars/endpointError3.jar");
	}
	catch (WSDeploymentDescException wsdde) {
	    String err = "service 'MyWebService' haven't got a unique name";
	    isOk = wsdde.getMessage().startsWith(err);
	    if (!isOk) {
		System.err.println("testServiceNameNotUnique");
		wsdde.printStackTrace(System.err);
	    }
	}	
	assertTrue("WSDeploymentDescException thrown : endpointError3.jar", isOk);
    }

    /**
     * Tests if the right exception is thrown when the
     * unicity of the port component name is not respected.
     */
    public void testPortComponentNameNotUnique() throws Exception {
	boolean isOk = false;
	try {
	    WSDeploymentDesc wsDD = getDeploymentDesc("output/webapps/endpointError4.war");
	}
	catch (WSDeploymentDescException wsdde) {
	    String err = "port-component 'WebappPortComp' has same name than a previously recorded port-component";
	    isOk = wsdde.getMessage().startsWith(err);
	    if (!isOk) {
		System.err.println("testPortComponentNameNotUnique");
		wsdde.printStackTrace(System.err);
	    }
	}	
	assertTrue("WSDeploymentDescException thrown : endpointError4.war", isOk);
    }

    /**
     * Tests if the right exception is thrown when the
     * unicity of the handler name is not respected.
     */
    public void testHandlerNameNotUnique() throws Exception {
	boolean isOk = false;
	try {
	    WSDeploymentDesc wsDD = getDeploymentDesc("output/ejbjars/endpointError5.jar");
	}
	catch (WSDeploymentDescException wsdde) {
	    String err = "handler 'MyHandler' haven't got a unique name";
	    isOk = wsdde.getMessage().startsWith(err);
	    if (!isOk) {
		System.err.println("testHandlerNameNotUnique");
		wsdde.printStackTrace(System.err);
	    }
	}	
	assertTrue("WSDeploymentDescException thrown : endpointError5.jar", isOk);
    }

    /**
     * Tests if the right exception is thrown when the
     * wsdl ports are not fully contains in the set of port-component.
     */
    public void testWsdlPortNotIncludedInPortComponentSet() throws Exception {
	boolean isOk = false;
	try {
	    WSDeploymentDesc wsDD = getDeploymentDesc("output/webapps/endpointError6.war");
	}
	catch (WSDeploymentDescException wsdde) {
	    String err = "WSDL 'WEB-INF/wsdl/warendpoint.wsdl' uses a port undefined as portComponent";
	    isOk = wsdde.getMessage().startsWith(err);
	    if (!isOk) {
		System.err.println("testWsdlPortNotIncludedInPortComponentSet");
		wsdde.printStackTrace(System.err);
	    }
	}	
	assertTrue("WSDeploymentDescException thrown : endpointError6.war", isOk);
    }

    /**
     * Tests if the right exception is thrown when the
     * soap-header QNames defined by a header doesn't exist in the WSDL
     * todo : implement this functionnality ...
     */
    public void testSoapHeaderNotInWsdl() throws Exception {
	boolean isOk = false;
	try {
	    WSDeploymentDesc wsDD = getDeploymentDesc("output/ejbjars/endpointError7.jar");
	}
	catch (WSDeploymentDescException wsdde) {
	    String err = "handler 'MyHandler' use an undefined soap header : '{http://ssbendpoint.beans.ws.objectweb.org}UnknownSoapHeader' in META-INF/wsdl/jarendpoint.wsdl";
	    isOk = wsdde.getMessage().startsWith(err);
	    if (!isOk) {
		System.err.println("testSoapHeaderNotInWsdl");
		wsdde.printStackTrace(System.err);
	    }
	}	
	assertTrue("WSDeploymentDescException thrown : endpointError7.jar", isOk);
    }

    /**
     * Tests if the right exception is thrown when a
     * port-component is linked (QName) to a port and the specified port
     * don't have a Soap Binding.
     */
    public void testPortDontUseSoapBinding() throws Exception {
	boolean isOk = false;
	try {
	    WSDeploymentDesc wsDD = getDeploymentDesc("output/webapps/endpointError8.war");
	}
	catch (WSDeploymentDescException wsdde) {
	    String err = "port-component 'WebappPortComp' associated with port " +
		"'{http://wsendpoint.servlets.ws.objectweb.org}wsendpoint' doesn't use a SOAP binding";
	    isOk = wsdde.getMessage().startsWith(err);
	    if (!isOk) {
		System.err.println("testPortDontUseSoapBinding");
		wsdde.printStackTrace(System.err);
	    }
	}	
	assertTrue("WSDeploymentDescException thrown : endpointError8.war", isOk);
    }

    /**
     * Tests if the right exception is thrown when a
     * wsdl-port don't exist in the WSDL.
     */
    public void testPortComponentSpecifiesAnUnknownPortInWsdl() throws Exception {
	boolean isOk = false;
	try {
	    WSDeploymentDesc wsDD = getDeploymentDesc("output/ejbjars/endpointError9.jar");
	}
	catch (WSDeploymentDescException wsdde) {
	    String err = "port-component 'ssbEndpointLocal' is linked with an unknown WSDL port " + 
		"'{http://ssbendpoint.beans.ws.objectweb.org}UnknownWsdlPort'";
	    isOk = wsdde.getMessage().startsWith(err);
	    if (!isOk) {
		System.err.println("testPortComponentSpecifiesAnUnknownPortInWsdl");
		wsdde.printStackTrace(System.err);
	    }
	}	
	assertTrue("WSDeploymentDescException thrown : endpointError9.jar", isOk);
    }

    /**
     * Tests if the right exception is thrown when a
     * jaxrpc endpoint is defined in an EjbJar
     */
    public void testServletLinkInEjbJar() throws Exception {
	boolean isOk = false;
	try {
	    WSDeploymentDesc wsDD = getDeploymentDesc("output/ejbjars/endpointError10.jar");
	}
	catch (WSDeploymentDescException wsdde) {
	    String err = "Error while reading/parsing";
	    isOk = wsdde.getMessage().startsWith(err);
	    if (!isOk) {
		System.err.println("testServletLinkInEjbJar");
		wsdde.printStackTrace(System.err);
	    }
	}	
	assertTrue("WSDeploymentDescException thrown : endpointError10.jar", isOk);
    }

    /**
     * Tests if the right exception is thrown when a
     * session bean endpoint is defined in a Webapp
     */
    public void testEjbLinkInWebapp() throws Exception {
	boolean isOk = false;
	try {
	    WSDeploymentDesc wsDD = getDeploymentDesc("output/webapps/endpointError11.war");
	}
	catch (WSDeploymentDescException wsdde) {
	    String err = "Unable to load EjbJar";
	    isOk = wsdde.getMessage().startsWith(err);
	    if (!isOk) {
		System.err.println("testEjbLinkInWebapp");
		wsdde.printStackTrace(System.err);
	    }
	}	
	assertTrue("WSDeploymentDescException thrown : endpointError11.war", isOk);
    }

    /**
     * Tests if the right exception is thrown when a
     * SEI class is not found.
     */
    public void testSeiClassNotFound() throws Exception {
	boolean isOk = false;
	try {
	    WSDeploymentDesc wsDD = getDeploymentDesc("output/ejbjars/endpointError12.jar");
	}
	catch (WSDeploymentDescException wsdde) {
	    String err = "SEI class not found org.objectweb.ws.beans.ssbendpoint.UnknownSeiClass" +
		" in port component :ssbEndpointLocal: org.objectweb.ws.beans.ssbendpoint.UnknownSeiClass";
	    isOk = wsdde.getMessage().startsWith(err);
	    if (!isOk) {
		System.err.println("testSeiClassNotFound");
		wsdde.printStackTrace(System.err);
	    }
	}	
	assertTrue("WSDeploymentDescException thrown : endpointError12.jar", isOk);
    }

    /**
     * Tests if the right exception is thrown when a
     * Handler class is not found.
     */
    public void testHandlerClassNotFound() throws Exception {
	boolean isOk = false;
	try {
	    WSDeploymentDesc wsDD = getDeploymentDesc("output/webapps/endpointError13.war");
	}
	catch (WSDeploymentDescException wsdde) {
	    String err = "handler class not found: org.objectweb.ws.handlers.UnknownHandler";
	    isOk = wsdde.getMessage().startsWith(err);
	    if (!isOk) {
		System.err.println("testHandlerClassNotFound");
		wsdde.printStackTrace(System.err);
	    }
	}	
	assertTrue("WSDeploymentDescException thrown : endpointError13.war", isOk);
    }


    /**
     * Tests if the right exception is thrown when
     * XML of webservices.xml is not well formed
     */
    public void testXmlNotWellFormed() throws Exception {
	boolean isOk = false;
	try {
	    WSDeploymentDesc wsDD = getDeploymentDesc("output/webapps/endpointError14.war");
	}
	catch (WSDeploymentDescException wsdde) {
	    String err = "WEB-INF/webservices.xml:26:16: Invalid XML: Error parsing XML " +
		"document: The end-tag for element type \"handler\" must end with a '>' delimiter.";
	    isOk = wsdde.getMessage().startsWith(err);
	    if (!isOk) {
		System.err.println("testXmlNotWellFormed");
		wsdde.printStackTrace(System.err);
	    }
	}	
	assertTrue("WSDeploymentDescException thrown : endpointError14.war", isOk);
    }

    public void testOkSimple() throws Exception {

	WSDeploymentDesc wsDD = getDeploymentDesc("output/webapps/endpoint.war");

	// Check if service-ref have been correctly loaded
	System.out.println("wsDD.getServiceDescs() = "+wsDD);
	assertEquals("wsDD ServiceRef(s)", 1, wsDD.getServiceDescs().size());

	ServiceDesc sd = (ServiceDesc) wsDD.getServiceDescs().get(0);

	assertNotNull("jaxrpc-mapping-file", sd.getMapping());
	assertNotNull("wsdl-file", sd.getWSDL());
	assertEquals("service-description-name", "simple web endpoint", sd.getName());

    }

    public void testOkComplex() throws Exception {

	WSDeploymentDesc wsDD = getDeploymentDesc("output/webapps/endpointComplex.war");

	// Check if service-ref have been correctly loaded
	FileInputStream wsdlis = getInputStream("etc/warEndpointComplex/WEB-INF/wsdl/warendpoint.wsdl");
	WSDLFile wsdlf = new WSDLFile(wsdlis, path + "etc/warEndpointComplex/WEB-INF/wsdl/warendpoint.wsdl");
	FileInputStream mfis = getInputStream("etc/warEndpoint/warEndpointMapping.xml");
	MappingFile mf = new MappingFile(mfis, path + "etc/warEndpoint/warEndpointMapping.xml");
	System.out.println("wsDD.getServiceDescs().size() = "+wsDD.getServiceDescs().size());
      
	assertEquals("wsDD ServiceDesc(s)", 2, wsDD.getServiceDescs().size());
      
	ServiceDesc sd = (ServiceDesc)wsDD.getServiceDescs().get(0);
	ServiceDesc sd2 = (ServiceDesc)wsDD.getServiceDescs().get(1);
	
      
	assertTrue("WSDL file 1", wsdlf.equals(sd.getWSDL()));
	assertTrue("Mapping file 1", mf.equals(sd.getMapping()));
	assertEquals("service-description 1", 2, sd.getPortComponents().size());
	assertNotNull("port component 1",sd.getPortComponent("WebappPortComp1"));
	assertNotNull("port component 2",sd.getPortComponent("WebappPortComp2"));
	PortComponentDesc pc1 = sd.getPortComponent("WebappPortComp1");
	PortComponentDesc pc2 = sd.getPortComponent("WebappPortComp2");
	assertEquals("port component 1 handler", 2, pc1.getHandlers().size());
	assertEquals("port component 2 handler", 1, pc2.getHandlers().size());
      
	assertTrue("WSDL file 2", wsdlf.equals(sd2.getWSDL()));
	assertTrue("Mapping file 2", mf.equals(sd2.getMapping()));
	assertEquals("service-description 2", 1, sd2.getPortComponents().size());
	assertNotNull("port component 3",sd2.getPortComponent("WebappPortComp3"));
	PortComponentDesc pc3 = sd2.getPortComponent("WebappPortComp3");
	assertEquals("port component 3 handler", 1, pc3.getHandlers().size());

    }


    /**
     * Return the webDeploymentDesc created from the filename.
     * @param filename The name of the file containning a web.xml/jonas-web.xml
     * @return the webDeploymentDesc created from the filename
     * @throws Exception when an error occurs(file not found, other exceptions ...).
     */
    private WSDeploymentDesc getDeploymentDesc(String filename) 
	throws Exception {	

	File f = new File(path + filename);
	URL jar = null;
	try {
	    jar = f.toURL();
	}
	catch(java.net.MalformedURLException neveruse){}

	ClassLoader loader = new URLClassLoader(new URL[] {jar}, Thread.currentThread().getContextClassLoader());
	return wsDDManager.getDeploymentDesc(jar, loader, null);
    }
    /**
     * Return the input stream from the filename.
     */
    private FileInputStream getInputStream(String filename) throws Exception {
	return new FileInputStream(path + filename);
    }
  
    public static void main(String[] args) {
	junit.textui.TestRunner.run(WSDeploymentDescManagerTest.class);
    }
}
